/**************************************************************************
 *
 * File: NosNecFilter.java
 * Author: Suresh Srinivasan
 *
 **************************************************************************/

import gov.nih.nlm.mms.filters.AbstractFilter;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;

import java.util.Properties;

import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;

/**
 * This Filter removes names with underspecification markers, "NOS", i.e.,
 * (not otherwise specified) and classification specific markers,
 * "NEC" (not elsewhere classified)
 *
 * @author Suresh Srinivasan
 **/
public class NosNecFilter extends AbstractFilter {

    static final boolean DEFAULT_REMOVENOS=false;
    static final boolean DEFAULT_REMOVENEC=false;
    boolean removenos;
    boolean removenec;
    JCheckBox checknos;
    JCheckBox checknec;

    /**
     * Instantiates an {@link NosNecFilter}.
     */
    public NosNecFilter() { }

    /**
     * Returns the help panel.
     * @return the help panel
     */
    public JPanel getHelpPanel(){
	return SwingToolkit.getPanelForHTML(
	    this.getClass().getClassLoader().getResource(
			 "NosNecHelp.html"));
    }

    /**
     * Returns the GUI configuration panel.
     * @return the GUI configuration panel
     */
    public JPanel getPanel(){
	JPanel p = new JPanel();
	checknos = new JCheckBox("Remove \"NOS\" names");
	checknec = new JCheckBox("Remove \"NEC\" names");
	removenos = DEFAULT_REMOVENOS;
	removenec = DEFAULT_REMOVENEC;
	checknos.setSelected(removenos);
	checknec.setSelected(removenec);

	// Use addActionListener instead of addItemListener
	// The action listener responds to a user click
	// but not to the programatic selection (i.e. checknos.setSelected(true)).
	// This is important because setPanelSelections should not trigger
	// a fireDataChaned() event, otherwise it poorly interacts with
	// the Undo/Redo mechanism.
	checknos.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	      fireDataChanged();
	      removenos = checknos.isSelected();
	  }
	});
	checknec.addActionListener(new ActionListener() {
	  public void actionPerformed(ActionEvent e) {
	      fireDataChanged();
	      removenec = checknec.isSelected();
	  }
	});

	p.setBorder(new CompoundBorder(new EtchedBorder(EtchedBorder.RAISED), new EmptyBorder(10, 10, 10, 10)));
	//	p.setLayout(new FlowLayout());
	p.setLayout(new GridLayout(0, 1));
        p.add(checknos);
        p.add(checknec);
	JPanel p1 = new JPanel();
	p1.setLayout(new BorderLayout());
	p1.add(p, BorderLayout.NORTH);
	p1.add(new JPanel(), BorderLayout.CENTER);

	//	return p;
	return p1;
    }

    /**
     * Returns the description
     * @return the description.
     */
    public String getTabDescription() {
	return "This is a sample filter provided to demonstrate the import filter feature.\n\nThis filter will remove atoms with \"NOS\" or \"NEC\" strings.";
    }

    /**
     * Reset to default configuration.
     */
    public void resetToDefaults() {
	removenos = DEFAULT_REMOVENOS;
	removenec = DEFAULT_REMOVENEC;
    }

    /**
     * Returns the filter state.
     * @return the filter state.
     */
    public Properties getProperties() {
	Properties p = new Properties();
	p.setProperty("removenos", String.valueOf(removenos));
	p.setProperty("removenec", String.valueOf(removenec));
	return p;
    }

    /**
     * Sets the filter state.
     */
    public void setProperties(Properties props) {
      removenos = Boolean.valueOf(props.getProperty("removenos","true")).booleanValue();
      removenec = Boolean.valueOf(props.getProperty("removenec","true")).booleanValue();
    }

    /**
     * Apply the filter logic to the {@link Concept}.
     * @param cui the {@link Concept}
     */
    public void applyFilter(Concept cui) {
	Atom a[] = cui.getAtoms();
	boolean deathrow[] = new boolean[a.length];

	for (int i=0; i<a.length; i++) {
	    String s = a[i].getStr();
	    if (removenos) {
		if (pattern_match(s, "NOS", true) ||
		    pattern_match(s, "not otherwise specified", false)) {
		    deathrow[i] = true;
		}
	    }
	    if (removenec) {
		if (pattern_match(s, "NEC", true) ||
		    pattern_match(s, "not elsewhere classified", false)) {
		    deathrow[i] = true;
		}
	    }
	}

	for (int i=0; i<a.length; i++) {
	    if (deathrow[i]) {
		System.err.println("CUI=" + cui.getCui() + ", SAB=" + a[i].getSource().toString() + ", SCD=" + a[i].getCode() + ": " + a[i].getStr());
		cui.removeAtom(a[i]);
	    }
	}
	return;
    }

    private boolean pattern_match(String x, String p, boolean exactcase) {
	int i;
	String s = (exactcase ? x : x.toLowerCase());
	int sl=0;
	int pl=0;

	if ((i=s.indexOf(p)) == (-1))
	    return false;
	if (i>0 && Character.isLetter(s.charAt(i-1)))
	    return false;
	sl = s.length();
	pl = p.length();
	if (i<(sl-pl) && Character.isLetter(s.charAt(i+pl)))
	    return false;
	return true;
    }

    /**
     * Sets the panel selections from the state passed in via {@link #setProperties(Properties)}.
     */
    public void setPanelSelections() {
	checknos.setSelected(removenos);
	checknec.setSelected(removenec);
    }

    /**
     * Return the name.
     * @return the name
     */
    public String getName() {
	return "NosNec (for Testing)";
    }

    /**
     * Returns the configuration details for the <code>mmsys.log</code>.
     * @return the configuration details for the <code>mmsys.log</code>
     */
    public String getConfigurationDetails() {
	String nos, nec;

	if (removenos)
	    nos = "Strings with underspecification markers (NOS) have been removed";
	else
	    nos = "Strings with underspecification markers (NOS) have not been removed";

	if (removenec)
	    nec = "Strings with classification markers (NEC) have been removed";
	else
	    nec = "Strings with classification markers (NEC) have not been removed";
	return nos + (nos != null ? System.getProperty("line.separator") : "") + nec;
    }

  /**
   * Indicates that the filter can be used in "efficient mode".
   * @return <code>true</code>
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}

