/**************************************************************************
 *
 * File: SampleMetaFilter.java
 * Author: Deborah Shapiro
 *
 **************************************************************************/
import gov.nih.nlm.mms.CuiList;
import gov.nih.nlm.mms.CuiSuiAuiList;
import gov.nih.nlm.mms.GUIConfigurable;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.SourcesToRemoveEvent;
import gov.nih.nlm.mms.SourcesToRemoveListener;
import gov.nih.nlm.mms.filters.AbstractFilter;
import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.swing.SuperJList;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.CoocurrenceType;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.RelationshipAttribute;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;

public class SampleMetaFilter extends AbstractFilter implements
    SourcesToRemoveListener {

  //
  // Filter state and GUI element
  //
  private Map checkbox_map = new HashMap();

  private boolean ignore_events = false;

  //
  // Track conditions that have been met already
  //
  private Set conditions_applied = new HashSet();

  //
  // Local copy of the cui list
  //
  private CuiList cui_list = new CuiSuiAuiList();

  //
  // All conditions we are looking for
  //
  private final Condition[] conditions = new Condition[36];

  private final Properties condition_name_map = new Properties();

  //
  // List of all languages
  //
  private final Language[] lats = Language.getLanguages();

  //
  // All coocurrence types
  //
  private final CoocurrenceType[] coc_types = CoocurrenceType
      .getCoocurrenceTypes();

  //
  // Used to write a "sample_meta.log" log file
  //
  private PrintWriter pw = null;

  //
  // List of sources to remove
  //
  private List sources_to_remove = new ArrayList();

  //
  // Sui saved for ambiguous SUI condition
  //
  private final Map saved_ams = new HashMap();

  //
  // CUIs added ONLY because they were CUI2 values
  //
  private final Set cui2s_added_by_user_only = new HashSet();

  //
  // List of sources for which SAMPLEMETA requires a concept
  //
  private Set sabs_required = new HashSet();

  //
  // List of sources meeting the KeepAtomsFromVariousSABs condition
  //
  private Set sabs_found = new HashSet();

  //
  // List of warnings indicated by validate() method
  //
  private List warning_msgs = new ArrayList();

  //
  // Constructor
  //

  /**
   * Instantiates a {@link SampleMetaFilter}.
   */
  public SampleMetaFilter() {

    //
    // Add default required SABs
    //
    sabs_required.add("SNOMEDCT");
    sabs_required.add("MSH");
    sabs_required.add("AIR");
    sabs_required.add("UWDA");
    sabs_required.add("CPT");
    sabs_required.add("ICD9CM");

    //
    // Create conditions to look for
    //
    conditions[0] = new KeepAtomFromEachLanguage();
    conditions[1] = new KeepUnicodeCharacter();
    conditions[2] = new KeepENGDiffLUIs();
    conditions[3] = new KeepENGDiffSUIs();
    conditions[4] = new KeepENGDiffAUIs();
    conditions[5] = new KeepENGDiffSABsLUIs();
    conditions[6] = new KeepENGDiffSABsSUIs();
    conditions[7] = new KeepENGDiffSABsAUIs();
    conditions[8] = new KeepNotENGDiffLUIs();
    conditions[9] = new KeepNotENGDiffSUIs();
    conditions[10] = new KeepNotENGDiffAUIs();
    conditions[11] = new KeepNotENGDiffSABsLUIs();
    conditions[12] = new KeepNotENGDiffSABsSUIs();
    conditions[13] = new KeepNotENGDiffSABsAUIs();
    conditions[14] = new KeepSNOMEDCTtoICD9CMMappingConcept();
    conditions[15] = new KeepMSHATXMappingConcept();
    conditions[16] = new KeepAllSRC();
    conditions[17] = new KeepAtomAttribute();
    conditions[18] = new KeepAtomRelationship();
    conditions[19] = new KeepAttributeConnectedToUnicode();
    conditions[20] = new KeepConceptLevelAttribute();
    conditions[21] = new KeepCoocurrenceTypes();
    conditions[22] = new KeepRelationshipWithRUIAttributes();
    conditions[23] = new KeepDEFWithUnicodeCharacter();
    conditions[24] = new KeepLongAttributeValue();
    conditions[25] = new KeepLongDEF();
    conditions[26] = new KeepRelationshipAttribute();
    conditions[27] = new KeepContextsWithSameSelfAtom();
    conditions[28] = new KeepContextWithUnicodeAncestor();
    conditions[29] = new KeepContextWithUnicodeSibling();
    conditions[30] = new KeepContextWithUnicodeSelf();
    conditions[31] = new KeepContextWithUnicodeChild();
    conditions[32] = new KeepAmbiguousSUIs();
    conditions[33] = new KeepAtomFromVariousSABs();
    conditions[34] = new KeepConceptLevelRelationship();
    conditions[35] = new KeepAllMSHTQ();
    for (int i = 0; i < conditions.length; i++)
      condition_name_map.put(conditions[i].getName(), "true");
  }

  //
  // Methods
  //

  /**
   * Returns a help {@link JPanel}.
   * 
   * @return a help {@link JPanel}
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(this.getClass().getClassLoader()
        .getResource("config/SampleMetaFilter.html"));
  }

  /**
   * Returns the description.
   * 
   * @return the description
   */
  public String getTabDescription() {
    return "This is a filter used to create a sample subset of the Metathesuaurus to be used for MetamorphoSys testing.";
  }

  /**
   * Sets the state of the filter.
   * 
   * @param props
   *          the state of the filter
   */
  public void setProperties(Properties props) {
    sabs_required = new HashSet();
    final Iterator iter = props.entrySet().iterator();
    while (iter.hasNext()) {
      final Map.Entry entry = (Map.Entry) iter.next();
      final String key = (String) entry.getKey();

      //
      // Handle required sabs list property
      //
      if (key.equals("required_sabs")) {
        final String[] tokens = FieldedStringTokenizer.split((String) entry
            .getValue(), ";");
        for (int i = 0; i < tokens.length; i++) {
          sabs_required.add(tokens[i]);
        }
      }

      //
      // Handle condition properties
      //
      else {
        condition_name_map.setProperty(key, props.getProperty(key));
      }
    }
  }

  /**
   * Returns the GUI configuration {@link JPanel}.
   * 
   * @return the GUI configuration {@link JPanel}
   */
  public JPanel getPanel() {

    //
    // Create panel with box layout and intial comment
    //
    final JPanel panel = new JPanel();
    panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
    final JLabel label = new JLabel(
        "<html><b>Choose from the following conditions:</b></html>");
    label.setBorder(new EmptyBorder(10, 0, 10, 0));
    panel.add(label);

    //
    // Sort conditions and add checkbox for each one
    //
    Arrays.sort(conditions, new ConditionsComparator());
    for (int i = 0; i < conditions.length; i++) {
      final Condition condition = conditions[i];
      final JCheckBox box = new JCheckBox(condition.getName());
      box.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          if (!ignore_events) {
            fireDataChanged();
            condition_name_map.setProperty(condition.getName(), String
                .valueOf(box.isSelected()));
          }
        }
      });
      checkbox_map.put(condition.getName(), box);
      panel.add(box);
      panel.setBorder(new EmptyBorder(15, 15, 15, 15));
    }
    return panel;
  }

  /**
   * Sets the panel selections from the state variables.
   */
  public void setPanelSelections() {
    ignore_events = true;
    final Iterator iter = condition_name_map.keySet().iterator();
    while (iter.hasNext()) {
      final String name = (String) iter.next();
      final JCheckBox box = (JCheckBox) checkbox_map.get(name);
      box.setSelected(Boolean.valueOf(condition_name_map.getProperty(name))
          .booleanValue());
    }
    ignore_events = false;
  }

  /**
   * Resets the filter state to the default.
   */
  public void resetToDefaults() {
    ignore_events = true;
    final Iterator iter = condition_name_map.keySet().iterator();
    while (iter.hasNext()) {
      final String name = (String) iter.next();
      condition_name_map.put(name, "true");
    }
    sabs_required.clear();
    sabs_found.clear();
    sabs_required.add("SNOMEDCT");
    sabs_required.add("MSH");
    sabs_required.add("AIR");
    sabs_required.add("UWDA");
    sabs_required.add("CPT");
    sabs_required.add("ICD9CM");
    ignore_events = false;
  }

  /**
   * Indicates whether or not this filter has advanced options
   * 
   * @return <code>true</code>
   */
  public boolean hasAdvancedOptions() {
    return true;
  }

  /**
   * Show advanced options GUI.
   */
  public void showOptions() {
    final SampleMetaFilterOptionsGUI source_options_gui = new SampleMetaFilterOptionsGUI();
    source_options_gui.setVisible(true);
  }

  /**
   * Returns a {@link Properties} object with the filter state
   * 
   * @return a {@link Properties} object with the filter state
   */
  public Properties getProperties() {
    final Properties p = new Properties();

    //
    // Condition properties
    //
    final Iterator i = condition_name_map.keySet().iterator();
    while (i.hasNext()) {
      String s = (String) i.next();
      p.setProperty(s, condition_name_map.getProperty(s));
    }

    //
    // Required SABs property
    //
    final StringBuffer line = new StringBuffer();
    final Iterator sab_iter = sabs_required.iterator();
    while (sab_iter.hasNext()) {
      String sab = (String) sab_iter.next();
      line.append(sab).append(";");
    }
    final Iterator sab_found_iter = sabs_found.iterator();
    while (sab_found_iter.hasNext()) {
      String sab = (String) sab_found_iter.next();
      line.append(sab).append(";");
    }
    final String prop = line.toString();
    p.setProperty("required_sabs", prop.substring(0, prop.length() - 1));

    return p;
  }

  /**
   * Indicates that the "Initalize CUI" phase of subsetting has started.
   */
  public void beginInitializeCuiList() {
    //
    // Initialize and open a log file
    //
    if (pw == null) {
      try {
        pw = new PrintWriter(new FileWriter(new File(user_configuration
            .getSubsetDirectory(), "sample_meta.log")));
        pw.println("Conditions met:");
      } catch (IOException e) {
        MMSToolkit
            .reportError("Problem creating log file, using standard out instead.");
        pw = new PrintWriter(System.out);
      }
      pw.flush();
    }

    //
    // Indicate that we are in init cui list phase
    //
    init_cui = true;
  }

  /**
   * Indicates that the "Initalize CUI" phase of subsetting has ended.
   */
  public void endInitializeCuiList() {

    //
    // Indicate that init cui list phase is over.
    //
    init_cui = false;

    //
    // Flush the log
    //
    pw.flush();
  }

  /**
   * Indicates that the "Subset content" phase of subsetting has ended.
   */
  public void endSubsetContent() {

    //
    // Indicate that we are no longer in subset content phase
    //
    subset_content = false;

    //
    // Write conditions not met to log
    //
    pw.println("");
    pw.println("");
    pw.println("Conditions not met:");
    for (int i = 0; i < conditions.length; i++) {
      if (!conditions_applied.contains(conditions[i]))
        pw.println(conditions[i].getClass().getName());
    }
    pw.flush();
    pw.close();
  }

  /**
   * Applies the filter to the specified {@link Concept}.
   * 
   * @param concept
   *          the {@link Concept}
   */
  public void applyFilter(Concept concept) {

    //
    // Check if we have seen this concept
    //
    boolean found = cui_list.inList(concept.getCui());

    //
    // If in init cui phase, check which conditions are satisfied
    // by this concept. Add it and relevant CUI2s as well
    //
    if (init_cui) {

      //
      // Iterate through all conditions
      //
      for (int i = 0; i < conditions.length; i++) {
        final Condition condition = conditions[i];

        //
        // Is condition met?
        //
        if (condition_name_map.get(condition.getName()).equals("true")
            && !conditions_applied.contains(condition)
            && condition.isMet(concept)) {

          //
          // Has this CUI already been added as a CUI2?
          // If so, remove it from that list because it now
          // exists by virtue of satisfying a condition
          //
          if (cui2s_added_by_user_only.contains(concept.getCui()))
            cui2s_added_by_user_only.remove(concept.getCui());

          //
          // Add this CUI to the cui list
          //
          cui_list.addCui(concept.getCui());

          //
          // Indicate that a matching condition was found
          //
          found = true;
        }
      }

      //
      // If this concept satisfies no conditions, clear it
      //
      if (!found)
        concept.clear();

    } else {

      //
      // If in subset content mode, make sure that we are properly
      // eliminating relationships to concepts that we will not be keeping
      //
      resolveReferences(concept);
    }
  }

  /**
   * Removes relationships if one of two conditions is met.
   * <ol>
   * <li>If the current concept was kept only as a CUI2 and the relationship's
   * CUI2 was only kept as a CUI2 then we do not want to keep the relationship
   * here because it will create an imbalance in the output. </li>
   * <li>If the current concept was kept only as a CUI2 and if the CUI2 of this
   * relationship is lower than the current CUI and the current relationship is
   * an "atom" relationship we do not want to keep it. This is because it has
   * already been removed in the reverse direction, so we do not want to create
   * an imbalance. </li>
   * </ol>
   * Without this logic, relationships would be kept in one direction only due
   * to the sequence of processing during the subset content phase.
   * 
   * @param concept
   *          {@link Concept} currently being filtered
   */
  private void resolveReferences(Concept concept) {

    final Relationship[] rels = concept.getRelationships();
    for (int i = 0; i < rels.length; i++) {
      if (cui2s_added_by_user_only.contains(concept.getCui())
          && cui2s_added_by_user_only.contains(rels[i].getRelatedConcept()
              .getCui())) {
        concept.removeRelationship(rels[i]);
      }
      if (rels[i] instanceof AtomRelationship
          && cui2s_added_by_user_only.contains(concept.getCui())
          && (rels[i].getRelatedConcept().getCui().compareTo(concept.getCui()) < 0)
          &&
          // identify auis being in list vs not being in list
          cui_list.addedByUserOnly(concept.getCui()))
        concept.removeRelationship(rels[i]);
    }
  }

  /**
   * Sets the {@link CuiList} reference.
   * 
   * @param cui_list
   *          {@link CuiList}
   */
  public void setCuiList(CuiList cui_list) {
    this.cui_list = cui_list;
  }

  /**
   * Returns the name.
   * 
   * @return the name
   */
  public String getName() {
    return "SAMPLEMETA (for Testing)";
  }

  /**
   * Returns the configuration details, as they should appear in the subset log
   * file. Multi-line comments should be separated by {@link System}.getProperty("line.separator")
   * to ensure platform independence.
   * 
   * @return the configuration details
   */
  public String getConfigurationDetails() {
    return "The SAMPLEMETA subset was run.";
  }

  /**
   * Indicates whether or not this filter can use the efficient mode during the
   * "Initialize Cuis" phase of the subsetting process. The efficient mode reads
   * only <code>MRCONSO.RRF</code> and <code>MRSTY.RRF</code> during the
   * "Initialize Cuis" phase.
   * 
   * @return <code>false</code>
   */
  public boolean isEfficientEnabled() {
    return false;
  }

  //
  // Interface
  //

  /**
   * Generically represents a condition that does or does not apply to a
   * {@link Concept}.
   */
  private interface Condition {
    /**
     * Indicates whether or not this condition applies to the specified
     * {@link Concept}.
     * 
     * @param concept
     *          the {@link Concept}
     * @return <code>true</code> if the condition is met; <code>false</code>
     *         otherwise
     */
    public boolean isMet(Concept concept);

    /**
     * Returns the name.
     * 
     * @return the name
     */
    public String getName();
  }

  //
  // ATOM RULES
  //

  /**
   * Keeps at least one atom with a non-ASCII UTF-8 character.
   */
  private class KeepUnicodeCharacter implements Condition {
    public String getName() {
      return "Atom with UTF-8 character";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      try {
        for (int i = 0; i < atoms.length; i++) {
          final Atom atom = atoms[i];
          if (MMSToolkit.containsUnicode(atom.getStr())) {
            conditions_applied.add(this);
            pw.println(concept.getCui() + " " + atom.getAui() + " "
                + this.getClass().getName());
            pw.flush();
            return true;
          }
        }
      } catch (UnsupportedEncodingException e) {
        MMSToolkit.reportError(e.getMessage());
      }
      return false;
    }
  }

  /**
   * Keeps at least two atoms with ambiguous SUI values (eg same SUI different
   * CUI) and keeps at least two atoms with ambiguous LUI values (eg same LUI
   * different CUI).
   */
  private class KeepAmbiguousSUIs implements Condition {
    public String getName() {
      return "Atom with ambiguous SUI";
    }

    public boolean isMet(Concept concept) {
      //
      // check if this concept has the same sui as a
      // previously processed concept
      //
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (saved_ams.keySet().contains(atom.getSui())
            && !saved_ams.get(((String) atom.getSui()))
                .equals(concept.getCui())) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + atom.getSui() + " "
              + saved_ams.get(((String) atom.getSui())) + " "
              + this.getClass().getName());
          return true;
        }
      }
      //
      // if an atom is present in the input AMBIG.SUI file
      // save a reference to it for future matching
      //
      List list = new ArrayList();
      try {
        list = user_configuration.getInputStream()
            .getAmbiguousStringIdentifiersAsList();
      } catch (Exception e) {
        e.printStackTrace();
      }
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (list.contains(atom.getSui())) {
          saved_ams.put(atom.getSui(), concept.getCui());
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two ENG atoms in the same CUI, LAT, SAB with different LUIs.
   */
  private class KeepENGDiffLUIs implements Condition {
    public String getName() {
      return "Atoms with different LUIs, same CUI, LAT, SAB";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom2.getLat().getAbbreviation().equals("ENG")
                && atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && !atom.getLui().equals(atom2.getLui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());

              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two ENG atoms in the same CUI, LAT, LUI, SAB with different
   * SUIs.
   */
  private class KeepENGDiffSUIs implements Condition {
    public String getName() {
      return "Atoms with different SUIs, same CUI, LAT, LUI, SAB";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom2.getLat().getAbbreviation().equals("ENG")
                && atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getLui().equals(atom2.getLui())
                && !atom.getSui().equals(atom2.getSui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());

              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two ENG atoms in the same CUI, LAT, SUI, SAB with different
   * AUIs.
   */
  private class KeepENGDiffAUIs implements Condition {
    public String getName() {
      return "Atoms with different AUIs, same CUI, LAT, SUI, SAB";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom2.getLat().getAbbreviation().equals("ENG")
                && atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getSui().equals(atom2.getSui())
                && !atom.getAui().equals(atom2.getAui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two ENG atoms in the same CUI, LAT with different SAB, LUIs.
   */
  private class KeepENGDiffSABsLUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different SAB, LUI (English only), same CUI, LAT";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom2.getLat().getAbbreviation().equals("ENG")
                && !atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && !atom.getLui().equals(atom2.getLui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two ENG atoms in the same CUI, LAT, LUI with different SAB,
   * SUIs.
   */
  private class KeepENGDiffSABsSUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different SAB, SUI (English only), same CUI, LAT, LUI";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom2.getLat().getAbbreviation().equals("ENG")
                && !atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getLui().equals(atom2.getLui())
                && !atom.getSui().equals(atom2.getSui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two ENG atoms in the same CUI, LAT, SUI with different SABs,
   * AUIs.
   */
  private class KeepENGDiffSABsAUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different SAB, AUI (English only), same CUI, LAT, SUI";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom2.getLat().getAbbreviation().equals("ENG")
                && !atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getSui().equals(atom2.getSui())
                && !atom.getAui().equals(atom2.getAui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two not ENG atoms in the same CUI, LAT, SAB with different
   * LUIs.
   */
  private class KeepNotENGDiffLUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different LUI (non-English), same CUI, LAT, SAB";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (!atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom.getLat().getAbbreviation().equals(
                atom2.getLat().getAbbreviation())
                && atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && !atom.getLui().equals(atom2.getLui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two not ENG atoms in the same CUI, LAT, LUI, SAB with
   * different SUIs.
   */
  private class KeepNotENGDiffSUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different SUI (non-English), same CUI, LAT, LUI, SAB";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (!atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom.getLat().getAbbreviation().equals(
                atom2.getLat().getAbbreviation())
                && atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getLui().equals(atom2.getLui())
                && !atom.getSui().equals(atom2.getSui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two not ENG atoms in the same CUI, LAT, SUI, SAB with
   * different AUIs.
   */
  private class KeepNotENGDiffAUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different AUI (non-English), same CUI, LAT, SUI, SAB";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (!atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom.getLat().getAbbreviation().equals(
                atom2.getLat().getAbbreviation())
                && atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getSui().equals(atom2.getSui())
                && !atom.getAui().equals(atom2.getAui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two not ENG atoms in the same CUI, LAT with different SAB,
   * LUIs.
   */
  private class KeepNotENGDiffSABsLUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different SAB, LUI (non-English), same CUI, LAT";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (!atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom.getLat().getAbbreviation().equals(
                atom2.getLat().getAbbreviation())
                && !atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && !atom.getLui().equals(atom2.getLui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two not ENG atoms in the same CUI, LAT, LUI with different
   * SAB, SUIs.
   */
  private class KeepNotENGDiffSABsSUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different SAB, SUI (non-English), same LAT, LUI";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (!atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom.getLat().getAbbreviation().equals(
                atom2.getLat().getAbbreviation())
                && !atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getLui().equals(atom2.getLui())
                && !atom.getSui().equals(atom2.getSui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least two not ENG atoms in the same CUI, LAT, SUI with different
   * SABs, AUIs.
   */
  private class KeepNotENGDiffSABsAUIs implements Condition {
    public String getName() {
      return "Atoms in the same concept with different SAB, AUI (English only), same LAT, SUI";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (!atom.getLat().getAbbreviation().equals("ENG")) {
          for (int j = 0; j < atoms.length; j++) {
            final Atom atom2 = atoms[j];
            if (atom.getLat().getAbbreviation().equals(
                atom2.getLat().getAbbreviation())
                && !atom.getSource().getVsab().equals(
                    atom2.getSource().getVsab())
                && atom.getSui().equals(atom2.getSui())
                && !atom.getAui().equals(atom2.getAui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + atom.getAui() + " "
                  + this.getClass().getName());
              pw.flush();
              return true;
            }
          }
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least one atom from each language.
   */
  private class KeepAtomFromEachLanguage implements Condition {
    public String getName() {
      return "Atom from each language";
    }

    private Set lats_found = new HashSet();

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (!lats_found.contains(atom.getLanguage())) {
          lats_found.add(atom.getLanguage());
          if (lats_found.size() == lats.length) {
            conditions_applied.add(this);
          }
          pw.println(concept.getCui() + " " + atom.getAui() + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least one atom from each of 6 SABs.
   */
  private class KeepAtomFromVariousSABs implements Condition {
    public String getName() {
      return "Atom from each required SAB (see options for list)";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        // look for concepts with the required sabs
        // remove sab from required list when found
        if (sabs_required.contains(atom.getSource().getRsab())) {
          sabs_required.remove(atom.getSource().getRsab());
          sabs_found.add(atom.getSource().getRsab());
          pw.println(concept.getCui() + " " + atom.getAui() + " "
              + this.getClass().getName());
          return true;
        }
        // only consider adding concepts with other sabs after the
        // required ones have been found
        else if (sabs_required.size() == 0
            && !sabs_found.contains(atom.getSource().getRsab())) {
          sabs_found.add(atom.getSource().getRsab());
          if (sabs_found.size() == 10)
            conditions_applied.add(this);
          pw.println(concept.getCui() + " " + atom.getAui() + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps all SRC concepts.
   */
  private class KeepAllSRC implements Condition {
    public String getName() {
      return "All SRC atoms";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getSource().getRsab().equals("SRC")) {
          pw.println(concept.getCui() + " " + atom.getAui() + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps all concepts with MSH|TQ atoms.
   */
  private class KeepAllMSHTQ implements Condition {
    public String getName() {
      return "All MSH/TQ atoms";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getSource().getRsab().equals("MSH")
            && atom.getTermType().getAbbreviation().equals("TQ")) {
          pw.println(concept.getCui() + " " + atom.getAui() + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps the "SNOMED to ICD9CM" mapping concept and keeps at least one mapping
   * from the "SNOMEDCT to ICD9CM" mapping concept.
   */
  private class KeepSNOMEDCTtoICD9CMMappingConcept implements Condition {
    public String getName() {
      return "Atoms from SNOMED-ICD9CM map set concept";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getStr().equals("SNOMEDCT mappings to ICD-9-CM")) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + atom.getAui() + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps the "MSH Associated Expressions" mapping concept and keeps at least
   * one mapping from the "MSH Associated Expressions" mapping concept.
   */
  private class KeepMSHATXMappingConcept implements Condition {
    public String getName() {
      return "Atoms from MSH ATX map set concept";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        if (atom.getStr().equals("MSH Associated Expressions")) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + atom.getAui() + " "
              + this.getClass().getName());
          keepFromExpression(concept);
          return true;
        }
      }
      return false;
    }

    private void keepFromExpression(Concept concept) {
      final Mapping[] mappings = concept.getMappings();
      for (int i = 0; i < mappings.length; i++) {
        String cui2 = "";
        if (mappings[i].getFrom() != null
            && mappings[i].getFrom().getExpression() != null) {
          cui2 = mappings[i].getFrom().getExpression();
          if (!cui_list.inList(cui2))
            cui2s_added_by_user_only.add(cui2);
          cui_list.addCui(cui2);
          pw.println(concept.getCui() + " " + cui2 + " "
              + this.getClass().getName());
        }
      }
    }

  }

  //
  // DEFINITION RULES
  //

  /**
   * Keeps at least one definition with length>4000 characters.
   */
  private class KeepLongDEF implements Condition {
    public String getName() {
      return "Definition with > 4000 characters";
    }

    public boolean isMet(Concept concept) {
      final Definition[] defs = concept.getDefinitions();
      for (int i = 0; i < defs.length; i++) {
        if (defs[i].getDef().length() > 4000) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + defs[i].getAtom().getAui() + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least one definition with a non-ASCII UTF-8 character.
   */
  private class KeepDEFWithUnicodeCharacter implements Condition {
    public String getName() {
      return "Definition with UTF-8 characer";
    }

    public boolean isMet(Concept concept) {
      final Definition[] defs = concept.getDefinitions();
      try {
        for (int i = 0; i < defs.length; i++) {
          if (MMSToolkit.containsUnicode(defs[i].getDef())) {
            conditions_applied.add(this);
            pw.println(concept.getCui() + " " + defs[i].getAtom().getAui()
                + " " + this.getClass().getName());
            return true;
          }
        }
      } catch (UnsupportedEncodingException e) {
        MMSToolkit.reportError(e.getMessage());
      }
      return false;
    }
  }

  //
  // ATTRIBUTE RULES
  //
  /**
   * Keeps at least one {@link AtomAttribute} (e.g. METAUI is an AUI).
   */
  private class KeepAtomAttribute implements Condition {
    public String getName() {
      return "Attribute attached to an atom";
    }

    public boolean isMet(Concept concept) {
      final Attribute[] atts = concept.getAttributes();
      for (int i = 0; i < atts.length; i++) {
        if (atts[i] instanceof AtomAttribute) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " "
              + ((AtomAttribute) atts[i]).getAtom().getAui() + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least one {@link RelationshipAttribute} (e.g. METAUI is a RUI).
   */
  private class KeepRelationshipAttribute implements Condition {
    public String getName() {
      return "Attribute attached to an relationship";
    }

    public boolean isMet(Concept concept) {
      final Attribute[] atts = concept.getAttributes();
      for (int i = 0; i < atts.length; i++) {
        if (atts[i] instanceof RelationshipAttribute) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " "
              + ((RelationshipAttribute) atts[i]).getRelationship().getRui()
              + " " + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least one concept level attribute (e.g. METAUI is empty).
   */
  private class KeepConceptLevelAttribute implements Condition {
    public String getName() {
      return "Attribute attached to an concept";
    }

    public boolean isMet(Concept concept) {
      final Attribute[] atts = concept.getAttributes();
      for (int i = 0; i < atts.length; i++) {
        final Attribute att = atts[i];
        if (!(att instanceof RelationshipAttribute)
            && !(att instanceof AtomAttribute)) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + "        " + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  /**
   * Keeps at least one attribute connected to a non-ASCII UTF-8 atom
   */
  private class KeepAttributeConnectedToUnicode implements Condition {
    public String getName() {
      return "Attribute with UTF-8 attribute value";
    }

    public boolean isMet(Concept concept) {
      final Attribute[] atts = concept.getAttributes();
      try {
        for (int i = 0; i < atts.length; i++) {
          if (atts[i] instanceof AtomAttribute) {
            Atom atom = ((AtomAttribute) atts[i]).getAtom();
            if (MMSToolkit.containsUnicode(atom.getStr())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + "        " + " "
                  + this.getClass().getName());
              return true;
            }
          }
        }
      } catch (UnsupportedEncodingException e) {
        MMSToolkit.reportError(e.getMessage());
      }

      return false;
    }
  }

  /**
   * Keeps at least one attribute value with length>4000 characters.
   */
  private class KeepLongAttributeValue implements Condition {
    public String getName() {
      return "Attribute with attribute value having > 4000 characters";
    }

    public boolean isMet(Concept concept) {
      final Attribute[] atts = concept.getAttributes();
      for (int i = 0; i < atts.length; i++) {
        if (atts[i].getAtv().length() > 4000) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + "        " + " "
              + this.getClass().getName());
          return true;
        }
      }
      return false;
    }
  }

  //
  // COOCURRENCE RULES
  //

  /**
   * Keeps at least one coocurrence with each coocurrence type and keeps all
   * coocurrences from CUIs kept and keeps all CUI2s from coocurrences kept.
   */
  private class KeepCoocurrenceTypes implements Condition {
    public String getName() {
      return "Coocurrences from each COT";
    }

    private Set coc_types_found = new HashSet();

    public boolean isMet(Concept concept) {
      final Coocurrence[] cocs = concept.getCoocurrences();
      for (int i = 0; i < cocs.length; i++) {
        final Coocurrence coc = cocs[i];
        if (!coc_types_found.contains(coc.getCoocurrenceType())) {
          coc_types_found.add(coc.getCoocurrenceType());
          if (coc_types_found.size() == coc_types.length) {
            conditions_applied.add(this);
          }
          pw.println(concept.getCui() + " " + this.getClass().getName());
          keepCUI2s(concept);
          return true;
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Coocurrence[] cocs = concept.getCoocurrences();
      for (int i = 0; i < cocs.length; i++) {
        final String cui2 = cocs[i].getRelatedConcept().getCui();
        if (!cui_list.inList(cui2))
          cui2s_added_by_user_only.add(cui2);
        cui_list.addCui(cui2);
        pw.println(concept.getCui() + " " + cui2 + " "
            + this.getClass().getName());
      }
    }
  }

  //
  // RELATIONSHIP RULES
  //

  /**
   * Keeps at least one AtomRelationship (e.g. AUI1, AUI2 not empty).
   */
  private class KeepAtomRelationship implements Condition {
    public String getName() {
      return "Relationship between two atoms";
    }

    public boolean isMet(Concept concept) {
      final Relationship[] rels = concept.getRelationships();
      for (int i = 0; i < rels.length; i++) {
        if (rels[i] instanceof AtomRelationship) {
          final AtomRelationship atom_rel = (AtomRelationship) rels[i];
          if (atom_rel.getAtom() != null && atom_rel.getRelatedAtom() != null) {
            conditions_applied.add(this);
            pw.println(concept.getCui() + " " + atom_rel.getAtom().getAui()
                + " " + this.getClass().getName());
            keepCUI2s(concept);
            return true;
          }
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Relationship[] rels = concept.getRelationships();
      for (int i = 0; i < rels.length; i++) {
        final String cui2 = rels[i].getRelatedConcept().getCui();
        if (!cui_list.inList(cui2))
          cui2s_added_by_user_only.add(cui2);
        cui_list.addCui(cui2);
        pw.println(concept.getCui() + " " + cui2 + " "
            + this.getClass().getName());
      }
    }
  }

  /**
   * Keeps at least one concept level relationship (e.g. AUI1, AUI2 are empty).
   */
  private class KeepConceptLevelRelationship implements Condition {
    public String getName() {
      return "Relationship between two concepts";
    }

    public boolean isMet(Concept concept) {
      final Relationship[] rels = concept.getRelationships();
      for (int i = 0; i < rels.length; i++) {
        if (!(rels[i] instanceof AtomRelationship)) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + "        " + " "
              + this.getClass().getName());
          keepCUI2s(concept);
          return true;
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Relationship[] rels = concept.getRelationships();
      for (int i = 0; i < rels.length; i++) {
        final String cui2 = rels[i].getRelatedConcept().getCui();
        if (!cui_list.inList(cui2))
          cui2s_added_by_user_only.add(cui2);
        cui_list.addCui(cui2);
        pw.println(concept.getCui() + " " + cui2 + " "
            + this.getClass().getName());
      }
    }

  }

  /**
   * Keeps at least one relationship with RUI attributes.
   */
  private class KeepRelationshipWithRUIAttributes implements Condition {
    public String getName() {
      return "Relationship with attributes";
    }

    public boolean isMet(Concept concept) {
      final Relationship[] rels = concept.getRelationships();
      final Attribute[] atts = concept.getAttributes();
      for (int i = 0; i < rels.length; i++) {
        for (int j = 0; j < atts.length; j++) {

          if (atts[j] instanceof RelationshipAttribute) {
            if (rels[i].getRui().equals(
                ((RelationshipAttribute) atts[j]).getRelationship().getRui())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " " + rels[i].getRui() + " "
                  + this.getClass().getName());
              keepCUI2s(concept);
              return true;
            }
          }
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Relationship[] rels = concept.getRelationships();
      for (int i = 0; i < rels.length; i++) {
        final String cui2 = rels[i].getRelatedConcept().getCui();
        if (!cui_list.inList(cui2))
          cui2s_added_by_user_only.add(cui2);
        cui_list.addCui(cui2);
        pw.println(concept.getCui() + " " + cui2 + " "
            + this.getClass().getName());
      }
    }

  }

  //
  // CONTEXT RULES
  //

  /**
   * Keeps at least one context where an ancestor has a non-ASCII UTF-8 string.
   */
  private class KeepContextWithUnicodeAncestor implements Condition {
    public String getName() {
      return "Context with UTF-8 ancestor atom";
    }

    public boolean isMet(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final ContextMember cxt_mem = cxt_mems[j];
          try {
            if (cxt_mem.getCxl().equals("ANC")
                && MMSToolkit
                    .containsUnicode(cxt_mem.getRelatedAtom().getStr())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " "
                  + cxt_mem.getRelatedConcept().getCui() + " "
                  + this.getClass().getName());
              keepCUI2s(concept);
              return true;
            }
          } catch (UnsupportedEncodingException e) {
            MMSToolkit.reportError(e.getMessage());
          }
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final String cui2 = cxt_mems[j].getRelatedConcept().getCui();
          if (!cui_list.inList(cui2))
            cui2s_added_by_user_only.add(cui2);
          cui_list.addCui(cui2);
          pw.println(concept.getCui() + " " + cui2 + " "
              + this.getClass().getName());
        }
      }
    }
  }

  /**
   * Keeps at least one context where a sibling has a non-ASCII UTF-8 string.
   */
  private class KeepContextWithUnicodeSibling implements Condition {
    public String getName() {
      return "Context with UTF-8 atom sibling";
    }

    public boolean isMet(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final ContextMember cxt_mem = cxt_mems[j];
          try {
            if (cxt_mem.getCxl().equals("SIB")
                && MMSToolkit
                    .containsUnicode(cxt_mem.getRelatedAtom().getStr())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " "
                  + cxt_mem.getRelatedConcept().getCui() + " "
                  + this.getClass().getName());
              keepCUI2s(concept);
              return true;
            }
          } catch (UnsupportedEncodingException e) {
            MMSToolkit.reportError(e.getMessage());
          }
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final String cui2 = cxt_mems[j].getRelatedConcept().getCui();
          if (!cui_list.inList(cui2))
            cui2s_added_by_user_only.add(cui2);
          cui_list.addCui(cui2);
          pw.println(concept.getCui() + " " + cui2 + " "
              + this.getClass().getName());
        }
      }
    }
  }

  /**
   * Keeps at least one context where a child has a non-ASCII UTF-8 string.
   */
  private class KeepContextWithUnicodeChild implements Condition {
    public String getName() {
      return "Context with UTF-8 child atom";
    }

    public boolean isMet(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final ContextMember cxt_mem = cxt_mems[j];
          try {
            if (cxt_mem.getCxl().equals("CHD")
                && MMSToolkit
                    .containsUnicode(cxt_mem.getRelatedAtom().getStr())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " "
                  + cxt_mem.getRelatedConcept().getCui() + " "
                  + this.getClass().getName());
              keepCUI2s(concept);
              return true;
            }
          } catch (UnsupportedEncodingException e) {
            MMSToolkit.reportError(e.getMessage());
          }
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final String cui2 = cxt_mems[j].getRelatedConcept().getCui();
          if (!cui_list.inList(cui2))
            cui2s_added_by_user_only.add(cui2);
          cui_list.addCui(cui2);
          pw.println(concept.getCui() + " " + cui2 + " "
              + this.getClass().getName());
        }
      }
    }

  }

  /**
   * Keeps at least one context where a self row has a non-ASCII UTF-8 string.
   */
  private class KeepContextWithUnicodeSelf implements Condition {
    public String getName() {
      return "Context with UTF-8 self atom";
    }

    public boolean isMet(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final ContextMember cxt_mem = cxt_mems[j];
          try {
            if (cxt_mem.getCxl().equals("CCP")
                && MMSToolkit
                    .containsUnicode(cxt_mem.getRelatedAtom().getStr())) {
              conditions_applied.add(this);
              pw.println(concept.getCui() + " "
                  + cxt_mem.getRelatedConcept().getCui() + " "
                  + this.getClass().getName());
              keepCUI2s(concept);
              return true;
            }
          } catch (UnsupportedEncodingException e) {
            MMSToolkit.reportError(e.getMessage());
          }
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final String cui2 = cxt_mems[j].getRelatedConcept().getCui();
          if (!cui_list.inList(cui2))
            cui2s_added_by_user_only.add(cui2);
          cui_list.addCui(cui2);
          pw.println(concept.getCui() + " " + cui2 + " "
              + this.getClass().getName());
        }
      }
    }

  }

  /**
   * Keeps at least two full contexts connected to the same atom.
   */
  private class KeepContextsWithSameSelfAtom implements Condition {
    public String getName() {
      return "Case of atom with more than one context";
    }

    public boolean isMet(Concept concept) {
      final Atom[] atoms = concept.getAtoms();
      for (int i = 0; i < atoms.length; i++) {
        if (atoms[i].getContextsAsList().size() > 1) {
          conditions_applied.add(this);
          pw.println(concept.getCui() + " " + atoms[i].getAui() + " "
              + this.getClass().getName());
          keepCUI2s(concept);
          return true;
        }
      }
      return false;
    }

    private void keepCUI2s(Concept concept) {
      final Context[] cxts = concept.getContexts();
      for (int i = 0; i < cxts.length; i++) {
        final ContextMember[] cxt_mems = cxts[i].getContextMembers();
        for (int j = 0; j < cxt_mems.length; j++) {
          final String cui2 = cxt_mems[j].getRelatedConcept().getCui();
          if (!cui_list.inList(cui2))
            cui2s_added_by_user_only.add(cui2);
          cui_list.addCui(cui2);
          pw.println(concept.getCui() + " " + cui2 + " "
              + this.getClass().getName());
        }
      }
    }

  }

  /**
   * Validates the filter. In this case, there are only warnings if the user
   * selects sources for removal, or opts to remove UTF-8 data.
   * 
   * @return <code>true</code> filter is valid; <code>false</code> otherwise
   */
  public boolean validate() {
    warning_msgs.clear();
    if (sources_to_remove.size() != 0)
      warning_msgs.add("WARNING: You are encouraged not to exclude any \n"
          + "sources when running the SAMPLEMETA filter.\n"
          + "Please go to the Source List tab and ensure\n"
          + "that no sources are selected to be removed.");
    if (user_configuration.getOutputStream().removeUnicode())
      warning_msgs.add("WARNING: You are encouranged not to remove records\n"
          + "containing extended UTF-8 characters when \n"
          + "running the SAMPLEMETA filter.\n"
          + "Please go to the Output Handler tab and uncheck\n"
          + "the Remove records with extended UTF-8 characters box.");
    return true;
  }

  /**
   * Returns the messages indicating why validation failed. Should only be
   * called if {@link #validate()} returns <code>false</code>.
   * 
   * @return the failure messages
   */
  public String[] getValidationErrors() {
    return new String[] { getValidateMessage() };
  }

  /**
   * Returns warnings that were indicated during validation.
   * 
   * @return the warning messages
   */
  public String[] getValidationWarnings() {
    if (warning_msgs.size() == 0)
      return null;
    return (String[]) warning_msgs.toArray(new String[0]);
  }

  /**
   * Returns the {@link GUIConfigurable} whose tab should next be displayed to
   * allow the user to correct an error or warning.
   * 
   * @return GUIConfigurable to be displayed
   */
  public GUIConfigurable getGUIConfigurableForValidationFailure() {
    if (sources_to_remove.size() != 0)
      return new SourcesToRemoveFilter();
    if (user_configuration.getOutputStream().removeUnicode())
      return user_configuration.getOutputStream();
    return null;
  }

  //
  // Comparator used to sort the Conditions.
  //
  public class ConditionsComparator implements Comparator {

    /**
     * Instantiates a comparator.
     */
    public ConditionsComparator() {
    }

    //
    // Implementation of Comparator interface
    //

    /**
     * Compares the specified field of each {@link String} object.
     * 
     * @param o1
     *          the first {@link Object} to be compared.
     * @param o2
     *          the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a
     *         positive integer as the first argument is less than, equal to, or
     *         greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null || !(o1 instanceof Condition)
          || !(o2 instanceof Condition))
        return 0;
      Condition s1 = (Condition) o1;
      Condition s2 = (Condition) o2;

      String source1 = s1.getName();
      String source2 = s2.getName();

      return source1.compareTo(source2);
    }
  }

  public class SampleMetaFilterOptionsGUI extends JDialog {

    //
    // Private Fields
    //
    private SuperJList sabs_required_list = null;

    //
    // Constructor
    //
    /**
     * Default constructor.
     */
    public SampleMetaFilterOptionsGUI() {
      super();
      setModal(true);
      initialize();
    };

    //
    // Methods
    //
    /**
     * This method sets up the GUI
     */
    private void initialize() {

      DataChangeListener dcl = new DataChangeListener();

      getContentPane().setLayout(new BorderLayout());

      FlowLayout done_layout = new FlowLayout();
      done_layout.setVgap(10);
      JPanel done_panel = new JPanel(done_layout);
      JPanel complete_panel = new JPanel(new BorderLayout());

      // init source list
      JPanel source_panel = new JPanel(new BorderLayout(5, 5));
      JLabel list_label = new JLabel("Select desired source abbreviations.");
      source_panel.add(list_label, BorderLayout.NORTH);
      sabs_required_list = new SuperJList(Source.getSources());

      String[] selected_value_array = (String[]) sabs_required
          .toArray(new String[0]);
      Source[] selected_sources = new Source[selected_value_array.length];
      for (int i = 0; i < selected_sources.length; i++) {
        selected_sources[i] = Source.getSource(selected_value_array[i]);
      }
      sabs_required_list.setSelectedValues(selected_sources, true);
      JScrollPane sp = new JScrollPane((Component)sabs_required_list,
          ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
          ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
      source_panel.add(sp, BorderLayout.CENTER);

      // configure done_panel
      JButton jdone = new JButton("Done");
      jdone.addActionListener(dcl);
      done_panel.add(jdone);

      // Add the sub-panels to complete_panel
      complete_panel.add(source_panel, BorderLayout.CENTER);
      complete_panel.add(done_panel, BorderLayout.SOUTH);
      EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
      complete_panel.setBorder(new CompoundBorder(etched_border,
          new EmptyBorder(10, 10, 10, 10)));

      getRootPane().setDefaultButton(jdone);
      getContentPane().add(complete_panel, BorderLayout.CENTER);
      this.pack();
      this.setLocationRelativeTo(null);
    };

    //
    // Inner classes
    //
    public class DataChangeListener implements java.awt.event.ActionListener {

      // constructor
      public DataChangeListener() {
      };

      // implementation of ActionListener interface
      public void actionPerformed(ActionEvent e) {
        if (e.getActionCommand().equals("Done")) {
          if (!ignore_events) {
            fireDataChanged();

            Object[] ret_values = sabs_required_list.getSelectedStringValues();
            sabs_required.clear();
            for (int i = 0; i < ret_values.length; i++)
              sabs_required.add(((Source) ret_values[i]).getRsab());
            SampleMetaFilterOptionsGUI.this.dispose();
          }
        }
      }
    }

  } // End of SampleMetaFilterOptionsGUI class

  //
  // SourcesToRemoveListener Implementation
  //

  /**
   * Updates list of sources to remove.
   * 
   * @param e
   *          the {@link SourcesToRemoveEvent}
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent e) {
    sources_to_remove = e.getSourcesToRemove();
  }

}
