/**************************************************************************
 * Package:	gov.nih.nlm.mms
 * Object:	Configuration.java
 *
 **************************************************************************/

package gov.nih.nlm.mms;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.HierarchicalSemanticType;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.STYMetaDescription;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.umls.meta.Suppressible;
import gov.nih.nlm.umls.meta.TermType;
import gov.nih.nlm.umls.meta.impl.MetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.STYMetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.SourceMetaDescriptionImpl;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

/**
 * Represents all application-level configuration data. This is the set of
 * static configuration data that does not change, such as the list of sources
 * present in the current release, as well as the set of default configurations
 * for the user data.
 * 
 * @author Deborah Shapiro, Brian Carlsen
 */

public class ApplicationConfiguration extends UserConfiguration {

  //
  // Private Fields
  //

  private List sources, termgroups, languages, semantic_types;

  private Set suppressed_terms;

  private Map ssab_tty_map, sab_tty_map, lang_map;

  private String home_dir = ".", config_dir = "config";

  //
  // Constructor
  //

  /**
   * Constructs an {@link ApplicationConfiguration} from the specified
   * properties file. Typically, this object is built from one of the
   * <code>METAMSYS/config/mmsys.{a,b,c}.prop</code> files.
   * 
   * @param config_file
   *          a java properties {@link File}
   * @throws LocalException
   *           if there is a problem loading the properties file
   */
  public ApplicationConfiguration(File config_file) throws LocalException {
    // Superclass constructor already calls load
    super(config_file);
  }

  /**
   * Constructs an {@link ApplicationConfiguration} from the specified
   * {@link Properties}.
   * 
   * @param p
   *          {@link Properties}
   * @throws LocalException
   *           if there is a problem loading the properties file
   */
  public ApplicationConfiguration(Properties p) throws LocalException {
    // Superclass constructor already calls load
    super(p);
  }

  //
  // Accessor Methods
  //

  /**
   * Returns the list of {@link HierarchicalSemanticType}s that are children of
   * a specified tree number.
   * 
   * @param stn
   *          a semantic type tree number
   * @return the {@link List} of {@link HierarchicalSemanticType}s that are
   *         children of
   */
  public List getSemanticTypeChildren(String stn) {
    List list = new ArrayList();
    for (int i = 0; i < semantic_types.size(); i++) {
      if (((HierarchicalSemanticType) semantic_types.get(i)).getStn()
          .startsWith(stn))
        list.add(semantic_types.get(i));
    }
    return list;
  }

  /**
   * Returns the set of {@link HierarchicalSemanticType}s that are children of
   * a specified tree number.
   * 
   * @param stn
   *          a semantic type tree number
   * @return the {@link Set} of {@link HierarchicalSemanticType}s that are
   *         children of
   */
  public Set getSemanticTypeChildrenAsSet(String stn) {
    return new TreeSet(getSemanticTypeChildren(stn));
  }

  /**
   * Returns a {@link List} of all {@link SourceTermType} objects.
   * 
   * @return a {@link List} of all {@link SourceTermType} objects.
   */
  // need this
  public List getSourceTermTypes() {
    return new ArrayList(termgroups);
  }

  /**
   * Returns the {@link SourceTermType} for the specified source abbreviation
   * and term type. If the source abbreviation and term type combination is not
   * valid, <code>null</code> is returned.
   * 
   * @param sab
   *          a source abbreviation, can be root or versioned
   * @param tty
   *          a term type
   * @return the {@link SourceTermType} for the specified source abbreviation
   *         and term type
   */
  public SourceTermType getSourceTermType(String sab, String tty) {
    String sab_tty = sab + tty;
    SourceTermType tg = (SourceTermType) sab_tty_map.get(sab_tty);
    if (tg != null)
      return tg;
    tg = (SourceTermType) ssab_tty_map.get(sab_tty);
    if (tg != null)
      return tg;
    return null;
  }

  /**
   * Returns a {@link List} of suppressible cui,sui values. Each element of the
   * list is a {@link String} with the format "cui|sui".
   * 
   * @param a
   *          {@link Atom}
   * @return a {@link List} of suppressible cui,sui values.
   */
  public boolean isSuppressedTerm(Atom a) {

    if (a.getConcept() != null) {
      // if RRF, don't check the list, already will be marked "E"
      if (a.getAtomIdentifier() != null) {
        if (a.getSuppress() == null)
          return false;
        return a.getSuppress().getAbbreviation().equals("E");
      }
      // only for ORF
      else {
        return suppressed_terms.contains(a.getConcept().getConceptIdentifier()
            + "|" + a.getStringIdentifier());
      }
    }
    return false;
  }

  //
  // Static methods
  //
  /**
   * Instantiates a {@link HierarchicalSemanticType} from fields in a
   * pipe-separated string.
   * 
   * @param line
   *          a {@link String} with three fields separated by "|" characters
   * @return {@link HierarchicalSemanticType}
   */
  private static STYMetaDescription newSTYMetaDescription(String line) {
    String[] tokens = FieldedStringTokenizer.split(line, "|");
    STYMetaDescriptionImpl sty = new STYMetaDescriptionImpl(tokens[0] + "|"
        + tokens[1] + "|" + tokens[2] + "|||||");
    return sty;
  }

  /**
   * Instantiates a {@link Source} from fields in a pipe-separated string.
   * 
   * @param row
   *          a {@link String} with fields separated by "|" characters
   * @return {@link Source}
   */
  private static SourceMetaDescription newSourceMetaDescription(String row) {
    String[] tokens = FieldedStringTokenizer.split(row, "|");
    String line = "||" + tokens[2] + "|" + tokens[0] + "|" + tokens[1] + "|"
        + tokens[3] + "||||" + tokens[7] + "||||" + tokens[4] + "||"
        + (tokens.length > 6 ? tokens[6] : "") + "||||"
        + (tokens.length > 5 ? tokens[5] : "") + "|||Y|||";
    SourceMetaDescriptionImpl source = new SourceMetaDescriptionImpl(line);
    return source;
  }

  /**
   * Instantiates a {@link SourceTermType} from fields in a pipe-separated
   * string.
   * 
   * @param row
   *          a {@link String} with three fields separated by "|" characters
   * @return {@link SourceTermType}
   */
  private SourceTermType newSourceTermType(String row) {
    String[] tokens = FieldedStringTokenizer.split(row, "|");
    SourceTermType tg = new SourceTermType();
    tg.setSource(Source.getSource(tokens[0]));
    tg.setTermType(TermType.getTermType(tokens[1]));
    if (tokens.length > 3)
      tg.setSuppress(Suppressible.getSuppressible(tokens[3]));
    return tg;
  }

  //
  // Protected Methods
  //

  /**
   * Configures the class by reading in {@link Properties}.
   * 
   * @param props
   *          default/application-level {@link Properties}
   * @return {@link Properties}
   * @throws LocalException
   */
  protected Properties loadProperties(Properties props) throws LocalException {
    // First load the user-level properties
    props = super.loadProperties(props);

    // Here we handle the application-level properties

    // Valid Languages
    if (props.containsKey("languages")) {
      String language_prop = props.getProperty("languages");
      String[] lines = FieldedStringTokenizer.split(language_prop, ";");
      languages = new ArrayList();
      lang_map = new HashMap();
      for (int i = 0; i < lines.length; i++) {
        String[] lat_language = FieldedStringTokenizer.split(lines[i], "|");
        MetaDescription meta_dsc = new MetaDescriptionImpl("LAT|"
            + lat_language[0] + "|expanded_form|" + lat_language[1] + "|");
        languages.add(meta_dsc);
      }
      Language.initialize(languages);
      Language[] lang_array = Language.getLanguages();
      for (int i = 0; i < lang_array.length; i++) {
        lang_map.put(lang_array[i].getAbbreviation(), lang_array[i]);
      }
    } else {
      throw new LocalException(
          "Default properties file does not have a languages property.");
    }

    // List of valid sources
    if (props.containsKey("sources")) {
      String source_prop = props.getProperty("sources");

      String[] lines = FieldedStringTokenizer.split(source_prop, ";");
      sources = new ArrayList();
      for (int i = 0; i < lines.length; i++) {
        SourceMetaDescription s = newSourceMetaDescription(lines[i]);
        sources.add(s);
      }
      Source.initialize(sources);
    } else {
      throw new LocalException(
          "Default properties file does not have a sources property.");
    }

    // Valid termgroups
    if (props.containsKey("termgroups")) {
      String tmgrp_prop = props.getProperty("termgroups");
      String[] lines = FieldedStringTokenizer.split(tmgrp_prop, ";");
      termgroups = new ArrayList();
      ssab_tty_map = new HashMap();
      sab_tty_map = new HashMap();
      // go through termgroups and initialize term types
      MetaDescription[] md = new MetaDescription[lines.length];
      for (int i = 0; i < lines.length; i++) {
        String[] term_type_fields = FieldedStringTokenizer.split(lines[i], "|");
        MetaDescription meta_dsc = new MetaDescriptionImpl("TTY|"
            + term_type_fields[1] + "|expanded_form||");
        md[i] = meta_dsc;
      }
      TermType.initialize(md);
      for (int i = 0; i < lines.length; i++) {
        SourceTermType tg = newSourceTermType(lines[i]);
        tg.setSource(Source.getSource(tg.getSource()
            .getVersionedSourceAbbreviation()));
        ssab_tty_map.put((tg.getSource().getRootSourceAbbreviation() + tg
            .getTermType().getAbbreviation()), tg);
        sab_tty_map.put((tg.getSource().getVersionedSourceAbbreviation() + tg
            .getTermType().getAbbreviation()), tg);
        termgroups.add(tg);
      }
      // alphabetize for the suppressible termgroups filter
      Collections.sort(termgroups);
    } else {
      throw new LocalException(
          "Default properties file does not have a termgroups property.");
    }

    // Suppressed suis list
    if (props.containsKey("suppressed_suis")) {
      String sup_terms_prop = props.getProperty("suppressed_suis");
      String[] lines = FieldedStringTokenizer.split(sup_terms_prop, ";");
      if (suppressed_terms == null)
        suppressed_terms = new HashSet();
      for (int i = 0; i < lines.length; i++)
        suppressed_terms.add(lines[i]);
    } else {
      throw new LocalException(
          "Default properties file does not have a suppressed_suis property.");
    }

    // Valid semantic types
    if (props.containsKey("semantic_types")) {
      String semantic_types_prop = props.getProperty("semantic_types");
      String[] lines = FieldedStringTokenizer.split(semantic_types_prop, ";");
      semantic_types = new ArrayList();
      for (int i = 0; i < lines.length; i++)
        semantic_types.add(newSTYMetaDescription(lines[i]));
      HierarchicalSemanticType.initialize(semantic_types);
      semantic_types = HierarchicalSemanticType
          .getHierarchicalSemanticTypesAsList();
    } else {
      throw new LocalException(
          "Default properties file does not have a semantic types property.");
    }

    return props;
  }

  /**
   * Gets called every time opened or changed config file.
   * 
   * @param ds
   *          {@link DefaultSubset}
   */
  public void setSelectedDefaultSubset(DefaultSubset ds) {
    super.setSelectedDefaultSubset(ds);
    this.setCurrentConfigFile(ds.getConfigurationFilename());
  }

  /**
   * Returns the path to the configuration directory. This is typically the
   * <code>MMSYS/config</code> directory.
   * 
   * @return the path to the configuration directory
   */
  public String getConfigDirectory() {
    return config_dir;
  }

  /**
   * Sets the path to the configuration directory. This is typically the
   * <code>MMSYS/config</code> directory.
   * 
   * @param config_d
   *          the path to the configuration directory
   */
  public void setConfigDirectory(String config_d) {
    File f = new File(config_d).getAbsoluteFile();
    config_dir = f.getName();
  }

  /**
   * Sets the path to the home directory. This is typically the current
   * directory, ".".
   * 
   * @param home_d
   *          the path to the home directory
   */
  public void setHomeDirectory(String home_d) {
    File f = new File(home_d).getAbsoluteFile();
    home_dir = f.getName();
  }

  /**
   * Returns the path to the home directory. This is typically the current
   * directory, ".".
   * 
   * @return the path to the home directory
   */
  public String getHomeDirectory() {
    return home_dir;
  }

}
