/*****************************************************************************
 * Package: gov.nih.nlm.mms
 * Object:  BeginSubsetAction.java
 *
 *****************************************************************************/

package gov.nih.nlm.mms;

import java.awt.Event;
import java.awt.event.ActionEvent;

import javax.swing.Action;
import javax.swing.KeyStroke;

/**
 * Used to initiate the subsetting process. Used for the menu item for the
 * <code>Done->Begin Subsetting</code> Menu of {@link MetamorphoSysGUI}.
 * 
 * @author Brian Carlsen, Deborah Shapiro
 */
public class BeginSubsetAction extends SaveConfigAction {

  //
  // Reference to GUI
  //
  private MetamorphoSysGUI gui;

  /**
   * Instantiates a {@link BeginSubsetAction}.
   * 
   * @param mms
   *          the {@link MetamorphoSys} application object
   * @param gui
   *          the {@link MetamorphoSysGUI}
   */
  public BeginSubsetAction(MetamorphoSys mms, MetamorphoSysGUI gui) {
    super(mms, gui);
    this.gui = gui;

    // configure action
    putValue(Action.NAME, "Begin Subset");
    putValue(Action.SHORT_DESCRIPTION, "Begin Subset");
    putValue(Action.MNEMONIC_KEY, new Integer((int) 'b'));
    putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke('B',
        Event.CTRL_MASK));
  }

  /**
   * Starts the subsetting operation and closes the GUI.
   * 
   * @param e
   *          An {@link java.awt.event.ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {
    gui.enableGlassPane();
  	
    //
    // Get user config reference
    //
    UserConfiguration user_configuration = gui.getUserConfiguration();

    //
    // Validate user selections one last time
    //
    if (!gui.validateGUIConfigurables()) {
      gui.disableGlassPane();
    	return;
    }
    
    //
    // call this to ensure that all source to remove listening filters
    // have updated their lists and will only print sources that will be
    // kept to the log
    //
    gui.fireSourcesToRemoveChanged();

    //
    // Save if data has changed
    //
    if (user_configuration.hasDataChanged()) {
      if (MMSToolkit.confirmRequest(MMSToolkit.getGui(),
          "You have made changes to this configuration.\n"
              + "Would you like to save the changes?")) {
        super.saveConfig();
      } else {
        backupConfig();
        user_configuration.setCurrentConfigFile(null);
      }
    }
    
    //
    // close the GUI
    //
    gui.dispose();

    // Begin subset operation
    Thread t = new Thread(new Runnable() {
      public void run() {
        try {
        	WelcomeGUI.getWelcomeGUI().enableGlassPane();
        	
          InstallConfiguration ic = InstallConfiguration
              .getInstallConfiguration();

          // install LEX and NET if requested by user
          boolean install_passed = InstallConfiguration.install(MMSToolkit
              .getGui());

          // only move on to subsetting if installation passed correctly
          if (install_passed)
            mms.subset();

        } catch (Exception e) {
          e.printStackTrace();
          gui.cancel();
          MMSToolkit.reportError(
              "Unexpected exception.  See console for more information", true);
          MMSToolkit.exitToWelcomeGUI(gui);
        }
      }
    });
    t.start();

  }
}
