/***************************************************************************
 *
 * Package:     gov.nih.nlm.mms
 * Interface:   ConceptInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.ReleaseMetadata;

/**
 *
 * This interface defines an API for reading Concepts from some input source.
 * This source can be MR Files, update MR Files, a database, or pretty
 * much anything else.
 *
 * @author Deborah Shapiro
 *
 */
public interface ConceptInputStream {

  /**
   * Returns a descriptive, yet short, name of the input stream
   * for display purposes. This name is to be used when prompting
   * the user for what input stream they would like to use.
   * @return a short name of the input stream for display purposes.
   */
  public String getName();

  /**
   * Sets the {@link UserConfiguration}.  In certain cases,
   * input streams may require information provided by the application
   * configuration.
   * @param config the {@link UserConfiguration}
   */
  public void setConfiguration(UserConfiguration config);

  /**
   * Opens the input stream.  Here is where the actual files
   * containing the Metathesaurus data should be opened.
   * @param input_loc locator information for the input stream
   * @throws Exception if there is a problem opening files
   */
  public void open(String input_loc) throws Exception;

  /**
   * Closes the input stream.  Here is where the actual files
   * containing the Metathesaurus data should be closed.
   * @throws Exception if there is a problem closing files
   */
  public void close() throws Exception;

  /**
   * Read a {@link Concept} from the input stream and return it.
   * This method should provide the mapping between the input
   * files and {@link Concept} objects.  Furthermore, it should
   * be implemented like a stream, in the sense that only the
   * content for the next {@link Concept} should be read into memory.
   * @return a {@link Concept} read from the input stream
   * @throws Exception if there is a problem reading the concept
   */
  public Concept readConcept() throws Exception;

  /**
   * Read an {@link IndexEntry} from the input stream and return it.
   * This method should provide the mapping between the input
   * files and {@link IndexEntry} objects.  Futhermore, it should be
   * implemented like a stream, in the sense that only the content
   * for the next {@link IndexEntry} should be read into memory.
   * @return an {@link IndexEntry}
   * @throws Exception if there is a problem reading the index entry
   */
  public IndexEntry readIndexEntry() throws Exception;

  /**
   * Read the {@link ReleaseMetadata} from the input stream and return it.
   * @return {@link ReleaseMetadata}
   * @throws Exception if there is a problem
   *         reading the release metadata
   */
  public ReleaseMetadata readReleaseMetadata() throws Exception;

  /**
   * Return an <code>int</code> between 0 and 100 indicating
   * how far along the input stream is in reading {@link Concept}s.
   * This information can be used for a progress monitor.
   * @return <code>int</code> representing percentage complete
   */
  public int getConceptProgress();

  /**
   * Return an <code>int</code> between 0 and 100 indicating
   * how far along the input stream is in reading {@link IndexEntry}
   * elements. This information can be used for a progress monitor.
   * @return <code>int</code> representing percentage complete
   */
  public int getIndexProgress();

  /**
   * Returns a {@link String} describing the configuration details.
   * This is a mechanism for the input stream to write details of
   * its operation to the log file that {@link MetamorphoSys}
   * eventually produces.
   * @return a {@link String} describing the configuration details
   */
  public String getConfigurationDetails();

  /**
   * Provides input stream access to the cui list.
   * @param cui_list {@link CuiList}
   */
  public void setCuiList(CuiList cui_list);

}
