/**************************************************************************
 *
 * Package:     gov.nih.nlm.mms
 * Interface:   ConceptOutputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.ReleaseMetadata;

/**
 * This interface defines an API for writing Concepts to the output
 * Metathesaurus, an output destination.
 * This destination can be MR Files, a database, or pretty
 * much anything else.
 *
 * @author Deborah Shapiro
 *
 */
public interface ConceptOutputStream {

  /**
   * Returns a descriptive, yet short, name of the output stream
   * for display purposes. This name is to be used when prompting
   * the user for what output stream they would like to use.
   * @return a short name of the output stream for display purposes.
   */
  public String getName();

  /**
   * Sets the {@link UserConfiguration}.  In certain cases,
   * output streams may require information provided by the user
   * configuration.
   * @param config the {@link UserConfiguration}
   */
  public void setConfiguration(UserConfiguration config);

  /**
   * Opens the output stream.  Here is where the actual files for
   * the output Metathesaurus should be opened.
   * @param output_loc locator information for the output stream
   * @throws Exception if there is a problem opening files
   */
  public void open(String output_loc) throws Exception;

  /**
   * Closes the output stream.  Here is where the actual files for
   * the output Metathesaurus should be closed.
   * @throws Exception if there is a problem closing files
   */
  public void close() throws Exception;

  /**
   * Write a {@link Concept} to the output stream.
   * @param concept a {@link Concept} to write to the output stream
   * @throws Exception if there is a problem writing the concept
   */
  public void writeConcept(Concept concept) throws Exception;

  /**
   * Write an {@link IndexEntry} to the output stream.
   * @param index_entry an {@link IndexEntry} to write out
   * @throws Exception if there is a problem writing the index entry
   */
  public void writeIndexEntry(IndexEntry index_entry) throws Exception;

  /**
   * Write the {@link ReleaseMetadata} to the output stream.
   * @param release_metadata the {@link ReleaseMetadata} to write out
   * @throws Exception if there is a problem
   *         writing the release metadata
   */
  public void writeReleaseMetadata(ReleaseMetadata release_metadata)
    throws Exception;

  /**
   * Returns a {@link String} describing the configuration details.
   * This is a mechanism for the output stream to write details of
   * its operation to the log file that {@link MetamorphoSys}
   * eventually produces.
   * @return a {@link String} describing the configuration details
   */
  public String getConfigurationDetails();

}

