/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  ConfigurationGUI.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.Action;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JPanel;

/**
 * This dialog is used to choose the configuration options during the
 * installation process. The choices are to make a new configuration or to open
 * a specified configuration file.
 * 
 * @author Deborah Shapiro
 */
public class ConfigurationGUI extends JFrame implements GlassPaneListener {

  private GlassComponent gc = null;

  /**
   * Instantiates the GUI. The constructor does not make it visible.
   * 
   * @param mms
   *          {@link MetamorphoSys}
   */
  public ConfigurationGUI(MetamorphoSys mms) {
    super();
    final InstallConfiguration ic = InstallConfiguration
        .getInstallConfiguration();

    //
    // Indicate which mode we are in and that the users's
    // selections now should override those in the config file chosen.
    //
    ic.setOverrideConfigurationFiles(true);
    ic.setInstallUMLS(true);

    //
    // Disable welcome gui
    //
    WelcomeGUI.getWelcomeGUI().enableGlassPane();

    //
    // Create glass component and install
    //
    gc = new GlassComponent(this);
    setGlassPane(gc);

    //
    // Set title
    //
    setTitle("MetamorphoSys Configuration");

    //
    // Configure layout
    //
    getContentPane().setLayout(new BorderLayout());
    // GridBagLayout gridbag = new GridBagLayout();
    // JPanel choice_panel = new JPanel(gridbag);
    GridBagConstraints constraints = new GridBagConstraints();
    constraints.fill = GridBagConstraints.BOTH;
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy = 0;
    constraints.gridx = 0;
    constraints.gridwidth = 1;
    constraints.gridy++;

    //
    // Compose menu bar
    //
    JMenuBar menu_bar = new JMenuBar();
    setJMenuBar(menu_bar);
    menu_bar.add(Box.createHorizontalGlue());
    JMenu help_menu = new JMenu("Help");
    menu_bar.add(help_menu);
    buildHelpMenu(help_menu);

    //
    // Button Panel
    //
    JPanel button_panel = new JPanel(new FlowLayout());

    // New Configuration Button
    Action local_new_config = new NewConfigAction(mms, WelcomeGUI
        .getWelcomeGUI()) {

      public void actionPerformed(ActionEvent e) {
        MMSProgressMonitor progress = ic.getProgressMonitor();
        progress.setOperation(MMSProgressMonitor.CONFIG_SUBSET);
        dispose();
        super.actionPerformed(e);
        if (isCancelled()) {
          disableGlassPane();
          return;
        }
      }
    };
    JButton jnew = new JButton(local_new_config);
    button_panel.add(jnew);

    // Open Configuration Button
    constraints.gridy++;
    Action local_open_config = new OpenConfigAction(mms, this) {

      public void actionPerformed(ActionEvent e) {
        super.actionPerformed(e);
        if (isCancelled()) {
          disableGlassPane();
          return;
        }
        dispose();
        MMSProgressMonitor progress = ic.getProgressMonitor();
        progress.setOperation(MMSProgressMonitor.CONFIG_SUBSET);
      }
    };
    JButton jopen = new JButton(local_open_config);
    button_panel.add(jopen);

    //
    // Cancel button
    //
    JButton jcancel = new JButton("Cancel");
    jcancel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        ic.clearProgressMonitor();
        dispose();
        return;
      }
    });
    button_panel.add(jcancel);

    getContentPane().add(button_panel, BorderLayout.SOUTH);
    FontSizeManager.adjustFontSize(this);
    this.pack();
    setLocationRelativeTo(null);

    //
    // Configure window listener
    //
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        ic.clearProgressMonitor();
        dispose();
        return;
      }
    });

  };

  /**
   * Builds the help menu items.
   * 
   * @param help_menu
   *          the help {@link JMenu}
   */
  private void buildHelpMenu(JMenu help_menu) {

    //
    // MetamorphoSys Configuration Help Menu Item
    //
    help_menu.add(MMSToolkit.getResourceHelpMenu("MetamorphoSys Configuration",
        "config/MetamorphoSysConfigurationHelp.html", this));
  }

  /**
   * Enable glass pane.
   */
  public void enableGlassPane() {
    gc.setVisible(true);
  }

  /**
   * Disable glass pane.
   */
  public void disableGlassPane() {
    gc.setVisible(false);
  }

  /**
   * Close this window and re-enable the {@link WelcomeGUI}.
   */
  public void dispose() {
    WelcomeGUI.getWelcomeGUI().disableGlassPane();
    super.dispose();
  }

} // End of ConfigurationGUI class
