/*****************************************************************************
 *
 * Package: gov.nih.nlm.mms
 * Object:  EnableFilterAction.java
 *
 *****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.filters.Filter;
import gov.nih.nlm.mms.filters.PrecedenceFilter;
import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.mms.filters.SuppressibleTermgroupsFilter;
import gov.nih.nlm.swing.ListDialog;

import java.awt.Event;
import java.awt.event.ActionEvent;
import java.util.*;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.KeyStroke;

/**
 * Handles enabling and disabling of filters. Used for the
 * <code>File->Enable/Disable Filters</code> menu of {@link MetamorphoSysGUI}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class EnableFilterAction extends AbstractAction {

  //
  // Reference to GUI
  //
  private MetamorphoSysGUI gui;

  /**
   * Instantiates an {@link EnableFilterAction}.
   *
   * @param gui
   *          the {@link MetamorphoSysGUI}
   * @param mms
   *          {@link MetamorphoSys}
   */
  public EnableFilterAction(MetamorphoSysGUI gui, MetamorphoSys mms) {
    super();
    this.gui = gui;

    // configure action
    putValue(Action.NAME, "Enable/Disable Filter");
    putValue(Action.SHORT_DESCRIPTION, "Enable/Disable Filter");
    putValue(Action.MNEMONIC_KEY, new Integer((int) 'f'));
    putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke('E',
        Event.CTRL_MASK));
  }

  /**
   * Present user with a dialog for selecting {@link Filter}s to enable or
   * disable. Once selections are made, enact the user selections.
   *
   * @param e
   *          an {@link ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {

    try {

      // Get configuration objects
      final UserConfiguration user_configuration = gui.getUserConfiguration();

      // Validate all existing filters before proceeding
      // if(!gui.validateGUIConfigurables())
      // return;

      // Before "before" state
      final List active_filters_before = (ArrayList) ((ArrayList) user_configuration
          .getActiveFilters()).clone();

      // remove required filters from display list
      List display_list = user_configuration.getValidFilters();
      SourcesToRemoveFilter strf = null;
      PrecedenceFilter pf = null;
      SuppressibleTermgroupsFilter stf = null;

      for (int i = 0; i < display_list.size(); i++) {
        if (display_list.get(i) instanceof SourcesToRemoveFilter)
          strf = (SourcesToRemoveFilter) display_list.get(i);
        else if (display_list.get(i) instanceof PrecedenceFilter)
          pf = (PrecedenceFilter) display_list.get(i);
        else if (display_list.get(i) instanceof SuppressibleTermgroupsFilter)
          stf = (SuppressibleTermgroupsFilter) display_list.get(i);
      }
      if (strf != null)
        display_list.remove(strf);
      if (pf != null)
        display_list.remove(pf);
      if (stf != null)
        display_list.remove(stf);

      SortedSet displaySet = new TreeSet();
      for (int i = 0; i < display_list.size(); i++) {
        displaySet.add(display_list.get(i));
      }

      final Filter[] display_filters = (Filter[]) displaySet
          .toArray(new Filter[0]);
      final String[] display_strings = new String[display_filters.length];
      List initial_selections = new ArrayList();
      for (int i = 0; i < display_filters.length; i++) {
        display_strings[i] = display_filters[i].getName();
      }
      for (int i = 0; i < active_filters_before.size(); i++) {
        initial_selections.add(((Filter) active_filters_before.get(i))
            .getName());
      }

      // Open Dialog
      final Object[] user_selections = ListDialog.showListMultipleMode(
          MMSToolkit.getGui(), "Choose Filters to be Enabled.",
          "Enable/Disable Filters", display_strings, initial_selections
              .toArray(new Object[0]));

      if (user_selections == null)
        return;

      // update user config based on dialog selections
      user_configuration.clearFilters();
      user_configuration.enableFilter(strf);
      user_configuration.enableFilter(pf);
      user_configuration.enableFilter(stf);

      for (int k = 0; k < user_selections.length; k++) {
        Filter enable_filter;
        String filter_name = user_selections[k].toString();
        for (int i = 0; i < display_strings.length; i++) {
          if (filter_name.equals(display_strings[i])) {
            enable_filter = display_filters[i];
            user_configuration.enableFilter(enable_filter);
          }
        }
      }

      // Get "after" state
      List active_filters_after = (ArrayList) ((ArrayList) user_configuration
          .getActiveFilters()).clone();

      // Identify filters that were "disabled"
      gui.setIgnoreEvents(true);
      for (int i = 0; i < active_filters_before.size(); i++) {
        Filter filter = (Filter) active_filters_before.get(i);
        if (!active_filters_after.contains(filter)) {
          gui.removeTab(filter);
          user_configuration.setProperties(filter);
          filter.removeDataChangeListener(gui);
        }
      }

      // Identify filters that were "enabled"
      for (int i = 0; i < active_filters_after.size(); i++) {
        Filter filter = (Filter) active_filters_after.get(i);
        if (!active_filters_before.contains(filter)) {
          gui.initializeGUIConfigurable(filter);
        }
      }
      gui.setIgnoreEvents(false);

      // Send undo event
      if (!active_filters_before.equals(active_filters_after)) {
        Object[] tuple = { new EnableFilterUndoHandler(gui),
            active_filters_before, active_filters_after };
        gui.pushOnUndoStack(tuple);
      }

    }

    catch (Exception ex) {
      ex.printStackTrace();
      MMSToolkit.reportError(MMSToolkit.getGui(),
          "Unexpected error while enabling filters.\n"
              + "See console for details.", false);
    }

  }

}