/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Object:     FilterEnableDialog.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.filters.Filter;
import gov.nih.nlm.mms.filters.PrecedenceFilter;
import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.mms.filters.SuppressibleTermgroupsFilter;
import gov.nih.nlm.swing.FontSizeManager;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;

/**
 * Prompts the user to select which filters should be enabled or disabled. The
 * valid and active filters lists from the {@link UserConfiguration} are used.
 * 
 * @author Deborah Shapiro
 */
public class FilterEnableDialog extends JDialog implements ActionListener {


  // List of checkboxes (for before/after comparisons)
  //
  private List checkbox_list = new ArrayList();

  //
  // User configuration reference
  //
  private UserConfiguration config = null;

  //
  // Constructor
  //

  /**
   * Instantiates a {@link FilterEnableDialog}.
   * 
   * @param config
   *          the {@link UserConfiguration}
   */
  public FilterEnableDialog(UserConfiguration config) {
    super(MMSToolkit.getGui());
    setModal(true);
    setTitle("Enable/Disable Filters");
    this.config = config;
    initialize();
  };

  /**
   * Initialize the GUI components in the panel
   */
  private void initialize() {

    // Configure Layout
    getContentPane().setLayout(new BorderLayout());
    JPanel data_panel = new JPanel(new GridBagLayout());
    GridBagConstraints constraints = new GridBagConstraints();
    constraints.fill = GridBagConstraints.BOTH;
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy = 0;
    constraints.gridwidth = 3;

    // For each filter, add a checkbox
    for (int i = 0; i < config.getValidFilters().size(); i++) {
      String valid_filter = ((Filter) (config.getValidFilters().get(i)))
          .getName();
      JCheckBox chk = new JCheckBox(valid_filter);
      constraints.gridy++;

      // If filter is presently active, mark the box as checked
      boolean set_selected = false;
      for (int j = 0; j < config.getActiveFilters().size(); j++) {
        if (((Filter) (config.getActiveFilters().get(j))).getName().equals(
            valid_filter)) {
          set_selected = true;
        }

        // If filter is one of the main three, disable it
        Object filter = config.getValidFilters().get(i);
        if (filter instanceof SourcesToRemoveFilter
            || filter instanceof PrecedenceFilter
            || filter instanceof SuppressibleTermgroupsFilter) {
          chk.setEnabled(false);
        } else {
          chk.setEnabled(true);
        }
      }
      chk.setSelected(set_selected);
      checkbox_list.add(chk);
      data_panel.add(chk, constraints);
    }

    // Configure buttons
    constraints.gridy++;
    JButton jdone = new JButton("Done");
    jdone.addActionListener(this);
    data_panel.add(jdone, constraints);

    // Configure border
    EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
    data_panel.setBorder(new CompoundBorder(etched_border, new EmptyBorder(10,
        10, 10, 10)));
    getRootPane().setDefaultButton(jdone);

    // Compose Panel & Pack dialog
    getContentPane().add(data_panel, BorderLayout.CENTER);
    FontSizeManager.adjustFontSize(this);
    this.pack();
  };

  //
  // ActionListener Implementatoin
  //

  /**
   * Applies user selections and disposes of the dialog.
   * 
   * @param e
   *          an {@link ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("Done")) {
      config.clearFilters();
      for (int i = 0; i < checkbox_list.size(); i++) {
        if (((JCheckBox) checkbox_list.get(i)).isSelected()) {
          config.enableFilter((Filter) (config.getValidFilters().get(i)));
        }
      }
      FilterEnableDialog.this.dispose();
    }
  }

}; // End of FilterEnableDialog class

