/*****************************************************************************
 *
 * Package: gov.nih.nlm.mms
 * Object:  ImportFilterAction
 *
 *****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.filters.Filter;
import gov.nih.nlm.swing.ListDialog;

import java.awt.Event;
import java.awt.event.ActionEvent;
import java.util.*;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.KeyStroke;

/**
 * Imports filters from the paths in the <code>/ext</code> directory Used for
 * the <code>File->Import Filter</code> menu of {@link MetamorphoSysGUI}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ImportFilterAction extends AbstractAction {

  //
  // GUI reference
  //
  private MetamorphoSysGUI gui;

  /**
   * Instantiates an {@link ImportFilterAction}.
   *
   * @param gui
   *          the {@link MetamorphoSysGUI}
   * @param mms
   *          {@link MetamorphoSys}
   */
  public ImportFilterAction(MetamorphoSysGUI gui, MetamorphoSys mms) {
    super();
    this.gui = gui;

    // configure action
    putValue(Action.NAME, "Import Filter..");
    putValue(Action.SHORT_DESCRIPTION, "Import Filter");
    putValue(Action.MNEMONIC_KEY, new Integer((int) 'i'));
    putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke('I',
        Event.CTRL_MASK));
  }

  /**
   * Present user with a dialog for selecting {@link Filter}s to import. These
   * {@link Filter} implementations must appear in jar files in the
   * <code>MMSYS/ext</code> directory.
   *
   * @param e
   *          an {@link ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {

    // Acquire configuration references
    UserConfiguration user_configuration = gui.getUserConfiguration();

    try {

      // Find filter implementations
      String[] filter_options = null;
      Filter import_filter = null;
      try {
        filter_options = MMSToolkit
            .getInterfaceImplementations("gov.nih.nlm.mms.filters.Filter");
      } catch (ClassNotFoundException cnfe) {
        cnfe.printStackTrace();
        MMSToolkit.reportError(gui,
            "Unexpected error looking for filters to import.\n"
                + "Make sure the filter interface exists in your classpath.",
            false);
        return;
      }

      // Do not display filters that are already valid
      List new_filters = new ArrayList();
      List valid_filters = user_configuration.getValidFilters();
      for (int i = 0; i < filter_options.length; i++) {
        // instantiate the filter in order to get its name for display
        Filter this_filter = (Filter) MMSToolkit
            .instantiateObject(filter_options[i]);
        boolean no_match = true;
        for (int j = 0; j < valid_filters.size(); j++) {
          if (((Filter) valid_filters.get(j)).getName().equals(
              this_filter.getName())
              || this_filter instanceof MetamorphoSysInputStream
              || this_filter instanceof MetamorphoSysOutputStream)
            no_match = false;
        }
        if (no_match)
          new_filters.add(this_filter);
      }

      SortedSet displaySet = new TreeSet();
      for (int i = 0; i < new_filters.size(); i++) {
        displaySet.add(new_filters.get(i));
      }

      final Filter[] display_filters = (Filter[]) displaySet
          .toArray(new Filter[0]);
      String[] display_options = new String[display_filters.length];
      for (int i = 0; i < display_options.length; i++) {
        display_options[i] = display_filters[i].getName();
      }

      // Open Dialog if there are any filters left to import
      if (display_options.length != 0) {
        Object[] user_selections = ListDialog.showListMultipleMode(MMSToolkit
            .getGui(), "Choose a Filter to Import.", "Import Filters",
            display_options, null);
        if (user_selections != null) {
          for (int k = 0; k < user_selections.length; k++) {
            String filter_name = user_selections[k].toString();
            for (int i = 0; i < display_options.length; i++) {
              if (filter_name.equals(display_options[i]))
                import_filter = display_filters[i];
            }

            // configure selected filter
            gui.initializeGUIConfigurable(import_filter);

            if (!(import_filter instanceof MetamorphoSysOutputStream)
                && !(import_filter instanceof MetamorphoSysInputStream))
              user_configuration.importFilter(import_filter);
          }
        }
      } else {
        MMSToolkit.reportError("There are no filters in the /ext directory "
            + "that have not already been installed.", false);
      }
    } catch (Exception ex) {
      ex.printStackTrace();
      MMSToolkit.reportError(gui, "Unexpected error while importing filters.\n"
          + "See console for details.", false);
    }
  }
}
