/*****************************************************************************
 *
 * Package: gov.nih.nlm.mms
 * Object:  IndexEntryQueue.java
 *
 ****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.impl.IndexEntryImpl;

import java.util.Date;

/**
 * First In First Out queue for {@link IndexEntry} objects.  A producer
 * adds index entries until the queue is full and then blocks.  A consumer
 * reads index entries out (notifying the producer when room is available
 * for adding more) and blocks when there are no more index entries to read.
 * The producer wakes up the consumer when there is more data to read.
 *
 * @author Brian Carlsen
 */
public class IndexEntryQueue {

  //
  // Fields
  //
  private IndexEntry[] index_entries = null;
  private int next_consumer_index = 0;
  private int next_producer_index = 0;
  private boolean producer_finished = false;
  private boolean is_producer_waiting = false;
  private boolean is_consumer_waiting = false;
  private Object block = new Object();

  //
  // Statistics for dynamic load balancing
  //
  private long producer_last_request = new Date().getTime();
  private long producer_request_ct = 1;
  private long producer_total_request_time = 300;
  private long consumer_last_request = new Date().getTime();
  private long consumer_request_ct = 1;
  private long consumer_total_request_time = 300;


  //
  // Constructor
  //

  /**
   * Instantiates a {@link IndexEntryQueue} with the specified size.
   * @param size the size of the queue
   */
  public IndexEntryQueue(int size) {
    index_entries = new IndexEntry[size];
  }
  
  /**
   * Indicates whether or not the producer is finished adding {@link IndexEntry}s.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean isProducerFinished() { return producer_finished; }

  /**
   * Sets the flag to <code>true</codE> indicating that the producer is finished
   * adding {@link IndexEntry}s.
   */
  public void setProducerFinished() {
    producer_finished = true;
    // Wake client if it is sleeping
    synchronized (block) {
      block.notifyAll();
    }
  }

  /**
   * Indicates whether or not the queue is full.
   * @return <code>true</code> if full, <codE>false</code> otherwise
   */
  public boolean isFull() {
    return next_consumer_index == next_producer_index && index_entries[next_producer_index] != null;
  }

  /**
   * Indicates whether or not the queue is empty.
   * @return <code>true</code> if empty, <codE>false</code> otherwise
   */
  public boolean isEmpty() {
    return next_consumer_index == next_producer_index && index_entries[next_consumer_index] == null;
  }

  /**
   * Adds a index entry to the queue, blocks if the queue is full afterwards.
   * Wakes up a sleeping consumer if the queue is at least 50% full.
   * @param index_entry the {@link IndexEntry} to add
   */
  public void put(IndexEntry index_entry) {
    producer_total_request_time += (new Date()).getTime() - producer_last_request;
    producer_request_ct++;

    //
    // Block if we need to wait.  When waking up, we will be ready to add more
    //
    checkProducerWait();
    producer_last_request = new Date().getTime();

    //
    //
    //
    index_entries[next_producer_index++] = index_entry;
    if (next_producer_index >= index_entries.length) next_producer_index = 0;

    checkConsumerNotify();

  }

  /**
   * Removes a {@link IndexEntry} from the queue, blocks if the queue is empty afterwards.
   * Wakes up a sleeping producer if the queue is at less than 50% full.
   * @return the {@link IndexEntry}
   */
  public IndexEntry get() {
    IndexEntry index_entry = null;
    consumer_total_request_time += (new Date()).getTime() - consumer_last_request;
    consumer_request_ct ++;

    //
    // Block if we need to wait.  When waking up, we will be ready to return
    //
    checkConsumerWait();
    if (isEmpty() && isProducerFinished()) return null;
    consumer_last_request = new Date().getTime();

    //
    // Get index_entry, null reference
    //
    index_entry = index_entries[next_consumer_index];
    index_entries[next_consumer_index++] = null;
    if (next_consumer_index >= index_entries.length) next_consumer_index = 0;

    checkProducerNotify();

    return index_entry;
  }

  /**
   * Check to see if producer should wait, if so, wait.
   */
  private void checkProducerWait() {
    try {
      synchronized (block) {
        if (isFull()) {
          is_producer_waiting = true;
          if (is_consumer_waiting)
            block.notifyAll();
          block.wait();
          is_producer_waiting = false;
        }
      }
    } catch (InterruptedException ie) {}
  }

  /**
   * Check to see if producer should wait, if so, wait.
   */
  private void checkConsumerWait() {
    try {
      synchronized (block) {
        if (isEmpty() && !isProducerFinished()) {
          is_consumer_waiting = true;
          if (is_producer_waiting)
            block.notifyAll();
          block.wait();
          is_consumer_waiting = false;
        }
      }
    } catch (InterruptedException ie) {}
  }

  /**
   * Check to see if consumer should awaken.
   */
  private void checkConsumerNotify() {
    //
    // Get consumer/producer average time stats
    //
    double producer_avg = (producer_total_request_time / producer_request_ct);
    double consumer_avg = (consumer_total_request_time / consumer_request_ct);
    int perf_gap = (int)(index_entries.length * (1 - (consumer_avg / (producer_avg+consumer_avg))));

    //
    // If consumer is waiting and producer is faster, wake consumer
    // OR if consumer is waiting and is faster, wake consumer if
    // the number of index_entries waiting to be processed = gap
    //
    if (is_consumer_waiting &&
        (producer_avg < consumer_avg ||
         (consumer_avg < producer_avg && getIndexEntryWaitingCount() >= perf_gap))) {
      synchronized (block) {
        block.notifyAll();
      }
    }
  }


  /**
   * Check to see if producer should awaken.
   */
  private void checkProducerNotify() {

    //
    // Get consumer/producer average time stats
    //
    double producer_avg = (producer_total_request_time / producer_request_ct);
    double consumer_avg = (consumer_total_request_time / consumer_request_ct);
    int perf_gap = (int)(index_entries.length * (producer_avg / (producer_avg+consumer_avg)));


    // If producer is waiting and consumer is faster, wake producer
    // OR if producer is waiting and is faster, wake producer if
    // the number of index_entries waiting to be processed < gap
    //
    if (is_producer_waiting &&
        (consumer_avg < producer_avg ||
         (producer_avg < consumer_avg && getIndexEntryWaitingCount() <= perf_gap))) {
      synchronized (block) {
        block.notifyAll();
      }
    }
  }

  /**
   * Returns the number of index_entries waiting to be processed.
   * @return the number of index_entries waiting to be processed
   */
  private int getIndexEntryWaitingCount() {
    if (next_producer_index > next_consumer_index) {
      return next_producer_index - next_consumer_index;
    } else if (next_producer_index == next_consumer_index) {
      if (index_entries[next_consumer_index] == null) return 0;
      else return index_entries.length;
    } else
      return next_producer_index + (index_entries.length - next_consumer_index);
  }

  /**
   * Self-qa test.
   * @param s command line arguments
   */
  public static void main (String[] s) {
    // Test the case where producer is faster
    final IndexEntryQueue queue = new IndexEntryQueue(100);
    final int producer_delay = 200;
    final int consumer_delay = 200;
    Thread producer = new Thread (
      new Runnable() {
        public void run() {
           for (int i = 0; i < 200; i ++) {
             try { Thread.sleep(producer_delay); } catch (InterruptedException e) {}
             IndexEntry c = new IndexEntryImpl();
             c.setCui("C"+i);
             queue.put(c);
           }
           queue.setProducerFinished();
        }
      }
    );

    Thread consumer = new Thread (
      new Runnable() {
        public void run() {
          while (true) {
            IndexEntry c = queue.get();
            if (c == null) break;
            try {Thread.sleep(consumer_delay);} catch (InterruptedException ie) {}
          }

        }
      }
    );

   producer.start();
   consumer.start();

   try {
     producer.join();
     consumer.join();
   } catch (InterruptedException ie) { ie.printStackTrace(); }

 }
}
