/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  InstallationGUI.java
 *
 *************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.browser.RRFBrowserGUI;
import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

import javax.swing.*;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;

/**
 * Used to choose the basic installation options.
 * 
 * @author Deborah Shapiro
 */
public class InstallationGUI extends JFrame implements GlassPaneListener {

  //
  // For enabling a glass pane
  //
  private GlassComponent gc = null;

  //
  // Indicates whether or not user cancelled the operation
  //
  private boolean is_cancelled = true;

  //
  // Constructor
  //

  /**
   * Instantiates the {@link InstallationGUI}. The constructor does not make it
   * visible.
   */
public InstallationGUI() {
    super();
    final InstallConfiguration ic = InstallConfiguration
        .getInstallConfiguration();

    //
    // Indicate to MetamorphoSysGUI to override config files with choices from
    // here
    //
    ic.setOverrideConfigurationFiles(true);

    //
    // Indicate that we are installing the full UMLS, not just Metathesaurus
    //
    ic.setInstallUMLS(true);

    //
    // disable welcome GUI
    //
    WelcomeGUI.getWelcomeGUI().enableGlassPane();

    //
    // Configure Installation GUI
    //
    setTitle("Install UMLS");
    gc = new GlassComponent(this);
    setGlassPane(gc);
    getContentPane().setLayout(new BorderLayout());
    GridBagLayout gridbag = new GridBagLayout();
    JPanel choice_panel = new JPanel(gridbag);
    GridBagConstraints constraints = new GridBagConstraints();
    constraints.fill = GridBagConstraints.HORIZONTAL;
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy = 0;
    constraints.gridx = 0;
    constraints.gridy++;

    //
    // Compose menu bar
    //
    JMenuBar menu_bar = new JMenuBar();
    setJMenuBar(menu_bar);
    menu_bar.add(Box.createHorizontalGlue());
    JMenu help_menu = new JMenu("Help");
    menu_bar.add(help_menu);
    buildHelpMenu(help_menu);

    //
    // Add header
    //
    constraints.gridy++;
    constraints.gridx = 2;
    constraints.gridwidth = 1;

    JTextPane text = new JTextPane();
    text.setEditable(false);
    text.setMinimumSize(new Dimension(200, 50));
    text.setPreferredSize(new Dimension(200, 50));
    text.setBackground(choice_panel.getBackground());
    SimpleAttributeSet set = new SimpleAttributeSet();
    StyleConstants.setAlignment(set, StyleConstants.ALIGN_CENTER);
    StyleConstants.setFontFamily(set, "Times");
    StyleConstants.setFontSize(set, 20);
    StyleConstants.setBold(set, true);
    text.setParagraphAttributes(set, true);
    text.setText("INSTALL UMLS");

    choice_panel.add(text, constraints);
    constraints.gridy++;

    //
    // Add source directory: label, field, browse button
    //
    constraints.gridx = 0;
    choice_panel.add(new JLabel("Source"), constraints);
    constraints.gridx = 2;
    final JTextField source_field = new JTextField(40);
    source_field.setEditable(false);

    // Try setting Source text field from system property, if available.
    final String source_prop =
        System.getProperty("gov.nih.nlm.mms.InstallationGUI.source");
    if (source_prop != null) {
      setUmlsDirectory(new File(source_prop), source_field, ic);
    }
    else {
      try {
        //
        // If "been_saved.txt" file exists, prepopulate Source field from saved
        // value
        //
        if (new File(ic.getMetamorphoSysConfigurationDirectory(),
                     "been_saved.txt").exists()) {
          Properties props = new Properties();
          FileInputStream propInputStream = new FileInputStream(new File(
              ic.getMetamorphoSysConfigurationDirectory(), "mmsys.prop.sav"));
          props.load(propInputStream);
          // Set the source path according to the source_paths property
          if (props.containsKey("source_paths")) {
            String source_path_list = props.getProperty("source_paths");
            String[] source_paths =
                FieldedStringTokenizer.split(source_path_list, ";");
            setUmlsDirectory(new File(source_paths[0]), source_field, ic);
          }
        }
        //
        // else set default metathesaurus file
        //
        else {
          source_field.setText(new File("..").getCanonicalPath());
        }
      } catch (IOException ioe) {
        // do nothing
      }
    }
    source_field.setBackground(Color.white);
    gridbag.setConstraints(source_field, constraints);
    choice_panel.add(source_field);
    constraints.gridx = 3;
    final JButton source_browse = new JButton("Browse..");
    source_browse.setPreferredSize(new Dimension(100, 22));
    source_browse.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        try {
          //
          // Select a source directory
          //
          File selected_file = null;
          if (!source_field.getText().equals("")) {
            selected_file = new File(source_field.getText()).getCanonicalFile();
          }
          File source_dir = MMSToolkit.chooseFile(
              "Select and Open the source directory", "Open..", ic
                  .getMetamorphoSysInputStream().getChooserType(),
              selected_file, selected_file, InstallationGUI.this);

          //
          // If selected, configure IC
          //
          if (source_dir != null) {

            InstallConfiguration ic = InstallConfiguration
                .getInstallConfiguration();
            setUmlsDirectory(source_dir, source_field, ic);
          }
        } catch (Exception ex) {
          ex.printStackTrace();
          MMSToolkit.reportError(InstallationGUI.this,
              "Unexpected error while setting source directory.\n"
                  + "See console for details.", false);
        }
      }
    });

    gridbag.setConstraints(source_browse, constraints);
    choice_panel.add(source_browse);

    //
    // Add source directory comment
    //
    constraints.gridy++;
    constraints.gridx = 2;
    constraints.insets = new Insets(0, 6, 10, 0);
    JLabel comment = new JLabel("This directory contains UMLS files.");
    choice_panel.add(comment, constraints);
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy++;

    //
    // Add destination directory: label, field, browse button
    //
    constraints.gridy++;
    constraints.gridx = 0;
    constraints.gridwidth = 1;
    choice_panel.add(new JLabel("Destination"), constraints);
    constraints.gridx = 2;
    final JTextField destination_field = new JTextField(40);
    destination_field.setEditable(false);
    destination_field.setBackground(Color.white);

    // Try setting Destination text field from system property, if available.
    final String destination_prop =
        System.getProperty("gov.nih.nlm.mms.InstallationGUI.destination");
    if (destination_prop != null) {
      setUmlsDestinationDirectory(new File(destination_prop), destination_field,
                                  ic);
    }
    else {
      try {
        //
        // If "been_saved.txt" file exists, prepopulate Destination field from
        // saved value
        //
        if (new File(ic.getMetamorphoSysConfigurationDirectory(),
                     "been_saved.txt").exists()) {
          Properties props = new Properties();
          FileInputStream propInputStream = new FileInputStream(new File(
              ic.getMetamorphoSysConfigurationDirectory(), "mmsys.prop.sav"));
          props.load(propInputStream);
          // Set the destination path according to the subsetname property
          if (props.containsKey("subsetname")) {
            File file = new File(props.getProperty("subsetname"));
            if (file.getName().equals("META") &&
                file.getParent().endsWith(MMSToolkit.getReleaseVersion())) 
              setUmlsDestinationDirectory(
                  file.getParentFile().getParentFile(), destination_field,
                  ic);
            else
              setUmlsDestinationDirectory(file, destination_field,
                  ic);
          }
        }
      } catch (IOException ioe) {
        // do nothing
      }
    }

    gridbag.setConstraints(destination_field, constraints);
    choice_panel.add(destination_field);
    constraints.gridx = 3;
    final JButton destination_browse = new JButton("Browse..");
    destination_browse.setPreferredSize(new Dimension(100, 22));
    destination_browse.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        try {
          File selected_file = null;

          if (!destination_field.getText().equals("")) {
            selected_file = new File(destination_field.getText())
                .getCanonicalFile();
          } else {
            selected_file = new File(source_field.getText()).getParentFile();
          }

          File destination_dir = MMSToolkit.chooseFile(
              "Select and Open the destination directory", "Open..",
              JFileChooser.DIRECTORIES_ONLY, selected_file, selected_file,
              InstallationGUI.this);
          if (destination_dir != null) {
            setUmlsDestinationDirectory(destination_dir, destination_field, ic);
          }
        } catch (Exception ex) {
          ex.printStackTrace();
          MMSToolkit.reportError(InstallationGUI.this,
              "Unexpected error while setting destination directory.\n"
                  + "See console for details.", false);
        }
      }
    });

    gridbag.setConstraints(destination_browse, constraints);
    choice_panel.add(destination_browse);

    //
    // Add destination comment
    //
    constraints.gridy++;
    constraints.gridx = 2;
    constraints.insets = new Insets(0, 6, 0, 0);
    comment = new JLabel("Select where to create the top-level UMLS directory.");
    choice_panel.add(comment, constraints);
    constraints.insets = new Insets(0, 6, 10, 0);
    constraints.gridy++;
    comment = new JLabel(
        "The directory name will correspond to the UMLS release, e.g.,  "
            + ic.getReleaseVersion().toUpperCase() + ".");
    choice_panel.add(comment, constraints);
    constraints.insets = new Insets(2, 2, 2, 2);
    constraints.gridy++;

    //
    // Add button panel label
    //
    constraints.gridx = 0;
    constraints.gridwidth = 3;
    constraints.gridy++;
    choice_panel
        .add(new JLabel("Install UMLS Knowledge Sources:"), constraints);

    //
    // Add Metathesaurus Installation Row
    //
    constraints.gridy++;
    constraints.gridx = 0;
    constraints.gridwidth = 3;
    final String metathesaurus_prop = System.getProperty(
        "gov.nih.nlm.mms.InstallationGUI.metathesaurus");
    if (metathesaurus_prop != null) {
      ic.setMetathesaurusChecked(metathesaurus_prop.equals("true"));
    }
    final JCheckBox meta_box = new JCheckBox("Metathesaurus", ic
        .isMetathesaurusChecked());
    meta_box.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        InstallConfiguration.getInstallConfiguration().setMetathesaurusChecked(
            meta_box.isSelected());
      }
    });
    choice_panel.add(meta_box, constraints);

    //
    // add NET Installation Row
    //
    constraints.gridy++;
    constraints.gridx = 0;
    constraints.gridwidth = 3;

    final String semantic_network_prop = System.getProperty(
        "gov.nih.nlm.mms.InstallationGUI.semantic_network");
    if (semantic_network_prop != null) {
      ic.setSemanticNetworkChecked(semantic_network_prop.equals("true"));
    }
    final JCheckBox net_box = new JCheckBox("Semantic Network", ic
        .isSemanticNetworkChecked());
    net_box.setMaximumSize(new Dimension(40, 22));
    choice_panel.add(net_box, constraints);
    net_box.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        InstallConfiguration.getInstallConfiguration()
            .setSemanticNetworkChecked(net_box.isSelected());
      }
    });

    //
    // add LEX Installation Row
    //
    constraints.gridy++;
    final String specialist_lexicon_prop = System.getProperty(
        "gov.nih.nlm.mms.InstallationGUI.specialist_lexicon");
    if (specialist_lexicon_prop != null) {
      ic.setLexiconChecked(specialist_lexicon_prop.equals("true"));
    }
    final JCheckBox lex_box = new JCheckBox(
        "SPECIALIST Lexicon & Lexical Tools", ic.isLexiconChecked());
    lex_box.setMaximumSize(new Dimension(40, 22));
    choice_panel.add(lex_box, constraints);
    lex_box.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        InstallConfiguration.getInstallConfiguration().setLexiconChecked(
            lex_box.isSelected());
      }
    });
    constraints.gridy++;

    //
    // add LVG Installation Row
    //
// constraints.gridy++;
// final String install_lvg_prop = System.getProperty(
// "gov.nih.nlm.mms.InstallationGUI.install_lvg");
// if (install_lvg_prop != null) {
// ic.setLvgChecked(install_lvg_prop.equals("true"));
// }
// final JCheckBox lvg_box = new JCheckBox("Install and Auto-Configure LVG",
// ic.isLvgChecked());
// lvg_box.setMaximumSize(new Dimension(40, 22));
// constraints.insets = new Insets(0, 18, 2, 2);
// choice_panel.add(lvg_box, constraints);
// constraints.insets = new Insets(2, 2, 2, 2);
// lvg_box.addActionListener(
// new ActionListener() {
// public void actionPerformed(ActionEvent e) {
// InstallConfiguration.getInstallConfiguration()
// .setLvgChecked(lvg_box.isSelected());
// }
// }
// );
// constraints.gridy++;
    
    //
    // add Install Button
    //
    constraints.gridx = 0;
    constraints.gridwidth = 1;
    constraints.weightx = 1;
    constraints.gridy++;
    JButton jinstall = new JButton("OK");
    jinstall.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        //
        // Indicate that "Cancel" was not pushed
        //
        is_cancelled = false;

        //
        // Disable the GUI
        //
        enableGlassPane();

        //
        // The installation should not happen in the Swing event handler
        // thread, so we put it off into a separate thread.
        //
        Thread t = new Thread(new Runnable() {
          public void run() {

            //
            // Get configuration information
            //
            InstallConfiguration ic = InstallConfiguration
                .getInstallConfiguration();

            //
            // validate user selections
            //
            boolean validate_passed = validateChoices();
            if (!validate_passed) {
              disableGlassPane();
              return;
            }

            //
            // Now, if selected, install the Metathesaurus
            //
            if (ic.isMetathesaurusChecked()) {

              //
              // Dispose of this GUI, keep WelcomeGUI disabled
              //
              InstallationGUI.this.dispose();

              ic.getProgressMonitor().setOperation(
                  MMSProgressMonitor.CONFIG_SUBSET);

              //
              // Let user choose a "New Configuration Action" or
              // an "Open Configuration Action"
              //
              new ConfigurationGUI(new MetamorphoSys()).setVisible(true);
            } else {
              InstallConfiguration.install(InstallationGUI.this);
              ic.clearProgressMonitor();
              MMSToolkit.exitToWelcomeGUI(InstallationGUI.this);
            }

          } // end public void run()

        }); // End of new Thread
        t.start();
      }
    }

    ); // end actionListener for install button

    //
    // Cancel button
    //
    JButton jcancel = new JButton("Cancel");
    jcancel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        is_cancelled = true;
        MMSToolkit.exitToWelcomeGUI(InstallationGUI.this);
        return;
      }
    });

    //
    // Add OK, Cancel buttons
    //
    JPanel button_panel = new JPanel(new FlowLayout());
    jinstall.setPreferredSize(jcancel.getPreferredSize());
    button_panel.add(jinstall);
    button_panel.add(jcancel);

    //
    // Add ok and cancel button panel
    //
    constraints.gridx = 0;
    constraints.gridwidth = 4;
    constraints.weightx = 1;
    constraints.gridy++;
    choice_panel.add(button_panel, constraints);

    // getContentPane().add(button_panel, BorderLayout.SOUTH);

    EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
    choice_panel.setBorder(new CompoundBorder(etched_border, new EmptyBorder(
        10, 10, 10, 10)));
    getContentPane().add(choice_panel, BorderLayout.CENTER);

    //
    // Prepare for display
    //
    FontSizeManager.adjustFontSize(this);
    this.pack();
    setLocationRelativeTo(null);

    //
    // Configure window listener
    //
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        is_cancelled = true;
        MMSToolkit.exitToWelcomeGUI(InstallationGUI.this);
        return;
      }
    });

  }
  /**
   * Updates the Source text field and sets the directory that contains the
   * Metathesaurus data.
   * 
   * @param source_dir
   *          source directory {@link File} of the Metathesaurus data
   * @param source_field
   *          Source text field
   * @param ic
   *          installation configuration
   */
  private static void setUmlsDirectory(File source_dir,
      JTextField source_field, InstallConfiguration ic) {
    source_field.setText(source_dir.getAbsolutePath());

    // also computes META, LEX, NET directories
    ic.setUmlsDirectory(source_dir);
  }

  /**
   * Updates the Destination text field and sets the directory into which the
   * UMLS will be installed.
   * 
   * @param destination_dir
   *          output directory {@link File} into which the UMLS will be
   *          installed.
   * @param destination_field
   *          Destination text field
   * @param ic
   *          installation configuration
   */
  private static void setUmlsDestinationDirectory(File destination_dir,
      JTextField destination_field, InstallConfiguration ic) {
    destination_field.setText(destination_dir.getAbsolutePath());
    ic.setUmlsDestinationDirectory(destination_dir);
  }

  /**
   * Builds the help menu items.
   * 
   * @param help_menu
   *          the help {@link JMenu}
   */
  private void buildHelpMenu(JMenu help_menu) {

    //
    // Install UMLS Help Menu Item
    //
    help_menu.add(MMSToolkit.getResourceHelpMenu("Install UMLS Help",
        "config/InstallUMLSHelp.html", this));

    //
    // UMLS Homepage menu item
    //
    help_menu.add(MMSToolkit.getURLHelpMenu("UMLS Homepage",
        "http://www.nlm.nih.gov/research/umls/", this));

    //
    // Docs Home Page Menu Item
    //
    help_menu.add(MMSToolkit.getURLHelpMenu("UMLS Documentation",
        "http://www.nlm.nih.gov/research/umls/umlsdoc.html", this));

    //
    // NLM Home Page Menu Item
    //
    help_menu.add(MMSToolkit.getURLHelpMenu("NLM Homepage",
        "http://www.nlm.nih.gov", this));
  }

  /**
   * Indicates whether or not the installation was cancelled.
   * 
   * @return <code>true</code> if the GUI was cancelled, <code>false</code>
   *         otherwise
   */
  public boolean isCancelled() {
    return is_cancelled;
  }

  /**
   * Enable glass pane.
   */
  public void enableGlassPane() {
    gc.setVisible(true);
  }

  /**
   * Disable glass pane.
   */
  public void disableGlassPane() {
    gc.setVisible(false);
  }

  /**
   * Close this window and re-enable the {@link WelcomeGUI}.
   */
  public void dispose() {
    super.dispose();
  }

  //
  // Private Methods
  //
  /**
   * Validates the user selections for directories.
   * 
   * @return <code>true</code> if validation passes <code>false</code>
   *         otherwise
   */
  private boolean validateChoices() {

    //
    // validate all the fields
    //
    InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();

    if (!ic.isSemanticNetworkChecked() && !ic.isLexiconChecked()
        && !ic.isMetathesaurusChecked() && !ic.isLvgChecked()) {
      MMSToolkit.notifyUser(InstallationGUI.this,
          "You must select at least one checkbox.");
      disableGlassPane();
      return false;
    }

    if (!ic.isLexiconChecked() && ic.isLvgChecked()) {
      MMSToolkit.notifyUser(InstallationGUI.this,
          "You must install the SPECIALIST Lexicon and Lexical Tools\n"
              + "when you choose to Install and Auto-Configure LVG.\n"
              + "Please also select the SPECIALIST Lexicon and Lexical Tools.");
      disableGlassPane();
      return false;
    }

    //
    // Create destination dir if not exists
    //
    if (ic.getUmlsDestinationDirectory() != null
        && !ic.getUmlsDestinationDirectory().exists()) {
      boolean confirm = MMSToolkit.confirmRequest(this,
          "The destination directory does not exist.\n"
              + "Would you like to create it?");
      if (confirm) {
        ic.getUmlsDestinationDirectory().mkdirs();
      } else {
        MMSToolkit.notifyUser(InstallationGUI.this,
            "Please select a destination directory that exists.");
        disableGlassPane();
        return false;
      }
    }

    //
    // Create the META dir as well if necessary
    //
    if (ic.getUmlsDestinationDirectory() != null
        && ic.getUmlsDestinationDirectory().exists()
        && ic.getMetaDestinationDirectory() != null
        && !ic.getMetaDestinationDirectory().exists()
        && ic.isMetathesaurusChecked()) {
      File meta_dir = ic.getMetaDestinationDirectory();
      if (!meta_dir.exists())
        meta_dir.mkdirs();

    }

    //
    // Validate destination dir
    //
    if (ic.getUmlsDestinationDirectory() == null) {
      MMSToolkit.notifyUser(InstallationGUI.this,
          "Please choose a destination directory.");
      disableGlassPane();
      return false;
    } else if (!ic.getUmlsDestinationDirectory().canWrite()) {
      MMSToolkit.notifyUser(InstallationGUI.this,
          "The destination directory is write protected.\n"
              + "Please choose another destination directory.");
      disableGlassPane();
      return false;
    }

    //
    // Validate META
    //
    if ((ic.isMetathesaurusChecked()) && ic.getMetathesaurusFiles() != null
        && ic.getMetathesaurusFiles().length > 0) {
      try {
        if (!MMSToolkit.archiveContainsDir(ic.getMetathesaurusFiles()[0],
            "META")) {
          disableGlassPane();
          return false;
        }
      } catch (Exception ex) {
        disableGlassPane();
        MMSToolkit.reportError(InstallationGUI.this, "Invalid source path:"
            + "\n\n" + ic.getMetathesaurusFiles()[0].getParent()
            + "\n\nMust be a directory containing" + "\na valid "
            + ic.getReleaseVersion().toUpperCase()
            + " Metathesaurus NLM data file."
            + "\nPlease choose another source path.", false);
        return false;
      }
    }

    //
    // Validate source paths against input stream
    //
    if (ic.isMetathesaurusChecked()) {

      String[] source_paths = new String[1];
      try {
        source_paths[0] = ic.getUmlsDirectory().getCanonicalPath();
      } catch (Exception e) {
        // do nothing
      }
      MetamorphoSysInputStream input_stream = new NLMFileMetamorphoSysInputStream();
      if (!input_stream.validateSourcePaths(source_paths)) {
        MMSToolkit.reportError(InstallationGUI.this, input_stream
            .getValidationErrors()[0], false);
        return false;
      }
    }

    //
    // Validate LEX
    //
    try {
      if ((ic.isLexiconChecked() || ic.isLvgChecked())
          && !MMSToolkit.archiveContainsDir(ic.getLexiconFile(), "LEX")) {
        disableGlassPane();
        return false;
      }
    } catch (Exception ex) {
      disableGlassPane();
      MMSToolkit.reportError(InstallationGUI.this, "Invalid source path:"
          + "\n\n" + ic.getLexiconFile().getParent()
          + "\n\nMust be a directory containing" + "\na valid "
          + ic.getLexiconFile().getName() + " file."
          + "\nPlease choose another source path.", false);
      return false;
    }

    //
    // Validate NET
    //
    try {
      if ((ic.isSemanticNetworkChecked())
          && !MMSToolkit.archiveContainsDir(ic.getSemanticNetworkFile(), "NET")) {
        disableGlassPane();
        return false;
      }
    } catch (Exception ex) {
      disableGlassPane();
      MMSToolkit.reportError(InstallationGUI.this, "Invalid source path:"
          + "\n\n" + ic.getSemanticNetworkFile().getParent()
          + "\n\nMust be a directory containing" + "\na valid "
          + ic.getSemanticNetworkFile().getName() + " file."
          + "\nPlease choose another source path.", false);
      return false;
    }

    //
    // Validate file byte counts in Metathesaurus source directory
    //
    try {
      final ChecksumValidator cv = new ChecksumValidator(ic.getUmlsDirectory(),
          this);
      boolean validByteCounts = cv.validateByteCountsOnly();
      if (!validByteCounts) {
        return false;
      }
    } catch (Exception e) {
      MMSToolkit.reportError(WelcomeGUI.getWelcomeGUI(),
          "Unexpected exception validating distribution\n"
              + "See console for details.", false);
    }
    return true;
  }

} // End of InstallationGUI class

