/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms
 * Object:	MMSProgressMonitor.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.util.ProgressEvent;
import gov.nih.nlm.util.ProgressListener;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JProgressBar;

/**
 * Progress monitor for installation and subsetting.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */ 
public abstract class MMSProgressMonitor extends JFrame implements ActionListener,
   ProgressListener {


  public static int CONFIG_SUBSET = 5;
  public static int INSTALL_META = 6;
  public static int INIT_CUIS = 7;
  public static int SUBSET_META = 8;
  public static int SUBSET_INDEXES = 9;
  public static int FINAL_PROCESSES = 10;
  public static int DONE = 11;

  public static boolean cancelled = false;
  
  //
  // Protected Fields
  //
  protected JCheckBox jconfig_subset = new JCheckBox("Configure Metathesaurus Subset");
  protected JCheckBox jinstall_meta = new JCheckBox("Installing Metathesaurus");
  protected JCheckBox jinit_cui = new JCheckBox("Initializing CUI List");
  protected JCheckBox jsubset_mr = new JCheckBox("Subsetting Content");
  protected JCheckBox jsubset_index = new JCheckBox("Subsetting Indexes");
  protected JCheckBox jfinal_process = new JCheckBox("Final Processes");
  protected JCheckBox jdone = new JCheckBox("Done");
  protected JLabel progress_label = new JLabel();
  protected JProgressBar jprogress_bar = new JProgressBar(0,100);
  protected JButton jcancel= new JButton("Cancel");

  //
  // The low_scale is set by the application for
  // scaling progress events properly.
  //
  private int low_scale = 0;
  private int high_scale = 0;


  /**
   * Configure the GUI.
   */
  protected abstract void initialize();

  //
  // Methods
  //

  /**
   * Returns the progress monitor with new value and repack.
   * @param value for progress of monitor
   */
  public void setProgress(int value) {
    jprogress_bar.setValue(value);
    setNote("Operation is " + value + " % Complete");
    this.pack();
  }

  /**
   * Update the progress monitor with new value and repack.
   * @return the percentage complete
   */
  public int getProgress() {
    return jprogress_bar.getValue();
  }

  /**
   * Update the message shown by the progress monitor and repack.
   * @param msg the message
   */
  public void setNote(String msg) {
    progress_label.setText(msg);
    this.pack();
  }

  /**
   * Set the indicated operation and repack.
   * @param oper the currently running operation
   */
  public abstract void setOperation(int oper);

  /**
   * Set the minimum progress value and repack.
   * @param min the minimum progress value
   */
  public void setMinimum(int min) {
    jprogress_bar.setMinimum(min);
    this.pack();
  }

  /**
   * Set the maximum progress value and repack.
   * @param max the maximum progress value
   */
  public void setMaximum(int max) {
    jprogress_bar.setMaximum(max);
    this.pack();
  }

  /**
   * Sets the {@link JProgressBar} to be visible or not.
   * @param visible <code>true</code> if should be visible
   *                <code>false</code> otherwise
   */
  protected void setProgressBarVisible(boolean visible) {
    if (visible)
      jprogress_bar.setVisible(true);
    else
      jprogress_bar.setVisible(false);
  }
  //
  // ActionListener Implementation
  //

  /**
   * Called when "Cancel" button is clicked.  Confirm
   * with user, then exit application.
   * @param e {@link ActionEvent}
   */
  public void actionPerformed (ActionEvent e) {
    confirmExit();
  }

  /**
   * Confirms with user before exiting the application.
   */
  protected void confirmExit() {
    boolean are_you_sure =
      MMSToolkit.confirmRequest(this,
        "Are you sure you want\nto cancel MetamorphoSys?");
    if (are_you_sure) {
    	cancelled = true;
      File f = new File(InstallConfiguration.getInstallConfiguration().getMetaDestinationDirectory(),"mmsys.log");
      if (f.exists()) {
      	this.dispose();
        MMSToolkit.showLogAndExitToWelcomeGUI("MetamorphoSys Log",f);
      } else
        MMSToolkit.exitToWelcomeGUI(this);
    }
  }

  /**
   * Indicates whether or not MetamorphoSys has been cancelled.
   * @return <code>boolean</code> true if MetamorphoSys has been cancelled
   *                              false otherwise
   */
  public boolean isCancelled() {
  	return cancelled;
  }
  
  //
  // ProgressListener interface
  //

  /**
   * Sets the low scale for scaled progress updates.
   * @param low_scale low scale
   */
  public void setLowScale(int low_scale) { this.low_scale = low_scale; }

  /**
   * Sets the high scale for scaled progress updates.
   * @param high_scale high scale
   */
  public void setHighScale(int high_scale) { this.high_scale = high_scale; }

  /**
   * Update progress.
   * @param pe the {@link ProgressEvent}
   */
  public void updateProgress(ProgressEvent pe) {
    int pct = pe.getScaledPercent(low_scale, high_scale);
    if (pct > getProgress()) setProgress(pct);
  }


}; // End of MMSProgressMonitor class

