/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms
 * Object:	MetamorphoSys.java
 *
 *Changes
 *  01/22/2006 BAC (1-74IJY): main() method now fakes suppress entries
 *     before constructor is called.  The initializeConfigurables is no
 *     longer explicitly called because it is done in the constructor.
 *  01/19/2006 BAC (1-74IJY): Re-introduced PrecedenceFilter code in subset()
 *     MRRANK now properly produced by output handlers.
 *  
 **************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.browser.BrowserState;
import gov.nih.nlm.mms.browser.RRFBrowserGUI;
import gov.nih.nlm.mms.filters.ConceptContentSelector;
import gov.nih.nlm.mms.filters.ConceptManipulator;
import gov.nih.nlm.mms.filters.ConceptSelector;
import gov.nih.nlm.mms.filters.Filter;
import gov.nih.nlm.mms.filters.PrecedenceFilter;
import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.swing.FontSizeManager;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.AtomCoocurrence;
import gov.nih.nlm.umls.meta.AtomMapping;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ConceptMapping;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.ReleaseMetadata;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.util.ProgressReporter;

import java.awt.Component;
import java.awt.Font;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;

/**
 * This class represents the MetamorphoSys application itself.
 * It primarily is responsible for implementing the subsetting functionality.
 *
 * This class can be directly called as an application if you do
 * not want to run MetamorphoSys with a GUI.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */
public class MetamorphoSys implements SourcesToRemoveListener {

  //
  // The default configuration
  //
  private ApplicationConfiguration default_configuration;

  //
  // The user configuration
  //
  private UserConfiguration user_configuration;

  //
  // The mmsys.log
  //
  private BufferedWriter log;

  //
  // Sources currently chosen to be removed
  //
  private List sources_to_remove = new ArrayList();

  //
  // Constructors
  //

  /**
   * Instantiates a {@link MetamorphoSys} object, initializes
   * the {@link ApplicationConfiguration} and {@link UserConfiguration}s from
   * the config/mmsys.a.prop and config/mmsys.prop.sav files.
   *
   * This constructor directly reports errors via {@link MMSToolkit#reportError}
   * rather than throwing exceptions.
   */
  public MetamorphoSys() {
    // Does the config directory exist and is it readable and writable?
    File config = new File(".", "config");
    if (!config.exists() ||
	!config.canRead() ||
	!config.isDirectory()) {
      MMSToolkit.reportError("Cannot find/read/write sub_directory " +
			     "config in directory " +
                             config.getParentFile(), true);
    }

    initializeApplicationConfiguration(new File("config", "mmsys.a.prop"));
    initializeUserConfiguration(new File("config", "mmsys.prop.sav"));
    initializeConfigurables();
  }

  /**
   * Instantiates a {@link MetamorphoSys} object, initializes
   * the {@link ApplicationConfiguration} and {@link UserConfiguration}s.
   *
   * This constructor directly reports errors via {@link MMSToolkit#reportError}
   * rather than throwing exceptions.
   * @param default_file {@link File}
   * @param user_file {@link File}
   */
  public MetamorphoSys(File default_file, File user_file) {
    initializeApplicationConfiguration(default_file);
    initializeUserConfiguration(user_file);
    initializeConfigurables();
  }

  /**
   * Instantiates a {@link MetamorphoSys} object.
   * @param ac {@link ApplicationConfiguration}
   * @param uc {@link UserConfiguration}
   */
  public MetamorphoSys(ApplicationConfiguration ac, UserConfiguration uc) {
    default_configuration = ac;
    user_configuration = uc;
    initializeConfigurables();
  }


  /**
   * Initializes the {@link ApplicationConfiguration}.
   * @param default_file the "mmsys.a.prop" file
   */
  private void initializeApplicationConfiguration(File default_file) {

    try {
      default_configuration = new ApplicationConfiguration(default_file);
    } catch (LocalException le) {
      le.printStackTrace();
      MMSToolkit.reportError(le.getMessage(), false);
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(
       "You have a badly formatted or incomplete default properties file.\n" +
       "Please replace your /config/mmsys.a.prop file with the\n" +
       "original default file and write protect it.\n", true);
    }
  }
    /**
     * Initializes the {@link UserConfiguration}.
     * @param user_file the "mmsys.prop.sav" file
     */
    private void initializeUserConfiguration(File user_file) {
      try {
        user_configuration = new UserConfiguration(user_file);
      }
      catch (LocalException le) {
        le.printStackTrace();
        MMSToolkit.reportError(le.getMessage(), false);
      }
      catch (Exception e) {
        e.printStackTrace();
        MMSToolkit.reportError(
            "You have a badly formatted or incomplete " +
            "mmsys.prop.sav properties file.\n" +
            "Please copy your /config/mmsys.a.prop file to mmsys.prop.sav\n" +
            "and restart the application.\n", true);
      }
  }

  /**
   * Instantiates a {@link MetamorphoSys} object, initializes
   * the {@link ApplicationConfiguration} and {@link UserConfiguration}s.
   * Used for on-the-fly subsetting, uses {@link Properties} objects rather
   * than configuration files.
   *
   * This constructor directly reports errors via {@link MMSToolkit#reportError}
   * rather than throwing exceptions.
   * @param user {@link Properties}
   * @param app {@link Properties}
   */
  public MetamorphoSys(Properties user, Properties app) {

    try {
      default_configuration = new ApplicationConfiguration(app);
    } catch (LocalException le) {
      le.printStackTrace();
      MMSToolkit.reportError(le.getMessage(), false);
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(
       "You have a badly formatted or incomplete " +
       "default properties parameter.\n", true);
    }

    try {
      user_configuration = new UserConfiguration(user);
    } catch (LocalException le) {
      le.printStackTrace();
      MMSToolkit.reportError(le.getMessage(), false);
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(
       "You have a badly formatted or incomplete " +
       "properties parameter.\n", true);
    };
    initializeConfigurables();
  }

  /**
   * Loads and configures the filters and input/output streams.
   */
  public void initializeConfigurables() {
    // Get active filter list
    List active_filters = user_configuration.getActiveFilters();
    if (active_filters == null)
      MMSToolkit.reportError(
       "There are no activated filters in the configuration", true);

    // Configure the active filters
    for (int i = 0; i < active_filters.size(); i++) {
      initializeConfigurable((Filter)active_filters.get(i));
    }
    // Configure the input/output streams
    initializeConfigurable(user_configuration.getInputStream());
    initializeConfigurable(user_configuration.getOutputStream());

    user_configuration.resetDataChanged();
  }

  /**
   * Configures a filter or input/output stream.
   * @param c {@link Configurable}
   */
  public void initializeConfigurable(Configurable c) {
    c.setConfiguration(user_configuration);
    c.setConfiguration(default_configuration);

    Properties p = user_configuration.getProperties(c);
    if (p.size() == 0)
      c.resetToDefaults();
    else
      c.setProperties(p);

  }

  //
  // Methods
  //

  /**
   * Subsets the {@link Concept}.  Here we remove content based on
   * the various installed filters but we do not try to use the {@link CuiList}
   * or try to resolve references.  In other words, related content (CUI2, AUI2)
   * is not part of the subsetting criteria.
   * @param concept the {@link Concept} to subset
   */
  public void subset(Concept concept) {

    //
    // fire this here b/c it won't have been fired if in batch mode
    //
    fireSourcesToRemoveChanged();

    final List concept_selectors = user_configuration.getConceptSelectors();
    for (int i = 0; i < concept_selectors.size(); i++) {
      if ( !((ConceptSelector) concept_selectors.get(i)).keep(concept)) {
        concept.clear();
        return;
      }
    }
    final List concept_content_selectors = user_configuration.
        getConceptContentSelectors();
    for (int i = 0; i < concept_content_selectors.size(); i++) {
      ( (ConceptContentSelector) concept_content_selectors.get(i)).applySelections(
          concept);
      if (concept.isEmpty()) {
        concept.clear();
        return;
      }
    }

    //
    // Apply remaining filters, or empty concept if it has no atoms
    //
    resolveInternalReferences(concept);

    // check again b/c resolveInternalReferences may have made concept empty
    if (!concept.isEmpty()) {
      //if (prec_filter != null)
      //  prec_filter.applyFilter(concept);
      final List concept_manipulators = user_configuration.getConceptManipulators();
      for (int i = 0; i < concept_manipulators.size(); i++) {
        ( (ConceptManipulator) concept_manipulators.get(i)).applyChanges(concept);
      }
    }
  }

  /**
   * Sets if {@link Source} is in current version.
   * @param source_md {@link Source}
   */
  public void subset(Source source_md) {
    if (!source_md.isCurrentVersion()) {
      source_md.setInCurrentSubset(false);
      return;
    }
    if (source_md.isCurrentVersion() &&
      sources_to_remove.contains(
      Source.getSource(source_md.getRootSourceAbbreviation()))) {
      source_md.setInCurrentSubset(false);
    } else {
      source_md.setInCurrentSubset(true);
    }
  }

  /**
   * Handle error.
   * @param e {@link Exception}
   * @param msg error string
   */
 private void handleError(Exception e, String msg) {
		InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();
		final MMSProgressMonitor progress = ic.getProgressMonitor();
		if (!progress.isCancelled()) {

			try {
				MMSToolkit.printStackTrace(e, log);
				log.close();
			} catch (Exception ex) {
				MMSToolkit.reportError("Unable to write to mmsys.log.");
			}
			try {
				user_configuration.getInputStream().close();
			} catch (Exception ioe) {
			}
			MMSToolkit.reportError(MMSToolkit.getGui(), msg, false);

			File subset_dir = InstallConfiguration.getInstallConfiguration()
					.getMetaDestinationDirectory();
			progress.dispose();
			MMSToolkit.showLogAndExitToWelcomeGUI("MetamorphoSys Subset Log",
					new File(subset_dir, "mmsys.log"));
		}
	}

  /**
	 * Subsets the Metathesaurus. This process has four main stages:
	 * <ol>
	 * <li> Initialize {@link CuiList}. This opens a {@link ConceptInputStream}
	 * and subsets all {@link Concept}s, maintaining a {@link CuiList} of those
	 * to keep, and those to get rid of.
	 * <li> Subset Concepts. We subset the {@link Concept}s in two stages because
	 * of the {@link Relationship}, {@link Coocurrence}, and {@link Context}
	 * data, all of which have cui1 and cui2 values. In this stage of subsetting
	 * we actually subset the {@link Concept}s, including this binary data and
	 * write the results to the subset destination.
	 * <li> Subset indexes. Once the core content is subsetted, we use the
	 * {@link CuiList} to keep those {@link IndexEntry} objects that are still
	 * valid, and these are also written to the subset destination.
	 * <li> Final Processes. Complete final processes which include writing MRCUI,
	 * MRAUI and ReleaseMetadata.
	 * </ol>
	 * 
	 */
  public void subset() {
    //
    // Set up progress monitor
    //
    InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();
    final MMSProgressMonitor progress = ic.getProgressMonitor();
    if (MMSToolkit.usingView()) {
      progress.setOperation(MMSProgressMonitor.INSTALL_META);
      progress.setProgressBarVisible(true);
    }

    //
    // Prep list of CUIS that are removed
    //
    final List subx_cuis = new ArrayList(1000000);

    //
    // Obtain and configure streams
    //
    final MetamorphoSysInputStream mmsys_input_stream =
        user_configuration.getInputStream();
    final MetamorphoSysOutputStream mmsys_output_stream =
        user_configuration.getOutputStream();

    //
    // Consult filters on use of efficient subsetting mode
    // Notify input stream so that reading concepts can function appropriately
    //
    boolean efficient_mode = true;
    final List active_filters = user_configuration.getActiveFilters();
    for (int i = 0; i < active_filters.size(); i++) {
      if (! ( (Filter) active_filters.get(i)).isEfficientEnabled()) {
        efficient_mode = false;
        break;
      }
    };
    mmsys_input_stream.setEfficientMode(efficient_mode);


    //
    // if not using view, tracks percent finished
    //
    int pct, prev_pct = 0;

    //
    // Obtain subset dir and config file references
    //
    final Date start_date = new Date();
    final String[] source_paths = user_configuration.getSourcePaths();
    final String subset_dir = user_configuration.getSubsetDirectory();
    final String config_file = user_configuration.getCurrentConfigurationFilename();

    //
    // Open mmsys.log file and write intro
    //
    try {
      log = new BufferedWriter(new FileWriter(
          new File(user_configuration.getSubsetDirectory(), "mmsys.log")));
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(
          "Exception opening mmsys.log: " + e.getMessage(), true);
    }

    //
    // Print log header
    //
    try {
    log.write("MetamorphoSys Version:................" + MMSToolkit.getApplicationVersion());
    log.newLine();
    log.write("MetamorphoSys Build Date:............." + MMSToolkit.getApplicationBuildDate());
    log.newLine();
    log.write("UMLS Build Date:......................" + MMSToolkit.getBuildDate());
    log.newLine();
    log.write("Release Version:......................" + MMSToolkit.getReleaseVersion());
    log.newLine();
    log.write("Release Date:........................." + MMSToolkit.getReleaseDate());
    log.newLine();
    log.write("Release Description:.................." + MMSToolkit.getReleaseDescription());
    log.newLine();
    log.write("Metathesaurus Source paths:..........." + source_paths[0]);
    log.newLine();
    for (int i = 1; i < source_paths.length; i++) {
      log.write("......................................" + source_paths[i]);
      log.newLine();
    }
    log.write("Subsetted Metathesaurus folder:......." + subset_dir);
    log.newLine();
    if (config_file != null) {
      log.write("Configuration file used:.............." + config_file);
      log.newLine();
    }
    log.write("Start at:............................." + start_date);
    log.newLine();
    log.flush();
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(
          "Exception writing to mmsys.log: " + e.getMessage(), true);
    }
    //
    // Inform listeners that CUI initialization has started.
    //
    for (int i = 0; i < active_filters.size(); i++)
      ( (Filter) active_filters.get(i)).beginInitializeCuiList();
    mmsys_input_stream.beginInitializeCuiList();

    //
    // INIT CUI LIST
    //
    final CuiSuiAuiList cui_list = new CuiSuiAuiList();

    //
    // Open input handler
    //
    try {
      mmsys_input_stream.open(source_paths);
    } catch (IOException ioe) {
      handleError(ioe, "Error opening input handler.");
      return;
    }

    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      try {
        SwingUtilities.invokeAndWait(
          new Runnable() {
            public void run() {
              progress.setMinimum(0); progress.setMaximum(100); progress.setProgress(0);
              progress.setOperation(MMSProgressMonitor.INIT_CUIS);
            }
          } );
      } catch (Exception e) {}
    } else
      System.out.println("["+ new Date() +"] "+"Begin Initialize CUI List");

    //
    // Inform listeners that CUI initialization has started.
    //
    for (int i = 0; i < active_filters.size(); i++)
      ( (Filter) active_filters.get(i)).setCuiList(cui_list);
    mmsys_input_stream.setCuiList(cui_list);


    final ConceptSelector[] concept_selectors = (ConceptSelector[])
        user_configuration.getConceptSelectors().toArray(new ConceptSelector[0]);
    final ConceptContentSelector[] concept_content_selectors = (ConceptContentSelector[])
        user_configuration.getConceptContentSelectors().toArray(new ConceptContentSelector[0]);
    final ConceptManipulator[] concept_manipulators = (ConceptManipulator[])
        user_configuration.getConceptManipulators().toArray(new ConceptManipulator[0]);
    final Filter[] old_style_filters = (Filter[])
    		user_configuration.getOldStyleFilters().toArray(new Filter[0]);
    
    //
    // Iterate through concept list
    //
    int counter = 0;
    final int reset_counter = 1000;
    Concept concept = null;
    OUTER_INIT:
    while (true) {

      //
      // Read concept
      //
      try {
        concept = mmsys_input_stream.readConcept();
        if (concept == null)
          break;
      }
      catch (IOException ioe) {
        handleError(ioe, "Error reading concept.");
        return;
      }

      counter++;

      //
      // Update progress monitor
      //
      if (counter == reset_counter) {
        counter = 0;
        if (MMSToolkit.usingView()) {
          if (progress.getProgress() < mmsys_input_stream.getConceptProgress()) {
            SwingUtilities.invokeLater(
                new Runnable() {
              public void run() {
                progress.setProgress(mmsys_input_stream.getConceptProgress());
              }
            });
          }
        }
        else if ( (pct = mmsys_input_stream.getConceptProgress()) % 10 == 0 &&
                 pct > 0 && pct != prev_pct) {
          System.out.println("["+ new Date() +"] "+"Initialize CUI List: " + pct + "% Finished");
          prev_pct = pct;
        }
      }

      //
      // Set unfiltered concept in the cui list
      //
      cui_list.setUnfilteredConcept(concept);

      for (int i = 0; i < concept_selectors.length; i++) {
        if (!concept_selectors[i].keep(concept))
          continue OUTER_INIT;
      }

      for (int i = 0; i < concept_content_selectors.length; i++) {
        concept_content_selectors[i].applySelections(concept);
        if (concept.isEmpty())
          continue OUTER_INIT;
      }
      
      for (int i = 0; i < old_style_filters.length; i++) {
        old_style_filters[i].applyFilter(concept);
        if (concept.isEmpty())
          continue OUTER_INIT;
      } 
      
      //
      // Used to remove UTF8 chars and implement MTH/MM logic
      //

      resolveInternalReferences(concept);

      for (int i = 0; i < concept_manipulators.length; i++) {
        concept_manipulators[i].applyChanges(concept);
      }

      // should never be the case that concept_manipulator removes content
      if (concept.isEmpty())
          continue;

      //
      // Add to cui list if concept has anything in it
      //
      cui_list.setFilteredConcept(concept);

    } // end while loop

    //
    // Inform listeners that CUI list initialization is done
    //
    for (int i = 0; i < active_filters.size(); i++)
      ( (Filter) active_filters.get(i)).endInitializeCuiList();
    mmsys_input_stream.endInitializeCuiList();

    //
    // Initialize CUI List is done
    //
    try {
      log.write("Initialize CUI List completed:........" + new Date());
      log.newLine();
    } catch (IOException ioe) {
      handleError(ioe, "Error writing to mmsys.log.\n" +
      		ioe.getMessage());
      return;
    }    
    
    //
    // Done initializing cui list.  Close/re-open stream
    //
    try {
      mmsys_input_stream.close();
      mmsys_input_stream.open(source_paths);
    } catch (IOException ioe) {
      handleError(ioe, "Error closing and re-opening input handler.");
      return;
    }

    //
    // Open output stream
    //
    try {
      mmsys_output_stream.open(user_configuration.getSubsetDirectory());
      if (MMSToolkit.usingView() &&
          mmsys_output_stream instanceof ProgressReporter) {
        ProgressReporter pr = (ProgressReporter)mmsys_output_stream;
        pr.addProgressListener(progress);
      }
    } catch (IOException ioe) {
      handleError(ioe, "Error opening output handler.");
      return;
    }

    //
    // Pass ambiguous SUI,LUI info from input to output
    //
    try {
      mmsys_output_stream.setCandidateAmbiguousStringIdentifiers(
          mmsys_input_stream.getAmbiguousStringIdentifiers());
      mmsys_output_stream.setCandidateAmbiguousTermIdentifiers(
          mmsys_input_stream.getAmbiguousTermIdentifiers());
    } catch (IOException ioe) {
      handleError(ioe,
          "Error looking up candidate ambiguous string/term identifiers.");
      return;
    }

    //
    // Inform listeners that subsetting of content is beginning
    //
    for (int i = 0; i < active_filters.size(); i++)
      ( (Filter) active_filters.get(i)).beginSubsetContent();
    mmsys_input_stream.beginSubsetContent();

    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      try {
        SwingUtilities.invokeAndWait(
          new Runnable() {
            public void run() {
              progress.setMinimum(0); progress.setMaximum(100); progress.setProgress(0);
              progress.setOperation(MMSProgressMonitor.SUBSET_META);
              System.gc();
            }
          } );
      } catch (Exception e) {}
    } else
      System.out.println("["+ new Date() +"] "+"Begin Subset Metathesaurus");

    //
    // Iterate through concept list
    //
    int cui_counter = 0;
    int subset_cui_counter = 0;
    counter = 0;
    OUTER_CONTENT:
    while (true) {

      //
      // Read concept
      //
      try {
        concept = mmsys_input_stream.readConcept();
        if (concept == null)
          break;
      }
      catch (IOException ioe) {
        handleError(ioe,
                    "Error reading concept.");
        return;
      }

      cui_counter++;
      counter++;

      //
      // Update progress
      //
      if (counter == reset_counter) {
        counter = 0;
        if (MMSToolkit.usingView()) {
          if (progress.getProgress() < mmsys_input_stream.getConceptProgress()) {
            SwingUtilities.invokeLater(
                new Runnable() {
              public void run() {
                progress.setProgress(mmsys_input_stream.getConceptProgress());
                System.gc();
              }
            });
          }
        }
        else if ( (pct = mmsys_input_stream.getConceptProgress()) % 10 == 0 &&
                 pct > 0 && pct != prev_pct) {
          System.out.println("["+ new Date() +"] "+"Subset Metathesaurus: " + pct + "% Finished");
          prev_pct = pct;
        }
      }

      //
      // recompute cui contents for index entry in case it wasn't done in
      // previous section
      //
      if (cui_list.addedByUserOnly(concept.getCui()))
        cui_list.setUnfilteredConcept(concept);

      boolean wrote_concept = false;

      //
      // Only process CUIs still on the list
      //
      if (cui_list.inList(concept.getCui())) {

        for (int i = 0; i < concept_selectors.length; i++) {
          if (!concept_selectors[i].keep(concept)) {
            continue OUTER_CONTENT;
          }
          if (concept.isEmpty())
            continue OUTER_CONTENT;
        }

        for (int i = 0; i < concept_content_selectors.length; i++) {
          concept_content_selectors[i].applySelections(concept);
          if (concept.isEmpty())
            continue OUTER_CONTENT;
        }
        
        for (int i = 0; i < old_style_filters.length; i++) {
          old_style_filters[i].applyFilter(concept);
          if (concept.isEmpty())
            continue OUTER_CONTENT;
        }        

        //
        // Resolve references (remove dangling pointers, etc)
        //
        resolveExternalReferences(cui_list, concept);

        //
        // Used to remove UTF8 chars and implement MTH/MM logic
        //

        resolveInternalReferences(concept);

        for (int i = 0; i < concept_manipulators.length; i++) {
          concept_manipulators[i].applyChanges(concept);
        }

        if (concept.isEmpty())
          continue;

        //
        // call method to remove based on cui2 and aui2
        //
        if (cui_list.addedByUserOnly(concept.getCui()))
          cui_list.setFilteredConcept(concept);

        try {
          mmsys_output_stream.writeConcept(concept);
        }
        catch (IOException ioe) {
          ioe.printStackTrace();
          handleError(ioe, "Error writing concept.\n" +
          		ioe.getMessage() + 
                      "\nSee console for details.");
          return;
        }

        wrote_concept = true;
        subset_cui_counter++;
      }


      //
      // hold integer values in subx_cuis list to be added to the
      // ReleaseMetadata at the end of subsetting indexes
      //
      if (!wrote_concept)
        subx_cuis.add(concept.getCui());

    }

    //
    // Inform listeners that subsetting of content is beginning
    //
    for (int i = 0; i < active_filters.size(); i++)
      ( (Filter) active_filters.get(i)).endSubsetContent();
    mmsys_input_stream.endSubsetContent();
    try {
      log.write("Subset Metathesaurus completed:......." + new Date());
      log.newLine();
    } catch (IOException ioe) {
      handleError(ioe, "Error writing to mmsys.log.\n" +
      		ioe.getMessage());
      return;
    }    
    
    cui_list.clearAuiSet();
    System.gc();
    
    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      try {
        SwingUtilities.invokeAndWait(
          new Runnable() {
            public void run() {
              progress.setMinimum(0); progress.setMaximum(100); progress.setProgress(0);
              progress.setOperation(MMSProgressMonitor.SUBSET_INDEXES);
            }
          } );
      } catch (Exception e) {}
    } else
      System.out.println("["+ new Date() +"] "+"Begin Subset Indexes");

    //
    // Subset index files
    //
    IndexEntry index_entry = null;
    counter = 0;
    final IndexEntryQueue ieq = new IndexEntryQueue(20000);
    final Exception[] exceptions = new Exception[1];
    final Thread producer = new Thread(
      new Runnable() {
        public void run() {
          try {
            IndexEntry ie = null;
            while ( (ie = mmsys_input_stream.readIndexEntry()) != null) {
              ieq.put(ie);
            }
            ieq.setProducerFinished();
          } catch (IOException ioe) {
            exceptions[0] = ioe;
          }
        }
      }
    );
    producer.start();

    //
    // Iterate through index entries
    //
    try {
      Thread.sleep(2000);
    } catch (InterruptedException ie) {} 
    
    while ( (index_entry = ieq.get()) != null) {  

      counter++;

      //
      // Write index entry if CUI,SUI are in the cui list
      //
      try {
        if (cui_list.inList(index_entry.getCui(), index_entry.getSui(), null))
          mmsys_output_stream.writeIndexEntry(index_entry);
      } catch (IOException ioe) {
        handleError(ioe, "Error writing index entry.\n" +
        		ioe.getMessage());
        return;
      }

      //
      // Update progress
      //
      if (counter == reset_counter) {
        counter = 0;
        if (MMSToolkit.usingView()) {
          if (progress.getProgress() < mmsys_input_stream.getIndexProgress()) {
            SwingUtilities.invokeLater(
              new Runnable() {
                public void run() {
                  progress.setProgress(mmsys_input_stream.getIndexProgress());
                }
              } );
          }
        } else if ( (pct = mmsys_input_stream.getIndexProgress()) % 10 == 0 &&
                   pct > 0 && pct != prev_pct) {
          System.out.println("["+ new Date() +"] "+"Subset Indexes: " + pct + "% Finished");
          prev_pct = pct;
        }
      }

    } // end while

    //
    // Check for exceptions
    //
    if (exceptions[0] != null) {
      handleError(exceptions[0], "Error writing index entries.");
      return;
    }

    try {
      log.write("Subset Index Files completed:........." + new Date());
      log.newLine();
    } catch (IOException ioe) {
      handleError(ioe, "Error writing to mmsys.log.\n" +
      		ioe.getMessage());
      return;
    }
    

    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      try {
        SwingUtilities.invokeAndWait(
          new Runnable() {
            public void run() {
              progress.setMinimum(0); progress.setMaximum(100); progress.setProgress(0);
               progress.setOperation(MMSProgressMonitor.FINAL_PROCESSES);
            }
          } );
      } catch (Exception e) {}
    } else
      System.out.println("["+ new Date() +"] "+"Final Processes");


    //
    //
    // FINAL PROCESSES
    //
    //

    //
    // Get release metadata
    //
    ReleaseMetadata release_metadata = null;
    try {
      release_metadata = mmsys_input_stream.readReleaseMetadata();
    } catch (IOException ioe) {
      handleError(ioe, "Error reading release metadata.\n" + ioe.getMessage());
      return;
    }

    //
    // Separate references to precedence and suppressible filters
    //
    PrecedenceFilter prec_filter = null;
    for (int i = 0; i < active_filters.size(); i++)
      if (active_filters.get(i) instanceof PrecedenceFilter)
        prec_filter = (PrecedenceFilter) active_filters.get(i);

    //
    // Get precedence list
    //
    final List precedence = (prec_filter == null) ? new ArrayList()
        : prec_filter.getPrecedence();

    //
    // Get termgroup precedence order and set suppressible flags for
    // termgroups
    //
    final List tgs = new ArrayList();
    final SourceTermType[] sabttys = (SourceTermType[]) precedence
        .toArray(new SourceTermType[0]);
    for (int i = 0; i < sabttys.length; i++) {
      if (!sources_to_remove.contains(sabttys[i].getSource()))
        tgs.add(sabttys[i]);
    }
    release_metadata.setSourceTermTypePrecedence((SourceTermType[]) tgs
        .toArray(new SourceTermType[0]));
        

    //
    // Handle Source metadata
    //
    final SourceMetaDescription[] source_metadata =
        release_metadata.getSourceMetadata();
    for (int i = 0; i < source_metadata.length; i++) {
      if (!source_metadata[i].isCurrentVersion()) {
        source_metadata[i].setInCurrentSubset(false);
        continue;
      }
      if (source_metadata[i].isCurrentVersion() &&
          sources_to_remove.contains(
            Source.getSource(source_metadata[i].getRootSourceAbbreviation()))) {
        source_metadata[i].setInCurrentSubset(false);
      } else
        source_metadata[i].setInCurrentSubset(true);
    }

    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      SwingUtilities.invokeLater(
        new Runnable() {
          public void run() {
            progress.setProgress(1);
          }
        } );
    } else
      System.out.println("["+ new Date() +"] "+"Final Processes 1% Finished");

      //
      // Handle cui mappings
      //
      int p = 0;
      // NOT final because we want to dump references when we are done
      List concept_mappings = release_metadata.getConceptMappingsAsList();
      for (int i = 0; i < concept_mappings.size(); i++) {
        final ConceptMapping this_mapping = (ConceptMapping)concept_mappings.get(i);
        if (!this_mapping.getCui2().equals(""))
          this_mapping.setInCurrentSubset(
              cui_list.inList(this_mapping.getCui2()));

        if (Math.floor((22*i) / concept_mappings.size()) > p) {
          p++;
          final int fp = p;
          if (MMSToolkit.usingView()) {
            SwingUtilities.invokeLater(
              new Runnable() {
                public void run() {
                  progress.setProgress(3+fp);
                }
              }
            );
          } else
            System.out.println("["+ new Date() +"] "+"Final Processes " + (3+p) + " Finished");
        }
      }
     concept_mappings = null;

      //
      // Update progress
      //
      if (MMSToolkit.usingView()) {
        SwingUtilities.invokeLater(
          new Runnable() {
            public void run() {
              progress.setProgress(25);
            }
          } );
      } else
        System.out.println("["+ new Date() +"] "+"Final Processes 25% Finished");

    //
    // keep track of subx rows from subx_cuis list
    //
    release_metadata.setReleaseVersion(default_configuration.getReleaseVersion());
    for (int i = 0; i < subx_cuis.size(); i++) {
      final String scui = (String) subx_cuis.get(i);
      release_metadata.addCUIRemovedFromSubset(scui);

      if (Math.floor((25*i) / subx_cuis.size()) > p) {
        p++;
        final int fp = p;
        if (MMSToolkit.usingView() && progress.getProgress() < 25+p) {
          SwingUtilities.invokeLater(
            new Runnable() {
              public void run() {
                progress.setProgress(25+fp);
              }
            }
          );
        } else
          System.out.println("["+ new Date() +"] "+"Final Processes " + (25+p) + " Finished");
      }
    }
    subx_cuis.clear();

    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      SwingUtilities.invokeLater(
        new Runnable() {
          public void run() {
            progress.setProgress(50);
          }
        } );
    } else
      System.out.println("["+ new Date() +"] "+"Final Processes 50% Finished");

      //
      // Handle cui mappings
      //
      p = 0;
      // NOT final because we want to dump references when we are done
      List atom_mappings = release_metadata.getAtomMappingsAsList();
      for (int i = 0; i < atom_mappings.size(); i++) {
        final AtomMapping this_mapping = (AtomMapping)atom_mappings.get(i);
        if (!this_mapping.getAui2().equals(""))
          this_mapping.setInCurrentSubset(
              cui_list.inList(this_mapping.getCui2(), null, this_mapping.getAui2()));

        if (Math.floor((50*i) / atom_mappings.size()) > p) {
          p++;
          final int fp = p;
          if (MMSToolkit.usingView()) {
            SwingUtilities.invokeLater(
              new Runnable() {
                public void run() {
                  progress.setProgress(3+fp);
                }
              }
            );
          } else
            System.out.println("["+ new Date() +"] "+"Final Processes " + (3+p) + " Finished");
        }
      }
      atom_mappings = null;
      cui_list.clear();
      System.gc();

      //
      // Update progress
      //
      if (MMSToolkit.usingView()) {
        SwingUtilities.invokeLater(
          new Runnable() {
            public void run() {
              progress.setProgress(53);
            }
          } );
      } else
        System.out.println("["+ new Date() +"] "+"Final Processes 53% Finished");


    //
    // Write the release metadata to output stream
    //
    try {
      if (MMSToolkit.usingView()) {
        progress.setLowScale(53);
        progress.setHighScale(80);
      }
      mmsys_output_stream.writeReleaseMetadata(release_metadata);
      release_metadata.clearAtomMappings();
      release_metadata.clearConceptMappings();
      release_metadata = null;
      System.gc();
    } catch (IOException ioe) {
      handleError(ioe, "Error writing release metadata.\n" + ioe.getMessage());
      return;
    }
    try {
      log.write("Subset Release Metadata completed:...." + new Date());
      log.newLine();
    } catch (IOException ioe) {
      handleError(ioe, "Error writing mmsys.log.\n" + ioe.getMessage());
      return;
    }    
    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      SwingUtilities.invokeLater(
        new Runnable() {
          public void run() {
            progress.setProgress(80);
          }
        } );
    } else
      System.out.println("["+ new Date() +"] "+"Final Processes 80% Finished");

    //
    // We are finished, close streams, sort&compute MD5s
    //
    try {
      mmsys_input_stream.close();
      if (MMSToolkit.usingView()) {
        progress.setLowScale(80);
        progress.setHighScale(99);
      }
      mmsys_output_stream.close();
    } catch (IOException ioe) {
      handleError(ioe, "Error closing input/output handlers.\n" + ioe.getMessage());
      return;
    }

    //
    // Finish writing to log
    //
    final Date end_date = new Date();
    String hms = MMSToolkit.timeToString( MMSToolkit.timeDifference(end_date, start_date));
    try {
    log.write("Finished at:.........................." + end_date);
    log.newLine();
    log.write("Concepts in source:..................." + new Integer(cui_counter));
    log.newLine();
    log.write("Concepts in subset:..................." + new Integer(subset_cui_counter));
    log.newLine();
    log.write("Time elapsed:........................." + hms);
    log.newLine();
    log.newLine();

    //
    // write configuration details out to log in attractive format
    //
    String details = mmsys_input_stream.getConfigurationDetails();
    if (details != null && !details.equals("")) {
      log.write("Details about MetamorphoSys Input Handler");
      log.newLine();
      log.write("-------------------------------");
      log.newLine();
      log.write(details);
      log.newLine(); log.newLine();
    }
    details = mmsys_output_stream.getConfigurationDetails();
    if (details != null && !details.equals("")) {
      log.write("Details about MetamorphoSys Output Handler");
      log.newLine();
      log.write("--------------------------------");
      log.newLine();
      log.write(details);
      log.newLine(); log.newLine();
    }

    for (int i = 0; i < active_filters.size(); i++) {
      if (active_filters.get(i) instanceof MetamorphoSysOutputStream)
        continue;
      details = ( (Filter) active_filters.get(i)).getConfigurationDetails();
      if (details != null && !details.equals("")) {
        log.write(details);
        log.newLine(); log.newLine();
      }
    }
    log.close();
  } catch (IOException ioe) {
    handleError(ioe, "Error writing to mmsys.log.\n" + ioe.getMessage());
    return;
  }
    //
    // Save configuration file
    //
    try {
      user_configuration.saveState(
          new File(user_configuration.getSubsetDirectory(), "config.prop"));
    } catch (LocalException le) {
      handleError(le,"Error writing configuration file");
      return;
    }

    //
    // Update progress
    //
    if (MMSToolkit.usingView()) {
      SwingUtilities.invokeLater(
        new Runnable() {
          public void run() {
            progress.setProgress(100);
            progress.setOperation(MMSProgressMonitor.DONE);
          }
        } );
    } else {
      System.out.println("["+ new Date() +"] "+"Final Processes 100% Finished");
    }

    //
    // Show log file
    //
    if (SwingToolkit.usingView()) {
      File log_file = new File(subset_dir, "mmsys.log");
      Object return_val = null;
      // RRF output format, so offer the RRF Browser upon subset completion
      if (!(user_configuration.getOutputStream() instanceof OriginalMRMetamorphoSysOutputStream ))
      	return_val = displayTextFile(WelcomeGUI.getWelcomeGUI(), "MetamorphoSys Subset Log",
                                          log_file, 20, 90);
      // ORF output format, so don't offer RRF Browser
      else 
    	  SwingToolkit.displayTextFile(progress, "MetamorphoSys Subset Log",
                                        log_file, 20, 90);
      // if user chooses Browse button
      if ( return_val != null && ( (String) return_val).startsWith("Browse")) {
        try {
          // make BrowserState to set subset directory to subset just completed
          BrowserState browser_state = BrowserState.getBrowserState();
          browser_state.setDirectoryHistory(new String[] {subset_dir});
          if (new File("config").canWrite()) {
            try {
            BrowserState.save(new File("config", "dir_saved.txt"));
            } catch (IOException ioe) {
              //
              // Treat this as if canWrite() returned true
              // for a DVD RW drive with a read-only DVD in it 
              //
            }
          }

          // set flag so user won't be queried to open the last saved config
          // which would be irrelevant b/c user just made new subset
          ic.setQueryUserToOpenSavedConfig(false);
          progress.dispose();
          
          RRFBrowserGUI rrf_browser_gui = RRFBrowserGUI.newRRFBrowserGUI();

          MMSToolkit.setReleaseDatLocation("../release.dat");
          //rrf_browser_gui = RRFBrowserGUI.getRRFBrowserGUI();
          rrf_browser_gui.setVisible(true);
          rrf_browser_gui.toFront();
          rrf_browser_gui.subsetDirectoryOpened(new File(subset_dir));
        }
        catch (Exception ex) {
          ex.printStackTrace();
          MMSToolkit.reportError(MMSToolkit.getGui(),
              "Unexpected exception.\nSee user console for details.", false);
        }
      } else { // user presses Done
      	MMSToolkit.exitToWelcomeGUI(progress);
      }
    }
  }

  /**
   * Responsible for making sure all data structures properly refer
   * to one another.  For example, if an {@link Atom} is removed from
   * a {@link Concept} any {@link AtomAttribute} connected to that
   * atom should also be removed.
   * @param cui_list {@link CuiSuiAuiList}
   * @param concept {@link Concept}
   */
  private void resolveExternalReferences(final CuiSuiAuiList cui_list, final Concept concept) {

    //
    // Remove Mappings if FROMEXPR CUI is not in cui list (for ATX mappings)
    //
    final Mapping[] mappings = concept.getMappings();
    for(int i = mappings.length-1; i>= 0; i--) {
      final Mapping mapping = mappings[i];
      if (mapping.getFrom() != null &&
          "ATX".equals(mapping.getType()) &&
          !cui_list.inList(mapping.getFrom().getExpression()))
        concept.removeMapping(mapping);
    }

    //
    // Remove RELs if CUI2, AUI2 not in cui list
    //
    final Relationship[] relationships = concept.getRelationships();
    for(int i = relationships.length-1; i >= 0; i--) {
      final Relationship rel = relationships[i];
      if (rel instanceof AtomRelationship) {
        final AtomRelationship atom_rel = (AtomRelationship)rel;
        if (atom_rel.getRelatedAtom() != null &&
            !cui_list.inList(relationships[i].getRelatedConcept().getCui(),
                           atom_rel.getRelatedAtom().getSui(),
                           atom_rel.getRelatedAtom().getAui()))
        concept.removeRelationship(rel);
      } else if (rel.getRelatedConcept() != null &&
                 !cui_list.inList(rel.getRelatedConcept().getCui()))
        concept.removeRelationship(rel);
    }

    //
    // Remove Coocurrences if CUI2, AUI2 not in cui list
    //
    final Coocurrence[] coocurrences = concept.getCoocurrences();
    for (int i = coocurrences.length-1; i >= 0; i--) {
      final Coocurrence coc = coocurrences[i];
      if (coocurrences[i] instanceof AtomCoocurrence) {
        final AtomCoocurrence atom_coc = (AtomCoocurrence)coc;
        if (atom_coc.getRelatedAtom() != null &&
            !coc.getRelatedConcept().getCui().equals("") &&
            !cui_list.inList(coc.getRelatedConcept().getCui(),
                             atom_coc.getRelatedAtom().getSui(),
                             atom_coc.getRelatedAtom().getAui()))
          concept.removeCoocurrence(coc);
      } else if (coc.getRelatedConcept() != null &&
                 (coc.getRelatedConcept().getCui().length() > 0) &&
                 !cui_list.inList(coc.getRelatedConcept().getCui()))
        concept.removeCoocurrence(coc);
    }

    // For contexts from MRHIER.RRF, there is the possibility that
    // atoms in the hierarchy will have been removed (such as due to
    // UTF-8 removal).  Therefore, check each atom in the path to root
    // to make sure it is still in the list.  If not, remove entire context.
    final Context [] contexts = concept.getContexts();
    for (int i=contexts.length-1; i>=0; i--) {
      String [] nodes = contexts[i].getPathToRoot();
      for (int j=0; j<nodes.length; j++) {
      	if (!cui_list.inList(null, null, nodes[j])) {
      		concept.removeContext(contexts[i]);
      	  break;
      	}
      }
    } 
  }

  /**
   * Resolve references within a concept.  Get rid of things
   * attached to atoms/rels no longer around.
   * @param concept {@link Concept}
   */
  private void resolveInternalReferences(final Concept concept) {
    final Set aui_set = new HashSet();

    //
    // Cache list of AUIs, remove those with non-ASCII unicode, if necessary
    //
    final Atom [] atoms = concept.getAtoms();
    try {
      for (int i = 0; i < atoms.length; i++) {
        final Atom atom = atoms[i];
        //
        // Remove atoms if output stream indicates to do so
        // and the string contains non-ASCII unicode characters
        //
        if (user_configuration.getOutputStream().removeUnicode() &&
            MMSToolkit.containsUnicode(atom.getStr())) {
          concept.removeAtom(atom);
        } else
          aui_set.add(atom.getAui());
      }
    } catch (UnsupportedEncodingException uee) {
      // Do nothing
    }

    if (concept.isEmpty()) return;

    //
    // if there are only MTH atoms left, remove all atoms from concept
    // so that the concept will be removed from the target set
    //
    if (user_configuration.getOutputStream().removeMTHOnlyConcepts()) {
      boolean only_mth = true;
      final Atom[] remaining_atoms = concept.getAtoms();
      for (int i = 0; i < remaining_atoms.length; i++) {
        final Atom atom = remaining_atoms[i];
        if (!atom.getSource().getVersionedSourceAbbreviation().equals("MTH")) {
          only_mth = false;
          break;
        }
      }
      if (only_mth) {
        concept.clearAtoms();
        return;
      }
    }

    //
    // Remove definitions if not connected to atom that still exists
    //
    final Definition [] defs = concept.getDefinitions();
    for (int i = defs.length-1; i >= 0; i--) {
      final Definition def = defs[i];
      if (def.getAtom() != null  &&
          def.getAtom().getAui() != null &&
          !aui_set.contains(def.getAtom().getAui()))
        concept.removeDefinition(def);
    }

    //
    // Remove locators if not connected to atom that still exists
    //
    final Locator [] locators = concept.getLocators();
    for (int i = locators.length-1; i >= 0; i--) {
      final Locator locator = locators[i];
      if (locator.getAtom() != null  &&
          locator.getAtom().getAui() != null  &&
          !aui_set.contains(locator.getAtom().getAui()))
        concept.removeLocator(locator);
    }

    //
    // Remove AtomRelationships if connected to atoms that do not exist
    //
    final Relationship [] relationships = concept.getRelationships();
    for (int i = relationships.length-1; i >= 0; i--) {
      if (relationships[i] instanceof AtomRelationship) {
        final AtomRelationship ar = (AtomRelationship)relationships[i];
        if (ar.getAtom() != null &&
            ar.getAtom().getAui() != null &&
            !aui_set.contains(ar.getAtom().getAui()))
          concept.removeRelationship(ar);
      }
    }

    //
    // Remove AtomCoocurrences if connected ot atoms that do not exist
    //
    final Coocurrence [] coocurrences = concept.getCoocurrences();
    for (int i = coocurrences.length-1; i >= 0; i--) {
      if (coocurrences[i] instanceof AtomCoocurrence) {
        final AtomCoocurrence ac = (AtomCoocurrence)coocurrences[i];
        if (ac.getAtom() != null  && ac.getAtom().getAui() != null  &&
            !aui_set.contains(ac.getAtom().getAui()))
          concept.removeCoocurrence(ac);
      }
    }

    final Attribute [] attributes = concept.getAttributes();
    Set sui_code_hs = null;
    for (int i = attributes.length-1; i >= 0; i--) {

      //
      // Remove AtomAttributes if connected to atoms that do not exist
      //
      if (attributes[i] instanceof AtomAttribute) {
        final AtomAttribute aa = (AtomAttribute) attributes[i];
        // additional information beyond what is available in the cui_list is needed
        // if we are using orig style files, so set of sui/code is intialized here
        if (sui_code_hs == null && aa.getAtom().getAui() == null) {
          sui_code_hs = new HashSet();
          Atom[] l_atoms = concept.getAtoms();
          for (int j = 0; j < l_atoms.length; j++) {
            sui_code_hs.add(l_atoms[j].getSui() + l_atoms[j].getCode());
          }
        }
        // if input is orig style files there may be attributes whose sui matches
        // an existing atom but whose sui/code does not
        if (aa.getAtom() != null && aa.getAtom().getAui() == null &&
            !sui_code_hs.contains(aa.getAtom().getSui() + aa.getAtom().getCode()))
          concept.removeAttribute(aa);
      }
    }

    final Context[] contexts = concept.getContexts();
    for (int i = contexts.length-1; i >= 0; i--) {
      final Context cxt = contexts[i];
      // additional information beyond what is available in the cui_list is needed
      // if we are using orig style files, so set of sui/code is intialized here
      if (sui_code_hs == null && cxt.getAtom().getAui() == null) {
        sui_code_hs = new HashSet();
        Atom[] l_atoms = concept.getAtoms();
        for (int j = 0; j < l_atoms.length; j++) {
          sui_code_hs.add(l_atoms[j].getSui() + l_atoms[j].getCode());
        }
      }
      // if input is orig style files there may be contexts whose sui matches
      // an existing atom but whose sui/code does not
      if (cxt.getAtom() != null && cxt.getAtom().getAui() == null &&
          !sui_code_hs.contains(cxt.getAtom().getSui()+cxt.getAtom().getCode()))
        concept.removeContext(cxt);
      else if (cxt.getAtom() != null && cxt.getAtom().getAui() != null &&
               !aui_set.contains(cxt.getAtom().getAui()))
        concept.removeContext(cxt);
    }
  }

  /**
   * Returns the {@link UserConfiguration}.
   * @return the {@link UserConfiguration}
   */
  public UserConfiguration getUserConfiguration() {
    return user_configuration;
  }

  /**
   * Sets the {@link UserConfiguration}.
   * @param uc the {@link UserConfiguration}.
   */
  public void setUserConfiguration(UserConfiguration uc) {
    user_configuration = uc;
  }

  /**
   * Returns the {@link ApplicationConfiguration}.
   * @return the {@link ApplicationConfiguration}
   */
  public ApplicationConfiguration getApplicationConfiguration() {
    return default_configuration;
  }

  /**
   * Sets the {@link ApplicationConfiguration}.
   * @param dc the {@link ApplicationConfiguration}.
   */
  public void setApplicationConfiguration(ApplicationConfiguration dc) {
    default_configuration = dc;
  }

  /**
   * Returns the {@link BufferedWriter} log.
   * @return the {@link BufferedWriter} log.
   */
  public BufferedWriter getLog() { return log; }


  //
  // Static Methods
  //


  /**
   * This method is called when there is a change to the selections
   * on the sources to exclude tab.
   * @param e {@link SourcesToRemoveEvent}
   */
  public void sourcesToRemoveChanged(SourcesToRemoveEvent e) {
    sources_to_remove = e.getSourcesToRemove();
  }

  /**
   * Fires a {@link SourcesToRemoveEvent} to inform all listening
   * filters to respond to the change in the sources to remove list.
   */
  protected void fireSourcesToRemoveChanged() {
    SourcesToRemoveFilter sources_to_exclude_filter = null;
    List active_filters = user_configuration.getActiveFilters();
    for (int i = 0; i < active_filters.size(); i++) {
      Filter filter = (Filter)active_filters.get(i);
      if(filter instanceof SourcesToRemoveFilter)
        sources_to_exclude_filter = (SourcesToRemoveFilter)filter;
    }

    SourcesToRemoveEvent e =
      new SourcesToRemoveEvent(this, sources_to_exclude_filter.getSourcesToRemove());
    sourcesToRemoveChanged(e);
  }

  /**
   * Displays a text file.  This method opens a dialog box displaying the
   * given {@link File}.  It returns upon an OK click or close.
   * @param parent the parent {@link Component}
   * @param title title for the display
   * @param file the message
   * @param rows the number of rows to be displayed at one time without scrolling
   * @param cols the number of columns to be displayed at one time without scrolling
   * @return user selected result
   */
  private static Object displayTextFile(Component parent, String title, File file,
                                     int rows, int cols) {

      BufferedReader reader = null;
      JTextArea text_area = new JTextArea(rows, cols);
      text_area.setEditable(false);
      text_area.setFont(new Font("Courier", Font.PLAIN, 12));
      try {
        reader = new BufferedReader(new FileReader(file));
        String line = "";
        while (true) {
          line = reader.readLine();
          if (line == null) {
            break;
          }
          text_area.append(line);
          text_area.append("\n");
        }
        text_area.setCaretPosition(0);
        reader.close();
      } catch (Exception ex) {
        MMSToolkit.notifyUser(parent, ex.getMessage());
      }

      String[] options  = {"Done", "Browse Subset.."};
      JOptionPane pane = new JOptionPane(new JScrollPane(text_area),
          JOptionPane.QUESTION_MESSAGE,
          JOptionPane.YES_NO_OPTION,
          null,
          options,
          options[1]);

      JDialog dialog = pane.createDialog(parent, title);
      FontSizeManager.adjustFontSize(dialog);
      dialog.pack();
      dialog.setVisible(true);
      return pane.getValue();
  }

  /**
   * Application entry point if no GUI is desired.
   * Running this main method will subset using the
   * MMSYS/config/mmsys.prop.sav configuration file
   * without prompting the user with additional configuration
   * options.
   * @param args {@link String}[]
   */
  public static void main(String[] args) {
    MetamorphoSys metamorphosys = null;
    try {
      // Instantiate Application
      MMSToolkit.setUsingView(false);
      MMSToolkit.setReleaseDatLocation("../release.dat");
      MMSToolkit.fakeMRDOCSuppressRows();
      metamorphosys = new MetamorphoSys();
      metamorphosys.subset();
    } catch (Exception e) {
      e.printStackTrace();
      // kill processes if input stream was using native unzip
      try { metamorphosys.getUserConfiguration().getInputStream().close(); } catch (Exception ex) {}
      MMSToolkit.reportError(MMSToolkit.getGui(),
                             "Unexpected exception.\nSee user console for details.",
                             false);
    }
    MMSToolkit.Exit(0);
  }
}
