/**************************************************************************
 *
 * Package:     gov.nih.nlm.mms
 * Object:      RichMRNLMMetamorphoSysInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.mms;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;

/**
 * Represents a {@link MetamorphoSysInputStream} which are
 * file(s) in the "NLM data file" format.
 *
 * @author Deborah Shapiro
 */
public class NLMFileMetamorphoSysInputStream extends RichMRMetamorphoSysInputStream {

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link NLMFileMetamorphoSysInputStream}.
   */
  public NLMFileMetamorphoSysInputStream() { }

  //
  // Methods
  //

  /**
   * Returns the short name for display purposes.
   * @return the short name for display purposes
   */
  public String getName() {
    return "NLM Data File Format";
  }

  /**
   * Returns a message for the log.
   * @return a message for the log
   */
  public String getConfigurationDetails() {
    return "MetamorphoSys Input: NLM Data File Format";
  }

  /**
   * Validates the source path.  This involves
   * checking that it both exists, is readable and that it has
   * access to MR data.
   * @param source_path path to be validated
   * @return <code>true</code> if validation passes
   *         <code>false</code> otherwise
   */
  public boolean validateSourcePath(String source_path) {
    InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();
    File source_dir = new File(source_path);
    if (!source_dir.exists()) {
      validate_msg = "Source location does not exist.";
      return false;
    }
    if (!source_dir.isDirectory()) {
      validate_msg = "Source location must be a directory.";
      return false;
    }
    File[] files = source_dir.listFiles();
    if (files == null) {
      validate_msg =
              "Invalid source location:" +
              "\n\n" + source_dir +
              "\n\nMust be a directory containing" +
              "\na valid " + ic.getReleaseVersion().toUpperCase() +
              " Metathesaurus NLM data file." +
              "\nPlease choose another source location.";
      return false;
    }
    File meta_nlm_file = null;
    boolean found = false;
    for (int i=0; i<files.length; i++) {
      if (files[i].getAbsolutePath().endsWith("meta.nlm")) {
        meta_nlm_file = files[i];
        found = true;
        break;
      }
    }

    if (!found) {
      validate_msg =
              "Invalid source location:" +
              "\n\n" + source_dir +
              "\n\nMust be a directory containing" +
              "\na valid " + ic.getReleaseVersion().toUpperCase() +
              " Metathesaurus NLM data file." +
              "\nPlease choose another source location.";
      return false;
    }

    //
    // Validate .zip file
    //
    try {
      found = MMSToolkit.archiveContainsDir(meta_nlm_file, "META");
      if (found) return true;
    } catch (Exception e) {
      validate_msg =
              "Invalid source location:" +
              "\n\n" + source_dir +
              "\n\nMust be a directory containing" +
              "\na valid " + ic.getReleaseVersion().toUpperCase() +
              " Metathesaurus NLM data file." +
              "\nPlease choose another source location.";
      return false;
    }
    validate_msg =
            "Invalid source location:" +
            "\n\n" + source_dir +
            "\n\nMust be a directory containing" +
            "\na valid " + ic.getReleaseVersion().toUpperCase() +
            " Metathesaurus NLM data file." +
            "\nPlease choose another source location.";
    return false;
  }

  /**
    * Attempts to determine the version of the release being
    * used and validate it against the version from the properties file.
    * If the version is determined to be a mismatch of if it cannot
    * be determined, <code>false</code> is returned.
    * @param source_path to be validated
   * @return <code>true</code> if validation passes
   *         <code>false</code> otherwise
    */
  public boolean validateVersion(String source_path) {
    InstallConfiguration ic = InstallConfiguration.getInstallConfiguration();

    File[] files = new File(source_path).listFiles();
    File meta_nlm_file = null;
    for (int i=0; i<files.length; i++) {
      if (files[i].getAbsolutePath().endsWith("meta.nlm")) {
        meta_nlm_file = files[i];
        break;
      }
    }

    try {
      //
      // Try unzip -l .zip and look for /dir/
      //
      String unzip_cmd = System.getProperty("unzip.path");
      String[] cmd = new String[] {
          unzip_cmd, "-l", meta_nlm_file.getAbsolutePath()};
      Process p = Runtime.getRuntime().exec(cmd);

      //
      // Read from input stream
      //
      BufferedReader in =
          new BufferedReader(
          new InputStreamReader(
          p.getInputStream()));
      String line;
      String match = "/" + ic.getReleaseVersion().toUpperCase() + "/";
      boolean found = false;
      while ( (line = in.readLine()) != null) {
        if (line.indexOf(match) != -1) {
          found = true;
        }
      }

      //
      // Wait for process to complete
      //
      p.waitFor();

      if (found) return true;

      validate_msg =
          "Warning: MetamorphoSys is not configured to run \n" +
          "against this version of the data, subsetting may be \n" +
          "incorrect. To correct this problem, use the\n" +
          ic.getReleaseVersion().toUpperCase() + " release of the Metathesaurus. ";

    } catch (Exception e) {
      validate_msg =
              "Invalid source location:" +
              "\n\n" + source_path +
              "\n\nMust be a directory containing" +
              "\na valid " + ic.getReleaseVersion().toUpperCase() +
              " Metathesaurus NLM data file." +
              "\nPlease choose another source location.";
    }
    return false;
  }

  /**
   * Indicates whether this input source has full context data, such as that
   * included in a MRCXT file.
   *
   * @return <code>true</code> if this input source has full context data;
   *         <code>false</code> otherwise.
   */
  public boolean hasFullContextData() {
    return false;
  }
}
