/*****************************************************************************
 *
 * Package: gov.nih.nlm.mms
 * Object:  NewConfigAction.java
 *
 *****************************************************************************/
package gov.nih.nlm.mms;

import java.awt.Component;
import java.awt.Event;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.io.File;

import javax.swing.Action;
import javax.swing.KeyStroke;

/**
 * Opens a new configuration file. Used for the
 * <code>File->New Configuration</code> menu of {@link MetamorphoSysGUI}.
 * 
 * @author Brian Carlsen, Deborah Shapiro
 */
public class NewConfigAction extends OpenConfigAction {

  //
  // Constructors
  //
  /**
   * Instantiates a {@link NewConfigAction}.
   * 
   * @param mms
   *          the {@link MetamorphoSys} object
   * @param owner parent {@link Component}
   */
  public NewConfigAction(MetamorphoSys mms, Frame owner) {
    super(mms, owner);

    // configure action
    putValue(Action.NAME, "New Configuration..");
    putValue(Action.SHORT_DESCRIPTION, "New Configuration");
    putValue(Action.MNEMONIC_KEY, new Integer((int) 'N'));
    putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke('N',
        Event.CTRL_MASK));
  }

  /**
   * Opens the specified configuration file.
   * 
   * @param e
   *          an {@link ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {

    // Save the current config first if it has changed
    if (mms.getUserConfiguration().hasDataChanged()) {
      if (MMSToolkit.confirmRequest(owner,
          "You have made changes to this configuration.\n"
              + "Would you like to save the changes?")) {
        super.saveConfig();
      } else {
        backupConfig();
      }
    }

    //
    // Indicate that a new configuration will be used, not a saved one
    //
    InstallConfiguration.getInstallConfiguration().setNewConfiguration(true);

    //
    // Present user with License agreement
    //
    LicenseAgreementGUI license_agreement_gui = new LicenseAgreementGUI(mms,
        (Frame) owner);
    license_agreement_gui.setVisible(true);

    if (license_agreement_gui.isCancelled()) {
      MMSToolkit.cancel();
      cancelled = true;
      return;
    }

    //
    // Have user select the "default configuration"
    //
    DefaultSubsetSelectionGUI default_subset_selection_gui = new DefaultSubsetSelectionGUI(
        mms.getApplicationConfiguration(), mms.getUserConfiguration(),
        (Frame) owner);
    default_subset_selection_gui.setVisible(true);

    if (default_subset_selection_gui.isCancelled()) {
      MMSToolkit.cancel();
      cancelled = true;
      return;
    }

    //
    // Acquire default subset info
    //
    DefaultSubset ds = default_subset_selection_gui.getSelectedDefaultSubset();

    //
    // Set the default config file
    //
    setFileToOpen(new File(mms.getApplicationConfiguration()
        .getConfigDirectory(), ds.getConfigurationFilename()));

    //
    // Open the user config files
    //
    openConfig();

    //
    // Close/open MetamorphoSysGUI
    //
    refreshGui();

    //
    // Override some settings: selected default subset and selected sources list
    //
    mms.getUserConfiguration().setSelectedDefaultSubset(ds);

    //
    // Indicate properties have changed (to initiate a "save" request)
    //
    mms.getUserConfiguration().setPropertiesChanged();

  }

}
