/*****************************************************************************
 *
 * Package: gov.nih.nlm.mms
 * Object:  OpenConfigAction.java
 *
 *****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.mms.filters.SourcesToRemoveFilter;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.awt.Component;
import java.awt.Event;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.swing.Action;
import javax.swing.JFileChooser;
import javax.swing.KeyStroke;
import javax.swing.filechooser.FileFilter;

/**
 * Opens a configuration file. Used for the
 * <code>File->Open Configuration</code> menu
 * of {@link MetamorphoSysGUI}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class OpenConfigAction extends SaveConfigAction {

    //
  // Private Fields
  //
  private File to_open;
  protected boolean cancelled = false;

  //
  // Constructors
  //

  /**
   * Instantiates an {@link OpenConfigAction}.
   * @param mms the {@link MetamorphoSys} object
   * @param owner parent {@link Component}
   */
  public OpenConfigAction(MetamorphoSys mms, Frame owner) {
    super(mms, owner);
    this.mms = mms;
    this.owner = owner;

    // configure action
    putValue(Action.NAME, "Open Configuration..");
    putValue(Action.SHORT_DESCRIPTION, "Open Configuration");
    putValue(Action.MNEMONIC_KEY, new Integer((int)'O'));
    putValue(Action.ACCELERATOR_KEY,
        KeyStroke.getKeyStroke('O', Event.CTRL_MASK));
  }

  /**
   * Sets the file to be opened.
   * @param to_open {@link File} to be opened
   */
  public void setFileToOpen(File to_open) {
    this.to_open = to_open;
  }

  /**
   * Opens the specified configuration file.
   * @param e an {@link ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {

    //
    // Present user with License agreement (if INSTALL)
    //
    if (InstallConfiguration.getInstallConfiguration().isInstallUMLS()) {
      LicenseAgreementGUI license_agreement_gui =
          new LicenseAgreementGUI(mms, (Frame)owner);
      license_agreement_gui.setVisible(true);

      if (license_agreement_gui.isCancelled()) {
        MMSToolkit.cancel();
        cancelled = true;
        return;
      }
    }

    //
    // Save the current config first if it has changed
    //
    if (mms.getUserConfiguration().hasDataChanged()) {
      if (MMSToolkit.confirmRequest(owner,
              "You have made changes to this configuration.\n" +
              "Would you like to save the changes?")) {
        super.saveConfig();
      } else {
        backupConfig();
      }
    }

    //
    // Open the configuration file
    //
    openConfig();

    //
    // If user cancels the file chooser, bail
    //
    if (isCancelled())
      return;

    UserConfiguration user_configuration = mms.getUserConfiguration();
    ApplicationConfiguration default_configuration = mms.getApplicationConfiguration();

    //
    // If no DefaultSubset is selected, choose the first one
    //
    if(user_configuration.getSelectedDefaultSubset() == null)
      user_configuration.setSelectedDefaultSubset(default_configuration.getDefaultSubsets()[0]);

    //
    // Set the selected DefaultSubset in application config to the
    // same thing as in the user config.
    //
    default_configuration.setSelectedDefaultSubset(
      user_configuration.getSelectedDefaultSubset());

    //
    // Indicate that a saved configuration will be used, not a new one
    //
    InstallConfiguration.getInstallConfiguration().setNewConfiguration(false);

    //
    // Refresh the GUI
    //
    refreshGui();

   }

  /**
   * Open the configuration, save if changed.
   */
  protected void openConfig() {

    MetamorphoSysInputStream in = mms.getUserConfiguration().getInputStream();
    String subset_dir = mms.getUserConfiguration().getSubsetDirectory();

    // Loop until "Cancel" or a file is successfully loaded
    UserConfiguration user_configuration = null;
    while (true) {
      if (to_open == null) {
        FileFilter propertiesFileFilter = new PropertiesFileFilter();
        to_open = MMSToolkit.chooseFile(
            "Select and Open the Configuration File", "Open",
            JFileChooser.FILES_ONLY, lastConfigFile.getParentFile(), null,
            propertiesFileFilter, owner);
        if (to_open == null) {
          cancelled = true;
          return; // if user choose "Cancel" on JFileChooser
        }
      }
      lastConfigFile = to_open;
      try {
        user_configuration = new UserConfiguration(to_open);
        user_configuration.setInputStream(in);
        mms.setUserConfiguration(user_configuration);

        // config file is bad - need to choose another
      } catch (LocalException le) {
        MMSToolkit.reportError(le.getMessage(), false);
        to_open = null;
        continue;
      }
      break; // configuration was opened successfully
    }

    //
    // Report differences if from different versions
    //
    String user_config_version = user_configuration.getReleaseVersion();
    String default_config_version = mms.getApplicationConfiguration().getReleaseVersion();
    if (!user_config_version.equals(default_config_version)) {
      MMSToolkit.notifyUser(owner, "You have opened a configuration from the \n" +
                            user_config_version + " release.  The current release\n" +
                            "version is " + default_config_version + ".  Please review the\n" +
                            "UMLS Metathesaurus configuration tabs to ensure\n" +
                            "that all settings are appropriately configured.");

      //
      // Open log file and write intro
      //
      BufferedWriter log = null;
      try {
        log = new BufferedWriter(new FileWriter(new File(subset_dir, "config_rpt.log")));
        log.write("This report can be found in: " + subset_dir + File.separator + "config_rpt.log");
        log.newLine();
        log.newLine();
      } catch (Exception e) {
        e.printStackTrace();
        MMSToolkit.reportError(owner,
            "Exception opening " + subset_dir + File.separator +
            "config_rpt.log: " + e.getMessage(), true);
      }
      Source[] sources = Source.getSources();
      List selected_sources = new ArrayList();
      List selected_rsabs = new ArrayList();
      List missing_sources = new ArrayList();
      List updated_selected_sources = new ArrayList();
      List updated_unselected_sources = new ArrayList();

      // get opened config's selected sources
      SourcesToRemoveFilter strf = new SourcesToRemoveFilter();
      String sources_prop = user_configuration.getProperty(strf, "selected_sources");
      String[] tokens = FieldedStringTokenizer.split(sources_prop, ";");
      for(int i=0; i<tokens.length; i++) {
        String[] f = FieldedStringTokenizer.split(tokens[i],"|");
        selected_sources.add(f[0]);
        selected_rsabs.add(f[1]);
      }

      // check for selected sources in user configuration that have been removed
      for (int i=0; i<selected_sources.size(); i++) {
        if (Source.getSource((String)selected_rsabs.get(i)) == null)
          missing_sources.add(selected_sources.get(i));
      }

      // check for sources that are new or updated
      for (int i=0; i<sources.length; i++) {
        if (sources[i].getImeta().equals(default_config_version)) {
          if (selected_rsabs.contains(sources[i].getRootSourceAbbreviation()))
            updated_selected_sources.add(sources[i].getVersionedSourceAbbreviation());
          else
            updated_unselected_sources.add(sources[i].getVersionedSourceAbbreviation());
        }
      }
      try {
      if (missing_sources.size() == 0) {
        log.write("None of the selected sources in the opened configuration have been removed.");
        log.newLine();
      } else {
        log.write("The following sources that were selected in the opened configuration");
        log.newLine();
        log.write("have been removed and are not in the " + default_config_version + " release version.");
        log.newLine();
        log.newLine();
        Collections.sort(missing_sources);
        for (int i=0; i<missing_sources.size(); i++) {
          log.write((String)missing_sources.get(i));
          log.newLine();
        }
      }
      log.newLine();

      if (updated_selected_sources.size() == 0) {
        log.write("None of the selected sources from the opened configuration are new or updated.");
        log.newLine();
      } else {
        log.write("The following sources are currently selected and are new or have been");
        log.newLine();
        log.write("updated in the " + default_config_version + " release version.");
        log.newLine();
        log.newLine();
        Collections.sort(updated_selected_sources);
        for (int i=0; i<updated_selected_sources.size(); i++) {
          log.write(updated_selected_sources.get(i).toString());
          log.newLine();
        }
      }
      log.newLine();

      if (updated_unselected_sources.size() != 0) {
        log.write("The following sources are NOT currently selected and new or have");
        log.newLine();
        log.write("been updated in the" + default_config_version + " release version.");
        log.newLine();
        log.newLine();
        Collections.sort(updated_unselected_sources);
        for (int i=0; i<updated_unselected_sources.size(); i++) {
          log.write(updated_unselected_sources.get(i).toString());
          log.newLine();
        }
      }

      if (user_configuration.didLoadCauseWarnings()) {
        log.newLine();
        log.newLine();
        log.write(user_configuration.getLoadWarningsReport());
        log.newLine();
      }

      log.flush();
      log.close();
      } catch (Exception ex) {
        ex.printStackTrace();
        MMSToolkit.reportError("Error writing to config_rpt.log.");
        return;
      }
      SwingToolkit.displayTextFile(owner, "Open Old Config File Report",
        new File(subset_dir, "config_rpt.log"));
    }
    return;
  }

  /**
   * Indicates whether or not the file chooser was cancelled.
   * @return <code>true</code> if so; <code>false</code> if not
   */
  public boolean isCancelled() {
    return cancelled;
  }

  /**
   * Refresh the GUI.
   */
  protected void refreshGui() {
    if (MMSToolkit.getGui() != null)
      MMSToolkit.getGui().dispose();
    MetamorphoSysGUI gui = new MetamorphoSysGUI(mms);
    gui.getUserConfiguration().resetDataChanged();
  }
}
