/*****************************************************************************
 * Package: gov.nih.nlm.mms
 * Object:  ValidateDistributionAction.java
 *
 *
 ****************************************************************************/
package gov.nih.nlm.mms;

import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;

/**
 * Used for the <code>File->Validate</code> menu of the {@link WelcomeGUI}.
 * 
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ValidateDistributionAction extends AbstractAction {

  private File source_dir = null;

  /**
   * Instantiates a {@link ValidateDistributionAction}.
   */
  public ValidateDistributionAction() {
    super();
    // configure action
    putValue(Action.NAME, "Validate Distribution");
    putValue(Action.SHORT_DESCRIPTION, "Validate Distribution");
    putValue(
        Action.LONG_DESCRIPTION,
        "Validates Distribution by comparing byte counts and MD5 hash values to a precomputed table.");
  }

  /**
   * Opens a small dialog informing the user about the application.
   * 
   * @param e
   *          an {@link java.awt.event.ActionEvent}
   */
  public void actionPerformed(ActionEvent e) {
    ValidateDistributionGUI validate_gui = new ValidateDistributionGUI();
    validate_gui.setVisible(true);
  }

  //
  // Inner Class
  //
  /**
   * GUI displayed by ValidateDistributionAction
   */
  public class ValidateDistributionGUI extends JFrame implements
      GlassPaneListener {

    private GlassComponent gc = null;

    /**
     * Instantiates the GUI. The constructor does not make it visible.
     */
    public ValidateDistributionGUI() {
      super();

      //
      // Disable welcome gui
      //
      WelcomeGUI.getWelcomeGUI().enableGlassPane();

      //
      // Create glass component and install
      //
      gc = new GlassComponent(this);
      setGlassPane(gc);

      //
      // Set title
      //
      setTitle("Validate Distribution");

      //
      // Configure layout
      //
      getContentPane().setLayout(new BorderLayout());
      GridBagLayout gridbag = new GridBagLayout();
      JPanel choice_panel = new JPanel(gridbag);
      GridBagConstraints constraints = new GridBagConstraints();
      constraints.fill = GridBagConstraints.BOTH;
      constraints.insets = new Insets(2, 2, 2, 2);
      constraints.gridy = 0;
      constraints.gridx = 0;
      constraints.gridwidth = 1;
      constraints.gridy++;

      //
      // Add header
      //
      constraints.gridy++;
      constraints.gridx = 2;
      constraints.gridwidth = 1;

      JTextPane text = new JTextPane();
      text.setEditable(false);
      text.setMinimumSize(new Dimension(200, 50));
      text.setPreferredSize(new Dimension(200, 50));
      text.setBackground(choice_panel.getBackground());
      SimpleAttributeSet set = new SimpleAttributeSet();
      StyleConstants.setAlignment(set, StyleConstants.ALIGN_CENTER);
      StyleConstants.setFontFamily(set, "Times");
      StyleConstants.setFontSize(set, 16);
      StyleConstants.setBold(set, true);
      text.setParagraphAttributes(set, true);
      text.setText("Validate Distribution");

      choice_panel.add(text, constraints);
      constraints.gridy++;

      //
      // add Source Directory Installation Row
      //
      constraints.gridx = 0;
      choice_panel.add(new JLabel("Source"), constraints);

      //
      // add Source JTextField
      //
      constraints.gridx = 2;
      final JTextField source_field = new JTextField(40);
      source_field.setEditable(false);

      //
      // Set default
      //
      try {
        source_dir = new File("..");
        source_field.setText(source_dir.getCanonicalPath());
      } catch (Exception e) {
        source_field.setText("..");
      }
      source_field.setBackground(Color.white);
      gridbag.setConstraints(source_field, constraints);
      choice_panel.add(source_field);

      //
      // add Source Browse JButton
      //
      constraints.gridx = 3;
      JButton source_browse = new JButton("Browse..");
      source_browse.setPreferredSize(new Dimension(100, 22));
      source_browse.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          try {
            File selected_file = null;
            if (!source_field.getText().equals(""))
              selected_file = 
                new File(source_field.getText()).getCanonicalFile();
            source_dir = MMSToolkit.chooseFile(
                "Select and Open the directory to validate", "Open..",
                JFileChooser.DIRECTORIES_ONLY, selected_file, selected_file,
                ValidateDistributionGUI.this);
            if (source_dir != null) {
              source_field.setText(source_dir.getCanonicalPath());
            }
          } catch (Exception ex) {
            ex.printStackTrace();
            MMSToolkit.reportError(WelcomeGUI.getWelcomeGUI(),
                "Unexpected error while setting source directory.\n"
                    + "See console for details.", false);
          }
        }
      });
      gridbag.setConstraints(source_browse, constraints);
      choice_panel.add(source_browse);

      //
      // add Source JLabel comment
      //
      constraints.gridy++;
      constraints.gridx = 2;
      constraints.insets = new Insets(0, 6, 10, 0);
      JLabel comment = new JLabel("(this directory contains UMLS files)");
      choice_panel.add(comment, constraints);
      constraints.insets = new Insets(2, 2, 2, 2);
      constraints.gridy++;

      //
      // Button Panel
      //
      JPanel button_panel = new JPanel(new FlowLayout());

      // Validation Button
      JButton jvalidate = new JButton("Validate");
      jvalidate.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          // enable glass pane first
          enableGlassPane();
          WelcomeGUI.getWelcomeGUI().enableGlassPane();
          Thread t = new Thread(new Runnable() {
            public void run() {
              try {
                final ChecksumValidator cv = 
                  new ChecksumValidator(source_dir,
                                        ValidateDistributionGUI.this);
                cv.validate();
              } catch (Exception e) {
                MMSToolkit.reportError(WelcomeGUI.getWelcomeGUI(),
                    "Unexpected exception validating distribution\n"
                        + "See console for details.", false);
              }
              WelcomeGUI.getWelcomeGUI().disableGlassPane();
            }
          });
          t.start();
          ValidateDistributionGUI.this.dispose();
        }
      });
      button_panel.add(jvalidate);

      // Cancel button
      JButton jcancel = new JButton("Cancel");
      jcancel.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          dispose();
          WelcomeGUI.getWelcomeGUI().disableGlassPane();
          return;
        }
      });
      button_panel.add(jcancel);

      EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
      choice_panel.setBorder(new CompoundBorder(etched_border, new EmptyBorder(
          10, 10, 10, 10)));

      getContentPane().add(choice_panel, BorderLayout.CENTER);
      getContentPane().add(button_panel, BorderLayout.SOUTH);
      this.pack();
      setLocationRelativeTo(null);

      //
      // Configure window listener
      //
      addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
          dispose();
          WelcomeGUI.getWelcomeGUI().disableGlassPane();
          return;
        }
      });

    }

    /**
     * Enable glass pane.
     */
    public void enableGlassPane() {
      gc.setVisible(true);
    }

    /**
     * Disable glass pane.
     */
    public void disableGlassPane() {
      gc.setVisible(false);
    }

  }
}