/*****************************************************************************
 *
 * Object:  BrowserState.java
 * Author:  Deborah Shapiro
 *
 *****************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.Configurable;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.util.ObjectXMLSerializer;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

/**
 * Data repository for historical information
 * on past opened directories and viewed concepts.
 *
 * @author  Deborah Shapiro (dshapiro@apelon.com)
 */
public class BrowserState  {

  private static BrowserState browser_state = null;

  private List dir_hist = new ArrayList();
  private Map dir_to_concept_list = new HashMap();
  private Map prop_map = new HashMap();
  private int font_point_size = 11;
  private String font_family = "Monospaced";


  /**
   * Constructor.
   */
  public BrowserState() {}

  /**
   * Returns the singleton reference to the {@link BrowserState}.
   * @return the singleton reference to the {@link BrowserState}
   */
  public static BrowserState getBrowserState() {
    if (browser_state == null)
      browser_state = new BrowserState();
    return browser_state;
  }

  /**
   * Returns the singleton reference to the {@link BrowserState}.
   * @param file where {@link BrowserState} was written
   * @return the singleton reference to the {@link BrowserState}
   * @throws Exception
   */
  public static BrowserState getBrowserState(File file) throws Exception {
    ObjectXMLSerializer xml_ser = new ObjectXMLSerializer();
    BufferedReader in = new BufferedReader(new InputStreamReader(
        new FileInputStream(file)));
    browser_state = (BrowserState) xml_ser.fromXML(in);
    in.close();
    return browser_state;
  }

  /**
   * Methods
   */
  /**
   * Saves the {@link BrowserState} to the given {@link File}.
   * @param file {@link File}
   * @throws java.lang.Exception
   */
  public static void save(File file) throws Exception {
    if (browser_state == null)
      return;
    ObjectXMLSerializer xml_ser = new ObjectXMLSerializer();
    BufferedWriter pw = new BufferedWriter(new FileWriter(file));
    pw.write(xml_ser.toXML(getBrowserState()));
    pw.close();
  }

  /**
   * Returns the last five opened directories.
   * @return {@link String}[] of directory paths
   */
  public String[] getDirectoryHistory() {
    return (String[])dir_hist.toArray(new String[0]);
  }

  /**
   * Resets the directory history to that given.
   * @param history <code>String</code>[]
   */
  public void setDirectoryHistory(String[] history) {
    dir_hist.clear();
    for (int i=0; i<history.length; i++) {
      dir_hist.add(history[i]);
    }
  }

  /**
   * Returns the map of directories to their {@link ConceptList}s.
   * @return {@link Map}
   */
  public Map getDirectoryToConceptListMap() {
    return dir_to_concept_list;
  }

  /**
   * Resets the {@link Map} of directories to their {@link ConceptList}s.
   * @param map {@link Map}
   */
  public void setDirectoryToConceptListMap(Map map) {
    dir_to_concept_list = map;
  }

  /**
   * Returns the {@link ConceptList} associated with the given
   * directory.
   * @param dir  subset directory of interest
   * @return associated {@link ConceptList}
   */
  public ConceptList getConceptListForDirectory(String dir) {
    return (ConceptList)dir_to_concept_list.get(dir);
  }

  /**
   * Returns the current subset directory.
   * @return <code>String</code> path to subset directory
   */
  public String getCurrentDirectory() {
    int index = dir_hist.size() - 1;
    return (String)dir_hist.get(index);
  }

  /**
   * Returns the list of viewed concepts in the currently
   * opened directory.
   * @return {@link Concept}[] list of concepts
   */
  public ConceptList getCurrentConceptList() {
    return (ConceptList)dir_to_concept_list.get(getCurrentDirectory());
  }

  /**
   * Adds a directory path to the history list of
   * opened directories.
   * @param dir {@link String}
   * @param cl {@link ConceptList}
   */
  public void addDirectory(String dir, ConceptList cl) {
    if (dir_to_concept_list.keySet().contains(dir))
      return;
    if (dir_hist.size() == 5)
      dir_hist.remove(0);
    dir_hist.add(dir);
    dir_to_concept_list.put(dir, cl);
  }

  /**
   * Returns the current font point size.
   * @return <code>int</code> font point size
   */
  public int getFontPointSize() {
    return font_point_size;
  }

  /**
   * Resets the font point size to that given.
   * @param font_point_size <code>int</code>
   */
  public void setFontPointSize(int font_point_size) {
  	this.font_point_size = font_point_size;
  }
  /**
   * Returns the current font family.
   * @return <code>String</code> font family
   */
  public String getFontFamily() {
    return font_family;
  }

  /**
   * Resets the font family to that given.
   * @param font_family <code>String</code>
   */
  public void setFontFamily(String font_family) {
  	this.font_family = font_family;
  }
  
  public void addProperties(Configurable f, Properties props) {
    if (props != null) {
      prop_map.put(f.getClass().getName(), props);
    }
  }

  public Properties getProperties(Configurable f) {
    return (Properties)prop_map.get(f.getClass().getName());
  }
}
