/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.browser;
 * Class:      CuiSearchFinder.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.browser;

import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.umls.io.RichMRFilesSingleConceptInputStream;
import gov.nih.nlm.umls.meta.Concept;

import java.awt.BorderLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.EmptyBorder;

/**
 * Represents a {@link Finder} that directly looks up concepts by CUI values.
 * 
 * @author Deborah Shapiro
 */
public class CuiSearchFinder extends AbstractFinder {

  //
  // Search field
  //
  private JTextField search_field;

  //
  // search button
  //
  private JButton go_button;

  /**
   * Instantiates a {@link CuiSearchFinder}.
   */
  public CuiSearchFinder() {
  }

  /**
   * Returns a {@link JPanel} containing configuration options.
   * 
   * @return a {@link JPanel} containing configuration options
   */
  public JPanel getPanel() {

    // top level panel
    JPanel cui_search_panel = new JPanel();
    cui_search_panel.setBorder(new EmptyBorder(5, 5, 5, 5));
    cui_search_panel.setLayout(new BorderLayout());

    // panel with label, field, button
    JPanel button_panel = new JPanel();
    BorderLayout bl = new BorderLayout(5, 5);
    button_panel.setLayout(bl);
    JLabel search_label = new JLabel("Enter a CUI:");
    button_panel.add(search_label, BorderLayout.NORTH);
    search_field = new JTextField(8);
    search_field.addKeyListener(new KeyAdapter() {
      public void keyTyped(KeyEvent e) {
        char c = e.getKeyChar();
        // hitting ENTER causes search
        if (c == KeyEvent.VK_ENTER) {
          CuiSearchAction csa = new CuiSearchAction();
          csa.actionPerformed(new ActionEvent(this, 0, ""));
        }
      }
    });
    button_panel.add(search_field, BorderLayout.CENTER);
    go_button = new JButton(new CuiSearchAction());

    button_panel.add(go_button, BorderLayout.SOUTH);
    cui_search_panel.add(button_panel, BorderLayout.NORTH);
    return cui_search_panel;
  }

  /**
   * Resets the configuration back to default values. When used with a GUI this
   * method should change the data model behind the GUI and allow
   * <code>setPanelSelections</code> to actually update the GUI.
   */
  public void resetToDefaults() {
    search_field.setText("");
  }

  /**
   * Highlights the text displayed in the cui text field.
   */
  public void highlightText() {
    search_field.requestFocus();
    search_field.setSelectionStart(0);
    search_field.setSelectionEnd(search_field.getText().length());
  }

  /**
   * Returns the name.
   * 
   * @return the name
   */
  public String getName() {
    return "CUI Search";
  }

  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions() {
  }

  /**
   * Indicates whether or not an options menu item should be supplied for
   * advanced configuration options.
   * 
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() {
    return false;
  }

  /**
   * Redraws the text portions of the {@link Finder}.
   */
  public void refresh() {
  	search_field.setFont(new Font(font_family, Font.PLAIN, font_point_size));
  }  
  
  /**
   * Action for the "Search" button.
   */
  private class CuiSearchAction extends AbstractAction {

    /**
     * Instantiates a {@link CuiSearchAction}.
     */
    public CuiSearchAction() {
      super();
      putValue(Action.NAME, "Get Concept..");
      putValue(Action.SHORT_DESCRIPTION, "Cui Lookup");
      putValue(Action.LONG_DESCRIPTION, "Cui Lookup");
    }

    /**
     * Prompts the user for a new subset directory and sets this value in the
     * text field.
     * 
     * @param e
     *          an {@link ActionEvent}
     */
    public void actionPerformed(ActionEvent e) {
      try {
        String cui = search_field.getText().toUpperCase();
        cui = cui.trim();
        // if cui starts with digit do this padding
        if (Character.isDigit(cui.toCharArray()[0])) {
          if (cui.startsWith("C")) {
            cui = cui.substring(1);
          }
          int icui = 0;
          int padcui = 10000000;
          try {
            icui = Integer.parseInt(cui);
            // handle future expansion of CUIs
            if (icui > padcui)
              padcui = 100000000;
          } catch (Exception ex) {
            highlightText();
            MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
                "The entry must be a CUI.", false);
            return;
          }
          cui = "C" + String.valueOf(padcui + icui).substring(1);
        }
        final RichMRFilesSingleConceptInputStream input = new RichMRFilesSingleConceptInputStream(
            cui, user_configuration.getSourcePaths()[0]);
        final Concept c = input.getConceptWithAtomsOnly();
        if (c != null && c.getCui() != null) {
          RRFBrowserGUI.getRRFBrowserGUI().enableGlassPane();
          Thread t = new Thread(new Runnable() {
            public void run() {
              clm.addConcept(c);
              RRFBrowserGUI.getRRFBrowserGUI().disableGlassPane();
            }
          });
          t.start();
        } else
          throw new NullPointerException();
      } catch (Exception ex) {
        highlightText();
        MMSToolkit.reportError(RRFBrowserGUI.getRRFBrowserGUI(),
            "Specified CUI is not in the subset. Try again.", false);
      }
    } // end actionPerformed

  } // end inner class

}
