/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.browser
 * Object:	MRCXTBuilderGUI.java
 *
 **************************************************************************/
package gov.nih.nlm.mms.cxt;

import gov.nih.nlm.mms.AboutAction;
import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.LocalException;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.MetamorphoSysGUI;
import gov.nih.nlm.mms.RichMRMetamorphoSysInputStream;
import gov.nih.nlm.mms.UserConfiguration;
import gov.nih.nlm.mms.WelcomeGUI;
import gov.nih.nlm.swing.GlassComponent;
import gov.nih.nlm.swing.GlassPaneListener;
import gov.nih.nlm.swing.NoEditTableModel;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.io.FileStatisticsConfiguration;
import gov.nih.nlm.umls.io.IndexBuilder;
import gov.nih.nlm.umls.io.InputStreamFileStatisticsConfiguration;
import gov.nih.nlm.umls.io.SuiCountIndexBuilder;
import gov.nih.nlm.umls.meta.MetaDescriptionInitializer;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.util.FileStatistics;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.ProgressMonitor;
import javax.swing.ScrollPaneConstants;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.table.AbstractTableModel;

/**
 * GUI application for building <tt>MRCXT.RRF</tt> from <tt>MRCONSO.RRF</tt>
 * and <tt>MRHIER.RRF</tt>
 * 
 * @author Deborah Shapiro
 */
public class MRCXTBuilderGUI extends JFrame implements GlassPaneListener {

  //
  // References to the configuration objects
  //
  private ApplicationConfiguration default_configuration = null;

  private UserConfiguration user_configuration = null;

  //
  // GUI components
  //
  private JButton go_button;

  private JMenu help_menu = null;

  private JCheckBox build_sibs_checkbox;

  private JCheckBox build_children_checkbox;

  private JCheckBox compute_xc_box;

  private JCheckBox max_contexts_checkbox;

  private JCheckBox versioned_sabs_checkbox;

  private JCheckBox add_unicode_bom_checkbox;

  private JTextField max_contexts_field;

  private JTextField subset_dir_field;

  private SortableJTable source_table;

  //
  // For glass pane
  //
  private GlassComponent gc = null;

  //
  // Static singleton instance
  //
  private static MRCXTBuilderGUI mrcxt_builder_gui = null;

  //
  // For building index files automatically
  //
  private FileStatisticsConfiguration fs_config;

  //
  // Directory of source files
  //
  private File source_dir;

  //
  // handler for building MRCXT.RRF
  //
  private MRCXTBuilderOutputStream output_stream = null;

  //
  // The log
  //
  private BufferedWriter log;
  
  //
  // Constructor
  //

  /**
   * Instantiates a {@link MetamorphoSysGUI} and builds the frame.
   */
  public MRCXTBuilderGUI() {
    super();

    output_stream = new MRCXTBuilderOutputStream();

    try {
      fs_config = new InputStreamFileStatisticsConfiguration(
          (InputStream) MRCXTBuilderGUI.class.getClassLoader().getResource(
              "config/mrpluscolsfiles.dat").getContent());
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError("Unexpected error loading mrpluscolsfiles.dat.",
          true);
    }

    //
    // Handle glass pane
    //
    gc = new GlassComponent(this);
    setGlassPane(gc);

    //
    // Get configuration references
    //

    try {
      default_configuration = new ApplicationConfiguration(new File("config",
          "mmsys.a.prop"));
      user_configuration = new UserConfiguration(new File("config",
          "mmsys.a.prop"));
      user_configuration.setInputStream(new RichMRMetamorphoSysInputStream());
    } catch (LocalException le) {
      le.printStackTrace();
      MMSToolkit.reportError(this, "Failed to load mmsys.a.prop.", false);

      MMSToolkit.exitToWelcomeGUI(this);
      return;
    }

    initialize();

  }

  /**
   * Initializes the GUI.
   */
  protected void initialize() {

    //
    // Compose menu bar
    //
    JMenuBar menu_bar = new JMenuBar();
    setJMenuBar(menu_bar);
    JMenu file_menu = new JMenu("File");
    menu_bar.add(file_menu);
    menu_bar.add(Box.createHorizontalGlue());
    help_menu = new JMenu("Help");
    menu_bar.add(help_menu);

    //
    // File Menu
    //

    // Exit application
    JMenuItem exit_menu_item = new JMenuItem(new ExitMRCXTAction(this));
    exit_menu_item.setAccelerator(KeyStroke.getKeyStroke('Q',
        java.awt.Event.CTRL_MASK, false));
    file_menu.add(exit_menu_item);

    //
    // Help Menu
    //

    // About the application
    JMenuItem about_menu_item = new JMenuItem(new AboutAction(
        "About MRCXT Builder", this));
    help_menu.add(about_menu_item);

    // RRF Browser Documentation menu item
    help_menu.add(MMSToolkit.getURLHelpMenu("MRCXT Builder Documentation  ",
        "http://www.nlm.nih.gov/research/umls/mrcxt_help.html", this));

    //
    // Create base panel for all elements
    //
    JPanel base_panel = new JPanel();
    base_panel.setLayout(new BoxLayout(base_panel, BoxLayout.Y_AXIS));
    FlowLayout f_layout = new FlowLayout(FlowLayout.LEFT, 5, 0);

    //
    // Create RRF Directory widgets
    //
    JPanel rrf_dir_panel = new JPanel(f_layout);
    rrf_dir_panel.add(new JLabel("RRF Directory"));
    subset_dir_field = new JTextField(45);
    subset_dir_field.setEditable(false);
    rrf_dir_panel.add(subset_dir_field);
    rrf_dir_panel.add(new JButton(
        new OpenSubsetAction(user_configuration, this)));
    base_panel.add(rrf_dir_panel);

    //
    // Choices Panel
    //
    JPanel p1 = new JPanel(f_layout);
    build_sibs_checkbox = new JCheckBox("Build Siblings");
    p1.add(build_sibs_checkbox);
    build_sibs_checkbox.setSelected(true);
    base_panel.add(p1);

    JPanel p2 = new JPanel(f_layout);
    build_children_checkbox = new JCheckBox("Build Children");
    p2.add(build_children_checkbox);
    build_children_checkbox.setSelected(true);
    base_panel.add(p2);

    JPanel p3 = new JPanel(f_layout);
    versioned_sabs_checkbox = new JCheckBox(
        "Use Versioned Source Abbreviations");
    p3.add(versioned_sabs_checkbox);
    base_panel.add(p3);

    JPanel p4 = new JPanel(f_layout);
    compute_xc_box = new JCheckBox("Compute XC Flags");
    p4.add(compute_xc_box);
    base_panel.add(p4);

    JPanel p5_panel = new JPanel(f_layout);
    add_unicode_bom_checkbox = new JCheckBox(
        "Prepend UTF-8 BOM characters to MRCXT.RRF");
    p5_panel.add(add_unicode_bom_checkbox);
    base_panel.add(p5_panel);

    JPanel max_cxts_panel = new JPanel(f_layout);

    max_contexts_checkbox = new JCheckBox("Limit Contexts to ");
    max_cxts_panel.add(max_contexts_checkbox);
    max_contexts_checkbox.setSelected(true);
    max_contexts_field = new JTextField("50", 5);
    max_cxts_panel.add(max_contexts_field);
    base_panel.add(max_cxts_panel);

    //
    // Source Table Panel
    //
    JPanel table_panel = new JPanel();
    TitledBorder titled_border = new TitledBorder(new EtchedBorder(
        EtchedBorder.RAISED), "Build Contexts for Selected Sources");
    titled_border.setTitleJustification(TitledBorder.CENTER);
    table_panel.setBorder(titled_border);

    List sources = Source.getSourcesAsList();
    for (int i = 0; i < sources.size(); i++) {
      if (!((Source) sources.get(i)).inCurrentSubset())
        sources.remove(i);
    }
    Source[] sorted_sources = (Source[]) sources.toArray(new Source[0]);
    Arrays.sort(sorted_sources);

    SourcesTableModel source_model = new SourcesTableModel(sorted_sources);  
    source_table = new SortableJTable(source_model);
    
    for (int i = 0; i < source_model.getRowCount(); i++) {
      source_table.addRowSelectionInterval(i, i);
    }
    source_table.getTableHeader().getColumnModel().getColumn(0)
        .setPreferredWidth(300);
    source_table.getTableHeader().getColumnModel().getColumn(1)
        .setPreferredWidth(120);
    source_table.getTableHeader().getColumnModel().getColumn(2)
        .setPreferredWidth(40);
    source_table.getTableHeader().getColumnModel().getColumn(3)
        .setPreferredWidth(60);

    JScrollPane sp = new JScrollPane(source_table,
        ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    sp.setPreferredSize(new Dimension(520, 200));

    table_panel.add(sp);
    base_panel.add(table_panel);

    //
    // Go panel
    //
    JPanel go_panel = new JPanel();
    go_panel.setLayout(new FlowLayout(FlowLayout.CENTER));
    go_button = new JButton("Go");
    go_button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if (source_dir == null)
          MMSToolkit.reportError(MRCXTBuilderGUI.this,
              "A RRF Directory with a valid\n" + "MRHIER.RRF must be opened.",
              false);
        else if (!new File(source_dir, "MRHIER.RRF").exists())
          MMSToolkit.reportError(MRCXTBuilderGUI.this, source_dir
              + " must contain a valid MRHIER.RRF file.", false);
        else if (new File(source_dir, "MRHIER.RRF").length() == 0)
          MMSToolkit.reportError(MRCXTBuilderGUI.this, source_dir
              + System.getProperty("file.separator")
              + "MRHIER.RRF must not be empty.", false);
        else if (source_table.getSelectedRowCount() == 0)
          MMSToolkit.reportError(MRCXTBuilderGUI.this,
              "At least one source must be selected.", false);
        else {
          final File mrcxt_file = new File(source_dir, "MRCXT.RRF");
          if (mrcxt_file.exists()) {
            boolean new_file = MMSToolkit
                .confirmRequest(
                    MRCXTBuilderGUI.this,
                    mrcxt_file.getAbsolutePath()
                        + "\n"
                        + "already exists.  Would you like to create a new MRCXT.RRF?");
            if (!new_file) {
              return;
            }
          }
          Thread runner = new Thread() {
            public void run() {
              //
              // Open log file and write intro
              //
              try {
                log = new BufferedWriter(new FileWriter(new File(source_dir,
                    "mrcxt_builder.log")));
              } catch (Exception e) {
                e.printStackTrace();
                MMSToolkit.reportError("Exception opening mrcxt_builder.log: "
                    + e.getMessage(), true);
              }

              final Date start_date = new Date();
              int mrhier_lines = 0;

              //
              // Print log header
              //
              try {
                log.write("Release Version:......................"
                    + MMSToolkit.getReleaseVersion());
                log.newLine();
                log.write("Release Date:........................."
                    + MMSToolkit.getReleaseDate());
                log.newLine();
                log.write("Release Description:.................."
                    + MMSToolkit.getReleaseDescription());
                log.newLine();
                log
                    .write("RRF Directory:........................"
                        + source_dir);
                log.newLine();
                log
                    .write("Start at:............................."
                        + start_date);
                log.newLine();
                log.flush();

                //
                // write MRCXT.RRF
                //
                mrhier_lines = output_stream.writeMRCXT(MRCXTBuilderGUI.this);
                if (!output_stream.isCancelled()) {
                  log.write("Start sort at:........................"
                      + new Date());
                  log.newLine();
                  output_stream.sortMRCXT(MRCXTBuilderGUI.this);
                }
              } catch (Exception e) {
                new File(source_dir, "MRCXT.RRF").delete();
                handleError(e, "");
              }

              try {
                if (output_stream.isCancelled()) {
                  File f = new File(source_dir, "MRCXT.RRF");
                  f.delete();
                  handleError(null, "Building "
                      + new File(source_dir, "MRCXT.RRF").getAbsolutePath()
                      + " was cancelled.");
                 
                  MMSToolkit.exitToWelcomeGUI(MRCXTBuilderGUI.this);
                  return;
                } else
                  MMSToolkit.exitToWelcomeGUI(MRCXTBuilderGUI.this);                  

                final Date end_date = new Date();
                String hms = MMSToolkit.timeToString(MMSToolkit.timeDifference(
                    end_date, start_date));
                log.write("Finished at:.........................." + end_date);
                log.newLine();
                log.write("Time elapsed:........................." + hms);
                log.newLine();
                log.write("Contexts written......................"
                    + mrhier_lines);
                log.newLine();
                log.write("Run parameters selected:");
                log.newLine();
                if (build_sibs_checkbox.isSelected()) {
                  log.write("Siblings were built.");
                  log.newLine();
                } else {
                  log.write("Siblings were not built.");
                  log.newLine();
                }
                if (build_children_checkbox.isSelected()) {
                  log.write("Children were built.");
                  log.newLine();
                } else {
                  log.write("Children were not built.");
                  log.newLine();
                }
                if (compute_xc_box.isSelected()) {
                  log.write("XC flags were computed.");
                  log.newLine();
                } else {
                  log.write("XC flags were not computed.");
                  log.newLine();
                }
                if (versioned_sabs_checkbox.isSelected()) {
                  log.write("Versioned source abbreviations were used.");
                  log.newLine();
                } else {
                  log.write("Root source abbreviations were used.");
                  log.newLine();
                }
                if (add_unicode_bom_checkbox.isSelected()) {
                  log.write("Unicode Byte Order Mark was inserted.");
                  log.newLine();
                } else {
                  log.write("Unicode Byte Order Mark was not inserted.");
                  log.newLine();
                }
                if (max_contexts_checkbox.isSelected()) {
                  log.write("Contexts were limited to "
                      + max_contexts_field.getText() + ".");
                  log.newLine();
                } else {
                  log.write("Contexts were not limited.");
                  log.newLine();
                }
                if (source_table.getSelectedRowCount() == source_table
                    .getRowCount()) {
                  log.write("All sources were included.");
                  log.newLine();
                } else {
                  log.write("\nThe following sources were included:");
                  log.newLine();
                  int[] selected_rows = source_table.getSelectedRows();
                  for (int i = 0; i < selected_rows.length; i++) {
                  	log.write((String)source_table.getValueAt(selected_rows[i],0));
                    log.newLine();
                  }
                }

                log.flush();
                log.close();
              } catch (Exception e) {
                e.printStackTrace();
                MMSToolkit
                    .reportError("Exception writing to mrcxt_builder.log: "
                        + e.getMessage(), true);
              }
              MMSToolkit.notifyUser(MRCXTBuilderGUI.this, "Building "
                  + new File(source_dir, "MRCXT.RRF").getAbsolutePath()
                  + " is complete.");

              SwingToolkit.displayTextFile(MRCXTBuilderGUI.this,
                  "MRCXT Builder Log",
                  new File(source_dir, "mrcxt_builder.log"));

              MMSToolkit.exitToWelcomeGUI(MRCXTBuilderGUI.this);
            }
          };
          runner.start();
        }
      }
    });
    go_panel.add(go_button);

    //
    // Cancel button
    //
    JButton jcancel = new JButton("Cancel");
    jcancel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        MMSToolkit.exitToWelcomeGUI(MRCXTBuilderGUI.this);
        return;
      }
    });

    //
    // Add OK, Cancel buttons
    //
    //JPanel button_panel = new JPanel(new FlowLayout());
    go_button.setPreferredSize(jcancel.getPreferredSize());
    go_panel.add(go_button);
    go_panel.add(jcancel);    
    
    base_panel.add(go_panel);
    EtchedBorder etched_border = new EtchedBorder(EtchedBorder.RAISED);
    base_panel.setBorder(new CompoundBorder(etched_border, new EmptyBorder(10,
        10, 10, 10)));

    //
    // Configure frame
    //
    setTitle("MRCXT Builder " + default_configuration.getReleaseVersion());
    setResizable(true);
    setContentPane(base_panel);

    //
    // Configure window listener
    //
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
      	MMSToolkit.exitToWelcomeGUI(MRCXTBuilderGUI.this);
        return;
      }
    });

    //
    // pack and make visible
    //
    pack();
    setLocationRelativeTo(null);
    getRootPane().setDefaultButton(go_button);

  }

  /**
   * Handles opening of a source directory.
   * 
   * @param source_dir
   *          the source directory
   */
  public void subsetDirectoryOpened(File source_dir) {
    this.source_dir = source_dir;
    try {
      File file = new File(source_dir, "config.prop");
      if (file.exists())
        user_configuration = new UserConfiguration(file);
      else {
        user_configuration = new UserConfiguration(new File("config",
            "mmsys.a.prop"));
      }
      user_configuration.setInputStream(new RichMRMetamorphoSysInputStream());

      MetaDescriptionInitializer.initialize(new File(source_dir, "MRDOC.RRF"),
          new File(source_dir, "MRSAB.RRF"), null, null);

      //
      // Update source list
      //
      final List sources = Source.getSourcesAsList();
      for (int i = 0; i < sources.size(); i++) {
        if (!((Source) sources.get(i)).getSabin().equals("Y"))
          sources.remove(i);
      }
      final Source[] sorted_sources = (Source[]) sources.toArray(new Source[0]);
      Arrays.sort(sorted_sources);
      final String[][] display_sources = new String[sorted_sources.length][4];
      for (int i = 0; i < display_sources.length; i++) {
        display_sources[i][0] = sorted_sources[i].getOfficialName();
        display_sources[i][1] = sorted_sources[i]
            .getVersionedSourceAbbreviation();
        display_sources[i][2] = sorted_sources[i].getRestrictionLevel();
        display_sources[i][3] =
            new Integer(sorted_sources[i].getCfr()).toString();
      }

      //
      // Change Data & select all rows
      //
      final NoEditTableModel model = new NoEditTableModel(display_sources,
          new Object[] {"Full Source Name", "Source Abbreviation", "Level", "Concepts"});
      source_table.setModel(model);
      for (int i = 0; i < display_sources.length; i++) {
        source_table.addRowSelectionInterval(i, i);
      }

    } catch (Exception le) {
      le.printStackTrace();
      MMSToolkit.reportError(this,
          "Unexpected error loading subset configuration.\n "
              + "The most likely problem is that you opened a \n "
              + "directory containing a subset from a previous version \n"
              + "of the Metathesaurus. \n"
              + "This may cause unexpected behavior. \n"
              + "See console for more details.", false);
    }
    user_configuration.setSourcePaths(new String[] { source_dir
        .getAbsolutePath() });

    setTitle("Rich Release Format MRCXT Builder "
        + default_configuration.getReleaseVersion());

    subset_dir_field.setText(source_dir.getAbsolutePath());

    buildIndexes();
  }

  private void buildIndexes() {

    //
    // Check AUI_MRCONSO.x
    //
    File index_file = new File(new File(source_dir, "indexes"), "AUI_MRCONSO.x");
    if (!index_file.exists()) {
      MMSToolkit.notifyUser(MRCXTBuilderGUI.this,
          "The browser has detected a missing index file\n"
              + "in your subset.  AUI_MRCONSO.x will now be built.");
      final IndexBuilder index_builder = new IndexBuilder(user_configuration
          .getSourcePaths()[0]);
      final FileStatistics mrconso_stats = fs_config
          .getFileStatistics("MRCONSO");
      final ProgressMonitor pm = new ProgressMonitor(this,
          "Building AUI_MRCONSO.x index", "Starting...", 0, 100);
      try {
        index_builder.buildIndex("MRCONSO", new String[] { "AUI" },
            mrconso_stats, pm);
      } catch (Exception e) {
        if (e.getMessage().indexOf("cancelled") != -1)
          MMSToolkit.reportError("Index build was cancelled.\n"
              + "Browser requires indexes to run properly.\n", true);
        e.printStackTrace();
        MMSToolkit.reportError("Problem building index.\n"
            + "Browser requires indexes to run properly.\n", true);
      }
      pm.setProgress(100);
    }

    //
    // PTR_AUI_RELA_MRHIER.x
    //
    index_file = new File(new File(source_dir, "indexes"),
        "PTR_AUI_RELA_MRHIER.x");
    if (!index_file.exists()) {
      MMSToolkit.notifyUser(MRCXTBuilderGUI.this,
          "The browser has detected a missing index file\n"
              + "in your subset.  PTR_AUI_RELA_MRHIER.x will now be built.");
      final IndexBuilder index_builder = new IndexBuilder(user_configuration
          .getSourcePaths()[0]);
      final FileStatistics mrhier_stats = fs_config.getFileStatistics("MRHIER");
      final ProgressMonitor pm = new ProgressMonitor(this,
          "Building PTR_AUI_RELA_MRHIER.x index", "Starting...", 0, 100);

      pm.setMillisToDecideToPopup(0);
      try {
        index_builder.buildIndex("MRHIER",
            new String[] { "PTR", "AUI", "RELA" }, mrhier_stats, pm);
      } catch (Exception e) {
        if (e.getMessage().indexOf("cancelled") != -1)
          MMSToolkit.reportError("Index build was cancelled.\n"
              + "Browser requires indexes to run properly.\n", true);
        e.printStackTrace();
        MMSToolkit.reportError("Problem building index.\n"
            + "Browser requires indexes to run properly.\n", true);

      }
      pm.setProgress(100);
    }

    //
    // SUI_CT_MRXW.x
    //
    index_file = new File(new File(source_dir, "indexes"), "SUI_CT_MRXW.x");
    if (!index_file.exists()) {
      MMSToolkit.notifyUser(MRCXTBuilderGUI.this,
          "The browser has detected a missing index file\n"
              + "in your subset.  SUI_CT_MRXW.x will now be built.");
      final SuiCountIndexBuilder index_builder = new SuiCountIndexBuilder(
          user_configuration.getSourcePaths()[0]);
      final ProgressMonitor pm = new ProgressMonitor(this,
          "Building SUI_CT_MRXW.x index", "Starting...", 0, 100);

      pm.setMillisToDecideToPopup(0);
      try {
        index_builder.buildSuiCountWordIndex(pm);
      } catch (Exception e) {
        if (e.getMessage().indexOf("cancelled") != -1)
          MMSToolkit.reportError("Index build was cancelled.\n"
              + "Browser requires indexes to run properly.\n", true);
        e.printStackTrace();
        MMSToolkit.reportError("Problem building index.\n"
            + "Browser requires indexes to run properly.\n", true);

      }
      pm.setProgress(100);
    }

  }

  /**
   * Indicates if context siblings should be written to MRCXT.RRF.
   * 
   * @return <code>true</code> if siblings should be built <code>false</code>
   *         otherwise
   */
  public boolean getBuildSiblings() {
    return build_sibs_checkbox.isSelected();
  }

  /**
   * Indicates if context children should be written to MRCXT.RRF.
   * 
   * @return <code>true</code> if children should be built <code>false</code>
   *         otherwise
   */
  public boolean getBuildChildren() {
    return build_children_checkbox.isSelected();
  }

  /**
   * Indicates if the XC flag should be computed for MRCXT.RRF.
   * 
   * @return <code>true</code> if yes <code>false</code> otherwise
   */
  public boolean getComputeXC() {
    return compute_xc_box.isSelected();
  }

  /**
   * Indicates the maximum number of contexts that will be written to MRCXT.RRF
   * for any given atom/sab. Returns -1 if the maximum number of contexts is
   * unlimited.
   * 
   * @return <code>int</code> maximum contexts
   */
  public int getMaxContexts() {
    if (!max_contexts_checkbox.isSelected()
        || max_contexts_field.getText().equals(""))
      return -1;
    else
      return new Integer(max_contexts_field.getText()).intValue();
  }

  /**
   * Indicates if the UTF8 Byte Order Mark (BOM) should be prepended to the
   * output MRCXT.RRF.
   * 
   * @return <code>true</code> if BOM is required <code>false</code> if not
   */
  public boolean getAddUnicodeBOM() {
    return add_unicode_bom_checkbox.isSelected();
  }

  /**
   * Indicates if MRCXT.RRF should be written with versioned or versionless
   * source abbreviations.
   * 
   * @return <code>true</code> if versioned <code>false</code> if
   *         versionless
   */
  public boolean getVersionedSabs() {
    return versioned_sabs_checkbox.isSelected();
  }

  /**
   * Returns the {@link File} where the MRHIER.RRF will be found and the
   * MRCXT.RRF should be written.
   * 
   * @return {@link File} directory
   */
  public File getSourceDir() {
    return source_dir;
  }

  /**
   * Returns {@link List} of {@link Source}s that have been selected to be
   * included in MRCXT.RRF.
   * 
   * @return {@link List} of included {@link Source}s
   */
  public List getIncludedSources() {
    List list = new ArrayList();
    int[] source_rows = source_table.getSelectedRows();
    for (int i = 0; i < source_rows.length; i++) {
      list.add(Source.getSource(source_table.getValueAt(source_rows[i], 1)
          .toString()));
    }
    return list;
  }

  /**
   * Handle error.
   * 
   * @param e
   *          {@link Exception}
   * @param msg
   *          error string
   */
  private void handleError(Exception e, String msg) {

    if (!msg.equals(""))
      MMSToolkit.reportError(this, msg, false);

    if (e != null)
      MMSToolkit.printStackTrace(e, log);
    try {
      log.newLine();
      log.write(msg);
      log.newLine();
      log.flush();
      log.close();

      user_configuration.getInputStream().close();
    } catch (Exception ex) {
      MMSToolkit.reportError("Unable to write to mrcxt_builder.log.");
    }
    SwingToolkit.displayTextFile(this, "MRCXT Builder Log", new File(
        source_dir, "mrcxt_builder.log"));

    return;
  }

  //
  // GlassPaneListener Implementation
  //

  /**
   * Enable glass pane.
   */
  public void enableGlassPane() {
    gc.setVisible(true);
  }

  /**
   * Disable glass pane.
   */
  public void disableGlassPane() {
    gc.setVisible(false);
  }
  
  private class SourcesTableModel extends AbstractTableModel {
    //
    // Private Fields
    //
    final private String[] column_names = {"Full Source Name", "Source Abbreviation", "Level", "Concepts"};

    final private Class[] column_classes = { String.class, String.class,
       Integer.class, Integer.class };

    private List sources = new ArrayList();

    //
    // Constructor
    //
    /**
     * This constructor gets the initial values for the sources table.
     * @param sorted_sources initial values for sources table
     */
    public SourcesTableModel(Source[] sorted_sources) {
    	for (int i=0; i<sorted_sources.length; i++)
        sources.add(sorted_sources[i]);
    }

    //
    // Methods
    //
    /**
     * Returns the number of columns in the sources to exclude table.
     * 
     * @return <code>int</code> number of columns
     */
    public int getColumnCount() {
      return column_names.length;
    }

    /**
     * Returns the number of rows in the sources to exclude table.
     * 
     * @return <code>int</code> number of rows
     */
    public int getRowCount() {
      return sources.size();
    }

    /**
     * Returns the name of a column in the sources table
     * 
     * @param column_number
     *          <code>int</code> number of the column in question
     * @return {@link String} name of the column in position column_number
     */
    public String getColumnName(int column_number) {
      return column_names[column_number];
    }

    /**
     * Returns the value at the cell of the sources table.
     * 
     * @param row
     *          <code>int</code> row number of the cell
     * @param column
     *          <code>int</code> column number of the cell
     * @return the value at the cell at row and column
     */
    public Object getValueAt(int row, int column) {
      Source s = (Source) sources.get(row);
      if (column == 0)
        return s.getOfficialName();
      else if (column == 1)
        return s.getVersionedSourceAbbreviation();
      else if (column == 2)
        return new Integer(s.getRestrictionLevel());
      else if (column == 3)
        return new Integer(s.getCfr());
      else
      	return null;
    }

    /**
     * Returns the {@link Class} of the objects in the given column.
     * 
     * @param column_number
     *          <code>int</code> indicates the column in question
     * @return the {@link Class} of the column
     */
    public Class getColumnClass(int column_number) {
      return column_classes[column_number];
    }

  }
  
  /**
   * Application entry point.
   * 
   * @param args
   *          command line arguments
   */
  public static void main(String[] args) {
    // Instantiate Application
    try {
      MMSToolkit.setUsingView(true);
      MMSToolkit.setReleaseDatLocation("../release.dat");

      mrcxt_builder_gui = new MRCXTBuilderGUI();
      mrcxt_builder_gui.setVisible(true);
      mrcxt_builder_gui.toFront();
    } catch (Exception e) {
      e.printStackTrace();
      MMSToolkit.reportError(mrcxt_builder_gui,
          "Unexpected exception.\nSee user console for details.", false);
    }
  }

}
