/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms;
 * Interface:  AbstractFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.CuiList;
import gov.nih.nlm.mms.GUIConfigurable;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.UserConfiguration;
import gov.nih.nlm.swing.DataChangeEvent;
import gov.nih.nlm.swing.DataChangeListener;
import gov.nih.nlm.umls.meta.Concept;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JPanel;

/**
 *
 * This is an abstract implementation of the filter class
 * that manages the mechanisms for handling data changes.
 *
 * @author Brian Carlsen
 *
 **/

public abstract class AbstractFilter implements Filter {

  //
  // flag indicating if data has changed
  //
  private boolean data_changed = false;

  //
  // User configuration
  //
  protected UserConfiguration user_configuration;
  protected ApplicationConfiguration default_configuration;

  //
  // State variables for SubsetListener
  //
  protected boolean init_cui = false, subset_content = false;

  //
  // Listener list
  //
  private List listeners = new ArrayList();

  //
  // Methods
  //

  /**
   * Method provides a {@link JPanel} that will be added to the main GUI
   * help menu to allow the user to consult a help window
   * that is particular to this filter.
   * @return {@link JPanel} to be added to the Help menu
   */
  public JPanel getHelpPanel() {
    return new JPanel();
  }

  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions() {
    MMSToolkit.notifyUser("No Advanced Options");
  };

  /**
   * Indicates whether or not an options menu item should be supplied
   * for advanced configuration options.
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() { return false; }

  /**
   * Enables filters to inform listeners of data changed event.
   */
  protected void fireDataChanged() {
    for (int i = 0; i < listeners.size(); i++)
      ((DataChangeListener)listeners.get(i)).dataChanged(
         new DataChangeEvent(this));
    data_changed = true;
  }

  /**
   * Enables filters to indicate that data has changed without
   * notifying the listeners or making this change undoable.
   */
  protected void fireNotUndoableDataChanged() {
    data_changed = true;
  }

  /**
   * Method returns a boolean indicating whether or not the values
   * in this filter have changed.
   * @return <code>true</code> if data has changed
   *         <code>false</code> otherwise
   */
  public boolean hasDataChanged() {
    return data_changed;
  }

  /**
   * Method allows the data_changed value to be set from the
   * main gui from the undo/redo mechanism.
   * @param b <code>boolean</code>
   */
  public void setDataChanged(boolean b)  {
    data_changed = b;
  }

  /**
   * Method resets the data_changed flag in the filter to indicate that
   * no data has been changed since the last save.
   */
  public void resetDataChanged() {
    data_changed = false;
  }

  /**
   * Method adds a {@link DataChangeListener} that will
   * be notified every time a {@link #fireDataChanged()} call is made
   * from a filter.  Allows main GUI to record all changes to
   * the filters in the undo/redo mechanism.
   * @param dcl {@link DataChangeListener}
   */
  public void addDataChangeListener(DataChangeListener dcl) {
    listeners.add(dcl);
  }

  /**
   * Method removes a {@link DataChangeListener} so that it
   * will no longer be notified every time a {@link #fireDataChanged()}
   * call is made from a filter.
   * @param dcl {@link DataChangeListener}
   */
  public void removeDataChangeListener(DataChangeListener dcl) {
    listeners.remove(dcl);
  }

  /**
   * Method provides the filter with values from the user configuration
   * file.
   * @param c {@link UserConfiguration}
   */
  public void setConfiguration(UserConfiguration c) {
    user_configuration = c;
  }

  /**
   * Method provides the filter with values from the default configuration
   * file.
   * @param c {@link UserConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_configuration = c;
  }

  /**
   * This method is the default method for filters that don't
   * check for validity and it returns true indicating a valid
   * filter.
   * @return <code>true</code> filter is valid
   */
  public boolean validate() {
    return true;
  }

  /**
   * This method is the default message for filters that don't
   * check for validity.  This won't be called on a valid filter.
   * @return {@link String}
   */
  public String getValidateMessage() {
    return "";
  }

  /**
   * Returns the messages indicating why validation failed.  Should only be
   * called if {@link #validate()} returns <code>false</code>.
   * @return the failure messages
   */
  public String[] getValidationErrors() {
    return new String[0];
  }

  /**
   * Returns warnings that were indicated during validation.
   * @return the warning messages
   */
  public String[] getValidationWarnings() {
    return new String[0];
  }

  /**
   * Returns the {@link GUIConfigurable} whose tab should next be displayed
   * to allow the user to correct an error or warning.
   * @return GUIConfigurable to be displayed
   */
  public GUIConfigurable getGUIConfigurableForValidationFailure() {
    return null;
  }

  /**
   * Provides filter access to the cui list.
   * @param cui_list {@link CuiList}
   */
  public void setCuiList(CuiList cui_list) {
  }

  /**
   * Informs {@link gov.nih.nlm.mms.SubsetListener} that the Initialize Cui List phase is starting.
   */
  public void beginInitializeCuiList() {
    init_cui = true;
  }

  /**
   * Informs {@link gov.nih.nlm.mms.SubsetListener} that the Initialize Cui List phase is ending.
   */
  public void endInitializeCuiList() {
    init_cui = false;
  }

  /**
   * Informs {@link gov.nih.nlm.mms.SubsetListener} that the Subset Content phase is starting.
   */
  public void beginSubsetContent() {
    subset_content = true;
  }

  /**
   * Informs {@link gov.nih.nlm.mms.SubsetListener} that the Subset Content phase is ending.
   */
  public void endSubsetContent() {
    subset_content = false;
  }

  /**
   * Returns the {@link Color} used to represent this filter.
   * Not used yet.
   * @return {@link Color}
   */
  public Color getColor() { return Color.black; }
  
  /**
   * Method which removes a given {@link Concept} or its components
   * from the release.  Called during the subsetting process.
   * @deprecated Use one of the sub-interfaces instead
   * @param concept {@link Concept}
   */
  public void applyFilter(Concept concept) {}

  /**
   * Compares this object with the specified object for order.  Returns a
   * negative integer, zero, or a positive integer as this object is less
   * than, equal to, or greater than the specified object.<p>
   * <p/>
   * Required method for implementing {@link Comparable} interface.
   * Indicates the means for comparison between two {@link Filter}s.
   * @param o the Object to be compared.
   * @return a negative integer, zero, or a positive integer as this object
   *         is less than, equal to, or greater than the specified object.
   * @throws ClassCastException if the specified object's type is not
   *         {@link GUIConfigurable}.
   */
  public int compareTo(Object o) {
    if (!(o instanceof GUIConfigurable)) {
      throw new ClassCastException();
    }
    return getName().compareTo(((GUIConfigurable) o).getName());
  }
}

