/***************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	AttributeTypesToRemoveListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.ListDialog;
import gov.nih.nlm.swing.SortableJTable;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * List selection listener for the Attribute Types To Exclude table.
 * Initializes and updates selections on the the attributes table.
 * Makes selections based on user's options settings.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class AttributeTypesToRemoveListSelectionListener
  implements ListSelectionListener, IgnorableListener {
  //
  // Private Fields
  //
  private SortableJTable table;
  private AttributeTypesToRemoveFilter filter;
  private boolean[] selection_before;
  private boolean ignore_events = false;

  //
  // Constructors
  //
  /**
   * Constructor - sets variables for parameters.
   * @param filter {@link AttributeTypesToRemoveFilter}
   * @param table {@link SortableJTable}
   */
  public AttributeTypesToRemoveListSelectionListener(
    AttributeTypesToRemoveFilter filter, SortableJTable table) {
    this.table = table;
    this.filter = filter;
  }

  /**
   * Sets selections on the {@link SortableJTable}
   * based on the attribute_types_to_remove data structure in the
   * {@link AttributeTypesToRemoveFilter}.
   * Records initial set of selected attribute types in order to determine
   * at later points whether selections have changed.
   * @param atr {@link List} of {@link AttType}s to be excluded from subset
   */
  public void setSelections(List atr) {
    setIgnoreEvents(true);
    table.clearSelection();

    selection_before = new boolean[table.getModel().getRowCount()];

    for(int i=0; i< atr.size(); i++) {
      AttType s = (AttType)atr.get(i);

      int row = filter.getRowForAttributeType(s);
      if(row != -1) {
        table.addRowSelectionInterval(row,row);
        selection_before[table.mapIndex(row)] = true;
      }
    }
    setIgnoreEvents(false);
  }

  //
  // Methods
  //
  /**
   * Method that recognizes a {@link ListSelectionEvent} and responds
   * by going through logic to determine if other {@link AttType}
   * rows should be selected as well.
   * @param e {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {

    if(e.getValueIsAdjusting()) return;

    // Only process events we care about
    if(ignoreEvents()) {
      return;
    }

    filter.fireDataChanged();

    //logic to avoid calling up multiple confirm_atttypes
    //dialog boxes when user doesn't Cntrl + click and only one
    //source row is selected
    int single_row = -1;
    if(table.getSelectedRowCount() == 1){
      single_row = table.getSelectedRow();
    }

    HashSet sources_seen = new HashSet();

    for(int i = e.getFirstIndex(); i<=e.getLastIndex(); i++) {

      AttType this_att= filter.getAttributeTypeAtRow(i);

      //Set current selection for this row and update filter
      boolean selected = table.isRowSelected(i);
      if(selected)
	filter.addAttributeTypeToRemove(this_att);
      else
	filter.removeAttributeTypeToRemove(this_att);

      //Determine if selection changed
      boolean selection_changed =
        (selection_before[table.mapIndex(i)] != selected);

      //Set selection_before for the next event
      selection_before[table.mapIndex(i)] = selected;


      HashSet set = new HashSet();
      Object[] related_atts = null;
      // Do not continue if selection for this row didn't change.
      // single_row variable prevents dialog box from
      // being opened repeatedly when a single row is clicked
      // thereby keeping all the attribute types previously selected

      if (selection_changed && (single_row == -1 || single_row == i)) {

        set.addAll(filter.getAttributeTypes(this_att.getSource()));

	if(set.contains(this_att))
	    set.remove(this_att);

	// related_atts should be given all sources in set b/c if
	// automatic selection is chosen, user will not be given a choice
	// of which related sources to select
	related_atts = set.toArray();

   	if(!filter.getAutomaticSelection() &&
	   !sources_seen.contains(this_att.getSource())) {

	  // get confirm_atttypes set
	  AttType[] confirm_atttypes =
	      (AttType[])set.toArray(new AttType[0]);
	  Arrays.sort(confirm_atttypes);

	  // if you are removing rel types and all related rows are already
	  // selected, don't raise the message
	  // likewise if you are keeping rel types and all related rows are
	  // already unselected, don't raise the message
	  boolean raise_message = false;
	  if(table.isRowSelected(i)) {
	    for(int j=0; j<confirm_atttypes.length; j++) {
	      int k = filter.getRowForAttributeType(confirm_atttypes[j]);
	      if(!table.isRowSelected(k))
	        raise_message = true;
	    }
	  } else {
	    for(int j=0; j<confirm_atttypes.length; j++) {
	      int k = filter.getRowForAttributeType(confirm_atttypes[j]);
	      if(table.isRowSelected(k))
	        raise_message = true;
	    }
	  }

	  if (!set.isEmpty() && raise_message) {
	    String message = "";
            if((table.isRowSelected(i) && filter.removeSelectedAttributes()) ||
              (!table.isRowSelected(i) && !filter.removeSelectedAttributes())) {
	      message = "<html><b>You have selected " +
                this_att.getSource().getRootSourceAbbreviation() + "|" +
                this_att.getType() +
                " for removal, the following attribute types</b><br> <b> typically should be removed with "
                + this_att.getSource().getRootSourceAbbreviation() + "|"
                + this_att.getType() +
                ".  Choose which ones should also be removed.</b></ul>";
	    } else {
	      message = "<html><b>You have chosen to keep attribute type " +
                this_att.getSource().getRootSourceAbbreviation() + "|" +
                this_att.getType() +
                ".  The following attribute types </b><br> <b> typically should be kept with " +
                this_att.getSource().getRootSourceAbbreviation() + "|" +
                this_att.getType() + ".  Choose which ones should also be kept.</b></ul>";
	    }

	    related_atts = ListDialog.showListMultipleMode(MMSToolkit.getGui(),
              message, "Confirm Selection", confirm_atttypes, null );
	  }

	  // Cancel was pressed in dialog or none were chosen
	  if (related_atts == null) continue;
	}

	sources_seen.add(this_att.getSource());

        //remove or maintain from the table based on dialog box results
        for(int j = 0; j < related_atts.length; j++) {

          int k = filter.getRowForAttributeType((AttType)related_atts[j]);

            // Unselect if selected
            if (k>-1 && table.isRowSelected(k) && !selected) {
              // prevent dialog box from being raised for each
              // of the other sources with the same rel type
              selection_before[table.mapIndex(k)] = false;
              removeAttributeTypeToRemove(k);
              // Select if unselected
            } else if(k>-1 && !table.isRowSelected(k) && selected) {
              selection_before[table.mapIndex(k)] = true;
              addAttributeTypeToRemove(k);
	    }
	}
      } // end (if selection changed or only a single row is selected)
    } // end from (first index, last index, etc.)
  }

  /**
   * Method returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code>
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Method that turns off responses to the {@link ListSelectionEvent}s
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }

  /**
   * Method selects row and adds {@link AttType} on that row to the list of
   * attribute types to be excluded.
   * @param row int indicating row in table to be added
   */
  private void addAttributeTypeToRemove(int row) {
    setIgnoreEvents(true);
    table.addRowSelectionInterval(row, row);
    filter.addAttributeTypeToRemove(filter.getAttributeTypeAtRow(row));
    setIgnoreEvents(false);
  }

  /**
   * Method deselects row and removes {@link AttType} on that row from the
   * list of attribute types to be excluded.
   * @param row int indicating row in table to be removed from list
   */
  private void removeAttributeTypeToRemove(int row) {
    setIgnoreEvents(true);
    table.removeRowSelectionInterval(row, row);
    filter.removeAttributeTypeToRemove(filter.getAttributeTypeAtRow(row));
    setIgnoreEvents(false);
  }
}

