/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	RelType.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.umls.meta.Source;

import java.util.StringTokenizer;

/**
 * Class contains all fields relevant to a relationship type.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class RelType implements Comparable {
  //
  // Private Fields
  //
  private Source source;
  private String rel_type;  //relationship type

  //
  // Constructors
  //
  /**
   * This is the default no-argument constructor
   */
  public RelType() {};

  /**
   * This constructor takes a {@link String} in the form
   * sab|type.
   * @param line {@link String} containing all fields needed for
   * a {@link RelType}
   */
  public RelType(String line) {
    String[] members = new String[2];
    StringTokenizer tokens = new StringTokenizer(line, "|");
    int i=0;
    while(tokens.hasMoreTokens()) {
      members[i++] = tokens.nextToken();
    }
    Source source = Source.getSource(members[0]);
    setSource(source);
    setType(members[1]);
  }

  /**
   * This constructor takes initial values for each of the fields in
   * the {@link RelType}.
   * @param sab {@link String} initial value for the source abbreviation
   * @param type {@link String}
   */
  public RelType(String sab, String type) {
    setSource(Source.getSource(sab));
    setType(type);
  }

  /**
   * This constructor takes initial values for each of the fields in
   * the {@link RelType}.
   * @param source {@link Source}
   * @param type {@link String}
   */
  public RelType(Source source, String type) {
    setSource(source);
    setType(type);
  }

  //
  // Methods
  //
  /**
   * Get hashcode based on source abbreviation and relationship type
   * @return hash code for this {@link RelType}
   */
  public int hashCode() {
    return (source.getRootSourceAbbreviation() + rel_type).hashCode();
  }

  /**
   * This method returns true if the parameter has a
   * source abbreviation and a relationship type that have the same
   * values as those in this {@link RelType}.
   * @param o  The {@link Object} to be compared with this {@link RelType}
   * @return <code>boolean</code> <code>true</code> if parameter equals this
   * {@link RelType}
   *   		     <code>false</code> if parameter is not equal to this
   * {@link RelType}
   */
  public boolean equals(Object o) {
    RelType rel_type = (RelType)o;
    if((rel_type.getSource().getRootSourceAbbreviation().equals(
      this.getSource().getRootSourceAbbreviation())) &&
       (rel_type.getType().equals(this.getType())))
      return true;
    else
      return false;
  }

  /**
   * This method compares the {@link RelType} with a parameter
   * {@link Object} based on the source abbreviation.
   * @param o  The {@link Object} to be compared with this {@link RelType}
   * @return int
   *
   *		 0 if the parameter is not a {@link RelType}
   */
  public int compareTo(Object o) {
    if(o instanceof RelType)
      return getSource().getRootSourceAbbreviation().compareTo(
        ((RelType)o).getSource().getRootSourceAbbreviation());
    else
      return 0;
  }

  /**
   * Returns a {@link String} representation of the {@link RelType}
   * that includes the two fields that make the {@link RelType} unique.
   * @return {@link String} representation of the {@link RelType}
   */
  public String toString() {
     return(this.getSource().getRootSourceAbbreviation() +
       "|" + this.getType());
  }

  //
  // Accessor Methods
  //
  /**
   * This method returns the name of the {@link Source} in this
   * {@link RelType}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * This method sets the {@link Source} of the {@link RelType}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
  }

  /**
   * This method returns the type of the {@link RelType}.
   * @return {@link String} representation of the type
   */
  public String getType() {
    return rel_type;
  }

  /**
   * This method sets the type of the {@link RelType}.
   * @param type {@link String} value for the type
   */
  public void setType(String type) {
    this.rel_type = type;
  }

}

