/***************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	RelationshipTypesToRemoveListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.ListDialog;
import gov.nih.nlm.swing.SortableJTable;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * List selection listener for the Relationship Types To Exclude table.
 * Initializes and updates selections on the the relationships table.
 * Makes selections based on user's options settings.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class RelationshipTypesToRemoveListSelectionListener
  implements ListSelectionListener, IgnorableListener {
  //
  // Private Fields
  //
  private SortableJTable table;
  private RelationshipTypesToRemoveFilter filter;
  private boolean[] selection_before;
  private boolean ignore_events = false;

  //
  // Constructors
  //
  /**
   * Initializes selections on the sources that will be excluded.
   * @param filter {@link RelationshipTypesToRemoveFilter}
   * @param table {@link SortableJTable}
   */
  public RelationshipTypesToRemoveListSelectionListener(
    RelationshipTypesToRemoveFilter filter, SortableJTable table) {
    this.table = table;
    this.filter = filter;
  }

  /**
   * Sets selections on the {@link SortableJTable}
   * based on the relationship_types_to_remove data structure in the
   * {@link RelationshipTypesToRemoveFilter}.
   * Records initial set of selected relationship types in order to determine
   * at later points whether selections have changed.
   * @param rtr {@link List} of relationship types to be excluded from subset
   */
  public void setSelections(List rtr) {
    setIgnoreEvents(true);
    table.clearSelection();

    selection_before = new boolean[table.getModel().getRowCount()];

    for(int i=0; i< rtr.size(); i++) {
      RelType s = (RelType)rtr.get(i);

      int row = filter.getRowForRelationshipType(s);
      if(row != -1) {
        table.addRowSelectionInterval(row,row);
        selection_before[table.mapIndex(row)] = true;
      }
    }
    setIgnoreEvents(false);
  }

  //
  // Methods
  //
  /**
   * Method that recognizes a {@link ListSelectionEvent} and responds
   * by going through logic to determine if other relationship type rows should
   * be selected as well.
   * @param e {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {

    if(e.getValueIsAdjusting()) return;

    // Only process events we care about
    if(ignoreEvents()) {
      return;
    }

    filter.fireDataChanged();

    //logic to avoid calling up multiple confirm_reltypes
    //dialog boxes when user doesn't Cntrl + click and only one
    //source row is selected
    int single_row = -1;
    if(table.getSelectedRowCount() == 1){
      single_row = table.getSelectedRow();
    }

    HashSet sources_seen = new HashSet();

    for(int i = e.getFirstIndex(); i<=e.getLastIndex(); i++) {

      RelType this_rel= filter.getRelationshipTypeAtRow(i);

      //Set current selection for this row and update filter
      boolean selected = table.isRowSelected(i);
      if(selected)
	filter.addRelationshipTypeToRemove(this_rel);
      else
	filter.removeRelationshipTypeToRemove(this_rel);

      //Determine if selection changed
      boolean selection_changed =
        (selection_before[table.mapIndex(i)] != selected);

      //Set selection_before for the next event
      selection_before[table.mapIndex(i)] = selected;


      HashSet set = new HashSet();
      Object[] related_rels = null;
      // Do not continue if selection for this row didn't change.
      // single_row variable prevents dialog box from
      // being opened repeatedly when a single row is clicked
      // thereby keeping all the relationship types previously selected

      if (selection_changed && (single_row == -1 || single_row == i)) {

        set.addAll(filter.getRelationshipTypes(this_rel.getSource()));

	if(set.contains(this_rel))
	    set.remove(this_rel);

	// related_rels should be given all sources in set b/c if
	// automatic selection is chosen, user will not be given a choice
	// of which related sources to select
	related_rels = set.toArray();

   	if(!filter.getAutomaticSelection() &&
	   !sources_seen.contains(this_rel.getSource())) {

	  // get confirm_reltypes set
	  RelType[] confirm_reltypes =
	      (RelType[])set.toArray(new RelType[0]);
	  Arrays.sort(confirm_reltypes);

	  // if you are removing rel types and all related rows are already
	  // selected, don't raise the message
	  // likewise if you are keeping rel types and all related rows are
	  // already unselected, don't raise the message
	  boolean raise_message = false;
	  if(table.isRowSelected(i)) {
	    for(int j=0; j<confirm_reltypes.length; j++) {
	      int k = filter.getRowForRelationshipType(confirm_reltypes[j]);
	      if(!table.isRowSelected(k))
	        raise_message = true;
	    }
	  } else {
	    for(int j=0; j<confirm_reltypes.length; j++) {
	      int k = filter.getRowForRelationshipType(confirm_reltypes[j]);
	      if(table.isRowSelected(k))
	        raise_message = true;
	    }
	  }

	  if (!set.isEmpty() && raise_message) {
	    String message = "";
            if((table.isRowSelected(i) && filter.removeSelectedRelationships()) ||
              (!table.isRowSelected(i) && !filter.removeSelectedRelationships())) {
	      message = "<html><b>You have selected " +
               this_rel.getSource().getRootSourceAbbreviation() + "|" +
               this_rel.getType() +
               " for removal, the following relationship types</b><br> <b> typically should be removed with " +
               this_rel.getSource().getRootSourceAbbreviation() + "|" +
               this_rel.getType() +
               ".  Choose which ones should also be removed.</b></ul>";
	    } else {
	      message = "<html><b>You have chosen to keep relationship type " +
              this_rel.getSource().getRootSourceAbbreviation() + "|" +
              this_rel.getType() +
              ".  The following relationship types </b><br> <b> typically should be kept with " +
              this_rel.getSource().getRootSourceAbbreviation() + "|" +
              this_rel.getType() +
              ".  Choose which ones should also be kept.</b></ul>";
	    }

	    related_rels =
            ListDialog.showListMultipleMode(MMSToolkit.getGui(), message,
              "Confirm Selection", confirm_reltypes, null );
	  }

	  // Cancel was pressed in dialog or none were chosen
	  if (related_rels == null) continue;
	}

	sources_seen.add(this_rel.getSource());

        //remove or maintain from the table based on dialog box results
        for(int j = 0; j < related_rels.length; j++) {

          int k =
            filter.getRowForRelationshipType((RelType)related_rels[j]);

            // Unselect if selected
            if (k>-1 && table.isRowSelected(k) && !selected) {
              // prevent dialog box from being raised for each
              // of the other sources with the same rel type
              selection_before[table.mapIndex(k)] = false;
              removeRelationshipTypeToRemove(k);
              // Select if unselected
            } else if(k>-1 && !table.isRowSelected(k) && selected) {
              selection_before[table.mapIndex(k)] = true;
              addRelationshipTypeToRemove(k);
	    }
	}
      } // end (if selection changed or only a single row is selected)
    } // end from (first index, last index, etc.)
  }

  /**
   * Method returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code>
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Method that turns off responses to the {@link ListSelectionEvent}s.
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }

  /**
   * Method selects row and adds {@link RelType} on that row to the list of
   * {@link RelType}s to be excluded.
   * @param row <code>int</code> indicating row to be excluded
   */
  private void addRelationshipTypeToRemove(int row) {
    setIgnoreEvents(true);
    table.addRowSelectionInterval(row, row);
    filter.addRelationshipTypeToRemove(filter.getRelationshipTypeAtRow(row));
    setIgnoreEvents(false);
  }

  /**
   * Method deselects row and removes {@link RelType} on that row from
   * the list of {@link RelType}s to be excluded.
   * @param row <code>int</code> indicating row to be included
   */
  private void removeRelationshipTypeToRemove(int row) {
    setIgnoreEvents(true);
    table.removeRowSelectionInterval(row, row);
    filter.removeRelationshipTypeToRemove(filter.getRelationshipTypeAtRow(row));
    setIgnoreEvents(false);
  }
}

