/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  SemanticTypesToRemoveFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.HierarchicalSemanticType;
import gov.nih.nlm.umls.meta.SemanticType;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.event.ListSelectionEvent;

/**
 *
 * This Filter removes all CUIs with a given semantic type
 *
 * @author Deborah Shapiro
 *
 **/

public class SemanticTypesToRemoveFilter extends AbstractFilter
  implements ConceptSelector {

  //
  // Flag to ignore valueChanged events
  //
  private boolean ignore_events = false;

  //
  // List of semantic_types to remove
  //
  private List semantic_types_to_remove = new ArrayList();

  //
  // Reference to the table (used by resetToDefaults)
  //
  private SemanticTypesToRemoveJTable semantic_type_table = null;

  private boolean remove_selected_stys = true;

  //
  // Indicates if any or all sys need to be removed to remove the concept or
  // indicates if any or all stys need to be included to include the concept
  //
  private boolean any_sty = true;

  private JPanel panel = null;
  private JPanel full_panel = null;
  private JRadioButton remove_stys = null;
  private JRadioButton keep_stys = null;

  //
  // Constructor
  //

  /**
   * Instantiates a {@link SemanticTypesToRemoveFilter}.
   */
  public SemanticTypesToRemoveFilter() { }

  //
  // Methods
  //

  /**
   * Returns the GUI panel that will display the filter's
   * screen as one of the tabs on the main GUI window.  The code can
   * assume that this method is called after
   * {@link #setConfiguration(ApplicationConfiguration)} and
   * {@link #setProperties(Properties)}.  There is no data
   * change as a result of calling this method.
   * @return {@link JPanel} that displays graphical interface
   * for the filter
   */
  public JPanel getPanel(){
    if (full_panel != null)
      return full_panel;

    full_panel = new JPanel();


    JPanel toggle_panel = new JPanel(new BorderLayout());
    remove_stys =
        new JRadioButton("Select semantic types to EXCLUDE from subset");
    keep_stys =
        new JRadioButton("Select semantic types to INCLUDE in subset");
    ButtonGroup group = new ButtonGroup();
    group.add(remove_stys);
    group.add(keep_stys);

    remove_stys.setSelected(removeSelectedSemanticTypes());
    remove_stys.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          setRemoveSelectedSemanticTypes(true);
          MMSToolkit.getGui().resetTitleForFilter(panel, "Semantic Types to Exclude");
        }
      }
    );

    toggle_panel.add(remove_stys, BorderLayout.CENTER);

    keep_stys.setSelected(!removeSelectedSemanticTypes());
    keep_stys.addActionListener(
      new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          setRemoveSelectedSemanticTypes(false);
          MMSToolkit.getGui().resetTitleForFilter(panel, "Semantic Types to Include");
        }
      }
    );

    toggle_panel.add(keep_stys, BorderLayout.SOUTH);

    semantic_type_table =
      new SemanticTypesToRemoveJTable(this);
    if (removeSelectedSemanticTypes())
      panel = MMSToolkit.makeCard(semantic_type_table, " Semantic Types to Exclude ");
    else
      panel = MMSToolkit.makeCard(semantic_type_table, " Semantic Types to Include ");
    full_panel.setLayout(new BorderLayout());
    full_panel.add(toggle_panel, BorderLayout.NORTH);
    full_panel.add(panel, BorderLayout.CENTER);
    return full_panel;

  }

  /**
   * Applies filter state to panel selections.
   */
  public void setPanelSelections() {
    semantic_type_table.setSelections(getSemanticTypesToRemove());

    if(removeSelectedSemanticTypes()) {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Semantic Types to Exclude");
      remove_stys.setSelected(true);
    } else {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Semantic Types to Include");
      keep_stys.setSelected(true);
    }
  }

  /**
   * Returns the description.
   * @return the description
   */
  public String getTabDescription() {
     return "Select one or more semantic types below.  Your customized " +
         "Metathesaurus subset will exclude or include any concepts assigned to " +
         "selected semantic types.";
 }

  /**
   * Resets state to default values.
   */
  public void resetToDefaults() {
    setSemanticTypesToRemove(new ArrayList());
    setAnySty(true);
    setRemoveSelectedSemanticTypes(true);
  }

  /**
   * Show Options GUI.
   */
  public void showOptions() {
    final SemanticTypesToRemoveOptionsGUI sem_types_options_gui =
      new SemanticTypesToRemoveOptionsGUI(this);
    sem_types_options_gui.setVisible(true);
  }

  /**
   * Indicates whether or not an options menu item should be supplied
   * for advanced configuration options.
   * @return <code>true</code>
   */
  public boolean hasAdvancedOptions() { return true; }

  /**
   * Returns the help {@link JPanel}.
   * @return the help {@link JPanel}
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(
      this.getClass().getClassLoader().getResource(
      "config/SemanticTypesToRemoveHelp.html"));
  }

  /**
   * Returns the filter state as a {@link Properties} object.
   * @return the filter state as a {@link Properties} object
   */
  public Properties getProperties() {
    //
    // Rebuild semantic_types_to_remove property
    //
    final Properties new_filter_props = new Properties();
    final List ltr = getSemanticTypesToRemove();
    final StringBuffer line = new StringBuffer();
    for (int i = 0; i < ltr.size(); i++) {
      line.append(((HierarchicalSemanticType)ltr.get(i)).getName());
      if (i < ltr.size() - 1)
	line.append(";");
    }
    new_filter_props.setProperty("selected_semantic_types", line.toString());
    new_filter_props.setProperty(
      "any_sty", new Boolean(anySty()).toString());

    new_filter_props.setProperty("remove_selected_stys",
      new Boolean(removeSelectedSemanticTypes()).toString());

    return new_filter_props;
  }

  /**
   * Sets the filter state via a {@link Properties} object.
   * @param props the {@link Properties} filter state
   */
  public void setProperties(Properties props) {

  	// make backwards compatible by still recognizing "remove_any_sty" property
    if (props.containsKey("any_sty") || props.containsKey("remove_any_sty")) {
      if (props.containsKey("any_sty"))
    	  setAnySty(new Boolean((String)props.get("any_sty")).booleanValue());
      else
      	setAnySty(new Boolean((String)props.get("remove_any_sty")).booleanValue());
    }

    if(props.containsKey("remove_selected_stys")) {
      setRemoveSelectedSemanticTypes(new Boolean((String)props.get(
        "remove_selected_stys")).booleanValue());
    }
    //
    // Get the selected_semantic_types property
    // make backward compatible with semantic_types_to_remove
    //
    if(props.containsKey("semantic_types_to_remove") ||
       props.containsKey("selected_semantic_types")) {

      //
      // Clear existing settings
      //
      semantic_types_to_remove = new ArrayList();

      // Set semantic_types to remove from property
      String selected_prop = "";
      if (props.containsKey("selected_semantic_types"))
        selected_prop = (String)props.get("selected_semantic_types");
      else
        selected_prop = (String)props.get("semantic_types_to_remove");
      final StringTokenizer tokens = new StringTokenizer(selected_prop, ";");
      while(tokens.hasMoreTokens()){
        String sem_type = tokens.nextToken();
        HierarchicalSemanticType local_hst =
            HierarchicalSemanticType.getHierarchicalSemanticType( sem_type);
        if (local_hst != null)
          addSemanticTypeToRemove(local_hst);
      }
    }
  }

  /**
   * Indicates if {@link Concept} should be selected to be kept for the
   * subset.
   * @param concept {@link Concept} to be evaluated for removal
   * @return <code>boolean</code> <code>true</code> if concept should be kept
   *          <code>false</code> otherwise
   */
  public boolean keep(Concept concept) {
    //
    // Get semantic types from concept
    //
    final SemanticType[] semantic_types_in_concept = concept.getSemanticTypes();

    if (removeSelectedSemanticTypes()) {
    //
    // Look through all concept STYs
    //
    for (int i = 0; i < semantic_types_in_concept.length; i++) {

      //
      // Try to find a match in STYs to remove
      //
      boolean found = false;
      for (int j = 0; j < semantic_types_to_remove.size(); j++) {

        //
        // Match found
        //
        if (semantic_types_in_concept[i].getSemanticType().getUi().equals(
            ((HierarchicalSemanticType)semantic_types_to_remove.get(j)).getUi())) {

          //
          // If "remove any", clear the concept and be done
          //
          if (any_sty) {
            //concept.clear();
            return false;
          }

          //
          // Otherwise break out of loop
          //
          found = true;
          break;
        }
      }

      //
      // If there is a concept STY not matching one on the list to remove
      // AND we are not doing a "remove any" check then we are done
      //
      if (!found && !any_sty)
        return true;
    }

    //
    // If either any STY had matched in ANY mode
    // or any STY had not matched in ALL mode, we would've
    // exited.  So the only option here is that we are in
    // all mode and all STYs matched or we are in ANY mode
    // and NO STYs matched.
    //
    if (!any_sty)
      //concept.clear();
      return false;
  }
  // selected semantic types are to be included in subset
  else {
     //
     // Look through all concept STYs
     //
     for (int i = 0; i < semantic_types_in_concept.length; i++) {

       //
       // Try to find a match in STYs to keep
       //
       boolean found = false;
       for (int j = 0; j < semantic_types_to_remove.size(); j++) {
         //
         // Match found
         //
         if (semantic_types_in_concept[i].getSemanticType().getUi().equals(
             ((HierarchicalSemanticType)semantic_types_to_remove.get(j)).getUi())) {
           found = true;
           break;
         }
       }

       //
       // If matching STY found in list to keep and "keep any"
       //
       if (found && any_sty) {
         return true;
       }

       //
       // If there is a concept STY not matching one on the list to keep
       // AND we are doing a "keep all" check then we clear concept
       //
       if (!found && !any_sty) {
         //concept.clear();
         return false;
       }
     }

     //
     // "keep any" but no STYs were on the list to keep - clear concept
     //
     if (any_sty) {
       //concept.clear();
       return false;
     }

     //
     // If we are doing "keep all" and all have been found
     //
     return true;
   }
   return true;
  }

  /**
   * Returns the name.
   * @return the name
   */
  public String getName() {
    return "Semantic Types List";
  }

  /**
   * Returns the configuration details for <code>mmsys.log</code>.
   * @return the configuration details for <code>mmsys.log</code>
   */
  public String getConfigurationDetails() {
    final StringBuffer sb = new StringBuffer();
    final String ls = System.getProperty("line.separator");
    if (removeSelectedSemanticTypes())
      sb.append("Concepts with the following Semantic Types were excluded from the subset").append(ls);
    else
      sb.append("Concepts with the following Semantic Types were included in the subset").append(ls);    final Iterator iter = semantic_types_to_remove.iterator();
    boolean empty_list = true;
    while (iter.hasNext()) {
      final HierarchicalSemanticType st = (HierarchicalSemanticType)iter.next();
      String tui = st.getUi();
      sb.append("  " +
	       tui +
	       "                                                                    ".substring(0,68 - tui.length()) +
	       st.getSty()).append(ls);
      empty_list = false;
    }

    if(empty_list)
      sb.append("  <none>").append(ls);
    return sb.toString();
  }

  /**
   * Adds a {@link HierarchicalSemanticType} to the data structure that holds
   * the list of semantic_types to be excluded.
   * @param semantic_type the {@link HierarchicalSemanticType}
   */
  public void addSemanticTypeToRemove(HierarchicalSemanticType semantic_type) {
    if(!semantic_types_to_remove.contains(semantic_type))
      semantic_types_to_remove.add(semantic_type);
  }

  /**
   * Removes a {@link HierarchicalSemanticType} from the data structure that
   * holds the list of semantic_types to be excluded.
   * @param semantic_type the {@link HierarchicalSemanticType}
   */
  public void removeSemanticTypeToRemove(HierarchicalSemanticType semantic_type) {
    semantic_types_to_remove.remove(semantic_type);
  }

  /**
   * Indicates whether or not the specified {@link HierarchicalSemanticType}
   * is of the type to be removed.
   * @param semantic_type the {@link HierarchicalSemanticType}
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean isSemanticTypeToRemove(HierarchicalSemanticType semantic_type) {
    if(semantic_type == null)
      return false;
    return semantic_types_to_remove.contains(semantic_type);
  }

  /**
   * Returns eturns the {@link List} containing all
   * {@link HierarchicalSemanticType}s to be excluded.
   * @return semantic_types_to_remove the {@link List}
   */
  public List getSemanticTypesToRemove() {
    return semantic_types_to_remove;
  }

  /**
   * Sets the {@link List} of {@link HierarchicalSemanticType}s to exclude.
   * @param str {@link List} of {@link HierarchicalSemanticType}s to exclude
   */
  public void setSemanticTypesToRemove(List str) {
    semantic_types_to_remove.clear();
    for(int i=0; i<str.size(); i++) {
      final HierarchicalSemanticType s = (HierarchicalSemanticType)str.get(i);
      addSemanticTypeToRemove(s);
    }
  }

  /**
   * Returns the {@link HierarchicalSemanticType} with the given
   * {@link String} representation.
   * @param sem_type the {@link String}
   * @return the corresponding {@link HierarchicalSemanticType}
   */
  public HierarchicalSemanticType getSTYForString(String sem_type) {
    for(int i=0; i<semantic_types_to_remove.size(); i++) {
      if(sem_type.equals(((HierarchicalSemanticType)semantic_types_to_remove.get(i)).getSty()))
	return ((HierarchicalSemanticType)semantic_types_to_remove.get(i));

    }
    return null;
  }

  /**
   * Indicates whether the automatic_selection
   * option has been chosen on the MetamorphoSys Options dialog box.
   * Automatic selection allows the application to select semantic types
   * sharing the same hierarchy without prompting the user.
   * @return automatic_selection <code>true</code> or
   *				 <code>false</code>
   */
  public boolean getAutomaticSelection() {
    return user_configuration.getAutomaticSelection();
  }

  /**
   * Indicates whether any or all of a concepts semantic types
   * must be on the removal list for the concept to be removed.  The user can
   * set this option on the Advanced Semantic Types to Exclude Options dialog box.
   * @return <code>boolean</code> <code>true</code> if concept will be removed
   *   if any of its terms have a semantic type to be removed
   *   <code>false</code> if concept will be removed only if all of its terms
   *   have semantic types on the removal list
   */
  public boolean anySty() {
    return any_sty;
  }

  /**
   * Sets whether any or all of a concepts semantic types
   * must be on the removal list for the concept to be removed.  The user can
   * set this option on the Advanced Semantic Types to Exclude Options dialog box.
   * @param b <code>boolean</code> <code>true</code> if concept should be removed
   *   if any of its terms have a semantic type to be removed
   *   <code>false</code> if concept should be removed only if all of its terms
   *   have semantic types on the removal list
   */
  public void setAnySty(boolean b) {
    fireNotUndoableDataChanged();
    any_sty = b;
  }

  /**
   * This method indicates whether the selections represent semantic types to
   * be excluded or semantic types to be kept.
   * @return <code>true</code> selected semantic types will be excluded
   * <code>false</code> selected semantic types will be kept
   */
  public boolean removeSelectedSemanticTypes() {
    return remove_selected_stys;
  }

  /**
   * This method sets whether or not the selections represent semantic types to
   * be excluded or semantic types to be kept.
   * @param b <code>boolean</code>
   */
  public void setRemoveSelectedSemanticTypes(boolean b) {
    remove_selected_stys = b;
  }

  //
  // ListSelectionListener Implentation
  //

  /**
   * Sets the list of semantic_types to removed when selection changes.
   * @param e {@link ListSelectionEvent}
   */
  public void valueChanged (ListSelectionEvent e) {

    if (ignore_events) return;

    fireDataChanged();

    //
    // Clear selections
    //
    semantic_types_to_remove = new ArrayList();

    //
    // Review entire table and re-build list of semantic_types to remove
    //
    for (int i = 0; i <= semantic_type_table.getRowCount(); i++)
      if (semantic_type_table.getSelectionModel().isSelectedIndex(i))
	addSemanticTypeToRemove(
          HierarchicalSemanticType.getHierarchicalSemanticType(
            semantic_type_table.getValueAt(i,0).toString()));
  }

  /**
   * Returns the {@link Color} that should be used to display
   * the name.
   * @return the {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates whether or not this filter can use the efficient mode during
   * the "Initialize Cuis" phase of the subsetting process.
   * The efficient mode reads only MRCONSO and MRSTY during the
   * "Initialize Cuis" phase.
   * @return <code>true</code>
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}

