/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	SemanticTypesToRemoveTableModel.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.umls.meta.HierarchicalSemanticType;

import java.util.Collections;
import java.util.List;

import javax.swing.table.AbstractTableModel;

/**
 * Manages and initializes the data for the Semantic Types To Exclude table.
 * 
 * @author Deborah Shapiro, Brian Carlsen
 */

public class SemanticTypesToRemoveTableModel extends AbstractTableModel {

  //
  // Private Fields
  //
  final private String[] column_names = { "TUI", "Semantic Type",
      "Semantic Hierarchy" };

  final private Class[] column_classes = { String.class, String.class,
      String.class };

  private List stys;

  //
  // Constructor
  //
  /**
   * This constructor gets the initial values for the
   * {@link SemanticTypesToRemoveJTable} table.
   */
  public SemanticTypesToRemoveTableModel() {
    stys = HierarchicalSemanticType.getHierarchicalSemanticTypesAsList();
    Collections.sort(stys);
  }

  //
  // Methods
  //
  /**
   * Returns the number of columns in the semantic types to exclude table.
   * 
   * @return <code>int</code> number of columns
   */
  public int getColumnCount() {
    return column_names.length;
  }

  /**
   * Returns the number of rows in the semantic types to exclude table.
   * 
   * @return <code>int</code> number of rows
   */
  public int getRowCount() {
    return stys.size();
  }

  /**
   * Returns the name of a column in the semantic types to exclude table.
   * 
   * @param column_number
   *          <code>int</code> number of the column in question
   * @return {@link String} name of the column in the table in position
   *         column_number
   */
  public String getColumnName(int column_number) {
    return column_names[column_number];
  }

  /**
   * Returns the value at the cell of the Semantic Types To Exclude table.
   * 
   * @param row
   *          <code>int</code> row number of the cell
   * @param column
   *          <code>int</code> column number of the cell
   * @return the {@link Object} of the cell at the given row and column
   */
  public Object getValueAt(int row, int column) {
    HierarchicalSemanticType s = (HierarchicalSemanticType) stys.get(row);
    if (column == 0)
      return s.getUi();
    else if (column == 1)
      return s.getSty();
    else if (column == 2)
      return s.getStn();
    else
      return null;
  }

  /**
   * Returns the {@link Class} of the {@link Object}s in column column_number.
   * 
   * @param column_number
   *          <code>int</code> indicates the column in question
   * @return the {@link Class} of the column
   */
  public Class getColumnClass(int column_number) {
    return column_classes[column_number];
  }

  /**
   * Returns the {@link HierarchicalSemanticType} at a given row in the Semantic
   * Types To Exclude Table.
   * 
   * @param row
   *          <code>int</code>
   * @return {@link HierarchicalSemanticType} at given row
   */
  public HierarchicalSemanticType getSemanticTypeAtRow(int row) {
    return (HierarchicalSemanticType) stys.get(row);
  }

  /**
   * Returns the row number for a given {@link HierarchicalSemanticType} in the
   * Semantic Types To Exclude Table.
   * 
   * @param sty
   *          {@link HierarchicalSemanticType}
   * @return <code>int</code> row number
   */
  public int getRowForSemanticType(HierarchicalSemanticType sty) {
    for (int i = 0; i < stys.size(); i++) {
      if (sty.equals(stys.get(i)))
        return i;
    }
    return -1;
  }
}
