/**************************************************************************
 *
 * Package:    gov.nih.nlm.mms.filters;
 * Interface:  SourcesToRemoveFilter.java
 *
 *************************************************************************/
package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.ApplicationConfiguration;
import gov.nih.nlm.mms.Dependency;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.mms.UserConfiguration;
import gov.nih.nlm.swing.SwingToolkit;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.ComponentHistory;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTable;

/**
 * 
 * This Filter removes all CUIs that are from sources chosen to be excluded from
 * the subset.
 * 
 * @author Deborah Shapiro
 * 
 */

public class SourcesToRemoveFilter extends AbstractFilter implements
    ConceptContentSelector {

  private ApplicationConfiguration default_config;

  private UserConfiguration user_config;

  private Properties filter_props;

  private boolean enforce_family_selection = true;

  private boolean enforce_dep_source_selection = true;

  private boolean remove_selected_sources = true;

  private List dependent_sources = new ArrayList();

  private List default_dependent_sources = new ArrayList();

  private Set selected_sources = new HashSet();

  private List default_sources_to_remove = new ArrayList();

  private SourcesToRemoveJTable source_table;

  private JPanel panel = null;

  private JPanel full_panel = null;

  private JRadioButton remove_sources = null;

  private JRadioButton keep_sources = null;

  //
  // Constructor
  //
  /**
   * Instantiates an empty {@link SourcesToRemoveFilter}.
   */
  public SourcesToRemoveFilter() {
  }

  //
  // Methods
  //
  /**
   * Set initial values for filter specific variables from user config file.
   */
  public void initialize() {
    if (filter_props.containsKey("enforce_family_selection")) {
      setEnforceFamilySelection(new Boolean((String) filter_props
          .get("enforce_family_selection")).booleanValue());
    }
    if (filter_props.containsKey("enforce_dep_source_selection")) {
      setEnforceDependentSourceSelection(new Boolean((String) filter_props
          .get("enforce_dep_source_selection")).booleanValue());
    }
    if (filter_props.containsKey("remove_selected_sources")) {
      setRemoveSelectedSources(new Boolean((String) filter_props
          .get("remove_selected_sources")).booleanValue());
    }
    if (filter_props.containsKey("selected_sources")) {
      selected_sources.clear();
      String remove_prop = (String) filter_props.get("selected_sources");
      String[] tokens = FieldedStringTokenizer.split(remove_prop, ";");
      for (int i = 0; i < tokens.length; i++) {
        Source local_source = Source.getSource(tokens[i].substring(0, tokens[i]
            .indexOf("|")));
        if (local_source != null)
          selected_sources.add(local_source);
      }
    }
    if (filter_props.containsKey("ip_associations")) {
      dependent_sources.clear();
      String dep_source_prop = (String) filter_props.get("ip_associations");
      StringTokenizer tokens = new StringTokenizer(dep_source_prop, ";");
      while (tokens.hasMoreTokens()) {
        String s = tokens.nextToken();
        Source source = Source.getSource(s.substring(0, s.indexOf("|")));
        Source dep_source = Source.getSource(s.substring(s.indexOf(":") + 1, s
            .lastIndexOf("|")));
        if (source != null && dep_source != null) {
          Dependency dep = new Dependency(source, dep_source);
          dependent_sources.add(dep);
        }
      }
    }
    if (default_config.getProperty(this, "ip_associations") != null) {
      default_dependent_sources.clear();
      String dep_source_prop = (String) default_config.getProperty(this,
          "ip_associations");
      StringTokenizer tokens = new StringTokenizer(dep_source_prop, ";");
      while (tokens.hasMoreTokens()) {
        String s = tokens.nextToken();
        Source source = Source.getSource(s.substring(0, s.indexOf("|")));
        Source dep_source = Source.getSource(s.substring(s.indexOf(":") + 1, s
            .lastIndexOf("|")));
        if (source != null && dep_source != null) {
          Dependency dep = new Dependency(source, dep_source);
          default_dependent_sources.add(dep);
        }
      }
    }
    if (default_config.getProperty(this, "selected_sources") != null) {
      default_sources_to_remove.clear();
      String remove_prop = (String) default_config.getProperty(this,
          "selected_sources");
      String[] tokens = FieldedStringTokenizer.split(remove_prop, ";");
      for (int i = 0; i < tokens.length; i++) {
        Source local_source = Source.getSource(tokens[i].substring(0, tokens[i]
            .indexOf("|")));
        if (local_source != null)
          default_sources_to_remove.add(local_source);
      }
    }
  }

  /**
   * Method provides the GUI panel that will display the filter's screen as one
   * of the tabs on the main GUI window.
   * 
   * @return {@link JPanel} that displays the graphical interface for the filter
   */
  public JPanel getPanel() {
    if (full_panel != null)
      return full_panel;

    full_panel = new JPanel();

    JPanel toggle_panel = new JPanel(new BorderLayout());
    remove_sources = new JRadioButton("Select sources to EXCLUDE from subset");
    keep_sources = new JRadioButton("Select sources to INCLUDE in subset");
    ButtonGroup group = new ButtonGroup();
    group.add(remove_sources);
    group.add(keep_sources);

    remove_sources.setSelected(removeSelectedSources());
    remove_sources.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        setRemoveSelectedSources(true);
        MMSToolkit.getGui().resetTitleForFilter(panel, "Sources to Exclude");
      }
    });

    toggle_panel.add(remove_sources, BorderLayout.CENTER);

    keep_sources.setSelected(!removeSelectedSources());
    keep_sources.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        setRemoveSelectedSources(false);
        MMSToolkit.getGui().resetTitleForFilter(panel, "Sources to Include");
      }
    });

    toggle_panel.add(keep_sources, BorderLayout.SOUTH);

    source_table = new SourcesToRemoveJTable(default_config, this);
    if (removeSelectedSources())
      panel = MMSToolkit.makeCard(source_table, " Sources to Exclude ");
    else
      panel = MMSToolkit.makeCard(source_table, " Sources to Include ");
    full_panel.setLayout(new BorderLayout());
    full_panel.add(toggle_panel, BorderLayout.NORTH);
    full_panel.add(panel, BorderLayout.CENTER);
    return full_panel;
  }

  /**
   * Method makes selections on the filter's table
   */
  public void setPanelSelections() {
    source_table.setSelections(getSelectedSources());
    if (removeSelectedSources()) {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Sources to Exclude");
      remove_sources.setSelected(true);
    } else {
      MMSToolkit.getGui().resetTitleForFilter(panel, "Sources to Include");
      keep_sources.setSelected(true);
    }
  }

  /**
   * Method provides the description of the filter's screen and includes
   * instructions on use.
   * 
   * @return {@link String} text of instructions for use of panel
   */
  public String getTabDescription() {
    Color c = getHighlightedColor();
    int b_red = c.getRed();
    int b_green = c.getGreen();
    int b_blue = c.getBlue();

    c = getHighlightedTextColor();
    int f_red = c.getRed();
    int f_green = c.getGreen();
    int f_blue = c.getBlue();

    return "<html>Indicate below to INCLUDE or EXCLUDE selected sources.  <br>"
        + "SELECTED sources appear on a dark background, e.g., <font style=\"background-color:rgb("
        + b_red
        + ","
        + b_green
        + ","
        + b_blue
        + "); color:rgb("
        + f_red
        + ","
        + f_green
        + ","
        + f_blue
        + ")\"> AI/RHEUM, 1993 </font>"
        + "."
        + "<br><br>"
        +

        "To undo selections and return to default source list, select \"Reset Source List\" "
        + "from Reset menu at top.</html>";
  }

  private Color getHighlightedColor() {
    JTable bogus_table = new JTable(1, 1);
    bogus_table.selectAll();
    return bogus_table.getSelectionBackground();
  }

  private Color getHighlightedTextColor() {
    JTable bogus_table = new JTable(1, 1);
    bogus_table.setValueAt("test", 0, 0);
    bogus_table.selectAll();
    return bogus_table.getSelectionForeground();
  }

  /**
   * Method allows the user to reset the filter's settings to defaults.
   */
  public void resetToDefaults() {
    setSelectedSources(default_sources_to_remove);
    setEnforceFamilySelection(true);
    setEnforceDependentSourceSelection(true);
    setRemoveSelectedSources(true);
    setSourceDependencies(default_dependent_sources);
    MMSToolkit.getGui().selectTabForFilter(this);
  }

  /**
   * Presents an advanced options GUI to the user.
   */
  public void showOptions() {
    final SourcesToRemoveOptionsGUI source_options_gui = new SourcesToRemoveOptionsGUI(
        this);
    source_options_gui.setVisible(true);
  }

  /**
   * Indicates whether or not an options menu item should be supplied for
   * advanced configuration options.
   * 
   * @return <code>true</code> if so; <code>false</code> otherwise
   */
  public boolean hasAdvancedOptions() {
    return true;
  }

  /**
   * Method provides a {@link JPanel} that will be added to the main GUI help
   * menu to allow the user to consult a help window that is particular to this
   * filter.
   * 
   * @return {@link JPanel} to be added to the Help menu
   */
  public JPanel getHelpPanel() {
    return SwingToolkit.getPanelForHTML(this.getClass().getClassLoader()
        .getResource("config/SourcesToRemoveHelp.html"));
  }

  /**
   * Method provides the filter with values from the default properties file.
   * 
   * @param c
   *          {@link ApplicationConfiguration}
   */
  public void setConfiguration(ApplicationConfiguration c) {
    default_config = c;
  }

  /**
   * Method returns the default {@link ApplicationConfiguration}.
   * 
   * @return {@link ApplicationConfiguration}
   */
  public ApplicationConfiguration getConfiguration() {
    return default_config;
  }

  /**
   * Method returns all properties specific to the filter.
   * 
   * @return {@link Properties} only those directly related to the filter. They
   *         will start with the filter's class name which will be concatenated
   *         to the name of the property. Filter.property_name
   */
  public Properties getProperties() {
    Properties new_filter_props = new Properties();
    new_filter_props.setProperty("enforce_family_selection", new Boolean(
        enforceFamilySelection()).toString());
    new_filter_props.setProperty("enforce_dep_source_selection", new Boolean(
        enforceDependentSourceSelection()).toString());
    new_filter_props.setProperty("remove_selected_sources", new Boolean(
        removeSelectedSources()).toString());
    List str = getSelectedSources();
    StringBuffer line = new StringBuffer();
    for (int i = 0; i < str.size(); i++) {
      line.append(((Source) str.get(i)).getRootSourceAbbreviation());
      line.append("|");
      line.append(((Source) str.get(i)).getSourceFamily());
      if (i <= str.size() - 2)
        line.append(";");
    }
    new_filter_props.setProperty("selected_sources", line.toString());

    ArrayList ds = (ArrayList) getSourceDependencies();
    StringBuffer line2 = new StringBuffer();
    for (int i = 0; i < ds.size(); i++) {
      line2.append(ds.get(i).toString());
      if (i <= ds.size() - 2)
        line2.append(";");
    }
    new_filter_props.setProperty("ip_associations", line2.toString());
    return new_filter_props;
  }

  /**
   * Method provides the filter with all properties specific to the filter.
   * 
   * @param p
   *          {@link Properties}
   */
  public void setProperties(Properties p) {
    filter_props = p;
    initialize();
  }

  /**
   * Method which determines if a given {@link Concept} or its components will
   * be removed from the release.
   * 
   * @param concept
   *          {@link Concept}
   */
  public void applySelections(Concept concept) {
    boolean remove_src_atoms_in_concept = false;

    //
    // Remove atoms based on source
    //
    final Atom[] atoms = concept.getAtoms();
    for (int i = atoms.length - 1; i >= 0; i--) {
      final Atom atom = atoms[i];
      final Source source = atom.getSource();
      if (isSourceToRemove(source)) {
        concept.removeAtom(atom);
      }
      if (source.getVersionedSourceAbbreviation().equals("SRC")
          && (atom.getTty().getAbbreviation().equals("RAB") || atom.getTty()
              .getAbbreviation().equals("VAB"))) {
        Source src = Source.getSource(atom.getStr());
        if (isSourceToRemove(src)) {
          remove_src_atoms_in_concept = true;
        }
      }
    }
    // if concept has a SRC|RAB or SRC|VAB atom remove all atoms with SRC from
    // concept
    if (remove_src_atoms_in_concept) {
      Atom[] src_concept_atoms = concept.getAtoms();
      for (int j = 0; j < src_concept_atoms.length; j++) {
        if (src_concept_atoms[j].getSource().getVersionedSourceAbbreviation()
            .equals("SRC")) {
          concept.removeAtom(src_concept_atoms[j]);
        }
      }
    }

    if (concept.isEmpty())
      return;

    //
    // Remove definitions based on source
    //
    final Definition[] definitions = concept.getDefinitions();
    for (int i = definitions.length - 1; i >= 0; i--) {
      final Definition def = definitions[i];
      final Source source = def.getSource();
      if (isSourceToRemove(source)) {
        concept.removeDefinition(def);
      }
    }

    //
    // Remove attributes based on source
    //
    final Attribute[] attributes = concept.getAttributes();
    for (int i = attributes.length - 1; i >= 0; i--) {
      final Attribute attribute = attributes[i];
      final Source source = attribute.getSource();
      if (isSourceToRemove(source)) {
        concept.removeAttribute(attribute);
      }
    }

    //
    // Remove coocurrences based on source
    //
    final Coocurrence[] coocurrences = concept.getCoocurrences();
    for (int i = coocurrences.length - 1; i >= 0; i--) {
      final Coocurrence coc = coocurrences[i];
      final Source soc = coc.getSource();
      if (isSourceToRemove(soc)) {
        concept.removeCoocurrence(coc);
      }
    }

    //
    // Remove relationships based on source
    //
    final Relationship[] relationships = concept.getRelationships();
    for (int i = relationships.length - 1; i >= 0; i--) {
      final Relationship rel = relationships[i];
      final Source source = rel.getSource();
      final Source sl = rel.getSl();
      if (isSourceToRemove(source) || isSourceToRemove(sl)) {
        concept.removeRelationship(rel);
      }
    }

    //
    // Remove contexts based on source
    //
    final Context[] contexts = concept.getContexts();
    for (int i = contexts.length - 1; i >= 0; i--) {
      final Context context = contexts[i];
      final Source source = context.getSource();
      if (isSourceToRemove(source)) {
        concept.removeContext(context);
      }
    }

    //
    // Remove mappings based on source
    //
    final Mapping[] mappings = concept.getMappings();
    for (int i = mappings.length - 1; i >= 0; i--) {
      final Mapping mapping = mappings[i];
      if (isSourceToRemove(mapping.getSource()))
        concept.removeMapping(mapping);
      // If to/from SABs are in current subset and are on list
      // to remove, dump the mappings.
      if (mapping.getFrom() != null && mapping.getFrom().getSource() != null
          && mapping.getFrom().getSource().inCurrentSubset()
          && isSourceToRemove(mappings[i].getFrom().getSource()))
        concept.removeMapping(mapping);
      if (mapping.getTo() != null && mapping.getTo().getSource() != null
          && mapping.getTo().getSource().inCurrentSubset()
          && isSourceToRemove(mapping.getTo().getSource()))
        concept.removeMapping(mapping);
    }

    //
    // Remove history based on source
    //
    final ComponentHistory[] chs = concept.getComponentHistories();
    for (int i = chs.length - 1; i >= 0; i--) {
      final ComponentHistory ch = chs[i];
      final Source source = ch.getSource();
      if (isSourceToRemove(source)) {
        concept.removeComponentHistory(ch);
      }
    }

  }

  /**
   * Returns the name.
   * 
   * @return {@link String}
   */
  public String getName() {
    return "Source List";
  }

  /**
   * Method returns a list of the sources to be excluded and sources to be kept
   * for display in the configuration file.
   * 
   * @return {@link String}
   */
  public String getConfigurationDetails() {
    final StringBuffer sb = new StringBuffer(5000);
    final String ls = System.getProperty("line.separator");

    sb.append("Excluded Sources").append(ls);
    boolean empty_list = true;
    Source[] sources = (removeSelectedSources()) ? (Source[]) getSelectedSources()
        .toArray(new Source[0])
        : (Source[]) getSelectedSourcesInverse().toArray(new Source[0]);

    Arrays.sort(sources);

    for (int i = 0; i < sources.length; i++) {
      final Source s = (Source) sources[i];
      final String sn = s.getOfficialName();
      if (sn.equals("") || s.getVersionedSourceAbbreviation().equals("SRC")  ||
      		!s.inCurrentSubset())
        continue;
      sb
          .append(
              "  "
                  + ((sn.length() > 67) ? sn.substring(0, 64) + "..." : sn)
                  + "                                                                    "
                      .substring(0,
                          68 - ((sn.length() > 67) ? 67 : sn.length()))
                  + s.getVersionedSourceAbbreviation()).append(ls);
      empty_list = false;
    }

    if (empty_list)
      sb.append("  <none>").append(ls);

    empty_list = true;
    sb.append(ls).append("Kept Sources").append(ls);
    sources = (removeSelectedSources()) ? (Source[]) getSelectedSourcesInverse()
        .toArray(new Source[0])
        : (Source[]) getSelectedSources().toArray(new Source[0]);
    Arrays.sort(sources);
    for (int i = 0; i < sources.length; i++) {
      final Source s = (Source) sources[i];
      final String sn = s.getOfficialName();
      if (sn.equals("") || s.getVersionedSourceAbbreviation().equals("SRC") ||
      		!s.inCurrentSubset())
        continue;
      sb
          .append(
              "  "
                  + ((sn.length() > 67) ? sn.substring(0, 64) + "..." : sn)
                  + "                                                                    "
                      .substring(0,
                          68 - ((sn.length() > 67) ? 67 : sn.length()))
                  + s.getVersionedSourceAbbreviation()).append(ls);
      empty_list = false;
    }
    if (empty_list)
      sb.append("  <none>").append(ls);

    return sb.toString();
  }

  /**
   * This method clears the data structure that holds the list of selected
   * sources.
   */
  public void clearSelectedSources() {
    selected_sources.clear();
  }

  /**
   * This method adds a {@link Source} to the data structure that holds the list
   * of selected sources.
   * 
   * @param source
   *          {@link Source}
   */
  public void addSelectedSource(Source source) {
    if (!selected_sources.contains(source)) {
      selected_sources.add(source);
    }
    // default_config.setSourcesToRemove(getSourcesToRemove());
  }

  /**
   * This method removes a {@link Source} from the data structure that holds the
   * list of selected sources.
   * 
   * @param source
   *          {@link Source}
   */
  public void removeSelectedSource(Source source) {
    selected_sources.remove(source);
    // default_config.setSourcesToRemove(getSourcesToRemove());
  }

  /**
   * This method returns the {@link List} containing all {@link Source}s
   * selected.
   * 
   * @return {@link List} selected_sources
   */
  public List getSelectedSources() {
    ArrayList list = new ArrayList(selected_sources);
    Collections.sort(list);
    return list;
  }

  /**
   * This method returns the {@link Set} containing all {@link Source}s
   * selected.
   * 
   * @return {@link Set} selected_sources
   */
  public Set getSelectedSourcesAsSet() {
    return selected_sources;
  }

  /**
   * This method is used by the {@link gov.nih.nlm.mms.SourcesToRemoveListener}
   * to determine which sources should be removed from the other {@link Filter}'s
   * displays.
   * 
   * @return {@link List} of {@link Source}s
   */
  public List getSourcesToRemove() {
    List str = null;
    if (removeSelectedSources()) {
      str = getSelectedSources();
    } else {
      str = getSelectedSourcesInverse();
    }
    str.remove(Source.getSource("SRC"));
    return str;
  }

  /**
   * This method returns the {@link List} of {@link Source}s that are not
   * selected.
   * 
   * @return {@link List} of {@link Source}s not selected
   */
  public List getSelectedSourcesInverse() {
    ArrayList selected_inverse = new ArrayList();
    ArrayList sources = (ArrayList) Source.getSourcesAsList();
    for (int i = 0; i < sources.size(); i++) {
      if (selected_sources.contains(sources.get(i)))
        continue;
      else
        selected_inverse.add(sources.get(i));
    }
    return selected_inverse;
  }

  /**
   * This method returns the {@link Set} of {@link Source}s that are not
   * selected.
   * 
   * @return {@link Set} of {@link Source}s not selected
   */
  public Set getSelectedSourcesInverseAsSet() {
    return new HashSet(getSelectedSourcesInverse());
  }

  /**
   * This method returns a {@link List} of all {@link Source}s to be excluded.
   * 
   * @return {@link List}
   */
  public List getDefaultSourcesToRemove() {
    return default_sources_to_remove;
  }

  /**
   * This method returns a {@link Set} containing the default list of all
   * {@link Source}s to be excluded.
   * 
   * @return {@link Set}
   */
  public Set getDefaultSourcesToRemoveAsSet() {
    return new HashSet(default_sources_to_remove);
  }

  /**
   * This method replaces the selected_sources list with the one provided.
   * 
   * @param str
   *          {@link List} of {@link Source}s
   */
  public void setSelectedSources(List str) {
    selected_sources.clear();
    for (int i = 0; i < str.size(); i++) {
      Source s = (Source) str.get(i);
      addSelectedSource(s);
    }
  }

  /**
   * This method replaces the selected_sources list with the one provided.
   * 
   * @param str
   *          {@link Set} of {@link Source}s
   */
  public void setSelectedSources(Set str) {
    setSelectedSources(new ArrayList(str));
  }

  /**
   * This method returns a <code>boolean</code> depending on whether or not
   * the {@link Source} is excluded from the subset.
   * 
   * @param s
   *          {@link Source}
   * @return <code>boolean</code> true if {@link Source} is to be removed
   *         false if {@link Source} is to be kept
   */
  public boolean isSourceToRemove(final Source s) {
    if (s == null) {
      if (removeSelectedSources())
        return false;
      else
        return true;
    }
    if (s.getVersionedSourceAbbreviation().equals("SRC"))
      return false;
    if (removeSelectedSources()) {
      return selected_sources.contains(s);
    } else {
      return !selected_sources.contains(s);
    }
  }

  /**
   * This method indicates whether or not the Enforce Family Selection option
   * has been chosen in the Advanced Source List Options dialog box.
   * 
   * @return <code>boolean</code> <code>true</code> all sources that share a
   *         source family will also be selected <code>false</code> otherwise
   */
  public boolean enforceFamilySelection() {
    return enforce_family_selection;
  }

  /**
   * This method sets whether or not the Enforce Family Selection option will be
   * executed on the gui.
   * 
   * @param b
   *          <code>boolean</code> <code>true</code> all sources that share
   *          a source family will also be selected <code>false</code>
   *          otherwise
   */
  public void setEnforceFamilySelection(boolean b) {
    enforce_family_selection = b;
  }

  /**
   * This method indicates whether the Enforce Dependent Source Selection option
   * has been chosen on the Advanced Source List Options dialog box.
   * 
   * @return <code>true</code> all sources that are dependent on the chosen
   *         source will also be selected <code>false</code> otherwise
   */
  public boolean enforceDependentSourceSelection() {
    return enforce_dep_source_selection;
  }

  /**
   * This method sets whether or not the Enforce Dependent Source Selection
   * option will be executed on the gui.
   * 
   * @param b
   *          <code>boolean</code>
   */
  public void setEnforceDependentSourceSelection(boolean b) {
    enforce_dep_source_selection = b;
  }

  /**
   * This method indicates whether the selections represent sources to be
   * excluded or sources to be kept.
   * 
   * @return <code>true</code> selected sources will be excluded
   *         <code>false</code> selected sources will be kept
   */
  public boolean removeSelectedSources() {
    return remove_selected_sources;
  }

  /**
   * This method sets whether or not the selections represent sources to be
   * excluded or sources to be kept.
   * 
   * @param b
   *          <code>boolean</code>
   */
  public void setRemoveSelectedSources(boolean b) {
    remove_selected_sources = b;
  }

  /**
   * This method returns whether or not the automatic_selection option has been
   * chosen on the MetamorphoSys Options dialog box. Automatic selection allows
   * the application to select dependent sources and sources sharing the source
   * family without prompting the user. This occurs only if the
   * enforce_dep_source_selection and enforce_family_selection flags are set to
   * true on the Advanced Source List Options dialog box.
   * 
   * @return <code>true</code> if automatic selection has been chosen
   *         <code>false</code> otherwise
   */
  public boolean getAutomaticSelection() {
    return user_config.getAutomaticSelection();
  }

  /**
   * Method provides the filter with values from the user configuration file.
   * 
   * @param c
   *          {@link UserConfiguration}
   */
  public void setConfiguration(UserConfiguration c) {
    user_config = c;
  }

  /**
   * Method returns a {@link List} of {@link Dependency}(s).
   * 
   * @return {@link List}
   */
  public List getSourceDependencies() {
    return dependent_sources;
  }

  /**
   * Method returns a {@link Set} of {@link Dependency}(s).
   * 
   * @return {@link Set}
   */
  public Set getSourceDependenciesAsSet() {
    return new HashSet(dependent_sources);
  }

  /**
   * Method gets all dependent sources for a given {@link Source}.
   * 
   * @param source
   *          {@link Source}
   * @return {@link List} of {@link Source}s
   */
  public List getDependentSources(Source source) {
    return new ArrayList(getDependentSourcesAsSet(source));
  }

  /**
   * Method gets all dependent sources for a given {@link Source}.
   * 
   * @param source
   *          {@link Source}
   * @return {@link Set} of {@link Source}s
   */
  public Set getDependentSourcesAsSet(Source source) {
    Set dep_sources = new HashSet();
    for (int i = 0; i < dependent_sources.size(); i++) {
      if (((Dependency) dependent_sources.get(i)).getSource().equals(source)) {
        dep_sources.add(((Dependency) dependent_sources.get(i))
            .getDependentSource());
      }
    }
    return dep_sources;
  }

  /**
   * Returns if given {@link Source} has a dependent source.
   * 
   * @param source
   *          {@link Source}
   * @return <code>boolean</code> <code>true</code> if source has a
   *         dependent source <code>false</code> otherwise
   */
  public boolean hasDependentSource(Source source) {
    for (int i = 0; i < dependent_sources.size(); i++) {
      if (((Dependency) dependent_sources.get(i)).getSource().equals(source)) {
        return true;
      }
    }
    return false;
  }

  /**
   * Clears data structure that holds {@link Dependency}(s).
   */
  public void clearSourceDependencies() {
    dependent_sources.clear();
  }

  /**
   * Replaces the struction that holds {@link Dependency}(s) with the one
   * given.
   * 
   * @param source_dependencies
   *          {@link List} of {@link Dependency}(s)
   */
  public void setSourceDependencies(List source_dependencies) {
    dependent_sources.clear();
    dependent_sources = source_dependencies;
  }

  /**
   * Returns the list of {@link Source}s from the default configuration.
   * 
   * @return {@link List} of sources
   */
  public List getSources() {
    return Source.getSourcesAsList();
  }

  /**
   * Method indicates the {@link Color} that should be used to display the name.
   * 
   * @return {@link Color}
   */
  public Color getColor() {
    return null;
  }

  /**
   * Indicates if {@link Filter} can use the efficient mode during the
   * "Initialize Cuis" phase of the subsetting process. The efficient mode reads
   * only MRCONSO and MRSTY during the "Initialize Cuis" phase.
   * 
   * @return <code>true</code> if efficient mode is acceptable
   *         <code>false</code> if efficient mode is not acceptable
   */
  public boolean isEfficientEnabled() {
    return true;
  }

}
