/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	SourcesToRemoveListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;

import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.ListDialog;
import gov.nih.nlm.umls.meta.Source;

import java.util.HashSet;
import java.util.List;
import java.util.Arrays;

import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;

/**
 * List selection listener for the Sources To Exclude table.
 * Initializes and updates selections on the the sources table.
 * Makes selections based on user's options settings.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class SourcesToRemoveListSelectionListener
  implements ListSelectionListener, IgnorableListener {
  //
  // Private Fields
  //
  private SourcesToRemoveJTable table;
  private SourcesToRemoveFilter filter;
  private boolean[] selection_before;
  private boolean ignore_events = false;

  //
  // Constructors
  //
  /**
   * Constructor - sets variables from parameters.
   * @param filter {@link SourcesToRemoveFilter}
   * @param table {@link SourcesToRemoveJTable}
   */
  public SourcesToRemoveListSelectionListener(SourcesToRemoveFilter filter,
    SourcesToRemoveJTable table) {
    this.table = table;
    this.filter = filter;
  }

  //
  // Methods
  //

  /**
   * Sets selections on the {@link SourcesToRemoveJTable}
   * based on the sources_to_remove data structure in the
   * {@link SourcesToRemoveFilter}.
   * Records initial set of selected sources in order to determine
   * at later points whether selections have changed.
   * @param str {@link List} of {@link Source}s to be excluded from subset
   */
  public void setSelections(List str) {
    setIgnoreEvents(true);
    table.clearSelection();

    selection_before = new boolean[table.getModel().getRowCount()];

    for(int i=0; i< str.size(); i++) {
      Source s = (Source)str.get(i);

      int row = table.getRowForSource(s);
      table.addRowSelectionInterval(row,row);

      selection_before[table.mapIndex(table.getRowForSource(s))] = true;
    }
    setIgnoreEvents(false);
  }

  /**
   * Method that recognizes a {@link ListSelectionEvent} and responds
   * by going through logic to determine if related sources should be selected
   * as well.
   * @param e {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {

    // prevents proceding through the code twice
    // once for the event related to deselecting a row(s)
    // once for the event related to selecting a row(s)
    if(e.getValueIsAdjusting()) return;

    // Only process events we care about
    //new Exception().printStackTrace();
    if(ignoreEvents()) {
      return;
    }
    filter.fireDataChanged();

    //logic to avoid calling up multiple confirm_sources
    //dialog boxes when user doesn't Cntrl + click and only one
    //source row is selected
    int single_row = -1;
    if(table.getSelectedRowCount() == 1 &&
       table.getSelectedRow() >= e.getFirstIndex() &&
       table.getSelectedRow() < e.getLastIndex()){
      single_row = table.getSelectedRow();
    }

    HashSet families_seen = new HashSet();

    for(int i = e.getFirstIndex(); i<=e.getLastIndex(); i++) {

      Source this_source = table.getSourceAtRow(i);

      //Set current selection for this row and update filter
      boolean selected = table.isRowSelected(i);
      if(selected)
	filter.addSelectedSource(this_source);
      else
	filter.removeSelectedSource(this_source);

      //Determine if selection changed
      boolean selection_changed =
        (selection_before[table.mapIndex(i)] != selected);

      //Set selection_before for the next event
      selection_before[table.mapIndex(i)] = selected;


      HashSet related_source_set = new HashSet();
      Object[] related_sources = null;
      // Do not continue if selection for this row didn't change.
      // single_row variable prevents dialog box from
      // being opened repeatedly when a single row is clicked
      // thereby keeping all the sources previously selected

      //if (selection_changed && (single_row == -1 || single_row == i)) {
      if ((selection_changed || single_row != -1) &&
          (single_row == -1 || single_row == i)) {
	// Get the source family set
	if (filter.enforceFamilySelection())
	  related_source_set.addAll(Source.
		       getSourcesForFamilyAsSet(this_source.getSourceFamily()));

	// Get the dependent source
	if (filter.enforceDependentSourceSelection())
	  related_source_set.addAll(filter.getDependentSourcesAsSet(this_source));

	if(related_source_set.contains(this_source))
	  related_source_set.remove(this_source);

	// related_sources should be given all sources in set b/c if
	// automatic selection is chosen, user will not be given a choice
	// of which related sources to select
	related_sources = related_source_set.toArray();

	if(!families_seen.contains(this_source.getSourceFamily())) {
	  // Handle manual selection
	  if (!filter.getAutomaticSelection())  {

            // get confirm_sources set
            Source[] confirm_sources =
	      (Source [])related_source_set.toArray(new Source[0]);

	    // if you are removing sources and all related rows are already
            // selected, don't raise the message
	    // likewise if you are keeping sources and all related rows are
	    // already unselected, don't raise the message
	    boolean raise_message = false;
	    if(table.isRowSelected(i)) {
	      for(int j=0; j<confirm_sources.length; j++) {
	        int k = table.getRowForSource(confirm_sources[j]);
	        if(!table.isRowSelected(k))
	          raise_message = true;
	      }
	    } else {
	      for(int j=0; j<confirm_sources.length; j++) {
	        int k = table.getRowForSource(confirm_sources[j]);
	        if(table.isRowSelected(k))
	          raise_message = true;
	      }
	    }

            Object[] end_source_names = null;
	    if (!related_source_set.isEmpty() && raise_message) {
	      String message = "";
	      if((table.isRowSelected(i) && filter.removeSelectedSources()) ||
              (!table.isRowSelected(i) && !filter.removeSelectedSources())) {
	        message = "<html><b>You have selected " +
                this_source.getVersionedSourceAbbreviation() +
                " for removal, the following sources</b><br> <b> typically should be removed with " +
                this_source.getVersionedSourceAbbreviation() +
                ".  Choose which ones should also be removed.</b></ul>";
              } else {
                  message = "<html><b>You have chosen to keep source " +
                  this_source.getVersionedSourceAbbreviation() +
                  ".  The following sources</b><br> <b> typically should be kept with " +
                  this_source.getVersionedSourceAbbreviation() +
                  ".  Choose which ones should also be kept.</b><ul>";
              }
              // Look up full source names
               String[] start_source_names = new String[confirm_sources.length];
               for (int k1 = 0; k1 < start_source_names.length; k1++)
                 start_source_names[k1] = confirm_sources[k1].getOfficialName();
               Arrays.sort(start_source_names);

               // Get user selections
               end_source_names = ListDialog.showListMultipleMode(
               MMSToolkit.getGui(), (Object)message,
                 "Confirm Selection", start_source_names, null);
             }


             // Cancel was pressed in dialog or none were chosen
             if (end_source_names == null) continue;

             // Take selected source names and look back up the source objects
             if(end_source_names != null) {
               related_sources = new Source[end_source_names.length];
               int ct = 0;
               for (int k1 = 0; k1 < confirm_sources.length; k1++) {
                 String name = confirm_sources[k1].getOfficialName();
                 for (int k2 = 0; k2 < end_source_names.length; k2++)
                   if (name.equals((String)end_source_names[k2]))
                      related_sources[ct++] = confirm_sources[k1];
               }
             }

	  }

	  families_seen.add(this_source.getSourceFamily());

	  //remove or maintain from the table based on dialog box results
	  for(int j = 0; j < related_sources.length; j++) {

	    int k = table.getRowForSource((Source)related_sources[j]);

  	    // If the table row matches, keep the row
	    if(table.getSourceAtRow(k).equals(
              Source.getSource(
              ((Source)related_sources[j]).getVersionedSourceAbbreviation()))) {
              //related_sources[j].toString()))) {

	      // Unselect if selected
	      if (table.isRowSelected(k) && !selected) {
	        // prevent dialog box from being raised for each
	        // of the related sources
	        selection_before[table.mapIndex(k)] = false;
	        removeSelectedSource(k);

	      // Select if unselected
	      } else if(!table.isRowSelected(k) && selected) {
	        selection_before[table.mapIndex(k)] = true;
	        addSelectedSource(k);
	      }
	    }
	  } // end for each related row update table
	} // end if not family_seen
      } // end (if selection changed or only a single row is selected)
    } // end from (first index, last index, etc.)
    table.tableChanged(new TableModelEvent(table.getModel()));
  }

  /**
   * Method returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code>
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Method that turns off responses to the {@link ListSelectionEvent}s.
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }

  /**
   * Method selects row and adds {@link Source} on that row to the list of
   * sources to be excluded.
   * @param row <code>int</code>
   */
  private void addSelectedSource(int row) {
    setIgnoreEvents(true);
    table.addRowSelectionInterval(row, row);
    filter.addSelectedSource(table.getSourceAtRow(row));
    setIgnoreEvents(false);
  }

  /**
   * Method deselects row and removes the {@link Source} on that row from the list of
   * sources to be excluded.
   * @param row <code>int</code>
   */
  private void removeSelectedSource(int row) {
    setIgnoreEvents(true);
    table.removeRowSelectionInterval(row, row);
    filter.removeSelectedSource(table.getSourceAtRow(row));
    setIgnoreEvents(false);
  }
}

