/**************************************************************************
 *
 * Package:	gov.nih.nlm.mms.filters
 * Object:	SuppressibleTermgroupsListSelectionListener.java
 *
 **************************************************************************/

package gov.nih.nlm.mms.filters;
import gov.nih.nlm.mms.MMSToolkit;
import gov.nih.nlm.swing.IgnorableListener;
import gov.nih.nlm.swing.SortableJTable;
import gov.nih.nlm.swing.TableSortEvent;
import gov.nih.nlm.swing.TableSortListener;
import gov.nih.nlm.umls.meta.SourceTermType;

import java.util.List;

import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

/**
 * List selection listener for the suppressible term types table.
 * Initializes and updates selections on the the suppressible term types table.
 *
 * @author Deborah Shapiro, Brian Carlsen
 */

public class SuppressibleTermgroupsListSelectionListener
  implements ListSelectionListener, IgnorableListener, TableSortListener{

  //
  // Private Fields
  //

  //
  // Table
  //
  private SortableJTable table;

  //
  // Filter
  //
  private SuppressibleTermgroupsFilter filter;

  //
  // flag indicating whether or not events are being ignored
  //
  private boolean ignore_events = false;

  //
  // Constructors
  //
  /**
   * Instantiates a {@link SuppressibleTermgroupsListSelectionListener}
   * with the specified {@link SuppressibleTermgroupsFilter} and {@link SortableJTable}.
   * @param filter the parent filter {@link SuppressibleTermgroupsFilter}
   * @param table the {@link SortableJTable} of {@link SourceTermType}s
   */
  public SuppressibleTermgroupsListSelectionListener(
    SuppressibleTermgroupsFilter filter, SortableJTable table) {
    this.table = table;
    this.filter = filter;
  }

  //
  // Methods
  //
  /**
   * Listens for changes in the list and validates them.
   * @param e the {@link ListSelectionEvent}
   */
  public void valueChanged(ListSelectionEvent e) {

    //
    // prevents proceding through the code twice
    // once for the event related to deselecting a row(s)
    // once for the event related to selecting a row(s)
    //
    if(e.getValueIsAdjusting()) return;

    //
    // Return if we are ignoring events
    //
    if(ignoreEvents()) return;

    //
    // Handle the deselection of default suppressible termgroups
    //
    boolean user_choice = false;
    for(int i = 0; i<table.getRowCount(); i++) {

      //
      // Is this a default suppressible termgroup?
      //
      if(!table.isRowSelected(i) &&
         filter.getConfirmDefaultSuppressibleTermgroups().
         contains(filter.getTermgroupAtRow(i))) {

        //
        // don't want to bring up dialog for each of the
        // unselected default termgroup lines
        //
        user_choice = MMSToolkit.confirmRequest(
          MMSToolkit.getGui(),
          "Your selection will cause the default selections to be \n" +
          "unselected.  Are you sure that you want to do this?");

        //
        // if user doesn't want to proceed to unselect defaults
        // update the GUI so the previously selected lines are selected again
        //
        if (!user_choice) {
          final List stgs = filter.getSuppressibleTermgroups();
          for (int j = 0; j < stgs.size(); j++) {
            final int row = filter.getRowForTermgroup((SourceTermType)stgs.get(j));
            final ListSelectionModel lsm =
                  (ListSelectionModel)table.getSelectionModel();
            setIgnoreEvents(true);
            lsm.addSelectionInterval(row, row);
            setIgnoreEvents(false);
          }
          return;
        }
        break;
      }
    }

    //
    // Push event onto undo stack
    //
    filter.fireDataChanged();

    //
    // Clear everything in filter and
    // add back in any rows that are selected
    //
    filter.clearSuppressibleTermgroups();

    final ListSelectionModel lsm = (ListSelectionModel)table.getSelectionModel();
    if (lsm.getMinSelectionIndex() == lsm.getMaxSelectionIndex())
      filter.clearConfirmDefaultSuppressibleTermgroups();
    for (int i = lsm.getMinSelectionIndex(); i <= lsm.getMaxSelectionIndex(); i++) {

      //
      // if nothing is selected, done
      //
      if (i == -1) break;

      //
      // If row is selected, add it as a suppressible termgroup
      //
      if (lsm.isSelectedIndex(i)) {
        filter.addSuppressibleTermgroup(filter.getTermgroupAtRow(i));
      }

      //
      // If row is unselected AND it is a default suppressible
      // remove it from the ConfirmDefaultSuppresibleTermgroups list
      //
      if (!lsm.isSelectedIndex(i) &&
          filter.getConfirmDefaultSuppressibleTermgroups().
          contains(filter.getTermgroupAtRow(i))) {
        filter.getConfirmDefaultSuppressibleTermgroups().
            remove(filter.getTermgroupAtRow(i));
      }
    }
  }

  /**
   * Returns <code>boolean</code> to indicate whether or not
   * {@link ListSelectionEvent}s are to be ignored.
   * @return <code>boolean</code>
   */
  public boolean ignoreEvents() {
    return ignore_events;
  }

  /**
   * Turns off responses to the {@link ListSelectionEvent}s.
   * @param b <code>boolean</code> <code>true</code> if events are to be ignored
   *				 <code>false</code> if events are to be processed
   */
  public void setIgnoreEvents(boolean b) {
    ignore_events = b;
  }

  /**
   * Selects a row and adds the {@link SourceTermType} on that row to the list of
   * suppressible {@link SourceTermType}s.
   * @param row number to be selected and suppressed
   */
  public void addSuppressibleTermgroup(int row) {
    table.addRowSelectionInterval(row, row);
    filter.addSuppressibleTermgroup(filter.getTermgroupAtRow(row));
  }

  /**
   * Deselects row and removes the {@link SourceTermType} on that row from the list of
   * suppressible {@link SourceTermType}s.
   * @param row number to be deselected and not suppressed
   */
  public void removeSuppressibleTermgroup(int row) {
    table.removeRowSelectionInterval(row, row);
    filter.removeSuppressibleTermgroup(filter.getTermgroupAtRow(row));
  }

  //
  // TableSortListener Implementation
  //

  /**
   * Handle start of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortStarted(TableSortEvent e) {
    setIgnoreEvents(true);
  }

  /**
   * Handle finish of table sorting.
   * @param e {@link TableSortEvent}
   */
  public void tableSortFinished(TableSortEvent e) {
    setIgnoreEvents(false);
  }

}

