/**************************************************************************
 *
 * Package:	gov.nih.nlm.swing
 * Object:	XTabbedPane.java
 *
 **************************************************************************/
package gov.nih.nlm.swing;

import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.event.MouseEvent;
import java.util.HashSet;
import java.util.Set;

import javax.swing.Icon;
import javax.swing.JTabbedPane;
import javax.swing.event.MouseInputAdapter;
import javax.swing.event.MouseInputListener;


/**
 * Extends {@link JTabbedPane} and enables the dragging and dropping of the tabs
 * to reorder them.
 * 
 * @author Deborah Shapiro
 */
public class XTabbedPane extends JTabbedPane {

  private MouseInputListener mouseHandler;

  private Cursor defaultCursor, handCursor;

  private int dragIndex = -1;

  private int lastDragIndex = -1;

  private int lastTabIndex = -1;

  private boolean dragCompleted = false;

  private Set frozen_panes = new HashSet();

  public XTabbedPane() {
    this(TOP);
  }

  public XTabbedPane(int tabPlacement) {
    this(tabPlacement, WRAP_TAB_LAYOUT);
  }

  public XTabbedPane(int tabPlacement, int tabLayoutPolicy) {
    super(tabPlacement, tabLayoutPolicy);
    mouseHandler = new MouseHandler();
    addMouseListener(mouseHandler);
    addMouseMotionListener(mouseHandler);
  }

  private void dragTab(int dragIndex, int tabIndex) {
    String title = getTitleAt(dragIndex);
    Icon icon = getIconAt(dragIndex);
    Component component = getComponentAt(dragIndex);
    String toolTipText = getToolTipTextAt(dragIndex);
    Color background = getBackgroundAt(dragIndex);
    Color foreground = getForegroundAt(dragIndex);
    Icon disabledIcon = getDisabledIconAt(dragIndex);
    int mnemonic = getMnemonicAt(dragIndex);
    int displayedMnemonicIndex = getDisplayedMnemonicIndexAt(dragIndex);
    boolean enabled = isEnabledAt(dragIndex);
    remove(dragIndex);
    insertTab(title, icon, component, toolTipText, tabIndex);
    setBackgroundAt(tabIndex, background);
    setForegroundAt(tabIndex, foreground);
    setDisabledIconAt(tabIndex, disabledIcon);
    setMnemonicAt(tabIndex, mnemonic);
    setDisplayedMnemonicIndexAt(tabIndex, displayedMnemonicIndex);
    setEnabledAt(tabIndex, enabled);
  }

  private Cursor getDefaultCursor() {
    if (defaultCursor == null) {
      defaultCursor = Cursor.getDefaultCursor();
    }
    return defaultCursor;
  }

  private Cursor getHandCursor() {
    if (handCursor == null) {
      handCursor = Cursor.getPredefinedCursor(Cursor.HAND_CURSOR);
    }
    return handCursor;
  }

  private int getTabIndex(int x, int y) {
    return getUI().tabForCoordinate(this, x, y);
  }

  private void maybeSetDefaultCursor() {
    Cursor cursor = getDefaultCursor();
    if (getCursor() != cursor) {
      setCursor(cursor);
    }
  }

  private void maybeSetHandCursor() {
    Cursor cursor = getHandCursor();
    if (getCursor() != cursor) {
      setCursor(cursor);
    }
  }

  class MouseHandler extends MouseInputAdapter {
    public void mouseDragged(MouseEvent e) {
      if (dragIndex != -1 && !frozen_panes.contains(new Integer(dragIndex))) {
        dragCompleted = false;
        if (getTabIndex(e.getX(), e.getY()) != -1) {
          maybeSetHandCursor();
        } else {
          maybeSetDefaultCursor();
        }
      }
    }

    public void mousePressed(MouseEvent e) {
      if (!e.isPopupTrigger() && e.getButton() == MouseEvent.BUTTON1) {
        dragCompleted = false;
        int tabIndex = getTabIndex(e.getX(), e.getY());
        if (tabIndex != -1) {
          dragIndex = tabIndex;
        }
      }
    }

    public void mouseReleased(MouseEvent e) {
      if (!e.isPopupTrigger() && e.getButton() == MouseEvent.BUTTON1) {
        if (dragIndex != -1) {
          int tabIndex = getTabIndex(e.getX(), e.getY());
          if (tabIndex != -1 && tabIndex != dragIndex
              && !frozen_panes.contains(new Integer(tabIndex))
              && !frozen_panes.contains(new Integer(dragIndex))) {

            lastDragIndex = dragIndex;
            lastTabIndex = tabIndex;

            dragTab(dragIndex, tabIndex);
            dragCompleted = true;
            setSelectedIndex(tabIndex);
            dragCompleted = false;
          }
          dragIndex = -1;
          maybeSetDefaultCursor();
        }
      }
    }
  }

  /**
   * Get {@link XTabbedPane} index of pane dragged.
   * 
   * @return index of dragged pane
   */
  public int getLastDragIndex() {
    return lastDragIndex;
  }

  /**
   * Get {@link XTabbedPane} index of where pane was dropped.
   * 
   * @return index of pane where dragged pane was dropped
   */
  public int getLastTabIndex() {
    return lastTabIndex;
  }

  /**
   * Indicates if drag and drop of pane has been completed.
   * 
   * @return <code>true</code> if completed <code>false</code> otherwise
   */
  public boolean isDragCompleted() {
    return dragCompleted;
  }

  /**
   * Indicates that the tabbed pane at the given index should be frozen and not
   * be permitted to be dragged and dropped.
   * 
   * @param pane_index
   *          int
   */
  public void freezePane(int pane_index) {
    frozen_panes.add(new Integer(pane_index));
  }

  /**
   * Indicates that the tabbed pane at the given index no longer should be
   * frozen and it should be permitted to be dragged and dropped.
   * 
   * @param pane_index
   *          int
   */
  public void unfreezePane(int pane_index) {
    frozen_panes.remove(new Integer(pane_index));
  }
}