/***************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Interface:   ConceptInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.Concept;

import java.io.IOException;

/**
 *
 * Generically represents a way of reading {@link Concept}s from some input source.
 * This source can be MR Files, update MR Files, a database, or pretty
 * much anything else that can be defined by one or more "locations" or "paths".
 *
 * @author Deborah Shapiro
 */
public interface ConceptInputStream {

  /**
   * Opens the input stream.  Here is where the actual files or database
   * containing the Metathesaurus data should be opened.
   * @param paths paths to the sources of Metathesaurus data
   * @throws IOException if there is a problem opening the input stream
   */
  public void open(String[] paths) throws IOException;

  /**
   * Closes the input stream.  Here is where the actual files or database
   * containing the Metathesaurus data should be closed.
   * @throws IOException if there is a problem closing the input stream
   */
  public void close() throws IOException;

  /**
   * Read a {@link Concept} from the input stream and return it.
   * This method should provide the mapping between the input
   * data and {@link Concept} objects.  Furthermore, it should
   * be implemented like a stream, in the sense that only the
   * content for the next {@link Concept} should be read into memory.
   * @return a {@link Concept} read from the input stream
   * @throws IOException if there is a problem reading a concept
   */
  public Concept readConcept() throws IOException;

  /**
   * Returns the percentage of {@link Concept}s that have been processed so far.
   * @return the percentage of {@link Concept}s that have been processed so far
   */
  public int getConceptProgress();


}
