/***************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Interface:   IndexBuilder.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.util.ColumnStatistics;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.FileStatistics;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8InputStreamReader;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;

import javax.swing.ProgressMonitor;

/**
 * Used for building an indexes for Metathesaurus files
 */
public class IndexBuilder  {

  //
  // Fields
  //

  /**
   * location of indexes directory
   */
  protected String dir;

  /**
   * Instantiates an {@link IndexBuilder}.
   * @param dir directory where indexes directory should be located
   */
  public IndexBuilder(String dir) {
    this.dir = dir;
  }


  /**
   * Builds an index file for the given file and column.
   * @param file name of file to be indexed
   * @param cols names of columns to be indexed
   * @param file_stats used to find column index in file
   * @param pm {@link ProgressMonitor} for reporting indexing progress
   * @throws java.lang.Exception
   */
  public void buildIndex(String file, String[] cols, FileStatistics file_stats,
                         final ProgressMonitor pm) throws Exception {

    // index directory
    final File idir = new File(dir, "indexes");
    if (!idir.exists())
      idir.mkdir();

    //
    // build index file name
    //
    String file_name = "";
    for (int i=0; i<cols.length; i++) {
      file_name += cols[i];
      file_name += "_";
    }
    file_name += file;
    file_name += ".x";
    File index = new File(idir, file_name);

    //
    // Determine line termination chars for file to index
    //
    // Use actual length of line termination chars in file
    // when computing file offset rather than assuming same
    // line termination as current platform.
    //
    String line_separator = null;

    final File in_file = new File(dir, file + ".RRF");
    long in_file_length = in_file.length();
    FileInputStream fis = new FileInputStream(in_file);
    final BufferedReader in_test = new BufferedReader(new InputStreamReader(
        fis, "UTF-8"));
    final char[] line_term_test = new char[5000];
    int charsRead = in_test.read(line_term_test);
    final String line_term = new String(line_term_test);
    if (line_term.indexOf("\r\n") != -1)
      line_separator = "\r\n";
    else if (line_term.indexOf("\n") != -1)
      line_separator = "\n";
    else if (line_term.indexOf("\r") != -1)
      line_separator = "\r";
    else if (charsRead == -1)
      line_separator = "";  // Use in case RRF file is empty
    else
      throw new RuntimeException("Could not find line termination characters in index file");
    fis.close();
    in_test.close();

    //
    // Open file to index and index file itself
    //
    fis = new FileInputStream(in_file);
    UTF8InputStreamReader uisr = new UTF8InputStreamReader(
        new BufferedInputStream(fis));
    final BufferedReader in = new BufferedReader(uisr, 512 * 1024);
    final BufferedWriter out = new BufferedWriter(new OutputStreamWriter(
        new FileOutputStream(index), "UTF-8"), 512 * 1024);

    //
    // Obtain columns to index
    //
    String line = null;
    final int[] col_numbers = new int[cols.length];
    for (int i=0; i<col_numbers.length; i++)
      col_numbers[i] = -1;
    final ColumnStatistics[] col_stats = file_stats.getAllColumnStatistics();
    for (int k = 0; k < cols.length; k++) {
      for (int i = 0; i < col_stats.length; i++) {
        if (col_stats[i].getColumnName().equals(cols[k])) {
          col_numbers[k] = i;
        }
      }
      if (col_numbers[k] == -1)
        throw new Exception("Unexpected error in making index file " + index + ".");
    }

    final int ls_length = line_separator.length();

    //
    // For each line of the file, create an index entry for columns to index
    //
    int counter = 0;
    int prev_progress = 0;
    //long prev_pos = uisr.hasByteOrderMark() ? 3 : 0;
    long fp = 0;
    while ( (line = in.readLine()) != null) {

      final String[] tokens = FieldedStringTokenizer.split(line, "|");
      for (int i=0; i<col_numbers.length; i++) {
        out.write(tokens[col_numbers[i]]);
        out.write("|");
      }
      //out.println(prev_pos);
      //prev_pos = fis.getChannel().position();
      out.write(String.valueOf(fp));
      out.newLine();
      //out.flush();

      //
      // Increment file pointer
      //
      fp += line.getBytes("UTF-8").length;
      fp += ls_length;

      //
      // Update progress monitor - this portion is 30% of total progress
      //
      counter++;
      if (counter == 10000 && pm != null) {
        counter = 0;
        final int i = (int)((fp * 29) / in_file_length) + 1;
        if ( i > prev_progress ) {
          pm.setProgress(i);
          pm.setNote("Operation is " + i + "% complete.");
          prev_progress = i;
        }
        if (pm.isCanceled()) {
          in.close();
          out.close();
          index.delete();
          throw new Exception("Index build cancelled.");
        }
      }
    }

    //
    // Close streams
    //
    in.close();
    out.flush();
    out.close();

    //
    // Connect progress monitor for sorting operation
    //
    ProgressMonitor pm2 = null;
    if (pm != null)
      pm2 = new ProgressMonitor(null, null, null, 0, 100) {

        private int prev_progress = 0;
        public void setProgress(int progress) {
          final int i = ( (int) ( ( (progress * 70) / 100) + 30));
          if (i > prev_progress) {
            pm.setProgress(i);
            pm.setNote("Operation is " + i + "% complete.");
            prev_progress = i;
          }
        }
      };

    //
    // Sort the index file
    //
    SystemToolkit.sort(index.getAbsolutePath(), true, pm2);

  }
}