/**************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Class:       InputStreamFileStatisticsConfiguration.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.util.ColumnStatistics;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.FileStatistics;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;


/**
 * Used to generate {@link FileStatistics} from an {@link InputStream} having
 * structured lines like a <tt>config/mrpluscolsfiles.dat</tt> file.
 *
 * @author Deborah Shapiro
 */
public class InputStreamFileStatisticsConfiguration implements FileStatisticsConfiguration {

  //
  // Map of file names -> FileStatistics
  //
  protected Map mr_stats = new HashMap();

  //
  // Map of file names to column lists
  //
  protected Map file_col_list_map = new HashMap();

  //
  // Map of file names to descriptions
  //
  protected Map file_desc_map = new HashMap();

  //
  // Map of column names to data types
  //
  protected Map col_data_type_map = new HashMap();

  //
  // Map of column names to descriptions
  //
  protected Map col_desc_map = new HashMap();

  /**
   * Instantiates a {@link InputStreamFileStatisticsConfiguration} from the
   * specified {@link InputStream}.
   * @param input_stream an {@link InputStream} containing data like that
   * in a <tt>config/mrpluscolsfiles.dat</tt> file
   * @throws IOException
   */
  public InputStreamFileStatisticsConfiguration(InputStream input_stream) throws IOException {
    initializeMaps(input_stream);
  }

  /**
   * Helper method for initializing MRFILES/MRCOLS configuration info.
   * @param input_stream {@link InputStream}
   * @throws IOException if anything goes wrong
   */
  protected void initializeMaps(InputStream input_stream) throws IOException{
    //
    // Prepare data for column/file statistics tracking
    //
    final BufferedReader in = new BufferedReader(
        new InputStreamReader(input_stream));

    String line = null;
    while ((line = in.readLine()) != null) {
      final String[] fields = FieldedStringTokenizer.split(line,"|");
      if (fields[1].equals("MRFILES")) {
        file_col_list_map.put(fields[0], fields[2]);
        file_desc_map.put(fields[0], fields[3]);
      } else if (fields[1].equals("MRCOLS")) {
        col_data_type_map.put(fields[0], fields[2]);
        col_desc_map.put(fields[0], fields[3]);
      }
    }
    in.close();
  }

  /**
   * Returns a {@link FileStatistics} object containing
   * {@link ColumnStatistics} objects describing the
   * specified file.
   * @param file a file
   * @return a {@link FileStatistics} for the specified file
   */
  public FileStatistics getFileStatistics(String file) {

    if (mr_stats.containsKey(file) && !file.equals("MRCOLS") &&
        !file.equals("MRFILES"))
      return (FileStatistics) mr_stats.get(file);

    final FileStatistics stats = new FileStatistics();
    stats.setFileName(file);
    stats.setDescription( (String) file_desc_map.get(file));

    //
    // Indicate that lines in these files may have UTF-8 non-ASCII characters.
    // (meaning that the number of characters != number of bytes).
    //
    if (file.equals("MRCONSO") || file.equals("MRSAT") || file.equals("MRDEF") ||
        file.equals("MRCXT") || file.startsWith("MRXW"))
      stats.setCharsEqualsBytes(false);

      //
      // Prepare column statistics objects for file
      //
    final String[] cols = FieldedStringTokenizer.split(
        ( (String) file_col_list_map.get(file)), ",");
    final ColumnStatistics[] col_stats = new ColumnStatistics[cols.length];
    for (int i = 0; i < cols.length; i++) {
      col_stats[i] = new ColumnStatistics();
      col_stats[i].setFileName(file);
      col_stats[i].setColumnName(cols[i]);
      col_stats[i].setDescription( (String) col_desc_map.get(cols[i]));
      col_stats[i].setDataType( (String) col_data_type_map.get(cols[i]));
      String data_type = col_stats[i].getDataType();
      if (data_type.indexOf('(') != -1 && data_type.indexOf(')') != -1) {
        col_stats[i].setDataTypeLength(
            data_type.substring(data_type.indexOf('(') + 1,
                                data_type.lastIndexOf(')')));
      }
    }
    stats.setAllColumnStatistics(col_stats);
    mr_stats.put(file, stats);
    return stats;
  }
 }
