/**************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Object:      OriginalMRFilesConceptInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.CoocurrenceType;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.HierarchicalSemanticType;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.MapObject;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.RelationshipLabel;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.StringType;
import gov.nih.nlm.umls.meta.Suppressible;
import gov.nih.nlm.umls.meta.TermStatus;
import gov.nih.nlm.umls.meta.TermType;
import gov.nih.nlm.umls.meta.impl.AtomAttributeImpl;
import gov.nih.nlm.umls.meta.impl.AtomImpl;
import gov.nih.nlm.umls.meta.impl.AttributeImpl;
import gov.nih.nlm.umls.meta.impl.ConceptImpl;
import gov.nih.nlm.umls.meta.impl.ContextImpl;
import gov.nih.nlm.umls.meta.impl.ContextMemberImpl;
import gov.nih.nlm.umls.meta.impl.CoocurrenceImpl;
import gov.nih.nlm.umls.meta.impl.DefinitionImpl;
import gov.nih.nlm.umls.meta.impl.LocatorImpl;
import gov.nih.nlm.umls.meta.impl.MapObjectImpl;
import gov.nih.nlm.umls.meta.impl.MappingImpl;
import gov.nih.nlm.umls.meta.impl.MetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.PlaceholderAtom;
import gov.nih.nlm.umls.meta.impl.PlaceholderConcept;
import gov.nih.nlm.umls.meta.impl.RelationshipImpl;
import gov.nih.nlm.umls.meta.impl.SemanticTypeImpl;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.SystemToolkit;
import gov.nih.nlm.util.UTF8InputStreamReader;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This {@link ConceptInputStream} builds {@link Concept} objects from a full
 * set of original format MR files.
 *
 * @author Deborah Shapiro
 */
public class OriginalMRFilesConceptInputStream implements ConceptInputStream {

  //
  // Readers for the various files
  //
  protected PushBackReader
    mrso_source,
    mrcon_source,
    mrsty_source,
    mrdef_source,
    mrrel_source,
    mrcxt_source,
    mrcoc_source,
    mratx_source,
    mrsat_source,
    mrlo_source;

  //
  // List of file names
  //
  protected String [] file_names;

  //
  // The directory containing the MR files
  //
  protected String mr_dir;

  //
  // The current concept being loaded
  //
  protected Concept this_concept = new ConceptImpl();

  //
  // Total number of MRCON lines
  //
  protected int mrcon_lines = 0;

  //
  // Number of MRCON lines read so far
  //
  protected int mrcon_counter = 0;

  //
  // FieldedStringTokenizer.split()
  //
  protected final String[] tokens = new String[50];

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link OriginalMRFilesConceptInputStream}.
   */
  public OriginalMRFilesConceptInputStream() { }

  //
  // Methods
  //

  /**
   * Opens all of the MR Files in preparation for subsetting.
   * @param install_paths the directory containing the files
   * @throws IOException if any files can not be opened
   */
  public void open(String[] install_paths) throws IOException {

    //
    // Set the directory
    //
    this.mr_dir = install_paths[0];

    MetaDescription[] mdsc_array = new MetaDescription[1];
    mdsc_array[0] = new MetaDescriptionImpl("STT|PF|expanded_form||");
    StringType.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("STT|VCW|expanded_form||");
    StringType.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("STT|VC|expanded_form||");
    StringType.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("STT|VO|expanded_form||");
    StringType.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("STT|VW|expanded_form||");
    StringType.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("STT|V|expanded_form||");
    StringType.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("TS|P|expanded_form||");
    TermStatus.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("TS|S|expanded_form||");
    TermStatus.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("TS|p|expanded_form||");
    TermStatus.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("TS|s|expanded_form||");
    TermStatus.initialize(mdsc_array);
    mdsc_array[0] = new MetaDescriptionImpl("REL|SUBX|expanded_form||");
    RelationshipLabel.initialize(mdsc_array);

    //
    // Open content files
    //
    mrso_source = openSourceFile("MRSO");
    mrcon_source = openSourceFile("MRCON");
    mrsty_source = openSourceFile("MRSTY");
    mrdef_source = openSourceFile("MRDEF");
    mrrel_source = openSourceFile("MRREL");
    mrcxt_source = openSourceFile("MRCXT");
    mrcoc_source = openSourceFile("MRCOC");
    mratx_source = openSourceFile("MRATX");
    mrsat_source = openSourceFile("MRSAT");
    mrlo_source = openSourceFile("MRLO");

    //
    // Count MRCON lines
    //
    getLineCounts();

  }

  /**
   * Close files opened during {@link #open(String[])}.
   * @throws IOException if files can not be closed
   */
  public void close() throws IOException {
    //
    // Close content files
    //
    mrso_source.close();
    mrcon_source.close();
    mrsty_source.close();
    mrdef_source.close();
    mrrel_source.close();
    mrcxt_source.close();
    mrcoc_source.close();
    mratx_source.close();
    mrsat_source.close();
    mrlo_source.close();
  }

  /**
   * Returns the percentage of <code>MRCON</code> lines that have been processed.
   * @return the percentage of <code>MRCON</code> lines that have been processed
   */
  public int getConceptProgress() {
    return (int)((mrcon_counter/(double)mrcon_lines)*100);
  }

  /**
   * Returns the next {@link Concept} from the MR Files.
   * @return the next {@link Concept}from the MR Files
   * @throws IOException if files could not be read
   */
  public Concept readConcept() throws IOException {
    String current_cui = "";
    this_concept = new ConceptImpl();

    //
    // Create Atoms for each group of rows from MRSO and MRCON with the given CUI
    //
    if (getAtoms(this_concept, current_cui) == null)
      return null;
    current_cui = this_concept.getCui();

    //
    // Create Definitions for each row in MRDEF with the given CUI
    //
    getDefinitions(this_concept, current_cui);

    //
    // Create SemanticTypes for each row in MRSTY with the given CUI
    //
    getSemanticTypes(this_concept, current_cui);

    //
    // Create Attributes for each row in MRSAT with the given CUI
    //
    getAttributes(this_concept, current_cui);

    //
    // Create Relationships for each row in MRREL with the given CUI
    //
    getRelationships(this_concept, current_cui);

    //
    // Create Locator for each row in MRLO with the given CUI
    //
    getLocators(this_concept, current_cui);

    //
    // Create Coocurrence for each row in MRCOC with the given CUI
    //
    getCoocurrences(this_concept, current_cui);

    //
    // Create Contexts and ContextMembers for rows in MRCXT
    //
    getContexts(this_concept, current_cui);

    //
    // Create Mappings for rows in MRATX
    //
    getMappings(this_concept, current_cui);

    //
    // Concept is fully populated, return it
    //
    return this_concept;
  }

  //
  // Protected Methods
  //

  /**
   * Opens the specified file in the directory passed to
   * {@link #open(String[])}.
   * @param name the filename
   * @return a {@link BufferedReader} on the file
   * @throws IOException if the file could not be opened
   */
  protected PushBackReader openSourceFile(String name) throws IOException{
    final File file = new File(mr_dir, name);
    if (file.exists())
      return new PushBackReader(new BufferedReader(new UTF8InputStreamReader(
        new BufferedInputStream(new FileInputStream(new File(mr_dir, name)))), 256 * 1024));
    else
      return new PushBackReader(new BufferedReader(new StringReader("")));
  }

  /**
   * Reads through <codE>MRFILES</code> to determine the size of <code>MRCON</code>.
   * This is used in the {@link #getConceptProgress()} method.
   * @return {@link Map} of file names to lines in file
   * @throws IOException if the file cannot be opened or read
   */
  protected Map getLineCounts() throws IOException {

    mrcon_lines = 0;
    mrcon_counter = 0;

    //
    // Read MRFILES into a 2-D string array
    //
    String[][] mrfiles = SystemToolkit.readFieldedFile(new File(mr_dir, "MRFILES"), "|");

    //
    // Initialize this class variable here
    // It is used by sub-classes which is why it is not local
    //
    final Map lines_in_file = new HashMap();

    //
    // Add entries to lines_in_file
    //
    for(int i=0; i<mrfiles.length; i++)
      lines_in_file.put(mrfiles[i][0], mrfiles[i][4]);

    //
    // Determine MRCON line count (assumes MRFILES is correct)
    //
    mrcon_lines = Integer.parseInt((lines_in_file.get("MRCON")).toString());

    return lines_in_file;
  }

  /**
   * Instantiates an {@link Atom} from a <code>MRSO</code> line.
   * @param line a line from <code>MRSO</code>
   * @return the {@link Atom} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Atom newInstanceFromMRSO(String line) throws IOException{

    //
    // Create a default atom
    //
    final Atom a = new AtomImpl();

    //
    // Parse the line and count the fields
    //
    FieldedStringTokenizer.split(line, "|", 8, tokens);

    //
    // Set atom fields
    //
    a.setLui(tokens[1]);
    a.setSui(tokens[2]);
    a.setSource(Source.getSource(tokens[3]));
    final TermType tty = TermType.getTermType(tokens[4]);
    if (tty == null) {
      //
      // We have no MRDOC, so fake it
      //
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("TTY|" +
                                              tokens[4] + "|expanded_form||");

      TermType.initialize(mdsc_array);
      a.setTermType(TermType.getTermType(tokens[4]));
    } else
      a.setTermType(tty);
    a.setCode(tokens[5]);
    // set to null for now, compute later
    a.setSuppress(null);

    return a;
  }

  /**
   * Returns {@link Mapping}s from the <code>MRATX</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept {@link Concept} that will contain the {@link Mapping}s
   * @param current_cui {@link String} representation of the cui
   * @throws IOException if there is a problem with the file
   */
  protected void getMappings(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through file
    //
    String line = null;
    while ((line = mratx_source.readLine()) != null) {

      //
      // Skip lines that don't match this CUI
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Build mappings from matching CUI rows
      //
      else if (line.startsWith(current_cui)) {

        //
        // Instantiate Mapping from the file
        //
        final Mapping mapping = newInstanceFromMRATX(line);

        //
        // Connect to concept
        //
        mapping.setConcept(this_concept);
        this_concept.addMapping(mapping);

        //
        // We fake the mapping type as ATX (this is how it works in RRF).
        //
        mapping.setType("ATX");
      }

      //
      // Went too far -- break
      //
      else {
        break;
      }
    }

    //
    // Reset for next read
    //
    mratx_source.push(line);
  }

  /**
   * Instantiates a {@link Mapping} from a MRATX line.
   * @param line a line from MRATX
   * @return the {@link Mapping} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Mapping newInstanceFromMRATX(String line)
    throws IOException {

    //
    // Instantiate a default mapping and map objects
    //
    final Mapping d = new MappingImpl();
    final MapObject to = new MapObjectImpl();
    final MapObject from = new MapObjectImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 5, tokens);

    //
    // Populate mapping
    //
    final RelationshipLabel rel = RelationshipLabel.getRelationshipLabel(tokens[2]);
    if (rel == null) {
      //
      // We have no MRDOC so fake it
      //
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("REL|" +
                                              tokens[2] + "|expanded_form||");

      RelationshipLabel.initialize(mdsc_array);
      d.setRel(RelationshipLabel.getRelationshipLabel(tokens[2]));
    } else
      d.setRel(rel);
    from.setExpression(tokens[0]);
    d.setFrom(from);
    to.setSource(Source.getSource(tokens[1]));
    to.setExpression(tokens[3]);
    d.setTo(to);
    d.setType("ATX");
    return d;
  }

  /**
   * Returns {@link Atom}s from the <code>MRCON</code>
   * and <code>MRSO</code> files with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @return {@link Concept} that will contain the {@link Atom}s
   * @throws IOException if there is a problem with the file
   */
  protected Concept getAtoms(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Prep variables
    //
    String line = null;
    final List current_concept_atoms = new ArrayList();

    //
    // First, determine what the next CUI in MRSO is
    //
    line = mrso_source.readLine();
    if (line == null)
      return null;
    current_cui = new String(line.substring(0, line.indexOf("|")));
    mrso_source.push(line);

    //
    // Create Atoms for each row in MRSO with the given CUI
    //
    while ((line = mrso_source.readLine()) != null &&
	   line.startsWith(current_cui)) {
      final Atom a = newInstanceFromMRSO(line);
      a.setConcept(this_concept);
      current_concept_atoms.add(a);
    }
    mrso_source.push(line);

    //
    // Set cui
    //
    this_concept.setCui(current_cui);

    //
    // For each row in MRCON with the given CUI, find the corresponding
    // Atom/row from MRSO and add the lat, ts, stt, and str to the Atom
    //
    while ((line = mrcon_source.readLine()) != null &&
	   line.startsWith(current_cui)) {
      mrcon_counter++;
      FieldedStringTokenizer.split(line, "|", 9, tokens);
      String mrcon_cui = tokens[0];
      String mrcon_lat = tokens[1];
      String mrcon_ts = tokens[2];
      //String mrcon_lui = tokens[3];
      String mrcon_stt = tokens[4];
      String mrcon_sui = tokens[5];
      String mrcon_str = tokens[6];

      //
      // Find matching Atom (from MRSO data)
      //
      for (int i = 0; i < current_concept_atoms.size(); i++) {
	final Atom a = (Atom)current_concept_atoms.get(i);

        //
        // Match if CUI & SUI are equal
        //
	if (mrcon_cui.equals(current_cui) && mrcon_sui.equals(a.getSui())) {

          //
          // Set language
          //
          final Language lat = Language.getLanguage(mrcon_lat);
          if (lat == null) {
            //
            // No MRDOC, so fake it
            //
            MetaDescription[] mdsc_array = new MetaDescription[1];
            mdsc_array[0] = new MetaDescriptionImpl("TTY|" +
                                              mrcon_lat + "|expanded_form||");

            Language.initialize(mdsc_array);
            a.setLat(Language.getLanguage(mrcon_lat));
          } else
            a.setLat(lat);

          //
          // Set TS
          //
          final TermStatus ts = TermStatus.getTermStatus(mrcon_ts);
          if (ts == null) {
            //
            // No MRDOC, so fake it
            //
            MetaDescription[] mdsc_array = new MetaDescription[1];
            mdsc_array[0] = new MetaDescriptionImpl("TS|" +
                                              mrcon_ts + "|expanded_form||");
            TermStatus.initialize(mdsc_array);
            a.setTs(TermStatus.getTermStatus(mrcon_ts));
          } else
            a.setTs(ts);

          //
          // Set STT
          //
          final StringType stt = StringType.getStringType(mrcon_stt);
          if (stt == null) {
            //
            // No MRDOC, so fake it
            //
            MetaDescription[] mdsc_array = new MetaDescription[1];
            mdsc_array[0] = new MetaDescriptionImpl("STT|" +
                                              mrcon_stt + "|expanded_form||");

            StringType.initialize(mdsc_array);
            a.setStt(StringType.getStringType(mrcon_stt));
          } else
            a.setStt(stt);

          //
          // Set the STR
          //
	  a.setStr(mrcon_str);

          //
          // Connect atoms and concept
          //
          if (this_concept == null)
            this_concept = new ConceptImpl();
          this_concept.addAtom(a);

        }
      }
    }
    mrcon_source.push(line);

    //
    // Verify that every atom has a string
    //  (This also verifies that every MRSO line
    //   connects to a MRCON line)
    //
    for (int i = 0; i < current_concept_atoms.size(); i++) {
      final Atom a = (Atom)current_concept_atoms.get(i);
      if (a.getStr() == null) {
	throw new IOException(
	  "MRSO row without matching MRCON row found (cui=" + current_cui +
          ", lui=" + a.getLui() + ", sui=" + a.getSui());
      }
    }

    this_concept.setCui(current_cui);
    return this_concept;
  }

  /**
   * Instantiates an {@link Attribute} from a <code>MRSAT</code> line.
   * @param line a line from <code>MRSAT</code>
   * @param current_concept the current {@link Concept}
   * @return the {@link Attribute} matching the line
   * @throws IOException if there is a problem with the file
   */
  private Attribute newInstanceFromMRSAT(String line, Concept current_concept)
    throws IOException {

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|",8, tokens);

    //
    // Determine if we are creating an Attribute, or an AtomAttribute
    //
    Attribute a = null;
    if (tokens[1].equals("") || tokens[2].equals("")) {

      //
      // Instantiate a default attribute
      //
      a = new AttributeImpl();
    } else {

      //
      // Instantiate an AtomAttribute and populate the atom
      //
      AtomAttribute aa = new AtomAttributeImpl();
      final Atom atom = new PlaceholderAtom();
      atom.setConcept(current_concept);
      atom.setLui(tokens[1]);
      atom.setSui(tokens[2]);
      atom.setCode(tokens[3]);
      aa.setAtom(atom);
      a = aa;
    }

    //
    // Populate shared fields
    //
    final AttributeName atn = AttributeName.getAttributeName(tokens[4]);
    if (atn == null) {
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("ATN|" +
                                              tokens[4] + "|expanded_form||");

      AttributeName.initialize(mdsc_array);
      a.setAtn(AttributeName.getAttributeName(tokens[4]));
    } else
      a.setAtn(atn);
    a.setSource(Source.getSource(tokens[5]));
    a.setAtv(tokens[6]);

    return a;
  }

  /**
   * Returns {@link Attribute}s from the <code>MRSAT</code> files with the
   * given cui and adds them to the current {@link Concept}
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if there is a problem with the file
   */
  protected void getAttributes(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through file
    //
    String line = null;
    while((line = mrsat_source.readLine()) != null) {

      //
      // Skip non-matching lines
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect attribute and concept
      //
      else if (line.startsWith(current_cui)) {
        final Attribute sat = newInstanceFromMRSAT(line, this_concept);
        sat.setConcept(this_concept);
        this_concept.addAttribute(sat);
      }

      //
      // went too far - break
      //
      else {
        break;
      }
    }
    mrsat_source.push(line);
  }

  /**
   * Instantiates an {@link Context} from a <code>MRCXT</code> line.
   * @param line a line from <code>MRCXT</code>
   * @param current_concept the current {@link Concept}
   * @return the {@link Context} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Context newInstanceFromMRCXT(String line, Concept current_concept)
    throws IOException {

    //
    // Instantiate a default context
    //
    final Context cxt = new ContextImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 13, tokens);

    //
    // Create atom for context and populate
    //
    final Atom atom = new PlaceholderAtom();
    atom.setCode(tokens[3]);
    atom.setConcept(current_concept);
    atom.setSui(tokens[1]);

    //
    // Connect context and atom
    //
    cxt.setAtom(atom);
    cxt.setSource(Source.getSource(tokens[2]));
    cxt.setCxn(tokens[4].equals("") ? 0 : Integer.parseInt(tokens[4]));

    return cxt;
  }

  /**
   * Returns {@link Context}s from the <code>MRCXT</code> files with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if anything wrong with file
   */
  protected void getContexts(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRCXT
    //
    String line = null;
    Context this_context = null, prev_context = null;
    while ((line = mrcxt_source.readLine()) != null) {

      //
      // Skip over non-matching lines
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, build context
      //
      else if (line.startsWith(current_cui)) {

        //
        // Create context and context member
        //
        this_context = newInstanceFromMRCXT(line, this_concept);
        this_context.setConcept(this_concept);
        final ContextMember cxn_rel = newInstanceFromMRCXTRel(line);

        //
        // Add member to the context
        //
        if (prev_context == null) {
          this_context.addContextMember(cxn_rel);
          prev_context = this_context;
        }

        //
        // If key changed, we're working with a new context
        //
        else if (this_context.getCxn() != prev_context.getCxn() ||
                 !this_context.getAtom().getSui().equals(prev_context.getAtom().getSui()) ||
                 !this_context.getAtom().getCode().equals(prev_context.getAtom().getCode()) ||
                 !this_context.getSource().getVsab().equals(prev_context.getSource().getVsab())) {
          this_context.addContextMember(cxn_rel);
          this_concept.addContext(prev_context);
          prev_context = this_context;
        } else {
          prev_context.addContextMember(cxn_rel);
        }
      }

      //
      // went too far - line_cui bigger than current_cui
      //
      else {
        break;
      }
    }

    if (prev_context != null) {
      this_concept.addContext(prev_context);
    }
    mrcxt_source.push(line);
  }

  /**
   * Instantiates an {@link ContextMember} from a <code>MRCXT</code> line.
   * @param line a line from <code>MRCXT</code>
   * @return the {@link ContextMember} matching the line
   * @throws IOException if there is a problem with the file
   */
  private ContextMember newInstanceFromMRCXTRel(String line) throws
      IOException {

    //
    // Instantiate an ContextMember
    //
    final ContextMember cm = new ContextMemberImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 13, tokens);

    //
    // Ignore fields 0-4, set context member data from other fields
    //
    cm.setCxl(tokens[5]);
    cm.setRank(tokens[6].equals("") ? 0 : Integer.parseInt(tokens[6]));
    final Atom atom = new PlaceholderAtom();
    atom.setStr(tokens[7]);
    cm.setRelatedAtom(atom);
    final Concept c = new PlaceholderConcept();
    c.setCui(tokens[8]);
    cm.setRelatedConcept(c);
    cm.setHcd(tokens[9]);

    //
    // Set RELA
    //
    final AdditionalRelationshipLabel rela = AdditionalRelationshipLabel.
        getAdditionalRelationshipLabel(tokens[10]);
    if (rela == null) {
      //
      // No MRDOC, so fake it
      //
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("RELA|" +
                                              tokens[10] + "|expanded_form||");

      AdditionalRelationshipLabel.initialize(mdsc_array);
      cm.setRela(AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[10]));
    } else
      cm.setRela(rela);
    cm.setXc(tokens[11]);
    return cm;
  }

  /**
   * Instantiates an {@link Coocurrence} from a <code>MRCOC</code> line.
   * @param line a line from <code>MRCOC</code>
   * @return the {@link Coocurrence} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Coocurrence newInstanceFromMRCOC(String line)
    throws IOException {

    //
    // Instantiate a default Coocurrence
    //
    final Coocurrence coc = new CoocurrenceImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 7, tokens);

    //
    // Set data from fields.
    //
    final Concept c = new PlaceholderConcept();
    c.setCui(tokens[1]);
    coc.setRelatedConcept(c);
    coc.setSource(Source.getSource(tokens[2]));

    //
    // Set coocurrence type
    //
    final CoocurrenceType cot = CoocurrenceType.getCoocurrenceType(tokens[3]);
    if (cot == null) {
      //
      // We have no MRDOC so fake it
      //
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("COT|" +
                                              tokens[3] + "|expanded_form||");

      CoocurrenceType.initialize(mdsc_array);
      coc.setCot(CoocurrenceType.getCoocurrenceType(tokens[3]));
    } else
      coc.setCot(cot);
    coc.setCof(tokens[4].equals("") ? 0 : Integer.parseInt(tokens[4]));
    coc.setCoa(tokens[5]);

    return coc;
  }

  /**
   * Returns a {@link Coocurrence}s from the <code>MRCOC</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from file
   */
  protected void getCoocurrences(Concept this_concept, String current_cui)
      throws IOException {
    //
    // Iterate through MRCOC
    //
    String line = null;
    while ((line = mrcoc_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching CUI, connect coocurrence and concept
      //
      else if (line.startsWith(current_cui)) {
        final Coocurrence coc = newInstanceFromMRCOC(line);
        coc.setConcept(this_concept);
        this_concept.addCoocurrence(coc);
      }

      //
      // went too far - break
      //
      else {
        break;
      }
    }
    mrcoc_source.push(line);
  }


  /**
   * Instantiates an {@link Definition} from a <code>MRDEF</code> line.
   * @param line a line from <code>MRDEF</code>
   * @return the {@link Definition} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Definition newInstanceFromMRDEF(String line)
    throws IOException {

    //
    // Instantiate a default Defintition
    //
    final Definition d = new DefinitionImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 4, tokens);

    //
    // Populate from fields
    //
    d.setSource(Source.getSource(tokens[1]));
    d.setDef(tokens[2]);

    return d;
  }

  /**
   * Returns {@link Definition}s from the <code>MRDEF</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from the file
   */
  protected void getDefinitions(Concept this_concept, String current_cui)
      throws IOException {

    //
    // Iterate through MRDEF
    //
    String line =  null;
    while ((line = mrdef_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Match found, connect defintion and concept
      //
      else if (line.startsWith(current_cui)) {
        final Definition d = newInstanceFromMRDEF(line);
        d.setConcept(this_concept);
        this_concept.addDefinition(d);
      }

      //
      // went too far - break
      //
      else {
        break;
      }
    }
    mrdef_source.push(line);
  }


  /**
   * Instantiates an {@link Locator} from a <code>MRLO</code> line.
   * @param line a line from <codE>MRLO</code>
   * @return the {@link Locator} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Locator newInstanceFromMRLO(String line)
    throws IOException {

    //
    // Instantiate a default Locator
    //
    final Locator loc = new LocatorImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 8, tokens);

    //
    // Populate locator
    //
    loc.setSource(Source.getSource(tokens[1]));
    loc.setFr(tokens[2].equals("") ? 0 : Integer.parseInt(tokens[2]));
    loc.setUn(tokens[3]);
    loc.setSui(tokens[4]);
    loc.setSna(tokens[5]);
    loc.setSoui(tokens[6]);

    return loc;
  }

  /**
   * Returns {@link Locator}s from the <code>MRLO</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from file
   */
  protected void getLocators(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRLO
    //
    String line = null;
    while((line = mrlo_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect locator and concept
      //
      else if (line.startsWith(current_cui)) {
        Locator lo = newInstanceFromMRLO(line);
        lo.setConcept(this_concept);
        this_concept.addLocator(lo);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }
    mrlo_source.push(line);
  }

  /**
   * Instantiates an {@link SemanticType} from a <code>MRSTY</code> line.
   * @param line a line from <code>MRSTY</code>
   * @return the {@link SemanticType} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  SemanticType newInstanceFromMRSTY(String line)
    throws IOException {

    //
    // Instantiate a default SemanticType
    //
    final SemanticType sty = new SemanticTypeImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 4, tokens);

    //
    // Populate semantic type
    //
    sty.setAtui(tokens[1]);
    sty.setSty(HierarchicalSemanticType.getHierarchicalSemanticType(tokens[2]));

    return sty;
  }

  /**
   * Returns {@link SemanticType}s from the <code>MRSTY</code> file with the
   * given cui and adds them to the current {@link Concept}.
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from file
   */
  protected void getSemanticTypes(Concept this_concept, String current_cui)
     throws IOException {

    //
    // Iterate through MRSTY
    //
    String line = null;
    while ((line = mrsty_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Match found, connect STY and concept
      //
      else if (line.startsWith(current_cui)) {
        final SemanticType sty = newInstanceFromMRSTY(line);
        sty.setConcept(this_concept);
        this_concept.addSemanticType(sty);
      }


      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrsty_source.push(line);
  }

  /**
   * Instantiates an {@link Relationship} from a <code>MRREL</code> line.
   * @param line a line from <code>MRREL</code>
   * @return the {@link Relationship} matching the line
   * @throws IOException if there is a problem with the file
   */
  private  Relationship newInstanceFromMRREL(String line)
    throws IOException{

    //
    // Instantiate a default Relationship
    //
    final Relationship r = new RelationshipImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 8, tokens);

    //
    // Populate relationship label
    //
    final RelationshipLabel rel = RelationshipLabel.getRelationshipLabel(tokens[1]);
    if (rel == null) {
      //
      // No MRDOC so fake it
      //
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("REL|" +
                                              tokens[1] + "|expanded_form||");

      RelationshipLabel.initialize(mdsc_array);
      r.setRel(RelationshipLabel.getRelationshipLabel(tokens[1]));
    } else
      r.setRel(rel);

    //
    // Populate other relationship fields
    //
    final Concept related_concept = new PlaceholderConcept();
    related_concept.setCui(tokens[2]);
    r.setRelatedConcept(related_concept);
    r.setSource(Source.getSource(tokens[4]));
    r.setSl(Source.getSource(tokens[5]));
    r.setSuppress(Suppressible.getSuppressible(tokens[6]));

    //
    // populate additional relationship label
    //
    final AdditionalRelationshipLabel rela = AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[3]);
    if (rela == null) {
      //
      // No MRDOC so fake it
      //
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("RELA|" +
                                              tokens[3] + "|expanded_form||");

      AdditionalRelationshipLabel.initialize(mdsc_array);
      r.setRela(AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[3]));
    } else
      r.setRela(rela);

    return r;
  }

 /**
  * Returns {@link Relationship}s from the <code>MRREL</code> file with the
   * given cui and adds them to the current {@link Concept}
   * @param this_concept the current {@link Concept}
   * @param current_cui the current CUI
   * @throws IOException if problem reading from file
   */
  protected void getRelationships(Concept this_concept, String current_cui)
    throws IOException {

    //
    // Iterate through MRREL
    //
    String line = null;
    while((line = mrrel_source.readLine()) != null) {

      //
      // Skip lines that do not match
      //
      if (line.compareTo(current_cui) < 0)
        continue;

      //
      // Matching line found, connect relationship and concept
      //
      else if (line.startsWith(current_cui)) {
        final Relationship rel = newInstanceFromMRREL(line);
        rel.setConcept(this_concept);
        this_concept.addRelationship(rel);
      }

      //
      // went too far - break and reset
      //
      else {
        break;
      }
    }

    mrrel_source.push(line);
  }
}
