/**************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Object:      OriginalMRMetathesaurusInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.ConceptMapping;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.RelationshipLabel;
import gov.nih.nlm.umls.meta.ReleaseMetadata;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.impl.ConceptMappingImpl;
import gov.nih.nlm.umls.meta.impl.LowercaseWordIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.MetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedStringIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedWordIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.ReleaseMetadataImpl;
import gov.nih.nlm.umls.meta.impl.SourceMetaDescriptionImpl;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This {@link MetathesaurusInputStream} builds
 * {@link IndexEntry}s, and {@link ReleaseMetadata} from a full set of standard
 * MR files.
 *
 * @author Deborah Shapiro
 */
public class OriginalMRMetathesaurusInputStream
  extends OriginalMRFilesConceptInputStream implements MetathesaurusInputStream {

  //
  // Private Fields
  //

  //
  // Index representing the current index file being processed
  //
  protected int current_index = 0;

  //
  // Total number of index lines
  //
  protected int index_lines;

  //
  // Number of index lines processed so far
  //
  protected int index_counter;

  //
  // Readers for index files
  //
  protected BufferedReader [] mrx_sources;

  //
  // Readers for MRCUI, MRSAB
  //
  protected BufferedReader mrcui_source, mrsab_source;

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link OriginalMRMetathesaurusInputStream};
   */
  public OriginalMRMetathesaurusInputStream() { }

  //
  // Methods
  //

  /**
   * Opens all target files in the specified directory.  The set of
   * target files is the standard MR files.
   * @param install_paths the subset directory
   * @throws IOException if the files cannot be opened
   */
  public void open(String[] install_paths) throws IOException {

    //
    // Borrow superclass functionality
    //
    super.open(install_paths);

    //
    // Open word index files
    //
    final Language[] languages = Language.getLanguages();
    int i = 0;
    mrx_sources =
      new BufferedReader[Language.getLanguagesAsList().size() + 2];
    file_names = new String[mrx_sources.length];
    for (; i < languages.length; i++) {
      file_names[i] = "MRXW." + languages[i].getAbbreviation();
      mrx_sources[i] =
	openSourceFile(file_names[i]);
    }

    //
    // Open norm index files
    //
    mrx_sources[i] = openSourceFile("MRXNW.ENG");
    file_names[i] = "MRXNW.ENG";
    mrx_sources[++i] = openSourceFile("MRXNS.ENG");
    file_names[i] = "MRXNS.ENG";

    //
    // Open metadata files
    //
    mrsab_source = openSourceFile("MRSAB");
    mrcui_source = openSourceFile("MRCUI");

    //
    // Cache line counts for MRCON, index files
    //
    getLineCounts();

    //
    // Initialize source data
    //
    String line = null;
    while ((line = mrsab_source.readLine()) != null &&
	   !line.equals("")) {
      //
      // update/complete Sources that were loaded from properties file
      //
      final SourceMetaDescription[] smd_array = new SourceMetaDescription[1];
      smd_array[0] = new SourceMetaDescriptionImpl(line);
      Source.initialize(smd_array);
    }
    mrsab_source.close();
    mrsab_source = openSourceFile("MRSAB");

  }

  /**
   * Close files opened during {@link #open(String[])}.
   * @throws IOException if files can not be closed
   */
  public void close() throws IOException {
    super.close();

    //
    // Close index files
    //
    for (int i = 0; i < mrx_sources.length; i++)
      mrx_sources[i].close();

    //
    // Close metadata files
    //
    mrsab_source.close();
    mrcui_source.close();
  }

  /**
   * Returns an {@link IndexEntry} from one of the index files.
   * We read through the index files one at a time, returning
   * all of the entries.
   * @return an {@link IndexEntry} from one of the index files
   * @throws IOException if the files cannot be read
   */
  public IndexEntry readIndexEntry() throws IOException {

    //
    // We have a current_index into mrx_sources
    // to indicate which reader should be used.
    //
    // If we read a line and it is null,
    // then increment current index
    //
    if (current_index >= mrx_sources.length)
      return null;
    final String line = mrx_sources[current_index].readLine();
    if (line == null) {
      current_index++;
      return readIndexEntry();
    } else {
      index_counter++;
      return newInstanceFromMRX(line, file_names[current_index]);
    }
  }

  /**
   * Returns the percentage of the index file lines that have been processed.
   * @return the percentage of the index file lines that have been processed
   */
  public int getIndexProgress() {
    return (int)((index_counter/(double)index_lines)*100);
  }

  /**
   * Reads through <code>MRFILES</code> to determine the size of the index files.
   * This is used in the {@link #getIndexProgress()} method.
   * @return {@link Map} of file names to lines in file
   * @throws IOException if the file cannot be opened or read
   */
  protected Map getLineCounts() throws IOException {

    //
    // Get MRCON line count
    //
    final Map lines_in_file = super.getLineCounts();

    //
    // Accumulate index file line counts
    //
    index_counter = 0;
    index_lines = 0;
    Iterator iter = lines_in_file.entrySet().iterator();
    while (iter.hasNext()) {
      Map.Entry me = (Map.Entry)iter.next();
      if (((String)me.getKey()).startsWith("MRX")) {
	  index_lines += Integer.parseInt(me.getValue().toString());
      }
    }

    return lines_in_file;
  }

  /**
   * Returns the {@link ReleaseMetadata} containing data from
   * MRSAB, MRDOC and MRCUI.
   * @return the {@link ReleaseMetadata} for this input stream
   * @throws IOException if the files could not be read
   */
  public ReleaseMetadata readReleaseMetadata()
    throws IOException {

    //
    // Instantiate ReleaseMetadata
    //
    final ReleaseMetadata release_metadata = new ReleaseMetadataImpl();
    String line = null;

    //
    // Iterate through MRSAB, add source info to release metadata
    //
    while ((line = mrsab_source.readLine()) != null && !line.equals("")) {
      final SourceMetaDescription[] smd_array = new SourceMetaDescription[1];
      smd_array[0] = new SourceMetaDescriptionImpl(line);
      release_metadata.addSourceMetaDescription(smd_array[0]);
    }

    //
    // Iterate through MRCUI, add cui mapping info to release metadata
    //
    while ((line = mrcui_source.readLine()) != null && !line.equals("")) {
      final ConceptMapping cui_mapping = newInstanceFromMRCUI(line);
      release_metadata.addConceptMapping(cui_mapping);
    }

    //
    // save references to ancillary and change files so they
    // can be copied to the target directory during output
    //
    release_metadata.addChangeFile("DELETED.CUI", openSourceFile("/CHANGE/DELETED.CUI"));
    release_metadata.addChangeFile("DELETED.LUI", openSourceFile("/CHANGE/DELETED.LUI"));
    release_metadata.addChangeFile("DELETED.SUI", openSourceFile("/CHANGE/DELETED.SUI"));
    release_metadata.addChangeFile("MERGED.CUI", openSourceFile("/CHANGE/MERGED.CUI"));
    release_metadata.addChangeFile("MERGED.LUI", openSourceFile("/CHANGE/MERGED.LUI"));

    return release_metadata;
  }

  /**
   * Returns the complete list of suis from the AMBIGSUI list.
   * They will be used in calculating entries for the output AMBIGSUI list.
   * @return list of suis
   * @throws IOException
   */
  public String[] getAmbiguousStringIdentifiers() throws IOException {
    return (String[])getAmbiguousStringIdentifiersAsList().toArray(new String[0]);
  }

  /**
   * Returns the complete list of suis from the AMBIGSUI list.
   * They will be used in calculating entries for the output AMBIGSUI list.
   * @return list of suis
   * @throws IOException
   */
  public List getAmbiguousStringIdentifiersAsList() throws IOException {
    final BufferedReader ambigsui = openSourceFile("AMBIG.SUI");
    final Set sui_set = new HashSet();
    String line = null;
    while ((line = ambigsui.readLine()) != null && !line.equals("")) {
      String[] sui_cui = FieldedStringTokenizer.split(line,"|");
      sui_set.add(sui_cui[0]);
    }
    ambigsui.close();
    return new ArrayList(sui_set);
  }

  /**
   * Returns the complete list of terms from the AMBIGLUI list.
   * They will be used in calculating entries for the output AMBIGLUI list.
   * @return list of terms
   * @throws IOException
   */
  public String[] getAmbiguousTermIdentifiers() throws IOException {
    return (String[])getAmbiguousTermIdentifiersAsList().toArray(new String[0]);
  }

  /**
   * Returns the complete list of terms from the AMBIGLUI list.
   * They will be used in calculating entries for the output AMBIGLUI list.
   * @return list of terms
   * @throws IOException
   */
  public List getAmbiguousTermIdentifiersAsList() throws IOException {
    final BufferedReader ambiglui = openSourceFile("AMBIG.LUI");
    final Set lui_set = new HashSet();
    String line = null;
    while ((line = ambiglui.readLine()) != null && !line.equals("")) {
      String[] lui_cui = FieldedStringTokenizer.split(line,"|");
      lui_set.add(lui_cui[0]);
    }
    ambiglui.close();
    return new ArrayList(lui_set);
  }


  //
  // Protected Methods
  //

  /**
   * Instantiates an {@link IndexEntry} from a MRX* line.
   * @param line a line from an index file
   * @param file_name the index file name
   * @return the {@link IndexEntry} matching the line
   * @throws IOException if there is a problem with the file
   */
  private IndexEntry newInstanceFromMRX(String line, String file_name)
    throws IOException {

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 6, tokens);

    //
    // Instantiate index entry from the right file
    //
    IndexEntry index_entry = null;
    if (file_name.equals("MRXNW.ENG"))
      index_entry = new NormalizedWordIndexEntryImpl();
    else if (file_name.equals("MRXNS.ENG"))
      index_entry = new NormalizedStringIndexEntryImpl();
    else
      index_entry = new LowercaseWordIndexEntryImpl();

    //
    // Populate index entry
    //
    index_entry.setLat(tokens[0]);
    index_entry.setValue(tokens[1]);
    index_entry.setCui(tokens[2]);
    index_entry.setLui(tokens[3]);
    index_entry.setSui(tokens[4]);

    return index_entry;
  }

  /**
   * Instantiates an {@link ConceptMapping} from a MRCUI line.
   * @param line a line from MRCUI
   * @return the {@link ConceptMapping} matching the line
   * @throws IOException if there is a problem with the file
   */
  private ConceptMapping newInstanceFromMRCUI(String line)
    throws IOException {

    //
    // Instantiate default ConceptMapping
    //
    final ConceptMapping mapping = new ConceptMappingImpl();

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 6, tokens);

    //
    // Populate concept mapping
    //
    mapping.setCui1(tokens[0]);
    mapping.setLastReleaseVersion(tokens[1]);
    final RelationshipLabel rel = RelationshipLabel.getRelationshipLabel(tokens[2]);
    if (rel == null) {
      //
      // No MRDOC so fake it
      //
      MetaDescription[] mdsc_array = new MetaDescription[1];
      mdsc_array[0] = new MetaDescriptionImpl("REL|" +
                                              tokens[2] + "|expanded_form||");

      RelationshipLabel.initialize(mdsc_array);
      mapping.setRel(RelationshipLabel.getRelationshipLabel(tokens[2]));
    } else
      mapping.setRel(rel);
    mapping.setCui2(tokens[3]);

    return mapping;
  }

}
