/**************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Object:      OriginalMRMetathesaurusOutputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ConceptMapping;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.ReleaseMetadata;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.umls.meta.TermType;
import gov.nih.nlm.umls.meta.impl.LowercaseWordIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedStringIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedWordIndexEntryImpl;
import gov.nih.nlm.util.ColumnStatistics;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.FileStatistics;
import gov.nih.nlm.util.SystemToolkit;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This {@link MetathesaurusOutputStream} generates standard MR files.
 *
 * @author Deborah Shapiro
 */
public class OriginalMRMetathesaurusOutputStream
    extends OriginalMRFilesConceptOutputStream
    implements MetathesaurusOutputStream {

  //
  // MRCOLS/MRFILES configuration inof
  //
  protected FileStatisticsConfiguration config;

  //
  // Change file names
  //
  protected String[] change_files;

  //
  // Ancillary file names
  //
  protected String[] ancillary_files;

  //
  // Maps to manage candidate ambiguous string and term identifiers
  //
  protected Map sui_to_cui_set = null;
  protected Map lui_to_cui_set = null;


  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link OriginalMRMetathesaurusOutputStream}.
   * @param config {@link FileStatisticsConfiguration}
   */
  public OriginalMRMetathesaurusOutputStream(FileStatisticsConfiguration config) {
    this.config = config;
  }

  //
  // Methods
  //

  /**
   * Opens the specified file in the directory passed to {@link #open(String)}.
   * @param name the filename
   * @return a {@link BufferedWriter} for the file
   * @throws IOException if the file could not be opened.
   */
  protected BufferedWriter openTargetFile(String name) throws IOException {
    //
    // Open file
    //
    final BufferedWriter out = super.openTargetFile(name);

    //
    // Prepare statistics
    //
    if (mr_targets.containsKey(name) || name.startsWith("MRX"))
      config.getFileStatistics(name);

    return out;
  }

  /**
   * Prints the given line to the file with the given name.
   * @param line the line to be printed
   * @param file_name printing destination
   * @throws IOException
   */
  protected void printLine(String line, String file_name) throws IOException {
    //
    // Print the line
    //
    super.printLine(line, file_name);

    //
    // Process the statistics
    //
    final FileStatistics stats = config.getFileStatistics(file_name);
    stats.processLine(line);
  }

  /**
   * Closes source files that were opened by this class's openTargets() method.
   * @throws IOException
   */
  public void close() throws IOException{
    super.close();

    //
    // Sort AMBIG files, MRCUI
    //
    SystemToolkit.sort(new File(mr_dir,"MRCUI").getAbsolutePath(), true);
    SystemToolkit.sort(new File(mr_dir,"AMBIG.LUI").getAbsolutePath());
    SystemToolkit.sort(new File(mr_dir,"AMBIG.SUI").getAbsolutePath());
  }

  /**
   * Writes the specified {@link Concept}.
   * @param c the specified {@link Concept}
   * @throws IOException if anything goes wrong
   */
  public void writeConcept(Concept c) throws IOException {

    super.writeConcept(c);

    //
    // Maintain ambiguous sui and lui maps
    //
    final Atom[] atoms = c.getAtoms();
    for (int i=0; i<atoms.length; i++)
      ambigCheck(atoms[i]);

  }

  /**
   * Writes an {@link IndexEntry} to the appropriate MR index file.
   * @param index_entry the {@link IndexEntry} to write
   * @throws IOException if the file could not be written
   */
  public void writeIndexEntry(IndexEntry index_entry)
    throws IOException {

    //
    // Prepare index entry line
    //
    final StringBuffer sb = new StringBuffer(50);
    sb.append(index_entry.getLat()); sb.append("|");
    sb.append(index_entry.getValue()); sb.append("|");
    sb.append(index_entry.getCui()); sb.append("|");
    sb.append(index_entry.getLui()); sb.append("|");
    sb.append(index_entry.getSui()); sb.append("|");

    //
    // Determine where to write it, write the line
    //
    if (index_entry instanceof LowercaseWordIndexEntryImpl)
      printLine(sb.toString(), "MRXW." + index_entry.getLat());
    else if (index_entry instanceof NormalizedWordIndexEntryImpl)
      printLine(sb.toString(),"MRXNW.ENG");
    else if (index_entry instanceof NormalizedStringIndexEntryImpl)
      printLine(sb.toString(),"MRXNS.ENG");
  }

  /**
   * Writes the specified {@link ReleaseMetadata} to <codE>MRCUI</code>.
   * @param release_metadata the specified {@link ReleaseMetadata}
   * @throws IOException if anything goes wrong
   */
  protected void writeMRCUI(ReleaseMetadata release_metadata)
    throws IOException {

    //
    // Iterate through concept mappings
    //
    final List cm_list = release_metadata.getConceptMappingsAsList();

    for (int i=0; i<cm_list.size(); i++) {
      //
      // Prepare MRCUI line
      //
      final StringBuffer sb = new StringBuffer(200);
      final ConceptMapping mapping = (ConceptMapping)cm_list.get(i);

      sb.append(mapping.getCui1());
      sb.append("|");
      sb.append(mapping.getVer());
      sb.append("|");
      sb.append(mapping.getRelationship());
      sb.append("|");
      if (mapping.getCui2() == null || mapping.getCui2().equals("")) {
        sb.append("||");
      }
      else {
        sb.append(mapping.getCui2());
        sb.append("|");
        if (mapping.inCurrentSubset())
          sb.append("Y|");
        else
          sb.append("N|");
      }

      //
      // Write MRCUI line
      //
      printLine(sb.toString(), "MRCUI");

    }

    //
    // Iterate over SUBX removed cuis
    //
    String[] subx_list = release_metadata.getCUIsRemovedFromSubset();

    for (int i = 0; i < subx_list.length; i++) {
      //
      // Prepare a line of MRCUI
      //
      final StringBuffer sb = new StringBuffer(50);
      sb.append(subx_list[i]); // CUI1
      sb.append("|");
      // requires that release_metadata.setReleaseVersion() is called
      sb.append(release_metadata.getReleaseVersion()).append("|");
      sb.append("SUBX|"); // REL
      sb.append("|"); // CUI2
      sb.append("|");

      //
      // Write a line and process file statistics
      //
      printLine(sb.toString(), "MRCUI");
    }

  }

  /**
   * Writes <code>MRSAB</code>.
   * @param release_metadata the {@link ReleaseMetadata} containing MRSAB data
   * @throws IOException if anything goes wrong
   */
  protected void writeMRSAB(ReleaseMetadata release_metadata)  throws IOException {

    //
    // Iterate through sources
    //
    final SourceMetaDescription[] source_metadata = release_metadata.getSourceMetadata();
    final List mrsab_lines = new ArrayList();
    for(int i = 0; i < source_metadata.length; i++) {

      //
      // Prepare MRSAB line
      //
      final StringBuffer sb = new StringBuffer(200);
      final SourceMetaDescription source = source_metadata[i];
      sb.append(source.getVcui()).append("|");
      sb.append(source.getRcui()).append("|");
      sb.append(source.getVersionedSourceAbbreviation()).append("|");
      sb.append(source.getRootSourceAbbreviation()).append("|");
      sb.append(source.getOfficialName()).append("|");
      sb.append(source.getSourceFamily()).append("|");
      sb.append(source.getVersion()).append("|");
      final Date start_date = source.getValidStartDate();
      final SimpleDateFormat df = new SimpleDateFormat("yyyy_MM_dd");
      sb.append(start_date == null ? "" : df.format(start_date)).append("|");
      final Date end_date = source.getValidEndDate();
      sb.append(end_date == null ? "" : df.format(end_date)).append("|");
      sb.append(source.getInsertMetaVersion()).append("|");
      sb.append(source.getRemoveMetaVersion()).append("|");
      sb.append(source.getLicenseContact()).append("|");
      sb.append(source.getContentContact()).append("|");
      if(Integer.parseInt(source.getRestrictionLevel()) != -1)
        sb.append(source.getRestrictionLevel()+"");
      sb.append("|");
      if(source.getTermFrequency() > 0)
        sb.append(source.getTermFrequency()+"");
      sb.append("|");
      if(source.getConceptFrequency() > 0)
        sb.append(source.getConceptFrequency() + "");
      sb.append("|");
      sb.append(source.getContextType()); sb.append("|");
      final TermType[] term_types = source.getTermTypes();
      if(term_types != null && term_types.length != 0) {
        for(int j = 0; j < term_types.length - 1; j++)
  	  sb.append(term_types[j].getAbbreviation()).append(",");
        sb.append(term_types[term_types.length-1].getAbbreviation());
      }
      sb.append("|");
      final AttributeName[] attributes = source.getAttributeNames();
      if(attributes != null && attributes.length != 0) {
        for(int j = 0; j < attributes.length - 1; j++)
  	  sb.append(attributes[j].getAbbreviation()).append(",");
        sb.append(attributes[attributes.length-1].getAbbreviation());
      }
      sb.append("|");
      sb.append(source.getLanguage() != null ? source.getLanguage().getAbbreviation() : "").append("|");
      sb.append(source.getCharacterEncoding()).append("|");
      sb.append(source.isCurrentVersion() ? "Y" : "N" ).append("|");
      sb.append(source.inCurrentSubset() ? "Y" : "N" ).append("|");

      //
      // Add line to list
      //
      mrsab_lines.add(sb.toString());
    }

    //
    // Sort lines and write MRSAB
    //
    Collections.sort(mrsab_lines);
    for(int i = 0; i < mrsab_lines.size(); i++) {
      printLine(mrsab_lines.get(i).toString(), "MRSAB");
    }

  }

  /**
   * Write <code>MRRANK</code>.
   * @param release_metadata the {@link ReleaseMetadata} containing <code>MRRANK</code> data
   * @throws IOException if anything goes wrong
   */
  protected void writeMRRANK(ReleaseMetadata release_metadata) throws IOException {

    //
    // Iterate through termgroups
    //
    final SourceTermType[] tg_list = release_metadata.getSourceTermTypePrecedence();
    for(int i = 0; i < tg_list.length; i++) {

      //
      // Prepare MRRANK line
      //
      final StringBuffer sb = new StringBuffer(200);
      final SourceTermType this_tg = tg_list[i];
      final String rank = "0000" + (tg_list.length -i) + "|";
      sb.append(rank.substring(rank.length()-5));
      if (versioned_output)
        sb.append(this_tg.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(this_tg.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(this_tg.getTermType().getAbbreviation()).append("|");
      sb.append(this_tg.isSuppressible() ? "Y" : "N").append("|");

      //
      // Write MRRANK line
      //
      printLine(sb.toString(),"MRRANK");
    }
  }

  /**
   * Writes an {@link ReleaseMetadata} to the appropriate MR metadata files.
   * @param release_metadata the {@link ReleaseMetadata} to write
   * @throws IOException if the file could not be written
   */
  public void writeReleaseMetadata(ReleaseMetadata release_metadata) throws
      IOException {

    //
    // copy change files from installation directory to target directory
    // keeping statistics information for MRFILES/COLS
    //
    change_files = release_metadata.getChangeFileNames();
    final File change_dir = new File(mr_dir, "CHANGE");
    if (!change_dir.exists() || !change_dir.isDirectory())
      change_dir.mkdir();
    
    //
    // Iterate through change files
    //
    for (int i = 0; i < change_files.length; i++) {

      final BufferedReader br = (BufferedReader) release_metadata.getChangeFile(
          change_files[i]);


      //
      // Determine file name
      //
      if (change_files[i].equals("DELETEDSUI.RRF")) change_files[i] = "DELETED.SUI";
      else if (change_files[i].equals("DELETEDCUI.RRF")) change_files[i] = "DELETED.CUI";
      else if (change_files[i].equals("DELETEDLUI.RRF")) change_files[i] = "DELETED.LUI";
      else if (change_files[i].equals("MERGEDLUI.RRF")) change_files[i] = "MERGED.LUI";
      else if (change_files[i].equals("MERGEDCUI.RRF")) change_files[i] = "MERGED.CUI";

      //
      // Write change file
      //
      final BufferedWriter pw = openTargetFile("CHANGE/" + change_files[i]);

      final FileStatistics stats = (FileStatistics)config.getFileStatistics(
          "CHANGE/" + change_files[i]);

      //
      // Copy change file data
      //
      String line;
      while ( (line = br.readLine()) != null) {
        pw.write(stats.processLine(line));
        pw.newLine();
      }
      pw.close();

    } // end of iteration through change files

    //
    // write out subsetted ancillary files to target directory
    // keeping statistics information for MRFILES/COLS
    //

    //
    // Open ancillary file AMBIG.SUI
    //
    final BufferedWriter as_pw = openTargetFile("AMBIG.SUI");
    final FileStatistics as_stats = config.getFileStatistics("AMBIG.SUI");

    //
    // Write AMBIG.SUI entries
    //
    final Iterator as_iter = sui_to_cui_set.entrySet().iterator();
    while (as_iter.hasNext()) {
      final Map.Entry me = (Map.Entry) as_iter.next();
      if (me.getValue () != null && ((HashSet) me.getValue()).size() > 1) {
        final String sui = (String) me.getKey();
        final HashSet cui_set = (HashSet) me.getValue();
        final String[] cuis = (String[])cui_set.toArray(new String[0]);
        for (int i = 0; i < cuis.length; i++) {
          as_pw.write(as_stats.processLine(sui + "|" + cuis[i] + "|"));
          as_pw.newLine();
        }
      }
    }
    as_pw.close();

    //
    // Open ancillary file AMBIG.LUI
    //
    final BufferedWriter al_pw = openTargetFile("AMBIG.LUI");
    final FileStatistics al_stats = config.getFileStatistics("AMBIG.LUI");

    //
    // Write AMBIG.LUI entries
    //
    final Iterator al_iter = lui_to_cui_set.entrySet().iterator();
    while (al_iter.hasNext()) {
      final Map.Entry me = (Map.Entry) al_iter.next();
      if (me.getValue () != null && ((HashSet) me.getValue()).size() > 1) {
        final String lui = (String) me.getKey();
        final HashSet cui_set = (HashSet) me.getValue();
        final String[] cuis = (String[])cui_set.toArray(new String[0]);
        for (int i = 0; i < cuis.length; i++) {
          al_pw.write(al_stats.processLine(lui + "|" + cuis[i] + "|"));
          al_pw.newLine();
        }
      }
    }
    al_pw.close();

    //
    // Write MRCUI
    //
    writeMRCUI(release_metadata);

    //
    // Write MRSAB
    //
    writeMRSAB(release_metadata);

    //
    // Write MRRANK
    //
    writeMRRANK(release_metadata);

    //
    // Write MRFILES/MRCOLS
    //
    writeFileColumnStats();

  }

  /**
   * Sets the list of suis that are candidates to be ambiguous.
   * @param suis the list of suis that are candidates to be ambiguous
   */
  public void setCandidateAmbiguousStringIdentifiers(String [] suis) {
    if (sui_to_cui_set == null)
      sui_to_cui_set = new HashMap();
    for (int i=0; i < suis.length; i++)
      sui_to_cui_set.put(suis[i], null);
  }

  /**
   * Sets the list of suis that are candidates to be ambiguous.
   * @param suis the list of suis that are candidates to be ambiguous
   */
  public void setCandidateAmbiguousStringIdentifiers(List suis) {
    setCandidateAmbiguousStringIdentifiers((String[])suis.toArray(new String[0]));
  }

  /**
   * Sets the list of terms that are candidates to be ambiguous.
   * @param luis candidate terms
   */
  public void setCandidateAmbiguousTermIdentifiers(String[] luis) {
    if (lui_to_cui_set == null)
      lui_to_cui_set = new HashMap();
    for (int i=0; i<luis.length; i++)
      lui_to_cui_set.put(luis[i], null);
  }

  /**
   * Sets the list of luis that are candidates to be ambiguous.
   * @param luis the list of luis that are candidates to be ambiguous
   */
  public void setCandidateAmbiguousTermIdentifiers(List luis) {
    setCandidateAmbiguousTermIdentifiers((String[])luis.toArray(new String[0]));
  }


  //
  // Protected Methods
  //

  /**
   * This method computes the file and columns statistics
   * for the MRFILES and MRCOLS tables themselves.
   * Should it be using "\n"?
   * @throws IOException if problem writing to files
   */
  protected void writeFileColumnStats() throws IOException {

    //
    // Iterate through file statistics for all files
    //
    final StringBuffer mrcols_static = new StringBuffer(200);
    final StringBuffer mrfiles_static = new StringBuffer(200);
    for(int i = 0; i < target_files.length; i++) {
      final FileStatistics file_stats = config.getFileStatistics(target_files[i]);
      //
      // Build up dummy in-memory versions of MRCOLS MRFILES
      //
      if(file_stats.getFileName().startsWith("MRCOLS") ||
         file_stats.getFileName().startsWith("MRFILES")) {
        continue;
      } else {
        mrfiles_static.append(getFileTextFromFileStatistics(file_stats));
        mrcols_static.append(getColumnTextFromFileStatistics(file_stats));
      }
    }

    final Language[] lats = Language.getLanguages();
    for(int i = 0; i < lats.length; i++) {
      final FileStatistics stats = config.getFileStatistics("MRXW." + lats[i]);
      mrfiles_static.append(getFileTextFromFileStatistics(stats));
      mrcols_static.append(getColumnTextFromFileStatistics(stats));
    }

    for (int i = 0; i < change_files.length; i++) {
      final FileStatistics stats = config.getFileStatistics("CHANGE/" +
          change_files[i]);
      mrfiles_static.append(getFileTextFromFileStatistics(stats));
      mrcols_static.append(getColumnTextFromFileStatistics(stats));
    }

    FileStatistics stats = config.getFileStatistics("AMBIG.SUI");
    mrfiles_static.append(getFileTextFromFileStatistics(stats));
    mrcols_static.append(getColumnTextFromFileStatistics(stats));

    stats = config.getFileStatistics("AMBIG.LUI");
    mrfiles_static.append(getFileTextFromFileStatistics(stats));
    mrcols_static.append(getColumnTextFromFileStatistics(stats));

    //
    // Iterate through, making guessing contents for MRFILES/COLS
    // When a revised guess matches an old one, we have converged
    // then write MRFILES/COLS
    //
    StringBuffer mrcols = new StringBuffer(200);
    StringBuffer mrfiles = new StringBuffer(200);
    StringBuffer mrcols_new = new StringBuffer(200);
    StringBuffer mrfiles_new = new StringBuffer(200);
    try {

      while(true) {

        //
        // Start with static data for MRCOLS
        //
	mrcols = new StringBuffer(200);
	mrcols.append(mrcols_new);
	mrcols_new = new StringBuffer(200);
	mrcols_new.append(mrcols_static);

        //
        // Start with static data for MRFILES
        //
	mrfiles = new StringBuffer(200);
	mrfiles.append(mrfiles_new);
	mrfiles_new = new StringBuffer(200);
	mrfiles_new.append(mrfiles_static);

        //
        // Compute MRFILES line for MRFILES
        //
	FileStatistics mrfiles_stats = config.getFileStatistics("MRFILES");
	String[] mrfiles_lines =
	  FieldedStringTokenizer.split(mrfiles.toString(), "\n");
	for(int i = 0; i < mrfiles_lines.length; i++) {
          if (!mrfiles_lines[i].equals(""))
  	    mrfiles_stats.processLine(mrfiles_lines[i]);
	}

        //
        // Compute MRFILES line for MRCOLS
        //
	FileStatistics mrcols_stats = config.getFileStatistics("MRCOLS");
	String[] mrcols_lines =
	  FieldedStringTokenizer.split(mrcols.toString(), "\n");
	for(int i = 0; i < mrcols_lines.length; i++) {
          if (!mrcols_lines[i].equals(""))
	    mrcols_stats.processLine(mrcols_lines[i]);
	}

        //
        // Append MRFILES,MRCOLS entries to MRFILES
        //
	mrfiles_new.append(getFileTextFromFileStatistics(mrfiles_stats));
	mrfiles_new.append(getFileTextFromFileStatistics(mrcols_stats));
	mrcols_new.append(getColumnTextFromFileStatistics(mrcols_stats));
	mrcols_new.append(getColumnTextFromFileStatistics(mrfiles_stats));

        //
        // If everything is the same, we are done.
        //
	if(mrfiles.toString().equals(mrfiles_new.toString()) &&
	   mrcols.toString().equals(mrcols_new.toString()))
	  break;
      }

      //
      // Split MRCOLS/MRFILES into lines
      //
      String[] mrfiles_output =
	FieldedStringTokenizer.split(mrfiles.toString(), "\n");
      String[] mrcols_output =
	FieldedStringTokenizer.split(mrcols.toString(), "\n");

      //
      // Sort MRCOLS/MRFILES
      //
      Arrays.sort(mrfiles_output);
      Arrays.sort(mrcols_output);

      //
      // Write MRFILES and MRCOLS
      //  (start at index 1 because FST.split causes a blank entry)
      //
      BufferedWriter mrfiles_target = (BufferedWriter)mr_targets.get("MRFILES");
      BufferedWriter mrcols_target = (BufferedWriter)mr_targets.get("MRCOLS");
      for (int i = 0; i < mrfiles_output.length; i++)
        if (!mrfiles_output[i].equals("")) {
          mrfiles_target.write(mrfiles_output[i]);
          mrfiles_target.newLine();
        }

      for (int i = 0; i < mrcols_output.length; i++)
        if (!mrcols_output[i].equals("")) {
          mrcols_target.write(mrcols_output[i]);
          mrcols_target.newLine();
        }
    } catch (Exception e) {
      //
      // Exception here means something weird happened
      //
      IOException ioe = new IOException(
       "Problem processing lines for FileStatistics ");
      ioe.initCause(e);
      throw ioe;
    }
  }

  /**
   * Returns a <code>MRFILES</codE> line.
   * @param file_stats {@link FileStatistics}
   * @return a <code>MRFILES</code> line
   */
  protected String getFileTextFromFileStatistics(FileStatistics file_stats) {

    //
    // Build MRFILES line
    //
    StringBuffer sb = new StringBuffer(200);
    ColumnStatistics [] col_stats = file_stats.getAllColumnStatistics();
    sb.append(file_stats.getFileName() + "|");
    sb.append(file_stats.getDescription() + "|");

    //
    // Build column list
    //
    for(int j=0; j<col_stats.length; j++) {
      ColumnStatistics this_col = col_stats[j];
      sb.append(this_col.getColumnName());
      sb.append(",");
    }

    sb.replace(sb.length()-1, sb.length(), "|");
    sb.append(col_stats.length + "|");
    sb.append(file_stats.getLineCount() + "|");
    sb.append(file_stats.getByteCount() + "|");

    //
    // For here always use \n as separator
    //
    sb.append("\n");

    return sb.toString();
  }


  /**
   * Returns a MRCOLS line.
   * @param file_stats {@link FileStatistics}
   * @return a MRCOLS line
   */
  protected String getColumnTextFromFileStatistics(FileStatistics file_stats) {

    //
    // Iterate through column statistics
    //
    StringBuffer sb = new StringBuffer(200);
    DecimalFormat formatter = new DecimalFormat("0.00");
    ColumnStatistics [] col_stats = file_stats.getAllColumnStatistics();
    for(int j=0; j<col_stats.length; j++) {

      //
      // build MRCOLS line
      //
      ColumnStatistics this_col = col_stats[j];
      sb.append(this_col.getColumnName() + "|");
      sb.append(this_col.getDescription() + "||");
      sb.append(this_col.getMinLength() + "|");
      sb.append(formatter.format(this_col.getAverageLength()));
      sb.append("|");
      sb.append(this_col.getMaxLength() + "|");
      sb.append(this_col.getFileName() + "|");
      sb.append(this_col.getDataType() + "|");

      //
      // For here always use \n as separator
      //
      sb.append("\n");
    }

    return sb.toString();
  }

  /**
   * Checks the maps to see if the given {@link Atom}s sui
   * or lui should be added to the ambiguous maps.
   * @param atom the {@link Atom}to check
   */
  protected void ambigCheck(Atom atom) {
    final String sui = atom.getSui();
    if (sui_to_cui_set.containsKey(sui)) {
      Set cui_set = (Set)sui_to_cui_set.get(sui);
      if (cui_set == null) {
        cui_set = new HashSet();
        sui_to_cui_set.put(sui,cui_set);
      }
      cui_set.add(atom.getConcept().getCui());
    }

    final String lui = atom.getLui();
    if (lui_to_cui_set.containsKey(lui)) {
      Set cui_set = (Set)lui_to_cui_set.get(lui);
      if (cui_set == null) {
        cui_set = new HashSet();
        lui_to_cui_set.put(lui, cui_set);
      }
      cui_set.add(atom.getConcept().getCui());
    }
  }

}
