/**************************************************************************
 *
 * Package:     gov.nih.nlm.umls.io
 * Object:      RichMRFilesIndexLookup.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.impl.ConceptImpl;
import gov.nih.nlm.umls.meta.impl.PlaceholderConcept;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.SystemToolkit;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Peforms a lookup in the index files to locate the
 * {@link Concept}s that contain given words.
 *
 * @author Deborah Shapiro
 */
public class RichMRFilesIndexLookup  {

  //
  // Private Fields
  //


  //
  // The directory containing the RRF files
  //
  private String dir = null;

  //
  // LAnguage to restrict by
  //
  private String lat = null;

  //
  // Maximum results to constrain searches
  //
  private int max_results = 50;

  //
  // Total result count from a search
  //
  private int total_result_ct;

  private double threshold;

  //
  // Caches
  //
  //private Map child_cache = null;
  //private Map sibling_cache = null;
  private Map ancestor_cache = null;
  //private Map count_child_cache = null;
  //private Map count_sibling_cache = null;
  private Map atom_cache = null;

  //
  // RandomAccessFile references (so as not not have to keep opening them)
  //
  private RandomAccessFile raf_mrconso = null;
  private RandomAccessFile raf_mrhier = null;
  private RandomAccessFile raf_ptr_rela_mrhier = null;
  private RandomAccessFile raf_sui_ct_mrxw = null;
  private RandomAccessFile raf_aui_mrconso = null;
  private RandomAccessFile raf_sui_ct_norm = null;
  private RandomAccessFile raf_mrxnw_eng = null;

  //
  // Constructors
  //

  /**
   * Instantiates an {@link RichMRFilesIndexLookup} from the
   * specified input directory.
   * @param dir subset directory where index files can be found
   * @throws IOException
   */
  public RichMRFilesIndexLookup(String dir) throws IOException {

    this.dir = dir;
    File f = new File(dir, "MRCONSO.RRF");
    if (f.exists())
      raf_mrconso = new RandomAccessFile(f, "r");

    f = new File(dir, "MRHIER.RRF");
    if (f.exists())
      raf_mrhier = new RandomAccessFile(f, "r");

    f = new File(dir, "indexes/PTR_AUI_RELA_MRHIER.x");
    if (f.exists())
      raf_ptr_rela_mrhier = new RandomAccessFile(f, "r");

    f = new File(dir, "indexes/SUI_CT_MRXW.x");
    if (f.exists())
      raf_sui_ct_mrxw = new RandomAccessFile(f, "r");

    f = new File(new File(dir, "indexes"),"AUI_MRCONSO.x");
    if (f.exists())
      raf_aui_mrconso = new RandomAccessFile(f, "r");

    f = new File(dir, "indexes/SUI_CT.x");
    if (f.exists())
      raf_sui_ct_norm = new RandomAccessFile(f, "r");

    f = new File(dir, "MRXNW_ENG.RRF");
    if (f.exists())
      raf_mrxnw_eng = new RandomAccessFile(f, "r");

    // Create caches
    final int MAX_ENTRIES = 16000;
    /*child_cache = new LinkedHashMap(MAX_ENTRIES+1, .75F, true) {
        // This method is called just after a new entry has been added
        public boolean removeEldestEntry(Map.Entry eldest) {
            return size() > MAX_ENTRIES;
        }
    };

    sibling_cache = new LinkedHashMap(MAX_ENTRIES+1, .75F, true) {
        // This method is called just after a new entry has been added
        public boolean removeEldestEntry(Map.Entry eldest) {
            return size() > MAX_ENTRIES;
        }
    };*/

    ancestor_cache = new LinkedHashMap(MAX_ENTRIES+1, .75F, true) {
        // This method is called just after a new entry has been added
        public boolean removeEldestEntry(Map.Entry eldest) {
            return size() > MAX_ENTRIES;
        }
    };

    /*count_child_cache = new LinkedHashMap(MAX_ENTRIES+1, .75F, true) {
        // This method is called just after a new entry has been added
        public boolean removeEldestEntry(Map.Entry eldest) {
            return size() > MAX_ENTRIES;
        }
    };*/

    /*count_sibling_cache = new LinkedHashMap(MAX_ENTRIES+1, .75F, true) {
        // This method is called just after a new entry has been added
        public boolean removeEldestEntry(Map.Entry eldest) {
            return size() > MAX_ENTRIES;
        }
    };*/

    atom_cache = new LinkedHashMap((5*MAX_ENTRIES)+1, .75F, true) {
        // This method is called just after a new entry has been added
        public boolean removeEldestEntry(Map.Entry eldest) {
            return size() > (5*MAX_ENTRIES);
        }
    };

  }

  /**
   * Release resources being used by this class.
   * @throws IOException
   */
  public void close() throws IOException {
    if (raf_mrconso != null)
      raf_mrconso.close();
    if (raf_mrhier != null)
      raf_mrhier.close();
   if (raf_ptr_rela_mrhier != null )
     raf_ptr_rela_mrhier.close();
   if (raf_sui_ct_mrxw != null)
     raf_sui_ct_mrxw.close();
   if (raf_aui_mrconso != null)
     raf_aui_mrconso.close();
   if (raf_sui_ct_norm != null)
     raf_sui_ct_norm.close();
   if (raf_mrxnw_eng != null)
     raf_mrxnw_eng.close();

  }

  //
  // Methods
  //
  /**
   * Looks up the given words in the appropriate index files and returns
   * the {@link Concept}s whose Strings contain all of the words.
   * Right-handed wildcards are supported via a '*' character.
   * @param p_words to be searched for
   * @return {@link Concept}<code>[]</code>
   * @throws IOException
   */
  public Concept[] findConceptsByWords(String[] p_words) throws IOException {
    //
    // variables
    //
    // modified list of search terms
    final List search_term_list = new ArrayList();

    // map of suis to the search terms found in that sui
    final Map sui_to_words = new HashMap(1000);

    // map of suis to their cuis
    final Map sui_to_cuis = new HashMap(1000);

    // map of suis to the number of words in their string
    final Map sui_to_sui_word_ct = new HashMap(1000);

    // map of search terms to the number of their appearance in the index
    final Map word_to_freq = new HashMap(1000);

    // cui -> RankedCui
    final Map ranked_cuis = new HashMap(1000);

    //
    // Sort words
    //
    String[] words = SystemToolkit.getIndexWords(p_words);
    final int search_string_word_ct = words.length;

    //
    // Iterate through all words
    //
    for (int i = 0; i < words.length; i++) {
      String word = words[i];

      //
      // skip empty words
      //
      if (word.equals(""))
        continue;

      //
      // apply wildcards (convert * to .*)
      //
      if (word.endsWith("*"))
        word = word.substring(0, word.length() - 1) + ".*";

      search_term_list.add(word);

      //
      // Look up results from this word
      //
      final StringBuffer sb = new StringBuffer(50);
      sb.append(getLanguage()).append("|").append(word).append("|");
      final PushBackReader sui_ct_index = openReader(raf_sui_ct_mrxw, sb.toString());

      int word_freq = 0;
      String line = null;
      while ( (line = sui_ct_index.readLine()) != null) {
        // Get CUI, SUI, word count
        final String[] tokens = FieldedStringTokenizer.split(line, "|", 6);
        final String cui = tokens[2];
        final String sui = tokens[4];

        //
        // If length of SUI string is to far away from
        // length of search string, bail
        //
        final Integer sui_word_count = new Integer(tokens[5]);
        if (sui_word_count.intValue() > (search_string_word_ct + getWordCountThreshold()))
          continue;

        if (!sui_to_sui_word_ct.containsKey(sui)) {
          sui_to_sui_word_ct.put(sui, sui_word_count);
        }

        // hash SUI->CUI
        if (!sui_to_cuis.containsKey(sui))
          sui_to_cuis.put(sui,new HashSet());
        ((Set)sui_to_cuis.get(sui)).add(cui);

        // hash SUI-> word set
        if (sui_to_words.containsKey(sui)) {
          final Set word_set = (HashSet) sui_to_words.get(sui);
          word_set.add(word);
        } else {
          final Set word_set = new HashSet();
          word_set.add(word);
          sui_to_words.put(sui, word_set);
        }
        word_freq++;
      }

      word_to_freq.put(word, new Integer(word_freq));

    } // end of for each word

    //
    // If words return 0 results, we are done
    //
    if (sui_to_words.size() == 0) {
      return new Concept[0];
    }

    //
    // Go through each sui, determine its rank based on search terms,
    // update ranked_cuis to have the highest rank for that cui
    //
    final Iterator iter = sui_to_words.entrySet().iterator();
    while (iter.hasNext()) {
      final Map.Entry entry = (Map.Entry)iter.next();
      final String sui = (String)entry.getKey();
      final Set sui_words = (HashSet)entry.getValue();
      double sui_rank = 0.0;

      final int sui_word_ct = ((Integer)sui_to_sui_word_ct.get(sui)).intValue();

      double matching_word_ct = 0.0;
      for (int j = 0; j < search_string_word_ct; j++) {
        final String word = (String)search_term_list.get(j);
        final int word_frequency =
            ((Integer)word_to_freq.get(word)).intValue();

        if (sui_words.contains(word)) {
          matching_word_ct ++;
          final double word_rank = 1.0 / word_frequency;
          sui_rank += word_rank;
        }
      }

      // Rank by % of words matched
      // Weight greater numbers of matched words higher
      final double multiplier = (matching_word_ct/sui_word_ct)*2;

      sui_rank *= multiplier;

      // Rank exact matches much higher
      if (matching_word_ct == sui_word_ct &&
          matching_word_ct == search_string_word_ct) {
        sui_rank += 100;
      }

      //
      // Rank the CUI according to the SUI
      //
      final String[] cuis_for_sui = (String[])((Set)sui_to_cuis.get(sui)).toArray(new String[0]);
      for (int i = 0; i < cuis_for_sui.length; i++) {
        final String cui = cuis_for_sui[i];
        if (ranked_cuis.containsKey(cui)) {
          RankedCui ranked_cui = (RankedCui) ranked_cuis.get(cui);
          if (sui_rank > ranked_cui.getRank()) {
            ranked_cui.setSui(sui);
            ranked_cui.setRank(sui_rank);
          }
        } else if (sui_rank > threshold) {
          RankedCui ranked_cui = new RankedCui(cui, sui, sui_rank);
          ranked_cuis.put(cui, ranked_cui);
        }
      }
    }

    //
    // Sort the cuis according to ranks and get Concepts to be returned
    //
    final List sorted_cuis = new ArrayList(ranked_cuis.values());
    final Comparator comp =
      new Comparator() {
        public int compare(Object o1, Object o2) {
          RankedCui r1 = (RankedCui)o1;
          RankedCui r2 = (RankedCui)o2;
          int x =  (int)((r2.getRank() - r1.getRank())*1000);
          if (x == 0) return r1.getCui().compareTo(r2.getCui());
          else return x;
        }
      };
    Collections.sort(sorted_cuis, comp);

    final Concept[] sorted_concepts = new Concept[sorted_cuis.size()];
    for (int i = 0; i < sorted_concepts.length; i++) {
      final RankedCui rc = (RankedCui)sorted_cuis.get(i);
      final String cui = rc.getCui();
      final double rank = rc.getRank();
      final RichMRFilesSingleConceptInputStream rrf =
          new RichMRFilesSingleConceptInputStream(cui, dir);
      final Concept concept =
        new DisplayableConcept(
          rrf.getConceptWithAtomsOnly(), rank);
      sorted_concepts[i] = concept;
    }

    total_result_ct = sorted_concepts.length;
    return sorted_concepts;

  }

  /**
   * Looks up the given normalized words in the appropriate index files and returns
   * the {@link Concept}s whose Strings contain all of the words.
   * Right-handed wildcards are supported via a '*' character.
   * @param p_words to be searched for
   * @return {@link Concept}<code>[]</code>
   * @throws IOException
   */
  public Concept[] findConceptsByNormWords(String[] p_words) throws IOException {
    //
    // variables
    //
    // modified list of search terms
    final List search_term_list = new ArrayList();

    // map of suis to the search terms found in that sui
    final Map sui_to_words = new HashMap(1000);

    // map of suis to their cuis
    final Map sui_to_cui = new HashMap(1000);

    // map of suis to the number of words in their string
    final Map sui_to_sui_word_ct = new HashMap(1000);

    // map of search terms to the number of their appearance in the index
    final Map word_to_freq = new HashMap(1000);

    // cui -> RankedCui
    final Map ranked_cuis = new HashMap(1000);

    //
    // Sort words
    //
    String[] words = SystemToolkit.getIndexWords(p_words);
    final int search_string_word_ct = words.length;

    //
    // Iterate through all words
    //
    for (int i = 0; i < words.length; i++) {
      String word = words[i];

      //
      // skip empty words
      //
      if (word.equals(""))
        continue;

      //
      // apply wildcards (convert * to .*)
      //
      if (word.endsWith("*"))
        word = word.substring(0, word.length() - 1) + ".*";

      search_term_list.add(word);

      //
      // Look up results from this word
      //
      final StringBuffer sb = new StringBuffer(50);
      sb.append("ENG|").append(word).append("|");
      final PushBackReader raf_mrxnw_eng_index = openReader(raf_mrxnw_eng, sb.toString());

      int word_freq = 0;
      String line = null;
      while ( (line = raf_mrxnw_eng_index.readLine()) != null) {
        // Get CUI, SUI
        final String[] tokens = FieldedStringTokenizer.split(line, "|", 5);
        final String cui = tokens[2];
        final String sui = tokens[4];
        final PushBackReader raf_sui_ct_norm_index = openReader(raf_sui_ct_norm, sui);
        String sui_ct_line = null;
        Integer sui_word_count = null;

        // attention add norm word|sui ct cache
        while ( ( sui_ct_line = raf_sui_ct_norm_index.readLine()) != null) {
        	final String[] sui_ct_tokens = FieldedStringTokenizer.split(sui_ct_line, "|", 2);
          sui_word_count = new Integer(sui_ct_tokens[1]);
        }

        //
        // If length of SUI string is too far away from
        // length of search string, bail
        //
        if (sui_word_count.intValue() > (search_string_word_ct + getWordCountThreshold()))
          continue;

        if (!sui_to_sui_word_ct.containsKey(sui)) {
          sui_to_sui_word_ct.put(sui, sui_word_count);
        }

        // hash SUI->CUI
        sui_to_cui.put(sui, cui);

        // hash SUI-> word set
        if (sui_to_words.containsKey(sui)) {
          final Set word_set = (HashSet) sui_to_words.get(sui);
          word_set.add(word);
        } else {
          final Set word_set = new HashSet();
          word_set.add(word);
          sui_to_words.put(sui, word_set);
        }
        word_freq++;
      }

      word_to_freq.put(word, new Integer(word_freq));

    } // end of for each word

    //
    // If words return 0 results, we are done
    //
    if (sui_to_words.size() == 0) {
      return new Concept[0];
    }

    //
    // Go through each sui, determine its rank based on search terms,
    // update ranked_cuis to have the highest rank for that cui
    //
    final Iterator iter = sui_to_words.entrySet().iterator();
    while (iter.hasNext()) {
      final Map.Entry entry = (Map.Entry)iter.next();
      final String sui = (String)entry.getKey();
      final Set sui_words = (HashSet)entry.getValue();
      double sui_rank = 0.0;

      final int sui_word_ct = ((Integer)sui_to_sui_word_ct.get(sui)).intValue();

      double matching_word_ct = 0.0;
      for (int j = 0; j < search_string_word_ct; j++) {
        final String word = (String)search_term_list.get(j);
        final int word_frequency =
            ((Integer)word_to_freq.get(word)).intValue();

        if (sui_words.contains(word)) {
          matching_word_ct ++;
          final double word_rank = 1.0 / word_frequency;
          sui_rank += word_rank;
        }
      }

      // Rank by % of words matched
      // Weight greater numbers of matched words higher
      final double multiplier = (matching_word_ct/sui_word_ct)*2;

      sui_rank *= multiplier;

      // Rank exact matches much higher
      if (matching_word_ct == sui_word_ct &&
          matching_word_ct == search_string_word_ct)
        sui_rank += 100;

      //
      // Rank the CUI according to the SUI
      //
      final String cui = (String) sui_to_cui.get(sui);
      if (ranked_cuis.containsKey(cui)) {
        RankedCui ranked_cui = (RankedCui)ranked_cuis.get(cui);
        if (sui_rank > ranked_cui.getRank()) {
          ranked_cui.setSui(sui);
          ranked_cui.setRank(sui_rank);
        }
      } else if (sui_rank > threshold) {
        RankedCui ranked_cui = new RankedCui(cui, sui, sui_rank);
        ranked_cuis.put(cui, ranked_cui);
      }
    }

    //
    // Sort the cuis according to ranks and get Concepts to be returned
    //
    final List sorted_cuis = new ArrayList(ranked_cuis.values());
    final Comparator comp =
      new Comparator() {
        public int compare(Object o1, Object o2) {
          RankedCui r1 = (RankedCui)o1;
          RankedCui r2 = (RankedCui)o2;
          return (int)((r2.getRank() - r1.getRank())*1000);
        }
      };
    Collections.sort(sorted_cuis, comp);

    final Concept[] sorted_concepts = new Concept[sorted_cuis.size()];
    for (int i = 0; i < sorted_concepts.length; i++) {
      final RankedCui rc = (RankedCui)sorted_cuis.get(i);
      final String cui = rc.getCui();
      final double rank = rc.getRank();
      final RichMRFilesSingleConceptInputStream rrf =
          new RichMRFilesSingleConceptInputStream(cui, dir);
      final Concept concept =
        new DisplayableConcept(
          rrf.getConceptWithAtomsOnly(), rank);
      sorted_concepts[i] = concept;
    }

    total_result_ct = sorted_concepts.length;
    return sorted_concepts;

  }

  /**
   * Returns the total number of concept matches found
   * on the previous search.
   * @return int number of concepts found on previous search
   */
  public int getPreviousTotalResultCount() {
    return total_result_ct;
  }

  /**
   * Looks up the given {@link Context}'s ptr + aui in the appropriate
   * index files and returns the {@link Context}s with matching ptrs.
   * @param cxt {@link Context}
   * @return {@link Context}<code>[]</code> children
   * @throws IOException
   */
  public Context[] findChildContexts(Context cxt) throws IOException {
    String ptr = "";
    if (!cxt.getPtr().equals(""))
      ptr = cxt.getPtr() + ".";
    ptr += cxt.getAtom().getAui();

    //final String ptr_rela = ptr + cxt.getRela();
    /*if (child_cache.containsKey(ptr_rela)) {
      return (Context[]) child_cache.get(ptr_rela);
    }*/
    final Context [] cxts = findContextsByPtr(cxt, ptr);
    //child_cache.put(ptr_rela, cxts);
    return cxts;
  }

  /**
   * Looks up the given {@link Context}'s ptr in the appropriate
   * index files and returns the {@link Context}s with matching ptrs.
   * @param cxt {@link Context}
   * @return {@link Context}<code>[]</code> children
   * @throws IOException
   */
  public Context[] findSiblingContexts(Context cxt) throws IOException {
    final String ptr = cxt.getPtr();
    //final String ptr_rela = ptr + cxt.getRela();
    /*if (sibling_cache.containsKey(ptr_rela)) {
      return (Context[]) sibling_cache.get(ptr_rela);
    }*/
    final Context [] cxts = findContextsByPtr(cxt, ptr);
    //sibling_cache.put(ptr_rela, cxts);
    return cxts;
  }

  /**
   * Looks up the given ptr in the appropriate
   * index files and returns the matching ancestor.
   * @param cxt {@link Context}
   * @param ptr <code>String</code>
   * @return {@link Context}<code>[]</code>
   * @throws IOException
   */
  public Context findAncestorContext(Context cxt, String ptr) throws IOException {
    final String ptr_rela = ptr +  cxt.getRela();
    if (ancestor_cache.containsKey(ptr_rela)) {
      return (Context) ancestor_cache.get(ptr_rela);
    }

    String parent = null;
    String self = null;
    if(ptr.indexOf(".") != -1) {
      parent = ptr.substring(0, ptr.lastIndexOf("."));
      self = ptr.substring(ptr.lastIndexOf(".") + 1);
    } else {
      parent = "";
      self = ptr;
    }

    final Context[] cxts = findContextsByPtr(cxt, parent, self);

    if (cxts == null || cxts.length == 0)
      return null;
    else {
      ancestor_cache.put(ptr_rela, cxts[0]);
      return cxts[0];
    }

  }

  /**
   * Looks up the given ptr in the appropriate
   * index files and returns the {@link Context}s with matching ptrs.
   * @param cxt {@link Context}
   * @param ptr <code>String</code>
   * @return {@link Context}<code>[]</code>
   * @throws IOException
   */
  private Context[] findContextsByPtr(Context cxt, String ptr) throws IOException {
    final List contexts = new ArrayList();
    String fp = "";

    String cxty = "";
    if (cxt.getSource() != null)
      cxty = cxt.getSource().getCxty();
    String rela = null;
    // Here, decide what is the rule, does RELA matter for children of the root? (YES!)
    if (cxty.indexOf("IGNORE-RELA") == -1 &&
        ptr.length() > 0 && !cxt.getPtr().equals("")) {
      if (cxt.getRela() != null)
        rela = cxt.getRela().getAbbreviation();
    }

    final String search_string = ptr + "|";

    final PushBackReader ptr_mrhier_index =
        openReader(raf_ptr_rela_mrhier,search_string);

    final RichMRFilesConceptInputStream rrf = new RichMRFilesConceptInputStream();

    String line = null;
    total_result_ct  = 0;
    while ( (line = ptr_mrhier_index.readLine()) != null) {
      final String[] tokens = FieldedStringTokenizer.split(line, "|", 5);
      if (max_results == -1 || total_result_ct < max_results) {
        if (ptr.length() > 0 && !cxt.getPtr().equals("")) {
//      if (ptr.indexOf(".") != -1) {
          String file_rela = tokens[2];
          if (rela != null && !file_rela.equals(rela))
            continue;
        }
        fp = tokens[3];
        raf_mrhier.seek(Long.parseLong(fp));
        line = SystemToolkit.readLine(raf_mrhier, "UTF-8");
        final Context this_cxt = rrf.newInstanceFromMRHIER(line);
        final Concept c = new PlaceholderConcept();
        c.setCui(line.substring(0, line.indexOf("|")));
        c.addContext(this_cxt);
        this_cxt.setConcept(c);
        contexts.add(this_cxt);
        total_result_ct++;
      }
    }
    return (Context[])contexts.toArray(new Context[0]);
  }

  /**
   * Looks up the given ptr in the appropriate
   * index files and returns the {@link Context}s with matching ptrs.
   * @param cxt {@link Context}
   * @param ptr <code>String</code>
   * @param aui <code>String</code>
   * @return {@link Context}<code>[]</code>
   * @throws IOException
   */
  private Context[] findContextsByPtr(Context cxt, String ptr, String aui) throws IOException {

    final List contexts = new ArrayList();
    String fp = "";

    String cxty = "";
    if (cxt.getSource() != null)
      cxty = cxt.getSource().getCxty();
    String rela = null;
    if (cxty.indexOf("IGNORE-RELA") == -1 &&
        ptr.length() > 0 && !cxt.getPtr().equals("")) {
      if (cxt.getRela() != null)
        rela = cxt.getRela().getAbbreviation();
    }

    String search_string = "";
    if (rela == null)
      search_string = ptr + "|"+ aui + "|";
    else
      search_string = ptr + "|" + aui + "|" + rela + "|";

    final PushBackReader ptr_mrhier_index =
        openReader(raf_ptr_rela_mrhier,search_string);

    final RichMRFilesConceptInputStream rrf = new RichMRFilesConceptInputStream();

    String line = null;
    total_result_ct  = 0;
    while ( (line = ptr_mrhier_index.readLine()) != null) {
      final String[] tokens = FieldedStringTokenizer.split(line, "|", 5);
      if (max_results == -1 || total_result_ct++ < max_results) {
        fp = tokens[3];
        raf_mrhier.seek(Long.parseLong(fp));
        line = SystemToolkit.readLine(raf_mrhier, "UTF-8");
        final Context this_cxt = rrf.newInstanceFromMRHIER(line);
        final Concept c = new ConceptImpl();
        c.setCui(line.substring(0, line.indexOf("|")));
        c.addContext(this_cxt);
        this_cxt.setConcept(c);
        contexts.add(this_cxt);
      }
    }
    return (Context[])contexts.toArray(new Context[0]);

  }

  /**
   * Looks up the given {@link Context} in the appropriate index files and returns
   * the count of {@link Context}s whose ptr matches the given one + the aui.
   * @param cxt {@link Context} parent
   * @return <code>int</code> children
   * @throws IOException
   */
  public int countChildContexts(Context cxt) throws IOException {
    String ptr = "";
    if (!cxt.getPtr().equals(""))
      ptr = cxt.getPtr() + ".";
    ptr +=  cxt.getAtom().getAui();

    //final String ptr_rela = ptr + cxt.getRela();
    /*if (child_cache.containsKey(ptr_rela)) {
      return ((Context[])child_cache.get(ptr_rela)).length;
    }
    else if (count_child_cache.containsKey(ptr_rela)) {
      return ( (Integer) count_child_cache.get(ptr_rela)).intValue();
    }*/

    int cnt = countContextsByPtr(cxt, ptr);
    //count_child_cache.put(ptr_rela, new Integer(cnt));
    return cnt;
  }

  /**
   * Looks up the given {@link Context} in the appropriate index files and returns
   * the count of {@link Context}s whose ptr matches the given one.
   * @param cxt {@link Context} parent
   * @return <code>int</code> siblings
   * @throws IOException
   */
  public int countSiblingContexts(Context cxt) throws IOException {
    final String ptr = cxt.getPtr();
    //final String ptr_rela = ptr + cxt.getRela();
    /*if (sibling_cache.containsKey(ptr_rela))
      return ((Context[])sibling_cache.get(ptr_rela)).length;
    else if (count_sibling_cache.containsKey(ptr_rela))
      return ((Integer)count_sibling_cache.get(ptr_rela)).intValue();*/

    int cnt = countContextsByPtr(cxt, ptr);
    //count_sibling_cache.put(ptr_rela, new Integer(cnt));
    return cnt;
  }

  /**
   * Looks up the given {@link Context} in the appropriate index files and returns
   * the count of {@link Context}s whose ptr matches the given one.
   * @param cxt {@link Context} parent
   * @param ptr path to root
   * @return <code>int</code> siblings
   * @throws IOException
   */
  private int countContextsByPtr(Context cxt, String ptr) throws IOException {

    String cxty = "";
    if (cxt.getSource() != null)
      cxty = cxt.getSource().getCxty();
    String rela = null;
    if (cxty.indexOf("IGNORE-RELA") == -1 &&
        ptr.length() > 0 && !cxt.getPtr().equals("")) {
      if (cxt.getRela() != null)
        rela = cxt.getRela().getAbbreviation();
    }

    final String search_string = ptr + "|";

    final PushBackReader ptr_mrhier_index =
        openReader(raf_ptr_rela_mrhier, search_string);

    String line = null;
    total_result_ct  = 0;
    while ( (line = ptr_mrhier_index.readLine()) != null) {
      final String[] tokens = FieldedStringTokenizer.split(line, "|", 4);
      String file_rela = tokens[2];
      if (rela != null && !file_rela.equals(rela))
        continue;

      total_result_ct++;
    }

    return total_result_ct;
  }

  /**
   * Looks up the given cui in MRCONSO and returns
   * the {@link Concept} associated with the given cui.
   * @param cui to be searched for
   * @return {@link Concept}
   * @throws IOException
   */
  public Concept findConceptByCui(String cui) throws IOException {
    final RichMRFilesSingleConceptInputStream rrf =
        new RichMRFilesSingleConceptInputStream(cui, dir);
    return new DisplayableConcept(rrf.getConceptWithAtomsOnly(),0);
  }

  /**
   * Looks up the given aui in the AUI_MRCONSO.x index file and returns
   * the {@link Concept} that contains the given aui.
   * @param aui to be searched for
   * @return {@link Concept} containing the aui
   * @throws IOException
   */
  public Concept findConceptByAui(String aui) throws IOException {
    Atom a = findAtomByAui(aui);
    Concept concept = new DisplayableConcept(a.getConcept(),0);
    return concept;
  }

  /**
   * Looks up the given aui in the AUI_MRCONSO.x index file and returns
   * the {@link Atom} that contains the given aui.
   * @param aui to be searched for
   * @return {@link Atom} containing the aui
   * @throws IOException
   */
  public Atom findAtomByAui(String aui) throws IOException {
    if (atom_cache.containsKey(aui)) {
      return (Atom) atom_cache.get(aui);
    }

    String fp = null;
    final PushBackReader aui_cui_index = openReader(raf_aui_mrconso,aui + "|");

    String line = null;
    while ( (line = aui_cui_index.readLine()) != null) {
      final String[] tokens = FieldedStringTokenizer.split(line, "|", 3);
      if (tokens[0].equals(aui)) {
        fp = tokens[1];
        break;
      }
    }
    //aui_cui_index.close();
    raf_mrconso.seek(Long.parseLong(fp));

    final RichMRFilesConceptInputStream rrf = new RichMRFilesConceptInputStream();
    line = SystemToolkit.readLine(raf_mrconso, "UTF-8");
    final Atom a = rrf.newInstanceFromMRCONSO(line);
    final Concept c = new PlaceholderConcept();
    c.setCui(line.substring(0, line.indexOf("|")));
    c.addAtom(a);
    a.setConcept(c);

    atom_cache.put(aui, a);
    return a;
  }

  /**
   * Clears the aui to atom cache.
   */
  private void clearCaches() {
    ancestor_cache.clear();
    //sibling_cache.clear();
    //child_cache.clear();
    atom_cache.clear();
    //count_child_cache.clear();
    //count_sibling_cache.clear();
  }

  /**
   * Sets the language of the index file that should be searched.
   * @param lat language abbreviation
   */
  public void setLanguage(String lat) {
    this.lat = lat;
  }

  /**
   * Returns the language of the index file that should be searched.
   * @return the language of the index file that should be searched
   */
  public String getLanguage() {
    return lat;
  }

  /**
   * Sets the maximum number of results that should be returned.
   * @param max_results number of results
   */
  public void setMaxResults(int max_results) {
    clearCaches();
    this.max_results = max_results;
  }

  /**
   * Sets the rank threshold value for searching.
   * @param threshold the rank threshold valud for searching
   */
  public void setThreshold(double threshold) {
    this.threshold = threshold;
  }

  /**
   * Returns the rank threshold value for searching.
   * @return the rank threshold value for searching
   */
  public double getThreshold() {
    return threshold;
  }

  /**
   * Returns the rank threshold value for searching.
   * @return the rank threshold value for searching
   */
  public int getWordCountThreshold() {
    // .0001 => 5
    // .001 => 0
    // 5 - (threshold*500)
    return 5 - (int)(threshold*500);
  }

  /**
   * Opens the index file of the given name. Returns a
   * {@link SearchStringPushBackReader} with the file pointer set
   * to the row that starts with the given LAT|word.
   * @param file the filename
   * @param word to be searched for in file
   * @return a {@link SearchStringPushBackReader} on the file
   * @throws IOException if the file could not be opened.
   */
  private PushBackReader openReader(RandomAccessFile file, String word) throws IOException {
    if (file != null)
      return new SearchStringPushBackReader(file, word);
    //
    // If we get here, the file was not found.
    //
    return new PushBackReader(new BufferedReader(new StringReader("")));
  }

  //
  // Inner classes
  //

  /**
   * Class representing a ranked concept.
   */
  private class RankedCui implements Comparable {

    //
    // CUI value
    //
    private String cui;

    //
    // SUI value
    //
    private String sui;

    //
    // rank value
    //
    private double rank;

    /**
     * Instantiates a {@link RankedCui} from the specified values
     * @param cui the CUI
     * @param sui the SUI
     * @param rank the rank
     */
    public RankedCui(String cui, String sui, double rank) {
      this.cui = cui;
      this.sui = sui;
      this.rank = rank;
    }

    /**
     * Returns the CUI value.
     * @return the CUI value
     */
    public String getCui() {
      return cui;
    }

    /**
     * Returns the rank.
     * @return the rank
     */
    public double getRank() {
      return rank;
    }

    /**
     * Returns the SUI value.
     * @return the SUI value
     */
    public String getSui() {
      return sui;
    }

    /**
     * Sets the rank.
     * @param rank the rank
     */
    public void setRank (double rank) {
      this.rank = rank;
    }

    /**
     * Sets the SUI value.
     * @param sui the SUI value
     */
    public void setSui (String sui) {
      this.sui = sui;
    }

    /**
     * Indicates the relative ordering of the specified object to this one.
     * @param o the object to compare to
     * @return int
     */
    public int compareTo(Object o) {
      if (o instanceof RankedCui) {
        if (rank < ((RankedCui)o).rank)
          return -1;
        if (rank > ((RankedCui)o).rank)
          return 1;
        if (rank == ((RankedCui)o).rank)
          return 0;
        return 0;
      }
      else
        return 0;
    }

  }

  /**
   * Extension of {@link Concept} to support a configurable display name.
   */
  private class DisplayableConcept extends ConceptImpl {

    /**
     * Instantiates a {@link DisplayableConcept} from the specified {@link Concept}
     * @param concept the {@link Concept}
     * @param rank the rank of the {@link Concept}
     */
    public DisplayableConcept(Concept concept, double rank) {
      super();
      setAtoms(concept.getAtoms());
      setCui(concept.getCui());
    }

    /**
     * Returns the preferred name of the concept.
     * @return the preferred name of the concept
     */
    public String getPreferredName() {
      Atom[] atoms = getAtoms();
      for (int i=0; i<atoms.length; i++) {
        if(atoms[i].hasPreferredAtomIdentifier() &&
           atoms[i].hasPreferredStringIdentifier() &&
           atoms[i].hasPreferredTermIdentifier() &&
           atoms[i].getLanguage().getAbbreviation().equals(lat)) {
          return atoms[i].getStr();
        }
      }
      return null;
    }

    /**
     * Returns a {@link String} representation of the concept based
     * on the prefererd name.
     * @return  a {@link String} representation of the concept
     */
    public String toString() {
      String preferred_string = null;
      preferred_string = getPreferredName();

      if (preferred_string != null)
        return super.getCui() + " " + preferred_string; // + " " + rank;
      else
        return super.getCui();
    }
  }
}
