/**************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Object:      RichMRMetathesaurusInputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.AtomMapping;
import gov.nih.nlm.umls.meta.ConceptMapping;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.RelationshipLabel;
import gov.nih.nlm.umls.meta.ReleaseMetadata;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.impl.AtomMappingImpl;
import gov.nih.nlm.umls.meta.impl.ConceptMappingImpl;
import gov.nih.nlm.umls.meta.impl.LowercaseWordIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.MetaDescriptionImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedStringIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedWordIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.ReleaseMetadataImpl;
import gov.nih.nlm.umls.meta.impl.SourceMetaDescriptionImpl;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This {@link ConceptInputStream} builds objects from a full set of
 * "Rich Release Format" MR files.
 *
 * @author Deborah Shapiro
 */
public class RichMRMetathesaurusInputStream extends RichMRFilesConceptInputStream
  implements MetathesaurusInputStream {

  //
  // Private Fields
  //

  //
  // Index representing the current index file being processed
  //
  protected int current_index = 0;

  //
  // Total number of index lines
  //
  protected int index_lines;

  //
  // Number of index lines processed so far
  //
  protected int index_counter;

  //
  // Readers for index files
  //
  protected BufferedReader [] mrx_sources;

  //
  // Readers for MRCUI, MRSAB
  //
  protected BufferedReader mrcui_source, mrsab_source, mraui_source;

  //
  // Constructors
  //

  /**
   * Instantiates an empty {@link RichMRMetathesaurusInputStream};
   */
  public RichMRMetathesaurusInputStream() { }

  //
  // Methods
  //

  /**
   * Opens all of the MR Files in preparation for subsetting.
   * @param install_paths the directory containing the files
   * @throws IOException if any files can not be opened
   */
  public void open(String[] install_paths) throws IOException {

    //
    // Open content files and MRDOC.RRF
    //
    super.open(install_paths);

    //
    // Open word index files
    //
    final Language[] languages = Language.getLanguages();
    int i = 0;
    mrx_sources =
      new BufferedReader[Language.getLanguagesAsList().size() + 2];
    file_names = new String[mrx_sources.length];
    for (; i < languages.length; i++) {
      file_names[i] = "MRXW_" + languages[i].getAbbreviation() + ".RRF" + add_suffix;
      mrx_sources[i] =
	openSourceFile(file_names[i]);
    }

    //
    // Open norm index files
    //
    mrx_sources[i] = openSourceFile("MRXNW_ENG.RRF" + add_suffix);
    file_names[i] = "MRXNW_ENG.RRF" + add_suffix;
    mrx_sources[++i] = openSourceFile("MRXNS_ENG.RRF" + add_suffix);
    file_names[i] = "MRXNS_ENG.RRF" + add_suffix;

    //
    // Open metadata files
    //
    mrsab_source = openSourceFile("MRSAB.RRF" + add_suffix);
    mrcui_source = openSourceFile("MRCUI.RRF" + add_suffix);
    mraui_source = openSourceFile("MRAUI.RRF" + add_suffix);

    //
    // Cache line counts for MRCONSO.RRF and index files
    //
    getLineCounts();

    //
    // Initialize sources from MRSAB.RRF
    //
    String line = null;
    while ((line = mrsab_source.readLine()) != null &&
	   !line.equals("")) {
      SourceMetaDescription[] smd_array = new SourceMetaDescription[1];
      smd_array[0] = new SourceMetaDescriptionImpl(line);
      Source.initialize(smd_array);
    }
    mrsab_source.close();
    mrsab_source = openSourceFile("MRSAB.RRF" + add_suffix);

  }

  /**
   * Close files opened during {@link #open(String[])}.
   * @throws IOException if files can not be closed
   */
  public void close() throws IOException {

    //
    // Close content files
    //
    super.close();

    //
    // Close index files
    //
    for (int i = 0; i < mrx_sources.length; i++)
      mrx_sources[i].close();

    //
    // Close metadata files
    //
    mrsab_source.close();
    mrdoc_source.close();
    mrcui_source.close();
    mraui_source.close();
  }

  /**
   * Reads through <code>MRFILES.RRF</code> to determine the size of the index files.
   * This is used in the {@link #getIndexProgress()} method.
   * @return {@link Map} of file names to lines in file
   * @throws IOException if the file cannot be opened or read
   */
  protected Map getLineCounts() throws IOException {

    //
    // Obtain MRCONSO.RRF line count
    //
    final Map lines_in_file = super.getLineCounts();

    //
    // Initialize counters
    //
    index_counter = 0;
    index_lines = 0;

    //
    // Aggregate index file line counts
    //
    Iterator iter = lines_in_file.entrySet().iterator();
    while (iter.hasNext()) {
      Map.Entry me = (Map.Entry)iter.next();
      if (((String)me.getKey()).startsWith("MRX")) {
	  index_lines += (Integer.parseInt(me.getValue().toString()));
      }
    }

    return lines_in_file;
  }

  /**
   * Returns an {@link IndexEntry} from one of the index files.
   * We read through the index files one at a time, returning
   * all of the entries.
   * @return an {@link IndexEntry} from one of the index files
   * @throws IOException if the files cannot be read
   */
  public IndexEntry readIndexEntry() throws IOException {

    //
    // We have a current_index into mrx_sources
    // to indicate which reader should be used.
    //
    // If we read a line and it is null,
    // then increment current index
    //

    if (current_index >= mrx_sources.length)
      return null;
    final String line = mrx_sources[current_index].readLine();
    if (line == null) {
      current_index++;
      return readIndexEntry();
    } else {
      index_counter++;
      return newInstanceFromMRX(line, file_names[current_index]);
    }

  }

  /**
   * Returns the percentage of the index file lines that have been processed.
   * @return the percentage of the index file lines that have been processed
   */
  public int getIndexProgress() {
    return (int)((index_counter/(double)index_lines)*100);
  }

  /**
   * Returns the {@link ReleaseMetadata} containing data from
   * <code>MRSAB.RRF</code>, <code>MRCUI.RRF</code>, <code>MRAUI.RRF</code>,
   * and <code>MRDOC.RRF</code>.
   * @return the {@link ReleaseMetadata}
   * @throws IOException if the files could not be read
   */
  public ReleaseMetadata readReleaseMetadata()
    throws IOException {
    //
    // Instantiate ReleaseMetadata objec
    //
    final ReleaseMetadata release_metadata = new ReleaseMetadataImpl();

    //
    // Iterate through sources
    //
    String line = null;
    while ((line = mrsab_source.readLine()) != null) {
      //
      // Add to release metadata
      //
      final SourceMetaDescription[] smd_array = new SourceMetaDescription[1];
      smd_array[0] = new SourceMetaDescriptionImpl(line);
      release_metadata.addSourceMetaDescription(smd_array[0]);
    }

    //
    // Iterate through MRDOC.RRF
    //
    while ((line = mrdoc_source.readLine()) != null) {
      //
      // Add to release metadata
      //
      final MetaDescription[] md_array = new MetaDescription[1];
      md_array[0] = new MetaDescriptionImpl(line);
      release_metadata.addMetadata(md_array[0]);
    }

    //
    // Iterate through concept mappings
    //
    while ((line = mrcui_source.readLine()) != null) {
      //
      // Add to release metadata
      //
      final ConceptMapping cui_mapping = newInstanceFromMRCUI(line);
      if (cui_mapping.getRel() != null &&
          cui_mapping.getRel().getAbbreviation().equals("SUBX"))
        release_metadata.addCUIRemovedFromSubset(line.substring(0, 8));
      else
        release_metadata.addConceptMapping(cui_mapping);
    }

    //
    // Iterate through atom mappings
    //
    while ((line = mraui_source.readLine()) != null) {

      //
      // Add to release metadata
      //
      final AtomMapping aui_mapping = newInstanceFromMRAUI(line);
      release_metadata.addAtomMapping(aui_mapping);
    }

    //
    // save references to ancillary and change files so they
    // can be copied to the target directory during output
    //
    release_metadata.addChangeFile("DELETEDCUI.RRF" + add_suffix, openSourceFile("/CHANGE/DELETEDCUI.RRF" + add_suffix));
    release_metadata.addChangeFile("DELETEDLUI.RRF" + add_suffix, openSourceFile("/CHANGE/DELETEDLUI.RRF" + add_suffix));
    release_metadata.addChangeFile("DELETEDSUI.RRF" + add_suffix, openSourceFile("/CHANGE/DELETEDSUI.RRF" + add_suffix));
    release_metadata.addChangeFile("MERGEDCUI.RRF" + add_suffix, openSourceFile("/CHANGE/MERGEDCUI.RRF" + add_suffix));
    release_metadata.addChangeFile("MERGEDLUI.RRF" + add_suffix, openSourceFile("/CHANGE/MERGEDLUI.RRF" + add_suffix));
    return release_metadata;
  }

  /**
   * Returns the complete list of suis from the AMBIGSUI list.
   * They will be used in calculating entries for the output AMBIGSUI list.
   * @return list of suis
   * @throws IOException
   */
  public String[] getAmbiguousStringIdentifiers() throws IOException {
    return (String[])getAmbiguousStringIdentifiersAsList().toArray(new String[0]);
  }

  /**
   * Returns the complete list of suis from the AMBIGSUI list.
   * They will be used in calculating entries for the output AMBIGSUI list.
   * @return list of suis
   * @throws IOException
   */
  public List getAmbiguousStringIdentifiersAsList() throws IOException {
    final BufferedReader ambigsui = openSourceFile("AMBIGSUI.RRF");
    final Set sui_set = new HashSet();
    String line = null;
    while ((line = ambigsui.readLine()) != null && !line.equals("")) {
      String [] sui_cui = FieldedStringTokenizer.split(line,"|");
      sui_set.add(sui_cui[0]);
    }
    ambigsui.close();
    return new ArrayList(sui_set);
  }

  /**
   * Returns the complete list of terms from the AMBIGLUI list.
   * They will be used in calculating entries for the output AMBIGLUI list.
   * @return list of terms
   * @throws IOException
   */
  public String[] getAmbiguousTermIdentifiers() throws IOException {
    return (String[])getAmbiguousTermIdentifiersAsList().toArray(new String[0]);
  }

  /**
   * Returns the complete list of terms from the AMBIGLUI list.
   * They will be used in calculating entries for the output AMBIGLUI list.
   * @return list of terms
   * @throws IOException
   */
  public List getAmbiguousTermIdentifiersAsList() throws IOException {
    final BufferedReader ambiglui = openSourceFile("AMBIGLUI.RRF");
    final Set lui_set = new HashSet();
    String line = null;
    while ((line = ambiglui.readLine()) != null && !line.equals("")) {
      String [] lui_cui = FieldedStringTokenizer.split(line,"|");
      lui_set.add(lui_cui[0]);
    }
    ambiglui.close();
    return new ArrayList(lui_set);
  }



  /**
   * Instantiates an {@link ConceptMapping} from a <code>MRCUI.RRF</code> line.
   * @param line a line from <code>MRCUI.RRF</code>
   * @return the {@link ConceptMapping} matching the line
   * @throws IOException if there is a problem with the file
   */
  private ConceptMapping newInstanceFromMRCUI(String line)
    throws IOException {

    //
    // Instantiate default ConceptMapping
    //
    final ConceptMapping a = new ConceptMappingImpl();

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 8, tokens);

    //
    // Populate concept mapping
    //
    a.setCui1(tokens[0]);
    a.setLastReleaseVersion(tokens[1]);
    a.setRel(RelationshipLabel.getRelationshipLabel(tokens[2]));
    a.setRela(AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[3]));
    a.setReason(tokens[4]);
    a.setCui2(tokens[5]);
    a.setMapin(tokens[6]);

    return a;
  }

  /**
   * Instantiates an {@link AtomMapping} from a <code>MRAUI.RRF</code> line.
   * @param line a line from <code>MRAUI.RRF</code>
   * @return the {@link AtomMapping} matching the line
   * @throws IOException if there is a problem with the file
   */
  private AtomMapping newInstanceFromMRAUI(String line)
    throws IOException {

    //
    // Instantiate default ConceptMapping
    //
    final AtomMapping a = new AtomMappingImpl();

    //
    // Parse line and count tokens
    //
    FieldedStringTokenizer.split(line, "|", 9, tokens);

    //
    // Populate concept mapping
    //
    a.setAui1(tokens[0]);
    a.setCui1(tokens[1]);
    a.setVer(tokens[2]);
    a.setRel(RelationshipLabel.getRelationshipLabel(tokens[3]));
    a.setRela(AdditionalRelationshipLabel.getAdditionalRelationshipLabel(tokens[4]));
    a.setMapReason(tokens[5]);
    a.setAui2(tokens[6]);
    a.setCui2(tokens[7]);
    a.setMapin(tokens[8]);

    return a;
  }

  /**
   * Instantiates an {@link IndexEntry} from a <code>MRX*</code> line.
   * @param line a line from an index file
   * @param file_name name of file
   * @return the {@link IndexEntry} matching the line
   * @throws IOException if there is a problem with the file
   */
  private IndexEntry newInstanceFromMRX(String line, String file_name)
    throws IOException {

    //
    // Parse line and count fields
    //
    FieldedStringTokenizer.split(line, "|", 6, tokens);

    //
    // Populate index entry
    //
    IndexEntry index_entry = null;
    if (file_name.startsWith("MRXNW_ENG")) {
      index_entry = new NormalizedWordIndexEntryImpl();
    } else if (file_name.startsWith("MRXNS_ENG")) {
      index_entry = new NormalizedStringIndexEntryImpl();
    } else
      index_entry = new LowercaseWordIndexEntryImpl();

    index_entry.setLat(tokens[0]);
    index_entry.setValue(tokens[1]);
    index_entry.setCui(tokens[2]);
    index_entry.setLui(tokens[3]);
    index_entry.setSui(tokens[4]);

    return index_entry;
  }

}
