/**************************************************************************
 * Package:     gov.nih.nlm.umls.io
 * Object:      RichMRMetathesaurusOutputStream.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.io;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomMapping;
import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ConceptMapping;
import gov.nih.nlm.umls.meta.IndexEntry;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.ReleaseMetadata;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.SourceTermType;
import gov.nih.nlm.umls.meta.TermType;
import gov.nih.nlm.umls.meta.impl.LowercaseWordIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedStringIndexEntryImpl;
import gov.nih.nlm.umls.meta.impl.NormalizedWordIndexEntryImpl;
import gov.nih.nlm.util.ColumnStatistics;
import gov.nih.nlm.util.FieldedStringTokenizer;
import gov.nih.nlm.util.FileStatistics;
import gov.nih.nlm.util.ProgressEvent;
import gov.nih.nlm.util.ProgressListener;
import gov.nih.nlm.util.ProgressReporter;
import gov.nih.nlm.util.SystemToolkit;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * This {@link MetathesaurusOutputStream} generates MR files in the "Rich
 * Release Format".  It is an extension of {@link RichMRFilesConceptOutputStream} with
 * additional code for the {@link ReleaseMetadata} content.
 *
 * @author Deborah Shapiro
 */
public class RichMRMetathesaurusOutputStream extends RichMRFilesConceptOutputStream
  implements MetathesaurusOutputStream, ProgressReporter {

  //
  // Listeners
  //
  protected List listeners = new ArrayList();

  //
  // MRFILES/MRCOLS configuration info
  //
  protected FileStatisticsConfiguration config;

  //
  // Change file names
  //
  protected String[] change_files;

  //
  // Ancillary file names
  //
  protected String[] ancillary_files;

  //
  // Maps to manage candidate ambiguous string and term identifiers
  //
  protected Map sui_to_cui_set = null;
  protected Map lui_to_cui_set = null;

  //
  // Atoms from the current concept.  Set in #writeConcept(Concept)
  //
  protected Atom[] atoms = null;

  //
  // Constructor
  //

  /**
   * Instantiates an empty {@link RichMRMetathesaurusOutputStream}.
   * @param config {@link FileStatisticsConfiguration}
   */
  public RichMRMetathesaurusOutputStream(FileStatisticsConfiguration config) {
    this.config = config;
  }

  //
  // Methods
  //

  /**
   * Closes source files that were opened by this class's openTargets() method.
   * @throws IOException
   */
  public void close() throws IOException{
    super.close();

    //
    // Sort MRCUI.RRF
    //
    SystemToolkit.sort(new File(mr_dir, "MRCUI" + file_ext).getAbsolutePath(), true);

    //
    // Sort MRAUI.RRF
    //
    SystemToolkit.sort(new File(mr_dir, "MRAUI" + file_ext).getAbsolutePath(), true);

    //
    // Sort AMBIG files
    //
    SystemToolkit.sort(new File(mr_dir,"AMBIGLUI.RRF").getAbsolutePath());
    SystemToolkit.sort(new File(mr_dir,"AMBIGSUI.RRF").getAbsolutePath());
  }

  /**
   * Prints the given line to the file with the given name.
   * @param line the line to be printed
   * @param file_name printing destination
   * @throws IOException
   */
  protected void printLine(String line, String file_name) throws IOException {

    //
    // Print the line
    //
    super.printLine(line, file_name);

    //
    // Handle stats
    //
    FileStatistics stats = config.getFileStatistics(file_name);
    stats.processLine(line);
  }

  /**
   * Writes a {@link Concept} to the appropriate MR files.
   * @param concept the {@link Concept} to write
   * @throws IOException if the files could not be written
   */
  public void writeConcept(Concept concept) throws IOException {

    super.writeConcept(concept);

    //
    // Maintain ambiguous sui and lui maps
    //
    atoms = concept.getAtoms();
    for (int i=0; i<atoms.length; i++)
      ambigCheck(atoms[i]);
  }


  /**
   * Writes an {@link IndexEntry} to the appropriate MR index file.
   * @param index_entry the {@link IndexEntry} to write
   * @throws IOException if the file could not be written
   */
  public void writeIndexEntry(IndexEntry index_entry)
    throws IOException {

    //
    // Prepare MRX* line
    //
    final StringBuffer sb = new StringBuffer(50);
    sb.append(index_entry.getLat()); sb.append("|");
    sb.append(index_entry.getValue()); sb.append("|");
    sb.append(index_entry.getCui()); sb.append("|");
    sb.append(index_entry.getLui()); sb.append("|");
    sb.append(index_entry.getSui()); sb.append("|");

    //
    // Determine appropriate file & write line
    //
    if (index_entry instanceof LowercaseWordIndexEntryImpl)
      printLine(sb.toString(),"MRXW_" + index_entry.getLat());
    else if (index_entry instanceof NormalizedWordIndexEntryImpl)
      printLine(sb.toString(),"MRXNW_ENG");
    else if (index_entry instanceof NormalizedStringIndexEntryImpl)
      printLine(sb.toString(),"MRXNS_ENG");

  }

  /**
   * Writes a line to the <code>MRCUI.RRF</code> file.
   * @param release_metadata the {@link ReleaseMetadata}
   * @throws IOException if anything goes wrong
   */
  protected void writeMRCUI(ReleaseMetadata release_metadata)  throws IOException {

    //
    // Iterate over concept mappings
    //
    List cm_list = release_metadata.getConceptMappingsAsList();

    for (int i=0; i<cm_list.size(); i++) {
      //
      // Prepare a line of MRCUI
      //
      final StringBuffer sb = new StringBuffer(50);
      final ConceptMapping mapping = (ConceptMapping)cm_list.get(i);
      sb.append(mapping.getCui1()); sb.append("|");
      sb.append(mapping.getVer()); sb.append("|");
      sb.append((mapping.getRel() != null) ? mapping.getRel().getAbbreviation() : ""); sb.append("|");
      sb.append((mapping.getRela() != null) ? mapping.getRela().getAbbreviation() : ""); sb.append("|");
      sb.append((mapping.getReason() != null) ? mapping.getReason() : ""); sb.append("|");
      if(mapping.getCui2() == null || mapping.getCui2().equals("")) {
        sb.append("||");
      } else {
	sb.append(mapping.getCui2()); sb.append("|");
        if(mapping.inCurrentSubset())
          sb.append("Y|");
        else
          sb.append("N|");
      }

      //
      // Write a line and process file statistics
      //
      printLine(sb.toString(),"MRCUI");
    }

    //
    // Iterate over SUBX removed cuis
    //
    String[] subx_list = release_metadata.getCUIsRemovedFromSubset();

    for (int i=0; i<subx_list.length; i++) {
      //
      // Prepare a line of MRCUI
      //
      final StringBuffer sb = new StringBuffer(50);
      sb.append(subx_list[i]); sb.append("|");
      // requires that release_metadata.setReleaseVersion() is called
      sb.append(release_metadata.getReleaseVersion()).append("|");
      sb.append("SUBX|"); // REL
      sb.append("|");  // RELA
      sb.append("|");  // REASON
      sb.append("|");  // CUI2
      sb.append("|");

      //
      // Write a line and process file statistics
      //
      printLine(sb.toString(),"MRCUI");
    }

    //
    // Sorting of the file now takes place at the end
    //
  }

  /**
   * Writes a line to the <code>MRAUI.RRF</code> file.
   * @param release_metadata the {@link ReleaseMetadata}
   * @throws IOException if anything goes wrong
   */
  protected void writeMRAUI(ReleaseMetadata release_metadata)  throws IOException {

    //
    // Iterate over atom mappings
    //
    List am_list = release_metadata.getAtomMappingsAsList();

    for (int i=0; i<am_list.size(); i++) {
      //
      // Prepare a line of MRAUI
      //
      final StringBuffer sb = new StringBuffer(50);
      final AtomMapping mapping = (AtomMapping)am_list.get(i);
      sb.append(mapping.getAui1()); sb.append("|");
      sb.append(mapping.getCui1()); sb.append("|");
      sb.append(mapping.getVer()); sb.append("|");
      sb.append((mapping.getRel() != null) ? mapping.getRel().getAbbreviation() : ""); sb.append("|");
      sb.append((mapping.getRela() != null) ? mapping.getRela().getAbbreviation() : ""); sb.append("|");
      sb.append((mapping.getMapReason() != null) ? mapping.getMapReason() : ""); sb.append("|");
      if(mapping.getAui2() == null || mapping.getAui2().equals("")) {
        sb.append("|||");
      } else {
        sb.append(mapping.getAui2()); sb.append("|");
        sb.append(mapping.getCui2()); sb.append("|");
        if(mapping.inCurrentSubset())
          sb.append("Y|");
        else
          sb.append("N|");
      }

      //
      // Write a line and process file statistics
      //
      printLine(sb.toString(),"MRAUI");
    }

    //
    // Sorting of the file now takes place at the end
    //
  }


  /**
   * Write a line to the <code>MRDOC.RRF</code> file.
   * @param release_metadata the {@link ReleaseMetadata}
   * @throws IOException if anything goes wrong
   */
  protected void writeMRDOC(ReleaseMetadata release_metadata)  throws IOException {

    //
    // Iterate through meta descriptions
    //
    final MetaDescription[] metadata = release_metadata.getMetadata();
    final List mrdoc_lines = new ArrayList();
    for(int i = 0; i < metadata.length; i++) {

      //
      // Prepare MRDOC.RRF line
      //
      final StringBuffer sb = new StringBuffer(50);
      final MetaDescription this_doc = metadata[i];
      sb.append(this_doc.getKey()); sb.append("|");
      sb.append(this_doc.getValue()); sb.append("|");
      sb.append(this_doc.getType()); sb.append("|");
      sb.append(this_doc.getExplain()); sb.append("|");

      //
      // Add line to list
      //
      mrdoc_lines.add(sb.toString());
    }

    //
    // Sort lines and write MRDOC.RRF
    //
    Collections.sort(mrdoc_lines);
    for(int i = 0; i < mrdoc_lines.size(); i++) {
      printLine(mrdoc_lines.get(i).toString(),"MRDOC");
    }
  }

  /**
   * Write a line to the <code>MRSAB.RRF</code> file.
   * @param release_metadata the {@link ReleaseMetadata}
   * @throws IOException if anything goes wrong
   */
  protected void writeMRSAB(ReleaseMetadata release_metadata)  throws IOException {

    //
    // Iterate through sources
    //
    final SourceMetaDescription[] source_metadata = release_metadata.getSourceMetadata();
    final List mrsab_lines = new ArrayList();
    for(int i = 0; i < source_metadata.length; i++) {

      //
      // Prepare MRSAB.RRF line
      //
      final StringBuffer sb = new StringBuffer(400);
      final SourceMetaDescription this_source = source_metadata[i];
      sb.append(this_source.getVcui()); sb.append("|");
      sb.append(this_source.getRcui()); sb.append("|");
      sb.append(this_source.getVersionedSourceAbbreviation()); sb.append("|");
      sb.append(this_source.getRootSourceAbbreviation()); sb.append("|");
      sb.append(this_source.getOfficialName()); sb.append("|");
      sb.append(this_source.getSourceFamily()); sb.append("|");
      sb.append(this_source.getVersion()); sb.append("|");
      Date start_date = this_source.getValidStartDate();
      SimpleDateFormat df = new SimpleDateFormat("yyyy_MM_dd");
      if(start_date != null) {
        sb.append(df.format(start_date));
      }
      sb.append("|");
      Date end_date = this_source.getValidEndDate();
      if(end_date != null) {
        sb.append(df.format(end_date));
      }
      sb.append("|");
      sb.append(this_source.getInsertMetaVersion()); sb.append("|");
      sb.append(this_source.getRemoveMetaVersion()); sb.append("|");
      sb.append(this_source.getLicenseContact()); sb.append("|");
      sb.append(this_source.getContentContact()); sb.append("|");
      if(Integer.parseInt(this_source.getRestrictionLevel()) != -1)
        sb.append(this_source.getRestrictionLevel() + "");
      sb.append("|");
      if(this_source.getTermFrequency() > 0)
        sb.append(this_source.getTermFrequency() + "");
      sb.append("|");
      if(this_source.getConceptFrequency() > 0)
        sb.append(this_source.getConceptFrequency() + "");
      sb.append("|");
      sb.append(this_source.getContextType()); sb.append("|");
      TermType[] term_types = this_source.getTermTypes();
      if(term_types != null && term_types.length != 0) {
        for(int j=0; j<term_types.length-1; j++) {
  	  sb.append(term_types[j].getAbbreviation()); sb.append(",");
        }
        sb.append(term_types[term_types.length-1].getAbbreviation());
      }
      sb.append("|");
      AttributeName[] attributes = this_source.getAttributeNames();
      if(attributes != null && attributes.length != 0) {
        for(int j=0; j<attributes.length-1; j++) {
          if (attributes[j] != null)
  	    sb.append(attributes[j].getAbbreviation()); sb.append(",");
        }
        if (attributes[attributes.length-1] != null)
          sb.append(attributes[attributes.length-1].getAbbreviation());
      }
      sb.append("|");
      sb.append(this_source.getLanguage() != null ?
        this_source.getLanguage().getAbbreviation() : ""); sb.append("|");
      sb.append(this_source.getCharacterEncoding()); sb.append("|");
      if(this_source.isCurrentVersion())
        sb.append("Y|");
      else
        sb.append("N|");
      if(this_source.inCurrentSubset())
        sb.append("Y|");
      else
        sb.append("N|");
      sb.append((this_source.getSsn() != null) ? this_source.getSsn() : ""); sb.append("|");
      sb.append((this_source.getScit() != null) ? this_source.getScit() : ""); sb.append("|");

      //
      // Add line to list
      //
      mrsab_lines.add(sb.toString());
    }

    //
    // Sort lines and write to MRSAB.RRF
    //
    Collections.sort(mrsab_lines);
    for(int i = 0; i < mrsab_lines.size(); i++) {
      printLine(mrsab_lines.get(i).toString(),"MRSAB");
    }
  }

  /**
   * Writes a line to the <code>MRRANK.RRF</code> file.
   * @param release_metadata the {@link ReleaseMetadata}
   * @throws IOException if anything goes wrong
   */
  protected void writeMRRANK(ReleaseMetadata release_metadata) throws IOException {

    //
    // Iterate through termgroups
    //
    final SourceTermType[] tg_list = release_metadata.getSourceTermTypePrecedence();
    for(int i = 0; i < tg_list.length; i++) {

      //
      // Prepare MRRANK.RRF line
      //
      final StringBuffer sb = new StringBuffer(20);
      final SourceTermType this_tg = tg_list[i];

      //
      // compute rank
      //
      String rank = "0000" + (tg_list.length -i) + "|";
      sb.append(rank.substring(rank.length()-5));
      if (versioned_output)
        sb.append(this_tg.getSource().getVersionedSourceAbbreviation());
      else
        sb.append(this_tg.getSource().getRootSourceAbbreviation());
      sb.append("|");
      sb.append(this_tg.getTermType().getAbbreviation()); sb.append("|");
      if(this_tg.isSuppressible())
        sb.append("Y|");
      else
        sb.append("N|");

      //
      // Write line, process file stats
      //
      printLine(sb.toString(),"MRRANK");
    }
  }

  /**
   * Writes an {@link ReleaseMetadata} to the appropriate MR metadata files.
   * @param release_metadata the {@link ReleaseMetadata} to write
   * @throws IOException if the file could not be written
   */
  public void writeReleaseMetadata(ReleaseMetadata release_metadata)
    throws IOException{

    String line;

    //
    // 1. copy change files from installation directory to target directory
    // keeping statistics information for MRFILES/COLS
    //
    change_files = release_metadata.getChangeFileNames();

    //
    // Iterate through change files
    //
    fireProgressEvent(1,"Start writing change files");
    final File change_dir = new File(mr_dir, "CHANGE");
    if (!change_dir.exists() || !change_dir.isDirectory())
      change_dir.mkdir();
    
    for (int i=0; i<change_files.length; i++) {

      final BufferedReader br = (BufferedReader)release_metadata.getChangeFile(change_files[i]);

      //
      // Write change file
      //
      final BufferedWriter pw = openTargetFile("CHANGE/" + change_files[i]);
      final FileStatistics stats = config.getFileStatistics(
        "CHANGE/" +change_files[i].substring(0, change_files[i].lastIndexOf('.')));

      //
      // Copy change file data
      //
      while ((line = br.readLine()) != null) {
        pw.write(stats.processLine(line));
        pw.newLine();
      }

      pw.close();
    }

    fireProgressEvent(14,"Done writing change files");


    //
    // 2. write out subsetted ancillary files to target directory
    // keeping statistics information for MRFILES/COLS
    //

    //
    // Open ancillary file AMBIGSUI.RRF
    //
    fireProgressEvent(15,"Start writing ancillary files");
    final BufferedWriter as_pw = openTargetFile("AMBIGSUI.RRF");
    final FileStatistics as_stats = config.getFileStatistics("AMBIGSUI");

    //
    // Write AMBIGSUI entries
    //
    final Iterator as_iter = sui_to_cui_set.entrySet().iterator();
    while (as_iter.hasNext()) {
      final Map.Entry me = (Map.Entry)as_iter.next();
      if (me.getValue() != null &&  ((HashSet) me.getValue()).size() > 1) {
        final String sui = (String) me.getKey();
        final HashSet cui_set = (HashSet) me.getValue();
        final String[] cuis = (String[])cui_set.toArray(new String[0]);
        for (int i = 0; i < cuis.length; i++) {
          String local_line = sui + "|" + cuis[i] + "|";
          as_pw.write(as_stats.processLine(local_line));
          as_pw.newLine();
        }
      }
    }
    as_pw.close();

    //
    // Open ancillary file AMBIGLUI.RRF
    //
    final BufferedWriter al_pw = openTargetFile("AMBIGLUI.RRF");
    final FileStatistics al_stats = config.getFileStatistics("AMBIGLUI");

    //
    // Write AMBIGLUI entries
    //
    final Iterator al_iter = lui_to_cui_set.entrySet().iterator();
    while (al_iter.hasNext()) {
      final Map.Entry me = (Map.Entry)al_iter.next();
      if (me.getValue() != null &&  ((HashSet) me.getValue()).size() > 1) {
        final String lui = (String) me.getKey();
        final HashSet cui_set = (HashSet) me.getValue();
        final String[] cuis = (String[])cui_set.toArray(new String[0]);
        for (int i = 0; i < cuis.length; i++) {
          String local_line = lui + "|" + cuis[i] + "|";
          al_pw.write(al_stats.processLine(local_line));
          al_pw.newLine();
        }
      }
    }
    al_pw.close();

    fireProgressEvent(28,"Done writing ancillary files");

    //
    // 3. Write MRCUI
    //
    fireProgressEvent(29,"Start writing MRCUI.RRF");
    writeMRCUI(release_metadata);
    fireProgressEvent(40,"Done writing MRCUI.RRF");

    //
    // 3a. Write MRAUI
    //
    fireProgressEvent(41,"Start writing MRAUI.RRF");
    writeMRAUI(release_metadata);
    fireProgressEvent(51,"Done writing MRAUI.RRF");

    //
    // 4. Write MRSAB
    //
    fireProgressEvent(52,"Start writing MRSAB.RRF");
    writeMRSAB(release_metadata);
    fireProgressEvent(63,"Done writing MRSAB.RRF");

    //
    // 5. Write MRRANK
    //
    fireProgressEvent(64,"Start writing MRRANK.RRF");
    writeMRRANK(release_metadata);
    fireProgressEvent(75,"Done writing MRRANK.RRF");

    //
    // 6. Write MRDOC
    //
    fireProgressEvent(76,"Start writing MRDOC.RRF");
    writeMRDOC(release_metadata);
    fireProgressEvent(87,"Done writing MRDOC.RRF");

    //
    // 7. Write MRFILES/MRCOLS
    //
    fireProgressEvent(88,"Start writing MRFILES.RRF, MRCOLS.RRF");
    writeFileColumnStats();
    fireProgressEvent(99,"Done writing MRFILES.RRF, MRCOLS.RRF");
  }

  /**
   * Sets the list of suis that are candidates to be ambiguous.
   * @param suis the list of suis that are candidates to be ambiguous
   */
  public void setCandidateAmbiguousStringIdentifiers(String [] suis) {
    if (sui_to_cui_set == null)
      sui_to_cui_set = new HashMap();
    for (int i=0; i<suis.length; i++)
      sui_to_cui_set.put(suis[i], null);
  }

  /**
   * Sets the list of suis that are candidates to be ambiguous.
   * @param suis the list of suis that are candidates to be ambiguous
   */
  public void setCandidateAmbiguousStringIdentifiers(List suis) {
    setCandidateAmbiguousStringIdentifiers((String[]) suis.toArray(new String[0]));
  }

  /**
   * Sets the list of terms that are candidates to be ambiguous.
   * @param luis list of terms that are candidates to be ambiguous
   */
  public void setCandidateAmbiguousTermIdentifiers(String[] luis) {
    if (lui_to_cui_set == null)
      lui_to_cui_set = new HashMap();
    for (int i = 0; i < luis.length; i++)
      lui_to_cui_set.put(luis[i], null);
  }

  /**
   * Sets the list of terms that are candidates to be ambiguous.
   * @param luis list of terms that are candidates to be ambiguous
   */
  public void setCandidateAmbiguousTermIdentifiers(List luis) {
    setCandidateAmbiguousTermIdentifiers((String[])luis.toArray(new String[0]));
  }


  /**
   * Opens the specified file in the directory passed to
   * {@link #open(String)}.
   * @param name the filename
   * @return a {@link BufferedWriter} for the file
   * @throws IOException if the file could not be opened.
   */
  protected BufferedWriter openTargetFile(String name) throws IOException {

    //
    // let superclass handle it first
    //
    BufferedWriter out;
    out = super.openTargetFile(name);

    //
    // Obtain list of target files.
    //
    ArrayList targets = new ArrayList();
    for (int i=0; i<target_files.length; i++) {
      targets.add(target_files[i] + file_ext);
    }

    //
    // Initialize file statistics for the current file
    //
    if (targets.contains(name) || name.startsWith("MRX") || name.endsWith(file_ext)) {
      config.getFileStatistics(name.substring(0, name.lastIndexOf(".")));
    }

    return out;
  }

  //
  // ProgressReporter API
  //

  /**
   * Adds a {@link ProgressListener}.
   * @param l thef{@link ProgressListener}
   */
  public void addProgressListener(ProgressListener l) {
    listeners.add(l);
  }

  /**
   * Removes a {@link ProgressListener}.
   * @param l thef{@link ProgressListener}
   */
  public void removeProgressListener(ProgressListener l) {
    listeners.remove(l);
  }

  /**
   * Fires a {@link ProgressEvent}.
   * @param pct percent done
   * @param note progress note
   */
  protected void fireProgressEvent(int pct, String note) {
    ProgressEvent pe = new ProgressEvent(this,pct,pct,note);
    for (int i = 0; i < listeners.size(); i++) {
      ((ProgressListener)listeners.get(i)).updateProgress(pe);
    }
  }

  //
  // Protected Methods
  //



  /**
   * This method computes the file and columns statistics
   * for the <code>MRFILES.RRF</code> and <code>MRCOLS.RRF</code> files themselves.
   * Should it be using "\n"?
   * @throws IOException if anything goes wrong
   */
  protected void writeFileColumnStats() throws IOException {

    //
    // Obtain entries for core files
    //
    StringBuffer mrcols_static = new StringBuffer(200);
    StringBuffer mrfiles_static = new StringBuffer(200);
    for(int i = 0; i < target_files.length; i++) {
      FileStatistics file_stats = config.getFileStatistics(target_files[i]);
      if(file_stats.getFileName().startsWith("MRCOLS") ||
         file_stats.getFileName().startsWith("MRFILES")) {
        continue;
      } else {
        mrfiles_static.append(getFileTextFromFileStatistics(file_stats));
        mrcols_static.append(getColumnTextFromFileStatistics(file_stats));
      }
    }

    //
    // Obtain entries for index files
    //
    Language[] lats = Language.getLanguages();
    for(int i = 0; i < lats.length; i++) {
      FileStatistics stats = config.getFileStatistics("MRXW_" + lats[i]);
      mrfiles_static.append(getFileTextFromFileStatistics(stats));
      mrcols_static.append(getColumnTextFromFileStatistics(stats));
    }

    //
    // Obtain entries for change files
    //
    for (int i = 0; i < change_files.length; i++) {
      FileStatistics stats = config.getFileStatistics("CHANGE/" +
          change_files[i].substring(0, change_files[i].lastIndexOf('.')));
      mrfiles_static.append(getFileTextFromFileStatistics(stats));
      mrcols_static.append(getColumnTextFromFileStatistics(stats));
    }

    //
    // Obtain entries for ancillary files
    //
    FileStatistics stats = config.getFileStatistics("AMBIGSUI");
    mrfiles_static.append(getFileTextFromFileStatistics(stats));
    mrcols_static.append(getColumnTextFromFileStatistics(stats));

    stats = config.getFileStatistics("AMBIGLUI");
    mrfiles_static.append(getFileTextFromFileStatistics(stats));
    mrcols_static.append(getColumnTextFromFileStatistics(stats));

    //
    // Attempt to converge by regenerating mrcols/mrfiles
    // until the statistics match
    //
    StringBuffer mrcols = new StringBuffer(200);
    StringBuffer mrfiles = new StringBuffer(200);
    StringBuffer mrcols_new = new StringBuffer(200);
    StringBuffer mrfiles_new = new StringBuffer(200);
    try {
      while(true) {

        //
        // mrcols_new is the current best guess
        // mrcols is the current computation
        //
	mrcols = new StringBuffer(200);
	mrcols.append(mrcols_new);
	mrcols_new = new StringBuffer(200);
	mrcols_new.append(mrcols_static);

	mrfiles = new StringBuffer(200);
	mrfiles.append(mrfiles_new);
	mrfiles_new = new StringBuffer(200);
	mrfiles_new.append(mrfiles_static);

        //
        // Compute MRFILES/COLS data for MRFILES itself
        //
	FileStatistics mrfiles_stats = config.getFileStatistics("MRFILES");
	String[] mrfiles_lines =
	  FieldedStringTokenizer.split(mrfiles.toString(), "\n");
	for(int i = 0; i < mrfiles_lines.length; i++) {
          if (!mrfiles_lines[i].equals(""))
  	    mrfiles_stats.processLine(mrfiles_lines[i]);
	}

        //
        // Compute MRFILES/COLS data for MRCOLS itself
        //
	FileStatistics mrcols_stats = config.getFileStatistics("MRCOLS");
	String[] mrcols_lines =
	  FieldedStringTokenizer.split(mrcols.toString(), "\n");
	for(int i = 0; i < mrcols_lines.length; i++) {
          if (!mrcols_lines[i].equals(""))
	  mrcols_stats.processLine(mrcols_lines[i]);
	}

        //
        // Compute "new" files
        //
	mrfiles_new.append(getFileTextFromFileStatistics(mrfiles_stats));
	mrfiles_new.append(getFileTextFromFileStatistics(mrcols_stats));
	mrcols_new.append(getColumnTextFromFileStatistics(mrcols_stats));
	mrcols_new.append(getColumnTextFromFileStatistics(mrfiles_stats));

        //
        // Compare to see if we have converged
        //
	if(mrfiles.toString().equals(mrfiles_new.toString()) &&
	   mrcols.toString().equals(mrcols_new.toString()))
	  break;
      }

      //
      // Once we have properly computed MRCOLS/FILES, we write them out
      //
      String[] mrfiles_output =
	FieldedStringTokenizer.split(mrfiles.toString(), "\n");
      String[] mrcols_output =
	FieldedStringTokenizer.split(mrcols.toString(), "\n");

      Arrays.sort(mrfiles_output);
      Arrays.sort(mrcols_output);

      BufferedWriter mrfiles_target = (BufferedWriter)mr_targets.get("MRFILES");
      BufferedWriter mrcols_target = (BufferedWriter)mr_targets.get("MRCOLS");
      for (int i = 0; i < mrfiles_output.length; i++)
        if (!mrfiles_output[i].equals("")) {
          mrfiles_target.write(mrfiles_output[i]);
          mrfiles_target.newLine();
        }

      for (int i = 0; i < mrcols_output.length; i++)
        if (!mrcols_output[i].equals("")) {
          mrcols_target.write(mrcols_output[i], 0, mrcols_output[i].length());
          mrcols_target.newLine();
        }

    } catch (Exception e) {
      IOException ioe = new IOException("Problem processing lines for MRFILES/MRCOLS.");
      ioe.initCause(e);
      throw ioe;
    }

  }

  /**
   * Returns the <code>MRFILES.RRF</code> line for the specified {@link FileStatistics}.
   * @param file_stats the {@link FileStatistics}
   * @return the <code>MRFILES.RRF</code> line for the specified {@link FileStatistics}
   */
  protected String getFileTextFromFileStatistics(FileStatistics file_stats) {
    StringBuffer sb = new StringBuffer(200);
    ColumnStatistics [] col_stats = file_stats.getAllColumnStatistics();
    sb.append(file_stats.getFileName() + file_ext + "|");
    sb.append(file_stats.getDescription() + "|");
    for(int j=0; j<col_stats.length; j++) {
      ColumnStatistics this_col = col_stats[j];
      sb.append(this_col.getColumnName());
      sb.append(",");
    }
    sb.replace(sb.length()-1, sb.length(), "|");
    sb.append(col_stats.length + "|");
    sb.append(file_stats.getLineCount() + "|");
    sb.append(file_stats.getByteCount() + "|");
    sb.append("\n");
    return sb.toString();
  }


  /**
   * Returns the <code>MRCOLS.RRF</code> lines for the specified {@link FileStatistics}.
   * @param file_stats the {@link FileStatistics}
   * @return the <code>MRCOLS.RRF</code> lines for the specified {@link FileStatistics}
   */
  protected String getColumnTextFromFileStatistics(FileStatistics file_stats) {
    StringBuffer sb = new StringBuffer(200);
    DecimalFormat formatter = new DecimalFormat("0.00");
    ColumnStatistics [] col_stats = file_stats.getAllColumnStatistics();
    for(int j=0; j<col_stats.length; j++) {
      ColumnStatistics this_col = col_stats[j];
      sb.append(this_col.getColumnName() + "|");
      sb.append(this_col.getDescription() + "||");
      sb.append(this_col.getMinLength() + "|");
      sb.append(formatter.format(this_col.getAverageLength()));
      sb.append("|");
      sb.append(this_col.getMaxLength() + "|");
      sb.append(this_col.getFileName() + file_ext + "|");
      sb.append(this_col.getDataType() + "|");
      sb.append("\n");
    }
    return sb.toString();
  }

  /**
   * Checks the maps to see if the given {@link Atom}s sui
   * or lui should be added to the ambiguous maps.
   * @param atom the {@link Atom}to check
   */
  protected void ambigCheck(Atom atom) {
    final String sui = atom.getSui();
    if (sui_to_cui_set.containsKey(sui)) {
      Set cui_set = (Set)sui_to_cui_set.get(sui);
      if (cui_set == null) {
        cui_set = new HashSet();
        sui_to_cui_set.put(sui,cui_set);
      }
      cui_set.add(atom.getConcept().getCui());
    }

    final String lui = atom.getLui();
    if (lui_to_cui_set.containsKey(lui)) {
      Set cui_set = (Set)lui_to_cui_set.get(lui);
      if (cui_set == null) {
        cui_set = new HashSet();
        lui_to_cui_set.put(lui, cui_set);
      }
      cui_set.add(atom.getConcept().getCui());
    }
  }
}
