/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Atom.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents information about a single named unit of
 * meaning within a {@link Concept}.
 * <p>
     * Following is a UML Class diagram showing the neighborhood around {@link Atom}
 * <p><center><img src="Atom.gif"></center>
 * <p>
 * All data in a row of the <code>MRCONSO</code> file is represented here.
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Atom
    extends Comparable, ConceptElement,
    ContentViewElement, SuppressibleElement, SourceElement {

  /**
   * Returns the source asserted identifier (SAUI) for this {@link Atom}.
   * This is shorthand for {@link #getSourceAtomIdentifier()}.
   * @return the SAUI
   */
  public String getSaui();

  /**
       * Returns the source asserted identifier (SAUI) for this {@link Atom}.
   * @return the SAUI
   */
  public String getSourceAtomIdentifier();

  /**
   * Sets the source asserted identifier (SAUI) for this {@link Atom}.
   * @param saui the SAUI
   */
  public void setSourceAtomIdentifier(String saui);

  /**
   * Sets the source asserted identifier (SAUI) for this {@link Atom}.
   * This is shorthand for {@link #setSourceAtomIdentifier(String)}.
   * @param saui the SAUI
   */
  public void setSaui(String saui);

  /**
   * Returns the {@link TermType} (TTY).
   * @return the {@link TermType}
   */
  public TermType getTermType();

  /**
   * Returns the {@link TermType} (TTY).
   * This is shorthand for {@link #getTermType()}.
   * @return the {@link TermType}
   */
  public TermType getTty();

  /**
   * Sets the {@link TermType} (TTY).
   * @param tty the {@link TermType}
   */
  public void setTermType(TermType tty);

  /**
   * Sets the {@link TermType} (TTY).
   * This is shorthand for {@link #setTermType(TermType)}.
   * @param tty the {@link TermType}
   */
  public void setTty(TermType tty);

  /**
   * Returns the source asserted descriptor (SDUI).
   * This is shorthand for {@link #getSourceDescriptor()}.
   * @return the SDUI
   */
  public String getSdui();

  /**
   * Returns the source asserted descriptor (SDUI).
   * @return the SDUI
   */
  public String getSourceDescriptor();

  /**
   * Sets the source asserted descriptor (SDUI).
   * @param sdui the SDUI
   */
  public void setSourceDescriptor(String sdui);

  /**
   * Sets the source asserted descriptor (SDUI).
   * This is shorthand for {@link #setSourceDescriptor(String)}.
   * @param sdui the SDUI
   */
  public void setSdui(String sdui);

  /**
   * Returns the source attributed concept identifier (SCUI).
   * This is shorthand for {@link #getSourceConceptIdentifier()}.
   * @return the SCUI
   */
  public String getScui();

  /**
   * Returns the source attributed concept identifier (SCUI).
   * @return the SCUI
   */
  public String getSourceConceptIdentifier();

  /**
   * Sets the source attributed concept identifier (SCUI).
   * This is shorthand for {@link #setSourceConceptIdentifier(String)}.
   * @param scui the SCUI
   */
  public void setScui(String scui);

  /**
   * Sets the source attributed concept identifier (SCUI).
   * @param scui the SCUI
   */
  public void setSourceConceptIdentifier(String scui);

  /**
   * Return the code (CODE).
   * @return the code
   */
  public String getCode();

  /**
   * Sets the code (CODE).
   * @param code the code
   */
  public void setCode(String code);

  /**
   * Returns the flag indicating whether or not this {@link Atom}
   * is preferred (ISPREF).
   * @return the flag indicating whether or not this {@link Atom} is preferred
   */
  public String getIspref();

  /**
   * Sets the flag indicating whether or not this {@link Atom}
   *  is preferred (ISPREF).
   * @param ispref the flag indicating whether or not this {@link Atom}
   * is preferred
   */
  public void setIspref(String ispref);

  /**
   * Indicates whether or not this {@link Atom} is the preferred atom identifier
   * for its {@link Concept} and string identifier.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredAtomIdentifier();

  /**
   * Indicates whether or not this {@link Atom} is the preferred atom identifier
   * for its {@link Concept} and string identifier.  This is shorthand for
   * {@link #hasPreferredAtomIdentifier()}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredAui();

  /**
   * Shorthand for {@link #getAtomIdentifier()}.
   * @return the AUI
   */
  public String getAui();

  /**
   * Shorthand for {@link #setAtomIdentifier(String)}.
   * @param aui the AUI
   */
  public void setAui(String aui);

  /**
   * Returns the atom unique identifier (AUI).
   * @return the AUI
   */
  public String getAtomIdentifier();

  /**
   * Sets the atom unique identifier (AUI).
   * @param aui the AUI
   */
  public void setAtomIdentifier(String aui);

  /**
   * Returns the {@link Language} (LAT).
   * @return the {@link Language}
   */
  public Language getLanguage();

  /**
   * Returns the {@link Language} (LAT).
   * This is shorthand for {@link #getLanguage()}.
   * @return {@link Language}
   */
  public Language getLat();

  /**
   * Sets the {@link Language} (LAT).
   * @param lat the {@link Language}
   */
  public void setLanguage(Language lat);

  /**
   * Sets the {@link Language} (LAT).
   * This is shorthand for {@link #setLanguage(Language)}.
   * @param lat the {@link Language}
   */
  public void setLat(Language lat);

  /**
   * Indicates whether or not this atom is part of the preferred {@link Term}
   * of the {@link Concept}.  In other words, does this {@link Atom} have
   * the same LUI as the preferred {@link Atom} of the {@link Concept}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredTermIdentifier();

  /**
   * Indicates whether or not this atom is part of the preferred {@link Term}
   * of the {@link Concept}.  This is shorthand for {@link
   * #hasPreferredTermIdentifier()}
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredLui();

  /**
   * Returns the {@link TermStatus} (TS).
   * @return the {@link TermStatus}
   */
  public TermStatus getTermStatus();

  /**
   * Returns the {@link TermStatus} (TS).
   * This is shorthand for {@link #getTermStatus()}.
   * @return the {@link TermStatus}
   */
  public TermStatus getTs();

  /**
   * Sets the {@link TermStatus} (TS).
   * @param ts the {@link TermStatus}
   */
  public void setTermStatus(TermStatus ts);

  /**
   * Sets the {@link TermStatus} (TS).
   * This is shorthand for {@link #setTermStatus(TermStatus)}.
   * @param ts the {@link TermStatus}
   */
  public void setTs(TermStatus ts);

  /**
   * Returns the term identifier (LUI).
   * This is shorthand for {@link #getTermIdentifier()}.
   * @return the LUI
   */
  public String getLui();

  /**
   * Sets the term identifier (LUI). This is shorthand
   * for {@link #setTermIdentifier(String)}.
   * @param lui the LUI
   */
  public void setLui(String lui);

  /**
   * Returns the term identifier (LUI).
   * @return the LUI
   */
  public String getTermIdentifier();

  /**
   * Sets the term identifier (LUI).
   * @param lui the LUI
   */
  public void setTermIdentifier(String lui);

  /**
   * Indicates whether or not this atom has the preferred string identifier
   * of the {@link Term} it is part of.  In other words, does this {@link Atom} have
   * the same SUI as the preferred {@link Atom} of the {@link Term}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredStringIdentifier();

  /**
   * Indicates whether or not this atom has the preferred string identifier
   * of the {@link Term} it is part of.  This is shorthand for
   * {@link #hasPreferredStringIdentifier()}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredSui();

  /**
   * Returns the {@link StringType} (STT).
   * @return the {@link StringType}
   */
  public StringType getStringType();

  /**
   * Returns the {@link StringType} (STT).
   * This is shorthand for {@link #getStringType()}.
   * @return the {@link StringType}
   */
  public StringType getStt();

  /**
   * Sets the {@link StringType} (STT).
   * @param stt the STT
   */
  public void setStringType(StringType stt);

  /**
   * Sets the {@link StringType} (STT).
   * This is shorthand for {@link #setStringType(StringType)}.
   * @param stt the STT
   */
  public void setStt(StringType stt);

  /**
   * Return the string identifier (SUI).
   * This is shorthand for {@link #getStringIdentifier()}.
   * @return the SUI
   */
  public String getSui();

  /**
   * Sets the string identifier (SUI).
   * This is shorthand for {@link #setStringIdentifier(String)}.
   * @param sui the SUI
   */
  public void setSui(String sui);

  /**
   * Return the string identifier (SUI).
   * @return the SUI
   */
  public String getStringIdentifier();

  /**
   * Sets the string identifier (SUI).
   * @param sui the SUI
   */
  public void setStringIdentifier(String sui);

  /**
   * Sets the name (STR).
   * @return the name
   */
  public String getName();

  /**
   * Returns the name (STR).
   * This is shorthand for {@link #getName()}.
   * @return the name
   */
  public String getStr();

  /**
   * Sets the name (STR).
   * @param str the name
   */
  public void setName(String str);

  /**
   * Sets the name (STR).
   * This is shorthand for {@link #setName(String)}.
   * @param str the name
   */
  public void setStr(String str);

  //
  // Definitions
  //

  /**
   * Returns the {@link Definition}s.
   * @return a {@link Definition} <code>[]</code>
   */
  public Definition[] getDefinitions();

  /**
   * Returns the {@link Definition}s as a {@link List}.
   * @return the {@link Definition}s as a {@link List}
   */
  public List getDefinitionsAsList();

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Atom} fields
   */
  public String getKeyFields();

  /**
   * Adds the specified {@link Definition}.
   * @param def the {@link Definition} to be added
   */
  public void addDefinition(Definition def);

  /**
   * Remove the specified {@link Definition}.
   * @param def the {@link Definition} to be removed
   */
  public void removeDefinition(Definition def);

  /**
   * Remove all {@link Definition}s.
   */
  public void clearDefinitions();

  /**
   * Replace {@link Definition}s with the ones in the specified array.
   * @param defs a {@link Definition} <code>[]</code>
   */
  public void setDefinitions(Definition defs[]);

  /**
   * Replace {@link Definition}s with the ones in the specified {@link List}.
   * @param defs a {@link List} of {@link Definition}s
   */
  public void setDefinitions(List defs);

  //
  // Locators
  //

  /**
   * Returns the {@link Locator}s.
   * @return a {@link Locator} <code>[]</code>
   */
  public Locator[] getLocators();

  /**
   * Returns the {@link Locator}s as a {@link List}.
   * @return the {@link Locator}s as a {@link List}
   */
  public List getLocatorsAsList();

  /**
   * Add the specified {@link Locator}.
   * @param loc the {@link Locator} to be added
   */
  public void addLocator(Locator loc);

  /**
   * Removes the specified {@link Locator}.
   * @param loc the {@link Locator} to be removed
   */
  public void removeLocator(Locator loc);

  /**
   * Remove all {@link Locator}s.
   */
  public void clearLocators();

  /**
   * Replace {@link Locator}s with the ones in the specified array.
   * @param locs a {@link Locator} <code>[]</code>
   */
  public void setLocators(Locator locs[]);

  /**
   * Replace {@link Locator}s with the ones in the specified {@link List}.
   * @param locs a {@link List} of {@link Locator}s
   */
  public void setLocators(List locs);

  //
  // Relationships
  //

  /**
   * Returns the {@link Relationship}s.
   * @return a {@link Relationship} <code>[]</code>
   */
  public Relationship[] getRelationships();

  /**
   * Returns the {@link Relationship}s as a {@link List}.
   * @return the {@link Relationship}s as a {@link List}
   */
  public List getRelationshipsAsList();

  /**
   * Adds the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be added
   */
  public void addRelationship(Relationship rel);

  /**
   * Removes the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be removed
   */
  public void removeRelationship(Relationship rel);

  /**
   * Remove all {@link Relationship}s.
   */
  public void clearRelationships();

  /**
   * Replace {@link Relationship}s with the ones in the specified array.
   * @param rels a {@link Relationship} <code>[]</code>
   */
  public void setRelationships(Relationship rels[]);

  /**
   * Replace {@link Relationship}s of the ones in the specified {@link List}.
   * @param rels a {@link List} of {@link Relationship}s
   */
  public void setRelationships(List rels);

  //
  // Coocurrences
  //

  /**
   * Returns the {@link Coocurrence}s.
   * @return a {@link Coocurrence} <code>[]</code>
   */
  public Coocurrence[] getCoocurrences();

  /**
   * Returns the {@link Coocurrence}s as a {@link List}.
   * @return the {@link Coocurrence}s as a {@link List}
   */
  public List getCoocurrencesAsList();

  /**
   * Adds the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be added
   */
  public void addCoocurrence(Coocurrence coc);

  /**
   * Removes the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be removed
   */
  public void removeCoocurrence(Coocurrence coc);

  /**
   * Remove all {@link Coocurrence}s.
   */
  public void clearCoocurrences();

  /**
   * Replace {@link Coocurrence}s with the ones in the specified array.
   * @param cocs a {@link Coocurrence} <code>[]</code>
   */
  public void setCoocurrences(Coocurrence cocs[]);

  /**
   * Replace {@link Coocurrence}s with the ones in the specified {@link List}.
   * @param cocs a {@link List} of {@link Coocurrence}s
   */
  public void setCoocurrences(List cocs);

  //
  // Contexts
  //

  /**
   * Return the {@link Context}s.
   * @return a {@link Context} <code>[]</code>
   */
  public Context[] getContexts();

  /**
   * Returns the {@link Context}s as a {@link List}.
   * @return the {@link Context}s as a {@link List}
   */
  public List getContextsAsList();

  /**
   * Adds the specified {@link Context}.
   * @param cxt the {@link Context} to be added
   */
  public void addContext(Context cxt);

  /**
   * Removes the specified {@link Context}.
   * @param cxt the {@link Context} to be removed
   */
  public void removeContext(Context cxt);

  /**
   * Remove all {@link Context}s.
   */
  public void clearContexts();

  /**
   * Replace {@link Context}s with the ones in the specified array.
   * @param cxts a {@link Context} <code>[]</code>
   */
  public void setContexts(Context cxts[]);

  /**
   * Replace {@link Context}s with the ones in the specified {@link List}.
   * @param cxts a {@link List} of {@link Context}s
   */
  public void setContexts(List cxts);

  //
  // AtomAttributes
  //

  /**
   * Returns the {@link AtomAttribute}s.
   * @return a {@link AtomAttribute} <code>[]</code>
   */
  public AtomAttribute[] getAttributes();

  /**
   * Returns the {@link AtomAttribute}s as a {@link List}.
   * @return the {@link AtomAttribute}s as a {@link List}
   */
  public List getAttributesAsList();

  /**
   * Adds the specified {@link AtomAttribute}.
   * @param att the {@link AtomAttribute} to be added
   */
  public void addAttribute(AtomAttribute att);

  /**
   * Removes the specified {@link AtomAttribute}.
   * @param att the {@link AtomAttribute} to be removed
   */
  public void removeAttribute(AtomAttribute att);

  /**
   * Remove all {@link AtomAttribute}s.
   */
  public void clearAttributes();

  /**
   * Replace {@link AtomAttribute}s with the ones in the specified array.
   * @param atts an {@link AtomAttribute} <code>[]</code>
   */
  public void setAttributes(AtomAttribute atts[]);

  /**
   * Replace {@link AtomAttribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link AtomAttribute}s
   */
  public void setAttributes(List atts);

  /**
   * Returns a subset of this {@link Atom}'s {@link Relationship}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Relationship[]
   */
  public Relationship[] getRelationships(Restrictor restrictor);


  /**
   * {@link Restrictor} that restricts by {@link Language}.
   */
  public final static class ByLanguageRestrictor implements Restrictor {
    private List language_list = new ArrayList();

    /**
     * Constructor
     * Keep only languages given as parameter.
     * @param languages {@link Language}[]
     */
    public ByLanguageRestrictor(Language[] languages) {
      for (int i=0; i<languages.length; i++) {
        language_list.add(languages[i]);
      }
    }

    /**
     * Compares the {@link Language} of each {@link AtomElement} object.
     * @param o1 the first {@link Object} to be compared.
     * @param o2 the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a positive integer
     *    as the first argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null ||
          !(o1 instanceof Atom) || !(o2 instanceof Atom))
         return 0;
      Language s1 = ((Atom) o1).getLanguage();
      Language s2 = ((Atom) o2).getLanguage();

      String lat1 = s1.getAbbreviation();
      String lat2 = s2.getAbbreviation();

      return lat1.compareTo(lat2);
    }

    /**
     * Returns whether or not the given {@link Atom} should be
     * kept or not
     * @param o {@link Object}
     * @return <code>boolean</code>
     */
    public boolean keep(Object o) {
      if (o instanceof Atom)
        return language_list.contains(((Atom)o).getLanguage());
      return true;
    }
  }

  /**
   * Sorts the atoms.
   */
  public final static class AtomRankComparator implements Comparator {

    //
    // Private fields
    //
    private boolean lat_flag = false;
    private boolean sui_flag = false;
    private boolean lui_flag = false;
    private HashMap atoms_rank = null;

    // Maps SourceTermType objects to strings representing the
    // ranking (for getRank).
    private Map termgroup_rank_map = null;
    private List lat_rank = new ArrayList();
    private List lui_rank = new ArrayList();
    private List sui_rank = new ArrayList();

    //
    // Constructors
    //

    /**
     * Constructor
     * @param atoms a {@link Atom}[] containing atoms to compare
     * @param prec {@link SourceTermType}[]
     * @param lat_flag a flag indicating whether or not to sort by LAT
     * @param sui_flag a flag indicating whether or not to sort by SUI
     * @param lui_flag a flag indicating whether or not to sort by LUI
     * @throws Exception
     */
    public AtomRankComparator(Atom[] atoms, SourceTermType[] prec, boolean lat_flag, boolean sui_flag,
                              boolean lui_flag) throws Exception {

      // Initialize lat_rank arraylist
      Language[] lats = Language.getLanguages();
      Arrays.sort(lats);
      for (int i=lats.length-1; i>=0; i--) {
        if (!lats[i].getAbbreviation().equals("ENG"))
          lat_rank.add(lats[i].getAbbreviation());
      }
      lat_rank.add("ENG");


      // Get list of source ttys in precedence order
      List precedence = new ArrayList();
      for (int i=0; i<prec.length; i++) {
        precedence.add(prec[i]);
      }

      // Initialize the termgroup_rank_map
      if (termgroup_rank_map == null) {
        termgroup_rank_map = new HashMap(400);
        for (int i = 0; i < precedence.size(); i++) {
          final SourceTermType tmgrp = (SourceTermType)precedence.get(i);
          final int rnk = 9999 - i;
          final String r = "0000" + rnk;
          final String pad_i = r.substring(r.length() - 4);
          termgroup_rank_map.put(tmgrp,pad_i);
        }
      }

      setAtoms(atoms);
      Arrays.sort(atoms, this);
      for (int i=0; i<atoms.length; i++) {
        if (!lui_rank.contains(atoms[i].getLui()))
          lui_rank.add(atoms[i].getLui());
      }
      for (int i=0; i<atoms.length; i++) {
        if (!sui_rank.contains(atoms[i].getSui()))
          sui_rank.add(atoms[i].getSui());
      }

      this.lat_flag = lat_flag;
      this.sui_flag = sui_flag;
      this.lui_flag = lui_flag;
      setAtoms(atoms);
    }

    //
    // Methods
    //

    /**
     * This method initializes the comparator with a new {@link Concept}.
     * @param atoms a new {@link Atom}[]
     */
    public void setAtoms(Atom[] atoms) {

      // clear hashmaps
      atoms_rank = new HashMap();

      // first go through non-mm atoms and make sure that the highest rank
      // for the atom's lower case string is in the string_maxrank hashmap
      for (int i = 0; i < atoms.length; i++) {
        // Get rank & cache it
        final Atom atom = atoms[i];
        final String rank = getRank(atom);
        atoms_rank.put(atoms[i].getAtomIdentifier(), rank);
      }
    }

    /**
     * Compare the two atoms based
     * on the ranks stored previously.
     * @param o1 {@link Object} 1 to compare
     * @param o2 {@link Object} 2 to compare
     * @return an <code>int</code> for the relative ordering
     */
    public int compare(Object o1, Object o2) {
      final String r1 = (String)atoms_rank.get(((Atom)o1).getAtomIdentifier());
      final String r2 = (String)atoms_rank.get(((Atom)o2).getAtomIdentifier());
      if (r1 == null || r2 == null)
        return 0;
      return r2.compareTo(r1);
    }

    /**
     * Ranking algorithm.
     * It goes: lat + termgroup_rank + ts_stt + sui + mm rank
     * or: lat + lui + termgroup_rank + ts_stt_rank + sui + mm rank
     * @param atom {@link Atom}
     * @return rank {@link String}
     */
    private String getRank(Atom atom) {

      // First compute the sab/tty rank
      //final SourceTermType tmrp = default_configuration.getSourceTermType(
      //    atom.getSource().getRsab(), atom.getTty().getAbbreviation());
      SourceTermType tmrp = new SourceTermType();
      tmrp.setSource(atom.getSource());
      tmrp.setTermType(atom.getTty());
      String pad_i = (String) termgroup_rank_map.get(tmrp);
      if (pad_i == null)
        pad_i = "0000";

        // Next compute the TS,STT rank
      String ts_stt = null;
      if (atom.getTs() == null || atom.getStt() == null || 
          atom.getTs().getAbbreviation().equals("") || 
          atom.getStt().getAbbreviation().equals(""))
        ts_stt = "";
      else
        ts_stt = atom.getTs().getAbbreviation() +
            atom.getStt().getAbbreviation().substring(0, 1);
      int ts_stt_value;
      if (ts_stt.equals("PP"))
        ts_stt_value = 4;
      else if (ts_stt.equals("SP"))
        ts_stt_value = 3;
      else if (ts_stt.equals("PV"))
        ts_stt_value = 2;
      else //ts_stt is "SV"
        ts_stt_value = 1;

        // ensure that MTH/MMs are in correct order from <1> to <n>
      //final String bn = (atom.getSource().getRsab().equals("MTH") &&
      //                   atom.getTermType().getAbbreviation().equals("MM")) ?
      //    String.valueOf(100 - getBracketNumber(atom.toString())) : "00";

      // Assemble the entire rank value and return
      final StringBuffer sb = new StringBuffer(40);
      // fool ENG atoms to be first by substituting "AAA"
      if (lat_flag)
        sb.append(getLanguageRank(atom.getLat().getAbbreviation()));
      if (lui_flag)
        sb.append(getLuiRank(atom.getLui()));
      if (sui_flag)
        sb.append(getSuiRank(atom.getSui()));
      sb.append(pad_i);
      //sb.append(bn);
      sb.append(ts_stt_value);
      sb.append(atom.getSui());
      sb.append(atom.getAui());

      return sb.toString();

    }

    private int getLuiRank(String lui) {
      return 9999 - lui_rank.indexOf(lui);
    }

    private int getSuiRank(String sui) {
      return 9999 - sui_rank.indexOf(sui);
    }

    private int getLanguageRank(String lat) {
      return lat_rank.indexOf(lat) + 100;
    }

    /**
     * Returns the bracket number of a bracket string.
     * For example, for "Apple <1>", it returns 1
     * @param string a bracket string
     * @return the bracket number
     */
    private int getBracketNumber(String string) {
      boolean gt_found = false;
      boolean digit_found = false;
      int multiplier = 1;
      int bracket_num = 0;

      for (int i = string.length() - 1; i > 0; i--) {
        if (i == string.length() - 1 && string.length() != '>')

          // last character must be '>'
          return 0;
        if (string.charAt(i) == '>') {
          // set gt_found
          gt_found = true;
        } else if (gt_found && string.charAt(i) > '0' && string.charAt(i) < '9') {
          // '>' was found and the current character is a digit
          // add the digit to bracket_num and increase the multiplier
          bracket_num += (string.charAt(i) - '0') * multiplier;
          multiplier *= 10;
          digit_found = true;
        } else if (gt_found && digit_found && string.charAt(i) == '<') {
          // digits followed by '>' found, '<' also found
          if (i > 0 && string.charAt(i - 1) == ' ') {
            // previous character is a ' '
            return bracket_num;
          } else {
            return 0;
          }
        } else {
          // matching pattern not found, no bracket string
          return 0;
        }
      }
      return 0;
    }

  }

}