/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   AtomElement.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.List;

/**
 * Represents information about a data element connected to an {@link Atom}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface AtomElement {

  /**
   * Returns the {@link Atom} to which this element is connected.
   * @return the {@link Atom} to which this element is connected
   */
  public Atom getAtom();

  /**
   * Sets the {@link Atom} to which this element is connected.
   * @param atom the {@link Atom} to which this element is connected.
   */
  public void setAtom(Atom atom);


  /**
   * {@link Restrictor} that restricts by {@link Language}.
   */
  public final static class ByLanguageRestrictor implements Restrictor {
    private List language_list = new ArrayList();

    /**
     * Constructor
     * Keep only languages given as parameter.
     * @param languages {@link Language}[]
     */
    public ByLanguageRestrictor(Language[] languages) {
      for (int i=0; i<languages.length; i++) {
        language_list.add(languages[i]);
      }
    }

    /**
     * Compares the {@link Language} of each {@link AtomElement} object.
     * @param o1 the first {@link Object} to be compared.
     * @param o2 the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a positive integer
     *    as the first argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null ||
          !(o1 instanceof AtomElement) || !(o2 instanceof AtomElement))
         return 0;
      Language s1 = ((AtomElement) o1).getAtom().getLanguage();
      Language s2 = ((AtomElement) o2).getAtom().getLanguage();

      String lat1 = s1.getAbbreviation();
      String lat2 = s2.getAbbreviation();

      return lat1.compareTo(lat2);
    }

    /**
     * Returns whether or not the given {@link AtomElement} should be
     * kept or not
     * @param o {@link Object}
     * @return <code>boolean</code>
     */
    public boolean keep(Object o) {
      if (o instanceof AtomElement)
        return language_list.contains(((AtomElement)o).getAtom().getLanguage());
      return true;
    }
  }


  /**
   * {@link Restrictor} that restricts by {@link Source}.
   */
  public final static class BySourceRestrictor implements Restrictor {
    private List source_list = new ArrayList();

    /**
     * Constructor
     * Keep only sources given as parameter.
     * @param sources {@link Source}[]
     */
    public BySourceRestrictor(Source[] sources) {
      for (int i=0; i<sources.length; i++) {
        source_list.add(sources[i]);
      }
    }

    /**
     * Compares the {@link Source} of each {@link AtomElement} object.
     * @param o1 the first {@link Object} to be compared.
     * @param o2 the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a positive integer
     *    as the first argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null ||
          !(o1 instanceof AtomElement) || !(o2 instanceof AtomElement))
         return 0;
      Source s1 = ((AtomElement) o1).getAtom().getSource();
      Source s2 = ((AtomElement) o2).getAtom().getSource();

      String source1 = s1.getRsab();
      String source2 = s2.getRsab();

      return source1.compareTo(source2);
    }

    /**
     * Returns whether or not the given {@link AtomElement} should be
     * kept or not
     * @param o {@link Object}
     * @return <code>boolean</code>
     */
    public boolean keep(Object o) {
      if (o instanceof AtomElement)
        return source_list.contains(((AtomElement)o).getAtom().getSource());
      return true;
    }
  }

}