/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   AtomMapping.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;


/**
 * Represents a mapping for a between two UMLS {@link Atom}
 * identifiers.  For the time being, we only represent changes in
 * CUI values from one release to another, but eventually this data
 * structure will represent things like "safe replacement".
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link AtomMapping}
 * <p><center><img src="AtomMapping.gif"></center>
 * <p>
 * All data in a row of the <code>MRAUI</code> file is represented here.
 * <p>
 * @author Brian Carlsen
 */
public interface AtomMapping extends Comparable {

  /**
   * Returns the atom identifier for the first atom (AUI1).
   * This method is shorthand for {@link #getAtomIdentifier()}.
   * @return the atom identifier for the first atom
   */
  public String getAui1();

  /**
   * Returns the atom identifier for the first atom (AUI1).
   * @return the atom identifier for the first atom
   */
  public String getAtomIdentifier();

  /**
   * Sets atom identifier for the first atom (AUI1).
   * This method is shorthand for {@link #setAtomIdentifier(String)}.
   * @param aui1 the atom identifier for the first atom
   */
  public void setAui1(String aui1);

  /**
   * Sets atom identifier for the first atom (AUI1).
   * @param aui1 the atom identifier for the first atom
   */
  public void setAtomIdentifier(String aui1);

  /**
   * Returns the concept identifier of the first atom as it was
   * in the specified release version (CUI1).
   * This method is shorthand for {@link #getConceptIdentifier()}.
   * @return the concept identifier for the first atom
   */
  public String getCui1();

  /**
   * Returns the concept identifier for the first atom (CUI1).
   * @return the concept identifier for the first atom
   */
  public String getConceptIdentifier();

  /**
   * Sets concept identifier for the first atom (CUI1).
   * This method is shorthand for {@link #setConceptIdentifier(String)}.
   * @param cui1 the concept identifier for the first atom
   */
  public void setCui1(String cui1);

  /**
   * Sets concept identifier for the first atom (CUI1).
   * @param cui1 the concept identifier for the first atom
   */
  public void setConceptIdentifier(String cui1);

  /**
   * Returns the release version in which the indicated change took place (VER).
   * @return the release version in which the indicated change took place
   */
  public String getReleaseVersion();

  /**
   * Returns the release version in which the indicated change took place (VER).
   * This method is shortand for {@link #getReleaseVersion()}.
   * @return the release version in which the indicated change took place
   */
  public String getVer();

  /**
   * Sets the release version in which the indicated change took place (VER).
   * @param ver the release version in which the indicated change took place
   */
  public void setReleaseVersion(String ver);

  /**
   * Sets the release version in which the indicated change took place (VER).
   * This method is shorthand for (@link #setReleaseVersion(String)}.
   * @param ver the release version in which the indicated change took place
   */
  public void setVer(String ver);

  /**
   * Returns the relationship between the two AUIs (REL).
   * @return the relationship between the two AUIs
   */
  public RelationshipLabel getRelationshipLabel();

  /**
   * Returns the relationship between the two AUIs (REL).
   * This method is shorthand for {@link #getRelationshipLabel}.
   * @return the relationship between the two AUIs
   */
  public RelationshipLabel getRel();

  /**
   * Sets the relationship between the two AUIs (REL).
   * @param rel the relationship between the two AUIs
   */
  public void setRelationshipLabel(RelationshipLabel rel);

  /**
   * Sets the relationship between the two AUIs (REL).
   * This method is shorthand for {@link #setRelationshipLabel(RelationshipLabel)}
   * @param rel the relationship between the two AUIs
   */
  public void setRel(RelationshipLabel rel);

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel();

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * This method is shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela();

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela then {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela);

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * This method is shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela then {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela);

  /**
   * Returns the atom identifier for the second atom (AUI2).
   * This method is shorthand for {@link #getRelatedAtomIdentifier()}.
   * @return the atom identifier for the second atom
   */
  public String getAui2();

  /**
   * Returns the atom identifier for the second atom (AUI2).
   * @return the atom identifier for the second atom
   */
  public String getRelatedAtomIdentifier();

  /**
   * Sets atom identifier for the second atom (AUI2).
   * This method is shorthand for {@link #setAtomIdentifier(String)}.
   * @param aui2 the atom identifier for the second atom
   */
  public void setAui2(String aui2);

  /**
   * Sets atom identifier for the second atom (AUI2).
   * @param aui2 the atom identifier for the second atom
   */
  public void setRelatedAtomIdentifier(String aui2);

  /**
   * Returns the concept identifier for the second concept (CUI2).
   * This method is shorthand for {@link #getRelatedConceptIdentifier()}.
   * @return the concept identifier for the second concept
   */
  public String getCui2();

  /**
   * Sets the concept identifier for the second concept (CUI2).
   * This method is shorthand for {@link #setRelatedConceptIdentifier(String)}.
   * @param cui2 the concept identifier for the second concept
   */
  public void setCui2(String cui2);

  /**
   * Returns the concept identifier for the second concept (CUI2).
   * @return the concept identifier for the second concept
   */
  public String getRelatedConceptIdentifier();

  /**
   * Sets the concept identifier for the second concept (CUI2).
   * @param cui2 the concept identifier for the second concept
   */
  public void setRelatedConceptIdentifier(String cui2);

  /**
   * Returns the "in current subset" value (MAPIN).
   * @return the "in current subset" value
   */
  public String getMapin();

  /**
   * Indicates whether or not this {@link AtomMapping} is in the current subset (MAPIN).
   * @return <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public boolean inCurrentSubset();

  /**
   * Sets the "in current subset" value (MAPIN).
   * @param mapin the "in current subset" value
   */
  public void setMapin(String mapin);

  /**
   * Sets the flag indicating whether or not
   * this {@link AtomMapping} is in the current subset (MAPIN).
   * @param mapin <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public void setInCurrentSubset(boolean mapin);

  /**
   * Returns the reason for this {@link AtomMapping} (MAPREASON).
   * @return the reason for this {@link AtomMapping}
   */
  public String getMapReason();

  /**
   * Sets the reason for this {@link AtomMapping} (MAPREASON).
   * @param reason the reason for this {@link AtomMapping}
   */
  public void setMapReason(String reason);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link AtomMapping} fields
   */
  public String getKeyFields();
}
