/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       AttributeName.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents an attribute name.  This class ensures that each unique
 * attribute name is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized, and it may
 * be initialized only once.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class AttributeName extends AbstractExpandedForm {

  // Map by abbreviation to the actual classes
  private static final Map atns = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiate an empty {@link AttributeName}.
   */
  public AttributeName() {};

  /**
   * Instantiates an {@link AttributeName} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing an attribute name
   */
  private AttributeName(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link AttributeName}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link AttributeName}s for those with a type of "ATN".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if ((meta_dsc[i].getKey().equals("ATN") ||
           meta_dsc[i].getKey().equals("MAPATN")) &&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (atns.containsKey(meta_dsc[i].getValue())) {
	   AttributeName a_name = (AttributeName)atns.get(meta_dsc[i].getValue());
           a_name.dsc = meta_dsc[i].getExplain();
        } else {
          atns.put(meta_dsc[i].getValue(), new AttributeName(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link AttributeName}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link AttributeName} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param atn the attribute name abbreviation to look up
   * @return the {@link AttributeName} for the specified abbreviation
   */
  public static AttributeName getAttributeName(String atn) {
    return (AttributeName)atns.get(atn);
  }

  /**
   * Returns all {@link AttributeName}s.
   * @return an array of {@link AttributeName}
   */
  public static AttributeName[] getAttributeNames() {
    return (AttributeName[])atns.values().toArray(new AttributeName[0]);
  }

  /**
   * Returns all {@link AttributeName}s as a {@link List}.
   * @return all {@link AttributeName}s as a {@link List}
   */
  public static List getAttributeNamesAsList() {
    return new ArrayList(atns.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}