/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   ConceptMapping.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;


/**
 * Represents a mapping between two UMLS {@link Concept}
 * identifiers.  Generally the first identifier is a concept that no
 * longer exists in the Metathesaurus and is being mapped to a {@link Concept}
 * in the current version (and current subset).
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link ConceptMapping}
 * <p><center><img src="ConceptMapping.gif"></center>
 * <p>
 * All data in a row of the <code>MRCUI</code> file is represented here.
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface ConceptMapping extends Comparable {

  /**
   * Returns the concept identifier for the first concept (CUI1).
   * This method is shorthand for {@link #getConceptIdentifier()}.
   * @return the concept identifier for the first concept
   */
  public String getCui1();

  /**
   * Returns the concept identifier for the first concept (CUI1).
   * @return the concept identifier for the first concept
   */
  public String getConceptIdentifier();

  /**
   * Sets concept identifier for the first concept (CUI1).
   * This method is shorthand for {@link #setConceptIdentifier(String)}.
   * @param cui1 the concept identifier for the first concept
   */
  public void setCui1(String cui1);

  /**
   * Sets concept identifier for the first concept (CUI1).
   * @param cui1 the concept identifier for the first concept
   */
  public void setConceptIdentifier(String cui1);

  /**
   * Returns the last release version in which the CUI1 was valid (VER).
   * @return the last release version in which the CUI1 was valid
   */
  public String getLastReleaseVersion();

  /**
   * Returns the last release version in which the CUI1 was valid (VER).
   * This method is shortand for {@link #getLastReleaseVersion()}.
   * @return the last release version in which the CUI1 was valid
   */
  public String getVer();

  /**
   * Sets the last release version in which the CUI1 was valid (VER).
   * @param ver the last release version in which the CUI1 was valid
   */
  public void setLastReleaseVersion(String ver);

  /**
   * Sets the last release version in which the CUI1 was valid (VER).
   * This method is shorthand for (@link #setLastReleaseVersion(String)}.
   * @param ver the last release version in which the CUI1 was valid
   */
  public void setVer(String ver);

  /**
   * Returns the relationship between the two CUIs (REL).
   * @return the relationship between the two CUIs
   */
  public RelationshipLabel getRelationship();

  /**
   * Returns the relationship between the two CUIs (REL).
   * This method is shorthand for {@link #getRelationship}.
   * @return the relationship between the two CUIs
   */
  public RelationshipLabel getRel();

  /**
   * Sets the relationship between the two CUIs (REL).
   * @param rel the relationship between the two CUIs
   */
  public void setRelationship(RelationshipLabel rel);

  /**
   * Sets the relationship between the two CUIs (REL).
   * This method is shorthand for {@link #setRelationship(RelationshipLabel)}
   * @param rel the relationship between the two CUIs
   */
  public void setRel(RelationshipLabel rel);

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel();

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * This method is shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela();

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela then {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela);

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * This method is shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela then {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela);

  /**
   * Returns the concept identifier for the second concept (CUI2).
   * This method is shorthand for {@link #getRelativeConceptIdentifier()}.
   * @return the concept identifier for the second concept
   */
  public String getCui2();

  /**
   * Sets the concept identifier for the second concept (CUI2).
   * This method is shorthand for {@link #setRelativeConceptIdentifier(String)}.
   * @param cui2 the concept identifier for the second concept
   */
  public void setCui2(String cui2);

  /**
   * Returns the concept identifier for the second concept (CUI2).
   * @return the concept identifier for the second concept
   */
  public String getRelativeConceptIdentifier();

  /**
   * Sets the concept identifier for the second concept (CUI2).
   * @param cui2 the concept identifier for the second concept
   */
  public void setRelativeConceptIdentifier(String cui2);

  /**
   * Returns the "in current subset" value (MAPIN).
   * @return the "in current subset" value
   */
  public String getMapin();

  /**
   * Indicates whether or not this {@link ConceptMapping} is in the current subset (MAPIN).
   * @return <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public boolean inCurrentSubset();

  /**
   * Sets the "in current subset" value (MAPIN).
   * @param mapin the "in current subset" value
   */
  public void setMapin(String mapin);

  /**
   * Sets the flag indicating whether or not
   * this {@link ConceptMapping} is in the current subset (MAPIN).
   * @param mapin <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public void setInCurrentSubset(boolean mapin);

  /**
   * Returns the reason for this {@link ConceptMapping} (MAPREASON).
   * @return the reason for this {@link ConceptMapping}
   */
  public String getMapReason();

  /**
   * Sets the reason for this {@link ConceptMapping} (MAPREASON).
   * @param reason the reason for this {@link ConceptMapping}
   */
  public void setMapReason(String reason);

  /**
   * Return the reason for this {@link ConceptMapping}.
   * @return reason
   */
  public String getReason();

  /**
   * Set the reason for this {@link ConceptMapping}.
   * @param reason reason for the mapping
   */
  public void setReason(String reason);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link ConceptMapping} fields
   */
  public String getKeyFields();
}
