/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Context.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.List;

/**
 * Represents a single "path to the root" of an {@link Atom}, including
 * the relevant tree-numbers and the various ancestors, children, and siblings.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link Context}
 * <p><center><img src="Context.gif"></center>
 * <p>
 * All data from <code>MRCXT<code> and <code>MRHIER<code> is represented here.
 * <p>
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Context
  extends AtomElement, Comparable, ConceptElement, ContentViewElement, SourceElement{

  /**
   * Returns the context number (CXN).
   * @return the <code>int</code> context number
   */
  public int getContextNumber();

  /**
   * Sets the context number (CXN).
   * @param cxn the <code>int</code> context number
   */
  public void setContextNumber(int cxn);

  /**
   * Returns the context number (CXN).
   * This is a shorthand for {@link #getContextNumber()}
   * @return cxn context number <code>int</code>
   */
  public int getCxn();

  /**
   * Sets the context number (CXN).
   * This is shorthand for {@link #setContextNumber(int)}.
   * @param cxn <code>int</code> context number
   */
  public void setCxn(int cxn);

  /**
   * Returns the parent atom identifier (PAUI).
   * This is shorthand for {@link #getParentAtomIdentifier()}.
   * @return the PAUI
   */
  public String getPaui();

  /**
   * Sets the parent atom identifier (PAUI).
   * This is shorthand for {@link #setParentAtomIdentifier(String)}.
   * @param paui the PAUI
   */
  public void setPaui(String paui);

  /**
   * Returns the parent atom identifier (PAUI).
   * @return the PAUI
   */
  public String getParentAtomIdentifier();

  /**
   * Sets the parent atom identifier (PAUI).
   * @param paui the PAUI
   */
  public void setParentAtomIdentifier(String paui);

  /**
   * Returns the path to the root (PTR) as a "." separated list of AUI values.
   * @return the path to the root
   */
  public String getPathToRootAsString();

  /**
   * Returns the path to the root (PTR) as a {@link List} of AUI values.
   * @return the path to the root as a {@link List} of AUI values
   */
  public List getPathToRootAsList();

  /**
   * Returns the path to the root (PTR) as a {@link String}<code>[]</code>
   * of AUI values.
   * @return the path to the root as a {@link String}<code>[]</code>
   * of AUI values
   */
  public String [] getPathToRoot();

  /**
   * Returns the path to the root (PTR) as a "." separated list of AUI values.
   * This is shorthand for {@link #getPathToRootAsString()}.
   * @return the path to the root
   */
  public String getPtr();

  /**
   * Sets the path to the root (PTR) as a "." separated list of AUI values.
   * @param ptr the path to the root
   */
  public void setPathToRoot(String ptr);

  /**
   * Sets the path to the root (PTR) as a "." separated list of AUI values.
   * This is shorthand for {@link #setPathToRoot(String)}.
   * @param ptr the path to the root
   */
  public void setPtr(String ptr);

  /**
   * Returns all {@link ContextMember}s.  These are the ancestors,
   * children, and siblings.
   * @return a sorted array of {@link ContextMember}s
   */
  public ContextMember [] getContextMembers();

  /**
   * Returns all {@link ContextMember}s as a {@link List}.  These are the ancestors,
   * children, and siblings.
   * @return a sorted {@link List} of {@link ContextMember}s
   */
  public List getContextMembersAsList();

  /**
   * Adds the specified {@link ContextMember}.
   * @param cm the {@link ContextMember} to be added
   */
  public void addContextMember(ContextMember cm);

  /**
   * Removes the specified {@link ContextMember}.
   * @param cm the {@link ContextMember} to be removed
   */
  public void removeContextMember(ContextMember cm);

  /**
   * Removes all {@link ContextMember}s.
   */
  public void clearContextMembers();

  /**
   * Replaces the {@link ContextMember}s with the ones in the specified array.
   * @param cms a {@link ContextMember}<code>[]</code>
   */
  public void setContextMembers(ContextMember [] cms);

  /**
   * Replaces the {@link ContextMember}s with the ones in the specified {@link List}.
   * @param cms a {@link List} of {@link ContextMember}s
   */
  public void setContextMembers(List cms);
  /**
   * Returns the hierarchical code (HCD).
   * @return the hierarchical code
   */
  public String getHierarchicalCode();

  /**
   * Returns the hierarchical code (HCD).
   * This is shorthand for {@link #getHierarchicalCode()}.
   * @return the hierarchical code
   */
  public String getHcd();

  /**
   * Sets the hierarchical code (HCD).
   * @param hcd the hierarchical code
   */
  public void setHierarchicalCode(String hcd);

  /**
   * Sets the hierarchical code (HCD).
   * This is shorthand for {@link #setHierarchicalCode(String)}.
   * @param hcd the hierarchical code
   */
  public void setHcd(String hcd);

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel();

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * This is shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela();

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela);

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * This is shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela);

  /**
   * Returns the {@link ContextMember} that has the context
   * member label (CXL) of "CCP" or self.
   * @return {@link ContextMember} with label "CCP"
   */
  public ContextMember getSelfMember();

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CHD" or child.
   * @return {@link ContextMember}[] of members with label "CHD"
   */
  public ContextMember[] getChildMembers();

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "CHD" or child.
   * @return {@link ContextMember}[] of members with label "CHD"
   */
  public List getChildMembersAsList();

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "ANC" or self.
   * @return {@link ContextMember}[] of members with label "ANC"
   */
  public ContextMember[] getAncestorMembers();

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "ANC" or self.
   * @return {@link ContextMember}[] of members with label "ANC"
   */
  public List getAncestorMembersAsList();

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "SIB" or self.
   * @return {@link ContextMember}[] of members with label "SIB"
   */
  public ContextMember[] getSiblingMembers();

  /**
   * Returns only the {@link ContextMember}s that have the context
   * member label (CXL) of "SIB" or self.
   * @return {@link ContextMember}[] of members with label "SIB"
   */
  public List getSiblingMembersAsList();

  /**
   * Indicates if context is of the top level.
   * @return <code>boolean</code> true if top level
   *                              false if otherwise
   */
  public boolean isTopLevel();

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Context} fields
   */
  public String getKeyFields();
}