/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   ContextMember.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

/**
 * Represents information about ancestors, children, or siblings
 * in a hierarchy.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link ContextMember}
 * <p><center><img src="ContextMember.gif"></center>
 * <p>
 * Some data in <code>MRCXT</code> is represented here.
 * <p>
 * @see Context
 * @see SourceContext
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface ContextMember extends Comparable,
  RelatedAtomElement, RelatedConceptElement {

  /**
   * Returns the rank (RNK) if the member is an ancestor.
   * @return the <code>int</code> rank
   */
  public int getRank();

  /**
   * Sets the rank (RNK) if the member is an ancestor
   * @param rnk the <code>int</code> rank
   */
  public void setRank(int rnk);

  /**
   * Returns the hierarchical code (HCD).
   * @return the hierarchical code
   */
  public String getHierarchicalCode();

  /**
   * Returns the hierarchical code (HCD).
   * This is shorthand for {@link #getHierarchicalCode()}.
   * @return the hierarchical code
   */
  public String getHcd();

  /**
   * Sets the hierarchical code (HCD).
   * @param hcd the hierarchical code
   */
  public void setHierarchicalCode(String hcd);

  /**
   * Sets the hierarchical code (HCD).
   * This is shorthand for {@link #setHierarchicalCode(String)}.
   * @param hcd the hierarchical code
   */
  public void setHcd(String hcd);

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel();

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * This is shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getRela();

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela);

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA) from the context member to its parent.
   * This is shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela);

  /**
   * Returns the "has child flag" value (XC).
   * @return the "has child flag" value
   */
  public String getXc();

  /**
   * Sets the "has child flag" value (XC).
   * @param xc the "has child flag" value
   */
  public void setXc(String xc);

  /**
   * Indicates whether or not this context member has children.
   * @return <code>true</code> if {@link ContextMember} has a child,
   *         <code>false</code> if {@link ContextMember} does not have a child
   */
  public boolean hasChild();

  /**
   * Sets the flag indicating whether or not this {@link ContextMember} has a child.
   * @param xc  <code>true</code> if {@link ContextMember} has a child,
   *            <code>false</code> if {@link ContextMember} does not have a child
   */
  public void setHasChild(boolean xc);

  /**
   * Returns the name for this {@link ContextMember} (CXS).
   * @return the name for this {@link ContextMember}
   */
  //public String getRelativeName();

  /**
   * Returns the name for this {@link ContextMember} (CXS).
   * This is shorthand for {@link #getRelativeName()}.
   * @return the name for this {@link ContextMember}
   */
  //public String getCxs();

  /**
   * Sets the string for this {@link ContextMember}.
   * @param cxs
   */
  //public void setRelativeName(String cxs);

  /**
   * Sets the name for this {@link ContextMember} (CXS).
   * This is shorthand for {@link #setRelativeName(String)}.
   * @param cxs the name for this {@link ContextMember}
   */
  //public void setCxs(String cxs);

  /**
   * Returns the label for this {@link ContextMember} (CXL).
   * @return the label for this {@link ContextMember}
   */
  public String getLabel();

  /**
   * Returns the label for this {@link ContextMember} (CXL).
   * This is shorthand for {@link #getLabel()}.
   * @return the label for this {@link ContextMember}
   */
  public String getCxl();

  /**
   * Sets the label for this {@link ContextMember}.
   * @param cxl label
   */
  public void setLabel(String cxl);

  /**
   * Sets the label for this {@link ContextMember} (CXL).
   * This is shorthand for {@link #setLabel(String)}.
   * @param cxl the label for this {@link ContextMember}
   */
  public void setCxl(String cxl);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link ContextMember} fields
   */
  public String getKeyFields();
}