/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Coocurrence.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.List;

/**
 * Represents a co-ocurring of two names ({@link Atom}s) or meanings
 * ({@link Concept}s) in a source of biomedical literature.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link Coocurrence}
 * <p><center><img src="Coocurrence.gif"></center>
 * <p>
 * All data in a row of the <code>MRCOC</code> file is represented here.
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Coocurrence extends
  Comparable, ConceptElement, ContentViewElement,
  RelatedConceptElement, SourceElement {

  /**
   * Returns the {@link CoocurrenceType} (COT).
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCoocurrenceType();

  /**
   * Returns the {@link CoocurrenceType} (COT).
   * This is shorthand for {@link #getCoocurrenceType()}.
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCot();

  /**
   * Sets the {@link CoocurrenceType} (COT).
   * @param cot the {@link CoocurrenceType}
   */
  public void setCoocurrenceType(CoocurrenceType cot);

  /**
   * Sets the {@link CoocurrenceType} (COT).
   * This is shorthand for {@link #setCoocurrenceType(CoocurrenceType)}.
   * @param cot the {@link CoocurrenceType}
   */
  public void setCot(CoocurrenceType cot);

  /**
   * Returns the coocurrence frequency (COF).
   * @return the coocurrence frequency
   */
  public int getCoocurrenceFrequency();

  /**
   * Returns the coocurrence frequency (COF).
   * This is shorthand for {@link #getCoocurrenceFrequency()}.
   * @return the coocurrence frequency
   */
  public int getCof();

  /**
   * Sets the coocurrence frequency (COF).
   * @param cof the coocurrence frequency
   */
  public void setCoocurrenceFrequency(int cof);

  /**
   * Sets the coocurrence frequency (COF).
   * This is shorthand for {@link #setCoocurrenceFrequency(int)}.
   * @param cof the coocurrence frequency
   */
  public void setCof(int cof);

  /**
   * Returns the coocurrence attribute value (COA).
   * This is shorthand for {@link #getCoocurrenceAttribute()}.
   * @return the coocurrence attribute value
   */
  public String getCoa();

  /**
   * Returns the coocurrence attribute value (COA).
   * @return the coocurrence attribute value
   */
  public String getCoocurrenceAttribute();

  /**
   * Sets the coocurrence attribute value (COA). This is a comma separated
   * list of nqualifier=frequency pairs.
   * @param coa the coocuurence attribute value
   */
  public void setCoocurrenceAttribute(String coa);

  /**
   * Sets the coocurrence attribute value (COA). This is a comma separated
   * list of name=value pairs. This is shorthand for
   * {@link #setCoocurrenceAttribute(String)}.
   * @param coa the coocuurence attribute value
   */
  public void setCoa(String coa);

  /**
   * Returns the {@link QualifierFrequency} list.
   * @return an array of {@link QualifierFrequency}s
   */
  public QualifierFrequency [] getQualifiers();

  /**
   * Returns the {@link QualifierFrequency} values as a {@link List}.
   * @return the {@link QualifierFrequency} values as a {@link List}
   */
  public List getQualifierAsList();

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified array.
   * @param coa an array of {@link QualifierFrequency} values
   */
  public void setQualifiers(QualifierFrequency [] coa);

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified {@link List}
   * @param coa a {@link List} of {@link QualifierFrequency} values
   */
  public void setQualifiers(List coa);

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Coocurrence} fields
   */
  public String getKeyFields();
}