/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       CoocurrenceType.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents a type of coocurrence.  This class ensures that each unique
 * coocurrence type is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class CoocurrenceType extends AbstractExpandedForm {

  // Map by abbreviation to the actual classes
  private static final Map cots = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link CoocurrenceType}.
   */
  public CoocurrenceType() {};

  /**
   * Instantiates an {@link CoocurrenceType} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing an coocurrence type
   */
  private CoocurrenceType(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link CoocurrenceType}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link CoocurrenceType}s for those with a type of "COT".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("COT") &&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (cots.containsKey(meta_dsc[i].getValue())) {
	   CoocurrenceType coc_type = (CoocurrenceType)cots.get(meta_dsc[i].getValue());
           coc_type.dsc = meta_dsc[i].getExplain();
        } else {
          cots.put(meta_dsc[i].getValue(), new CoocurrenceType(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link CoocurrenceType}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link CoocurrenceType} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param cot the coocurrence type abbreviation to look up
   * @return the {@link CoocurrenceType} for the specified abbreviation
   */
  public static CoocurrenceType getCoocurrenceType(String cot) {
    return (CoocurrenceType)cots.get(cot);
  }

  /**
   * Return an array of all {@link CoocurrenceType}s.
   * @return {@link CoocurrenceType} []
   */
  public static CoocurrenceType[] getCoocurrenceTypes() {
    return (CoocurrenceType[])cots.values().toArray(new CoocurrenceType[0]);
  }

  /**
   * Return an {@link List} of all {@link CoocurrenceType}s.
   * @return an {@link List} of all {@link CoocurrenceType}s
   */
  public static List getCoocurrenceTypesAsList() {
    return new ArrayList(cots.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }
}