/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   HierarchicalSemanticType.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Represents information about a semantic type in the semantic network.
 * This class ensures that each unique hierarchical semantic type is
 * represented by a singleton object.  In order for this to work correctly,
 * the class must be properly initialized.
 *
 * Data from the SRDEF file (STY rows) can be found here.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class HierarchicalSemanticType
  implements Comparable, STYMetaDescription{

  // Map by abbreviation to the actual classes
  private static final Map stys = new HashMap();

  // Abbreviation and description for this individual instance.
  private String def;
  private String ex;
  private String ui;
  private String sty;
  private String nh;
  private String stn;
  private String un;

  /**
   * Instantiates an empty {@link HierarchicalSemanticType}.
   */
  public HierarchicalSemanticType() {};

  /**
   * Instantiates an {@link HierarchicalSemanticType} from the
   * specified {@link STYMetaDescription}.
   * @param meta_dsc a {@link STYMetaDescription} representing an
   * hierarchical semantic type
   */
  private HierarchicalSemanticType (STYMetaDescription meta_dsc) {
    def = meta_dsc.getDefinition();
    ex = meta_dsc.getExamplesAsString();
    ui = meta_dsc.getIdentifier();
    sty = meta_dsc.getName();
    nh = meta_dsc.getNH();
    stn = meta_dsc.getTreeNumber();
    un = meta_dsc.getUsageNote();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link HierarchicalSemanticType}s that exist at one time.
   * @param meta_dsc an array of {@link STYMetaDescription}s
   */
  public static synchronized void initialize(STYMetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
	if (stys.containsKey(meta_dsc[i].getName())) {
	   HierarchicalSemanticType hst = (HierarchicalSemanticType)stys.get(meta_dsc[i].getName());
           hst.def = meta_dsc[i].getDefinition();
           hst.ex = meta_dsc[i].getExamplesAsString();
           hst.ui = meta_dsc[i].getIdentifier();
           hst.nh = meta_dsc[i].getNH();
           hst.stn = meta_dsc[i].getTreeNumber();
           hst.un = meta_dsc[i].getUsageNote();
        } else {
          stys.put(meta_dsc[i].getName(), new HierarchicalSemanticType(meta_dsc[i]));
        }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link HierarchicalSemanticType}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(STYMetaDescription[])}.
   * @param meta_dsc an array of {@link STYMetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((STYMetaDescription[])meta_dsc.toArray(new STYMetaDescription[0]));
  }

  /**
   * Returns the {@link HierarchicalSemanticType} for the specified semantic type name.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified name
   * was not one of those used during initialization.
   * @param name the semantic type name to look up
   * @return the {@link HierarchicalSemanticType} for the specified name
   */
   public static HierarchicalSemanticType getHierarchicalSemanticType(String name) {
    return (HierarchicalSemanticType)stys.get(name);
  }

  /**
   * Return all {@link HierarchicalSemanticType}s.
   * @return an array of {@link HierarchicalSemanticType}
   */
   public static HierarchicalSemanticType[] getHierarchicalSemanticTypes() {
    return (HierarchicalSemanticType[])stys.values().toArray(new HierarchicalSemanticType[0]);
  }

  /**
   * Return all {@link HierarchicalSemanticType}s.
   * @return a {@link List} of {@link HierarchicalSemanticType}
   */
   public static List getHierarchicalSemanticTypesAsList() {
    return new ArrayList(stys.values());
  }

  /**
   * Returns the unique identifier (UI).
   * This is shorthand for {@link #getIdentifier()}.
   * @return the unique identifier
   */
  public String getUi() { return getIdentifier(); };

  /**
   * Returns the unique identifier (UI).
   * @return the unique identifier
   */
  public String getIdentifier() { return ui; };

  /**
   * Returns the name (STY).
   * This is shorthand for {@link #getName()}.
   * @return the name
   */
  public String getSty() { return getName(); };

  /**
   * Returns the name (STY).
   * @return the name
   */
  public String getName() { return sty; };

  /**
   * Returns the tree number (STN).
   * This is shorthand for {@link #getTreeNumber()}.
   * @return the tree number
   */
  public String getStn() { return getTreeNumber(); };

  /**
   * Returns the tree number (STN).
   * @return the tree number
   */
  public String getTreeNumber() { return stn; };

  /**
   * Returns the definition (DEF).
   * This is shorthand for {@link #getDefinition()}.
   * @return the definition
   */
  public String getDef() { return getDefinition(); };

  /**
   * Returns the definition (DEF).
   * @return the definition
   */
  public String getDefinition() { return def; };

  /**
   * Returns a semi-colon (;) separated list of Metathesaurus strings (EX)
   * which are examples of this semantic type.
   * This is shorthand for {@link #getExamplesAsString()}.
   * @return a value listing Metathesaurus examples
   */
  public String getEx() { return getExamplesAsString(); };

  /**
   * Returns a {@link String} array of of Metathesaurus strings (EX)
   * which are examples of this semantic type.
   * @return an {@link String} array of Metathesaurus examples
   */
  public String[] getExamples() {
    StringTokenizer st = new StringTokenizer(getEx(),";");
    String[] x = new String[st.countTokens()];
    int ct = 0;
    while (st.hasMoreTokens())
      x[ct++] = st.nextToken().trim();
    return x;
  };

  /**
   * Returns a semi-colon (;) separated list of Metathesaurus strings (EX)
   * which are examples of this semantic type.
   * This is shorthand for {@link #getExamplesAsString()}.
   * @return a value listing Metathesaurus examples
   */
  public String getExamplesAsString() { return ex; };

  /**
   * Returns a {@link List} array of of Metathesaurus strings (EX)
   * which are examples of this semantic type.
   * @return an {@link List} of Metathesaurus examples
   */
  public List getExamplesAsList() {
    StringTokenizer st = new StringTokenizer(getEx(),";");
    List x = new ArrayList();
    while (st.hasMoreTokens())
      x.add(st.nextToken().trim());
    return x;
  }

  /**
   * Returns the usage note (UN).
   * This is shorthand for {@link #getUsageNote()}.
   * @return the usage note
   */
  public String getUn() { return getUsageNote(); };

  /**
   * Returns the usage note (UN).
   * @return the usage note
   */
  public String getUsageNote() { return un; };


  /**
   * Indicates whether or not this semantic type or its descendents allow
   * the non human flag.
   * @return <code>true</code> if allowed, <code>false</code> if not
   */
  public boolean allowsNonHumanFlag() { return "Y".equals(nh); };

  /**
   * Returns the non human flag value (NH).
   * @return the non human flag value
   */
  public String getNH() { return nh; };

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getStn().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link Source} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof HierarchicalSemanticType)) return false;
    return this == o;
  }

  /**
   * Compares this {@link HierarchicalSemanticType} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof HierarchicalSemanticType)
      return getStn().compareTo(
        ((HierarchicalSemanticType)o).getStn());
    else
      return 0;
  }

}
