/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       Language.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents information about a language. This class ensures that each
 * unique language is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class Language extends AbstractExpandedForm {

  // Map by abbreviation to the actual classes
  private static final Map lats = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link Language}.
   */
  public Language() {};

  /**
   * Instantiates a {@link Language} from the specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a language
   */
  private Language(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link Language}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link Language}s for those with a type of "LAT".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("LAT")&&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (lats.containsKey(meta_dsc[i].getValue())) {
	   Language lang = (Language)lats.get(meta_dsc[i].getValue());
           lang.dsc = meta_dsc[i].getExplain();
        } else {
          lats.put(meta_dsc[i].getValue(), new Language(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link Language}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
   public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link Language} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param lat the language abbreviation to look up
   * @return the {@link Language} for the specified abbreviation
   */
   public static Language getLanguage(String lat) {
    return (Language)lats.get(lat);
  }

  /**
   * Return all {@link Language}s.
   * @return an array of {@link Language}
   */
   public static Language[] getLanguages() {
    return (Language[])lats.values().toArray(new Language[0]);
  }

  /**
   * Returns all {@link Language}s as a {@link List}.
   * @return all {@link Language}s as a {@link List}
   */
  public static List getLanguagesAsList() {
    return new ArrayList(lats.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}