/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       RelationshipLabel.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Represents a relationship label value. This class ensures that each unique
 * relationship label is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class RelationshipLabel extends AbstractExpandedForm  {

  // Map by abbreviation to the actual classes
  private static final Map rels = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link RelationshipLabel}.
   */
  public RelationshipLabel() {};

  /**
   * Instantiates a {@link RelationshipLabel} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a
   * relationship label
   */
  private RelationshipLabel(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link RelationshipLabel}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link RelationshipLabel}s for those with a type of "REL".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("REL")&&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (rels.containsKey(meta_dsc[i].getValue())) {
	   RelationshipLabel rel_type = (RelationshipLabel)rels.get(meta_dsc[i].getValue());
           rel_type.dsc = meta_dsc[i].getExplain();
        } else {
          rels.put(meta_dsc[i].getValue(), new RelationshipLabel(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link RelationshipLabel}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link RelationshipLabel} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param rel the relationship label abbreviation to look up
   * @return the {@link RelationshipLabel} for the specified abbreviation
   */
  public static RelationshipLabel getRelationshipLabel(String rel) {
    return (RelationshipLabel)rels.get(rel);
  }

  /**
   * Return all {@link RelationshipLabel}s.
   * @return an array of {@link RelationshipLabel}s
   */
  public static RelationshipLabel[] getRelationshipLabels() {
    return (RelationshipLabel[])rels.values().toArray(new RelationshipLabel[0]);
  }

  /**
   * Return all {@link RelationshipLabel}s as a {@link List}.
   * @return all {@link RelationshipLabel}s as a {@link List}
   */
  public static List getRelationshipLabelsAsList() {
    ArrayList rel_list = new ArrayList();
    Iterator iter = rels.values().iterator();
    while (iter.hasNext()) {
      rel_list.add(iter.next());
    }
    return rel_list;
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}