/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   ReleaseMetadata.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.io.Reader;
import java.util.Date;
import java.util.List;

/**
 * Represents information about a release of the Metathesaurus.
 *
 * Corresponds to release.dat, plus some data in MRSAB, MRRANK, MRDOC, MRCUI, MRAUI.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface ReleaseMetadata extends Comparable {

  /**
   * Returns the release version.
   * @return tbe release version
   */
  public String getReleaseVersion();

  /**
   * Sets the release version.
   * @param rv tbe release version
   */
  public void setReleaseVersion(String rv);

  /**
   * Returns the release {@link Date}.
   * @return tbe release {@link Date}.
   */
  public Date getReleaseDate();

  /**
   * Sets the release {@link Date}.
   * @param rd tbe release {@link Date}
   */
  public void setReleaseDate(Date rd);

  /**
   * Returns the release description.
   * @return tbe relase description
   */
  public String getReleaseDescription();

  /**
   * Sets the release description.
   * @param rd tbe release description
   */
  public void setReleaseDescription(String rd);

  /**
   * Returns {@link SourceMetaDescription}s in the current subset.
   * @return an array of all {@link SourceMetaDescription} in the current subset
   */
  public SourceMetaDescription[] getSourceMetadata();

  /**
   * Returns {@link SourceMetaDescription} in the current subset as a {@link List}.
   * @return {@link SourceMetaDescription} in the current subset as a {@link List}
   */
  public List getSourceMetadataAsList();

  /**
   * Sets {@link SourceMetaDescription} in the current subset.
   * @param sabs the {@link SourceMetaDescription} in the current subset
   */
  public void setSourceMetadata(SourceMetaDescription[] sabs);

  /**
   * Sets {@link SourceMetaDescription}s in the current subset.
   * @param sabs a {@link List} of {@link SourceMetaDescription}s in the current subset
   */
  public void setSourceMetadata(List sabs);

  /**
   * Adds the specified {@link SourceMetaDescription}.
   * @param smd the {@link SourceMetaDescription} to be added
   */
  public void addSourceMetaDescription(SourceMetaDescription smd);

  /**
   * Removes the specified {@link SourceMetaDescription}.
   * @param smd the {@link SourceMetaDescription} to be removed
   */
  public void removeSourceMetaDescription(SourceMetaDescription smd);

  /**
   * Removes all {@link SourceMetaDescription}s.
   */
  public void clearSourceMetaDescription();

  /**
   * Returns all {@link ConceptMapping}s.
   * @return an array of all {@link ConceptMapping}s
   */
  public ConceptMapping[] getConceptMappings();

  /**
   * Returns all {@link ConceptMapping}s as a {@link List}.
   * @return all {@link ConceptMapping}s
   */
  public List getConceptMappingsAsList();

  /**
   * Sets the {@link ConceptMapping}s.
   * @param cm the {@link ConceptMapping}s
   */
  public void setConceptMappings(ConceptMapping[] cm);

  /**
   * Sets the {@link ConceptMapping}s.
   * @param cm a {@link List} of the {@link ConceptMapping}s
   */
  public void setConceptMappings(List cm);

  /**
   * Adds the specified {@link ConceptMapping}.
   * @param cm the {@link ConceptMapping} to be added
   */
  public void addConceptMapping(ConceptMapping cm);

  /**
   * Removes the specified {@link ConceptMapping}.
   * @param cm the {@link ConceptMapping} to be removed
   */
  public void removeConceptMapping(ConceptMapping cm);

  /**
   * Removes all {@link ConceptMapping}s.
   */
  public void clearConceptMappings();

  /**
   * List of SUBX cuis.
   * @return array of SUBX cuis
   */
  public String[] getCUIsRemovedFromSubset();

  /**
   * Add SUBX cui to list of cuis removed from subset.
   * @param cui {@link String}
   */
  public void addCUIRemovedFromSubset(String cui);

  /**
   * Returns all {@link AtomMapping}s.
   * @return an array of all {@link AtomMapping}s
   */
  public AtomMapping[] getAtomMappings();

  /**
   * Returns all {@link AtomMapping}s as a {@link List}.
   * @return all {@link AtomMapping}s
   */
  public List getAtomMappingsAsList();

  /**
   * Sets the {@link AtomMapping}s.
   * @param cm the {@link AtomMapping}s
   */
  public void setAtomMappings(AtomMapping[] cm);

  /**
   * Sets the {@link AtomMapping}s.
   * @param cm a {@link List} of the {@link AtomMapping}s
   */
  public void setAtomMappings(List cm);

  /**
   * Adds the specified {@link AtomMapping}.
   * @param cm the {@link AtomMapping} to be added
   */
  public void addAtomMapping(AtomMapping cm);

  /**
   * Removes the specified {@link AtomMapping}.
   * @param cm the {@link AtomMapping} to be removed
   */
  public void removeAtomMapping(AtomMapping cm);

  /**
   * Removes all {@link AtomMapping}s.
   */
  public void clearAtomMappings();

  /**
   * Returns all {@link SourceTermType}s in rank order.
   * @return an array of all {@link SourceTermType}s in rank order
   */
  public SourceTermType[] getSourceTermTypePrecedence();

  /**
   * Returns all {@link SourceTermType}s in rank order as a {@link List}.
   * @return all {@link SourceTermType}s in rank order as a {@link List}
   */
  public List getSourceTermTypePrecedenceAsList();

  /**
   * Sets all {@link SourceTermType}s in rank order.
   * @param sab_ttys an array of all {@link SourceTermType}s in rank order
   */
  public void setSourceTermTypePrecedence(SourceTermType[] sab_ttys);

  /**
   * Sets all {@link SourceTermType}s in rank order.
   * @param sab_ttys a {@link List} of all {@link SourceTermType}s in rank order
   */
  public void setSourceTermTypePrecedence(List sab_ttys);


  /**
   * Returns {@link MetaDescription}s in the current subset.
   * @return an array of all {@link MetaDescription}s in the current subset
   */
  public MetaDescription[] getMetadata();

  /**
   * Returns {@link MetaDescription} in the current subset as a {@link List}.
   * @return {@link MetaDescription} in the current subset as a {@link List}
   */
  public List getMetadataAsList();

  /**
   * Sets {@link MetaDescription} in the current subset.
   * @param docs the {@link MetaDescription} in the current subset
   */
  public void setMetadata(MetaDescription[] docs);

  /**
   * Sets {@link MetaDescription}s in the current subset.
   * @param docs a {@link List} of {@link MetaDescription}s in the current subset
   */
  public void setMetadata(List docs);

  /**
   * Adds the specified {@link MetaDescription}.
   * @param md the {@link MetaDescription} to be added
   */
  public void addMetadata(MetaDescription md);

  /**
   * Removes the specified {@link MetaDescription}.
   * @param md the {@link MetaDescription} to be removed
   */
  public void removeMetadata(MetaDescription md);

  /**
   * Removes all {@link MetaDescription}s.
   */
  public void clearMetadata();

  /**
   * Returns a {@link String}[] of all the file names in the CHANGE directory.
   * @return array of all the file names in the CHANGE directory
   */
  public String[] getChangeFileNames();

  /**
   * Returns the {@link Reader} for the given CHANGE file.
   * @param file_name CHANGE file
   * @return {@link Reader} for the given CHANGE file
   */
  public Reader getChangeFile(String file_name);

  /**
   * Returns a {@link String}[] of all the ancillary file names.
   * @return array of all the ancillary file names
   */
  public String[] getAncillaryFileNames();

  /**
   * Returns a {@link Reader} over the specified ancillary file.
   * @param file_name ancillary file
   * @return a {@link Reader} over the specified ancillary file
   */
  public Reader getAncillaryFile(String file_name);

  /**
   * Adds change file to a list of change files.
   * @param file_name name of change file
   * @param reader a {@link Reader} over the file
   */
  public void addChangeFile(String file_name, Reader reader);

  /**
   * Adds ancillary file to a list of ancillary files.
   * @param file_name name of ancillary file
   * @param reader a {@link Reader} over the file
   */
  public void addAncillaryFile(String file_name, Reader reader);

  /**
   * Returns a {@link String} key which can be used by equals, compareTo, and
   * hashCode methods.
   *
   * @return a {@link String} composed from key {@link ReleaseMetadata} fields.
   */
  public String getKeyFields();

}
