/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   SourceElement.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

/**
 * Represents information about a data element
 * asserted by a {@link Source}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface SourceElement {

  /**
   * Returns the source for this {@link SourceElement}.
   * @return the {@link Source}
   */
  public Source getSource();


  /**
   * Sets the source for this {@link SourceElement}.
   * @param source {@link Source}
   */
  public void setSource(Source source);

  /**
   * Sorts the sources.
   */
  public final static class BySourceComparator implements Comparator {
    /**
     * Instantiates a comparator.
     */
    public BySourceComparator() { }

    //
    // Implementation of Comparator interface
    //
    /**
     * Compares the specified field of each {@link String} object.
     * @param o1 the first {@link Object} to be compared.
     * @param o2 the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a positive integer
     *    as the first argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null ||
          !(o1 instanceof SourceElement) || !(o2 instanceof SourceElement))
         return 0;
      Source s1 = ((SourceElement) o1).getSource();
      Source s2 = ((SourceElement) o2).getSource();

      String source1 = s1.getRsab();
      String source2 = s2.getRsab();

      return source1.compareTo(source2);
    }
  }

  public final static class BySourceRestrictor implements Restrictor {
    private List source_list = new ArrayList();

    /**
     * Constructor
     * Keep only sources given as parameter.
     * @param sources {@link Source}[]
     */
    public BySourceRestrictor(Source[] sources) {
      for (int i=0; i<sources.length; i++) {
        source_list.add(sources[i]);
      }
    }

    /**
     * Compares the {@link Source} of each {@link SourceElement} object.
     * @param o1 the first {@link Object} to be compared.
     * @param o2 the second {@link Object} to be compared.
     * @return An <code>int</code> which might be negative integer, zero, or a positive integer
     *    as the first argument is less than, equal to, or greater than the second.
     */
    public int compare(Object o1, Object o2) {
      if (o1 == null || o2 == null ||
          !(o1 instanceof SourceElement) || !(o2 instanceof SourceElement))
         return 0;
      Source s1 = ((SourceElement) o1).getSource();
      Source s2 = ((SourceElement) o2).getSource();

      String source1 = s1.getRsab();
      String source2 = s2.getRsab();

      return source1.compareTo(source2);
    }

    /**
     * Returns whether or not the given {@link SourceElement} should be
     * kept or not
     * @param o {@link Object}
     * @return <code>boolean</code>
     */
    public boolean keep(Object o) {
      if (o instanceof SourceElement)
        return source_list.contains(((SourceElement)o).getSource());
      return true;
    }
  }
}