/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Source.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.Date;
import java.util.List;

/**
 * Represents information about a UMLS source.
 *
 * All of the data in a row of MRSAB is represented here.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface SourceMetaDescription extends Comparable {

  /**
   * Returns the concept identifier for the versioned SRC concept (VCUI).
   * This is shorthand for {@link #getVersionedConceptIdentifier()}.
   * @return the concept identifier for the versioned SRC concept
   */
  public String getVcui();

  /**
   * Returns the concept identifier for the versioned SRC concept (VCUI).
   * @return the concept identifier for the versioned SRC concept
   */
  public String getVersionedConceptIdentifier();

  /**
   * Returns the concept identifier for the root SRC concept (VCUI).
   * This is shorthand for {@link #getRootConceptIdentifier()}.
   * @return the concept identifier for the root SRC concept
   */
  public String getRcui();

  /**
   * Returns the concept identifier for the root SRC concept (VCUI).
   * @return the concept identifier for the root SRC concept
   */
  public String getRootConceptIdentifier();

  /**
   * Returns the versioned source abbreviation (VSAB).
   * This is shorthand for (@ link #getVersionedSourceAbbreviation()}.
   * @return the versioned source abbreviation
   */
  public String getVsab();

  /**
   * Returns the versioned source abbreviation (VSAB).
   * @return the versioned source abbreviation
   */
  public String getVersionedSourceAbbreviation();

  /**
   * Returns the root source abbreviation (RSAB).
   * This is shorthand for {@link #getRootSourceAbbreviation()}.
   * @return the root source abbreviation
   */
  public String getRsab();

  /**
   * Returns the root source abbreviation (RSAB).
   * @return the root source abbreviation
   */
  public String getRootSourceAbbreviation();

  /**
   * Returns the source official name (SON).
   * This is shorthand for {@link #getOfficialName()}.
   * @return the source official name
   */
  public String getSon();

  /**
   * Returns the source official name (SON).
   * @return the source official name
   */
  public String getOfficialName();

  /**
   * Returns the source short name (SSN).
   * This is shorthand for {@link #getShortName()}.
   * @return the source short name
   */
  public String getSsn();

  /**
   * Returns the source short name (SSN).
   * @return the source short name
   */
  public String getShortName();

  /**
   * Returns the source citation (SCIT).
   * This is shorthand for {@link #getCitation()}.
   * @return the source citation
   */
  public String getScit();

  /**
   * Returns the source citation (SCIT).
   * @return the source citation
   */
  public String getCitation();

  /**
   * Returns the source family (SF).
   * This is shorthand for {@link #getSourceFamily()}.
   * @return the source family
   */
  public String getSf();

  /**
   * Returns the source family (SF).
   * @return the source family
   */
  public String getSourceFamily();

  /**
   * Returns source version (SVER).
   * This is shorthand for {@link #getVersion()}.
   * @return the source version
   */
  public String getSver();

  /**
   * Returns source version (SVER).
   * @return the source version
   */
  public String getVersion();

  /**
   * Returns the date when this source becomes valid in the Metathesaurus (MSTART).
   * This is shorthand for {@link #getValidStartDate()}.
   * @return the date when this source becomes valid in the Metathesaurus
   */
  public Date getMstart();

  /**
   * Returns the date when this source becomes valid in the Metathesaurus (MSTART).
   * @return the date when this source becomes valid in the Metathesaurus
   */
  public Date getValidStartDate();

  /**
   * Returns the date when this source is no longer valid in the Metathesaurus (MEND).
   * This is shorthand for {@link #getValidEndDate()}.
   * @return the date when this source is no longer valid in the Metathesaurus
   */
  public Date getMend();

  /**
   * Returns the date when this source is no longer valid in the Metathesaurus (MEND).
   * @return the date when this source is no longer valid in the Metathesaurus
   */
  public Date getValidEndDate();

  /**
   * Returns the version of the Metathesaurus in which this source
   * first appeared (IMETA).
   * @return the version of the Metathesaurus in which this source
   * first appeared
   */
  public String getInsertMetaVersion();

  /**
   * Returns the version of the Metathesaurus in which this source
   * first appeared (IMETA).
   * This is shorthand for {@link #getInsertMetaVersion()}
   * @return the version of the Metathesaurus in which this source
   * first appeared
   */
  public String getImeta();

  /**
   * Returns the first version of the Metathesaurus in which this source
   * no longer (RMETA).
   * This is shorthand for {@link #getRemoveMetaVersion()}
   * @return the first version of the Metathesaurus in which this source
   * no longer
   */
  public String getRmeta();

  /**
   * Returns the first version of the Metathesaurus in which this source
   * no longer (RMETA).
   * @return the first version of the Metathesaurus in which this source
   * no longer
   */
  public String getRemoveMetaVersion();

  /**
   * Returns the license contact information (SLC).
   * This is shorthand for {@link #getLicenseContact()}.
   * @return the license contact information
   */
  public String getSlc();

  /**
   * Returns the license contact information (SLC).
   * @return the license contact information
   */
  public String getLicenseContact();

  /**
   * Returns the content contact information (SCC).
   * This is shorthand for {@link #getContentContact()}.
   * @return the content contact information
   */
  public String getContentContact();

  /**
   * Returns the content contact information (SCC).
   * @return the content contact information
   */
  public String getScc();

  /**
   * Returns the restriction level (SRL).
   * This is shorthand for {@link #getRestrictionLevel()}.
   * @return the restriction level
   */
  public String getSrl();

  /**
   * Returns the restriction level (SRL).
   * @return the restriction level
   */
  public String getRestrictionLevel();

  /**
   * Returns the frequency of terms (TFR).
   * This is shorthand for {@link #getTermFrequency()}
   * @return the frequency of terms
   */
  public int getTfr();

  /**
   * Returns the frequency of terms (TFR).
   * @return the frequency of terms
   */
  public int getTermFrequency();

  /**
   * Returns the frequency of concepts (CFR).
   * This is shorthand for {@link #getConceptFrequency()}
   * @return the frequency of concepts
   */
  public int getCfr();

  /**
   * Returns the frequency of concepts (CFR).
   * @return the frequency of concepts
   */
  public int getConceptFrequency();

  /**
   * Returns the context type (CXTY).
   * This is shorthand for {@link #getContextType()}.
   * @return the context type
   */
  public String getCxty();

  /**
   * Returns the context type (CXTY).
   * @return the context type
   */
  public String getContextType();

  /**
   * Returns the {@link TermType}s.
   * @return an array of {@link TermType}s
   */
  public TermType[] getTermTypes();

  /**
   * Returns the {@link TermType}s as a {@link List}.
   * @return the {@link TermType}s as a {@link List}
   */
  public List getTermTypesAsList();

  /**
   * Returns the term types as a comma (,) separated
   * list of term type abbreviations (TTYL).
   * This is shorthand for {@link #getTermTypesAsString()}.
   * @return the term type values
   */
  public String getTtyl();

  /**
   * Returns the term types as a comma (,) separated
   * list of term type abbreviations (TTYL).
   * @return the term type values
   */
  public String getTermTypesAsString();

  /**
   * Get the {@link AttributeName}s for this {@link Source}.
   * Returns the {@link AttributeName}s.
   * @return an array of {@link AttributeName}s
   */
  public AttributeName[] getAttributeNames();

  /**
   * Returns the {@link AttributeName}s as a {@link List}.
   * @return the {@link AttributeName}s as a {@link List}
   */
  public List getAttributeNamesAsList();

  /**
   * Returns the attribute names as a comma (,) separated
   * list of attribute name abbreviations (ATNL).
   * This is shorthand for {@link #getAttributeNamesAsString()}.
   * @return the attribute name values
   */
  public String getAtnl();

  /**
   * Returns the attribute names as a comma (,) separated
   * list of attribute name abbreviations (ATNL).
   * @return the attribute name values
   */
  public String getAttributeNamesAsString();

  /**
   * Returns the {@link Language} (LAT).
   * This is shorthand for {@link #getLanguage()}.
   * @return the {@link Language}
   */
  public Language getLat();

  /**
   * Returns the {@link Language} (LAT).
   * @return the {@link Language}
   */
  public Language getLanguage();

  /**
   * Returns the ISO character encoding (CENC).
   * This is shorthand for {@link #getCharacterEncoding()}.
   * @return the character encoding
   */
  public String getCenc();

  /**
   * Returns the ISO character encoding (CENC).
   * @return the character encoding
   */
  public String getCharacterEncoding();

  /**
   * Returns the "is current version" value (CURVER).
   * @return the "is current version" value
   */
  public String getCurver();

  /**
   * Indicates whether or not this represents the current version
   * of the source.
   * @return <code>true</code> if it does, <code>false</code> otherwise.
   */
  public boolean isCurrentVersion();

  /**
   * Returns the "in current subset" value (SABIN).
   * @return the "in current subset" value
   */
  public String getSabin();

  /**
   * Indicates whether or not this source is represented
   * in the current subset.
   * @return <code>true</code> if it is, <code>false</code> otherwise.
   */
  public boolean inCurrentSubset();

  /**
   * Sets whether or not this represents the current version
   * of the source.
   * @param sabin <code>boolean</code>
   */
  public void setInCurrentSubset(boolean sabin);

  /**
   * Sets the "in current subset" value (SABIN).
   * @param sabin "in current subset" value
   */
  public void setSabin(String sabin);

}
