/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   StringInfo.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.List;

/**
 * Represents string information for a {@link Term}.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link StringInfo}
 * <p><center><img src="StringInfo.gif"></center>
 * <p>
 * @see Term
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface StringInfo extends Comparable, TermElement {

  /**
   * Returns the {@link StringType} (STT).
   * @return the {@link StringType}
   */
  public StringType getStringType();

  /**
   * Returns the {@link StringType} (STT).
   * This is shorthand for {@link #getStringType()}.
   * @return the {@link StringType}
   */
  public StringType getStt();

  /**
   * Sets the {@link StringType} (STT).
   * @param stt the STT
   */
  public void setStringType(StringType stt);

  /**
   * Sets the {@link StringType} (STT).
   * This is shorthand for {@link #setStringType(StringType)}.
   * @param stt the STT
   */
  public void setStt(StringType stt);

  /**
   * Return the string identifier (SUI).
   * This is shorthand for {@link #getStringIdentifier()}.
   * @return the SUI
   */
  public String getSui();

  /**
   * Sets the string identifier (SUI).
   * This is shorthand for {@link #setStringIdentifier(String)}.
   * @param sui the SUI
   */
  public void setSui(String sui);

  /**
   * Return the string identifier (SUI).
   * @return the SUI
   */
  public String getStringIdentifier();

  /**
   * Sets the string identifier (SUI).
   * @param sui the SUI
   */
  public void setStringIdentifier(String sui);

  /**
   * Sets the name (STR).
   * @return the name
   */
  public String getName();

  /**
   * Returns the name (STR).
   * This is shorthand for {@link #getName()}.
   * @return the name
   */
  public String getStr();

  /**
   * Sets the name (STR).
   * @param str the name
   */
  public void setName(String str);

  /**
   * Sets the name (STR).
   * This is shorthand for {@link #setName(String)}.
   * @param str the name
   */
  public void setStr(String str);

  /**
   * Returns all {@link Atom}s.
   * @return an array of {@link Atom}s
   */
  public Atom[] getAtoms();

  /**
   * Returns all {@link Atom}s as a {@link List}.
   * @return all {@link Atom}s as a {@link List}
   */
  public List getAtomsAsList();

  /**
   * Adds the specified {@link Atom}.
   * @param str_source the {@link Atom} to add
   */
  public void addAtom(Atom str_source);

  /**
   * Removes the specified {@link Atom}.
   * @param str_source the {@link Atom} to remove
   */
  public void removeAtom(Atom str_source);

  /**
   * Removes all {@link Atom}s.
   */
  public void clearAtoms();

  /**
   * Replaces the {@link Atom}s with those in the specified array.
   * @param str_source an array of {@link Atom}s
   */
  public void setAtoms(Atom [] str_source);

  /**
   * Replaces the {@link Atom}s with those in the specified {@link List}
   * @param str_source a {@link List} of {@link Atom}s
   */
  public void setAtoms(List str_source);

  /**
   * Indicates whether or not this atom has the preferred string identifier
   * of the {@link Term} it is part of.  In other words, does this {@link Atom} have
   * the same SUI as the preferred {@link Atom} of the {@link Term}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredStringIdentifier();

  /**
   * Indicates whether or not this atom has the preferred string identifier
   * of the {@link Term} it is part of.  This is shorthand for
   * {@link #hasPreferredStringIdentifier()}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredSui();

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link StringInfo} fields
   */
  public String getKeyFields();
}