/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       StringType.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Represents an string type.  This class ensures that each unique
 * string type is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */

 public final class StringType extends AbstractExpandedForm  {

   // Map by abbreviation to the actual classes
  private static final Map stts = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link StringType}.
   */
  public StringType() {};

  /**
   * Instantiates an {@link StringType} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a string type
   */
  private StringType(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link StringType}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link StringType}s for those with a type of "STT".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("STT")&&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (stts.containsKey(meta_dsc[i].getValue())) {
	   StringType str_type = (StringType)stts.get(meta_dsc[i].getValue());
           str_type.dsc = meta_dsc[i].getExplain();
        } else {
          stts.put(meta_dsc[i].getValue(), new StringType(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link StringType}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link StringType} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param stt the string type abbreviation to look up
   * @return the {@link StringType} for the specified abbreviation
   */
  public static StringType getStringType(String stt) {
    return (StringType)stts.get(stt);
  }

  /**
   * Returns all {@link StringType}s.
   * @return an array of {@link StringType}
   */
  public static StringType[] getStringTypes() {
    return (StringType[])stts.values().toArray(new StringType[0]);
  }

  /**
   * Returns all {@link StringType}s as a {@link List}.
   * @return all {@link StringType}s as a {@link List}
   */
  public static List getStringTypesAsList() {
    return new ArrayList(stts.values());
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}