/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       Suppressible.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Represents a suppressible value. This class ensures that each unique
 * suppressible value is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class Suppressible extends AbstractExpandedForm  {

  // Map by abbreviation to the actual classes
  private static final Map suppressibles = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link Suppressible}.
   */
  public Suppressible() {};

  /**
   * Instantiates a {@link Suppressible} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a
   * suppressible value
   */
  private Suppressible(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getValue();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link Suppressible}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link Suppressible}s for those with a type of "SUPPRESS".
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      if (meta_dsc[i].getKey().equals("SUPPRESS")&&
          meta_dsc[i].getType().equals("expanded_form")) {
	if (suppressibles.containsKey(meta_dsc[i].getValue())) {
	   Suppressible suppress_type = (Suppressible)suppressibles.get(meta_dsc[i].getValue());
           suppress_type.dsc = meta_dsc[i].getExplain();
        } else {
          suppressibles.put(meta_dsc[i].getValue(), new Suppressible(meta_dsc[i]));
        }
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link Suppressible}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link Suppressible} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param sup_abrev the suppressible abbreviation to look up
   * @return the {@link Suppressible} for the specified abbreviation
   */
  public static Suppressible getSuppressible(String sup_abrev) {
    return (Suppressible)suppressibles.get(sup_abrev);
  }

  /**
   * Return all {@link Suppressible}s.
   * @return an array of {@link Suppressible}s
   */
  public static Suppressible[] getSuppressibles() {
    return (Suppressible[])suppressibles.values().toArray(new Suppressible[0]);
  }

  /**
   * Return all {@link Suppressible}s as a {@link List}.
   * @return all {@link Suppressible}s as a {@link List}
   */
  public static List getSuppressiblesAsList() {
    ArrayList suppress_list = new ArrayList();
    Iterator iter = suppressibles.values().iterator();
    while (iter.hasNext()) {
      suppress_list.add(iter.next());
    }
    return suppress_list;
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}