/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Interface:   Term.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.List;

/**
 * Represents {@link Concept} term information.  This is a higher level
 * view of a {@link Concept}'s names than {@link Atom}.
 * <p>
 * Following is a UML Class diagram showing the neighborhood around {@link Term}
 * <p><center><img src="Term.gif"></center>
 * <p>
 * @author Brian Carlsen, Deborah Shapiro
 */
public interface Term
 extends Comparable, ConceptElement {

  /**
   * Returns the {@link Language} (LAT).
   * @return the {@link Language}
   */
  public Language getLanguage();

  /**
   * Returns the {@link Language} (LAT).
   * This is shorthand for {@link #getLanguage()}.
   * @return {@link Language}
   */
  public Language getLat();

  /**
   * Sets the {@link Language} (LAT).
   * @param lat the {@link Language}
   */
  public void setLanguage(Language lat);

  /**
   * Sets the {@link Language} (LAT).
   * This is shorthand for {@link #setLanguage(Language)}.
   * @param lat the {@link Language}
   */
  public void setLat(Language lat);

  /**
   * Returns the {@link TermStatus} (TS).
   * @return the {@link TermStatus}
   */
  public TermStatus getTermStatus();

  /**
   * Returns the {@link TermStatus} (TS).
   * This is shorthand for {@link #getTermStatus()}.
   * @return the {@link TermStatus}
   */
  public TermStatus getTs();

  /**
   * Sets the {@link TermStatus} (TS).
   * @param ts the {@link TermStatus}
   */
  public void setTermStatus(TermStatus ts);

  /**
   * Sets the {@link TermStatus} (TS).
   * This is shorthand for {@link #setTermStatus(TermStatus)}.
   * @param ts the {@link TermStatus}
   */
  public void setTs(TermStatus ts);

  /**
   * Returns the term identifier (LUI).
   * This is shorthand for {@link #getTermIdentifier()}.
   * @return the LUI
   */
  public String getLui();

  /**
   * Sets the term identifier (LUI). This is shorthand
   * for {@link #setTermIdentifier(String)}.
   * @param lui the LUI
   */
  public void setLui(String lui);

  /**
   * Returns the term identifier (LUI).
   * @return the LUI
   */
  public String getTermIdentifier();

  /**
   * Sets the term identifier (LUI).
   * @param lui the LUI
   */
  public void setTermIdentifier(String lui);

  /**
   * Sets the name (STR).
   * @return the name
   */
  public String getName();

  /**
   * Returns the name (STR).
   * This is shorthand for {@link #getName()}.
   * @return the name
   */
  public String getStr();

  /**
   * Sets the name (STR).
   * @param str the name
   */
  public void setName(String str);

  /**
   * Sets the name (STR).
   * This is shorthand for {@link #setName(String)}.
   * @param str the name
   */
  public void setStr(String str);

  /**
   * Returns all {@link StringInfo} term variants.
   * @return an array of {@link StringInfo} objects
   */
  public StringInfo [] getTermVariants();

  /**
   * Returns all {@link StringInfo} term variants as a {@link List}.
   * @return all {@link StringInfo} term variants as a {@link List}
   */
  public List getTermVariantsAsList();

  /**
   * Adds the specified term variant.
   * @param tv a {@link StringInfo} term variant to add
   */
  public void addTermVariant(StringInfo tv);

  /**
   * Removes all term variants.
   */
  public void clearTermVariants();

  /**
   * Removes the specified term variant.
   * @param tv a {@link StringInfo} term variant to remove
   */
  public void removeTermVariant(StringInfo tv);

  /**
   * Replaces term variants with the specified {@link StringInfo} objects.
   * @param tvs an array of {@link StringInfo} objects
   */
  public void setTermVariants(StringInfo [] tvs);

  /**
   * Replaces term variants with the specified {@link List}
   * @param tvs a {@link List} of {@link StringInfo} objects
   */
  public void setTermVariants(List tvs);

  /**
   * Returns all {@link TermAttribute}s.
   * @return an array of {@link TermAttribute} objects
   */
  public TermAttribute [] getTermAttributes();

  /**
   * Returns all {@link TermAttribute}s as a {@link List}.
   * @return all {@link TermAttribute}s as a {@link List}
   */
  public List getTermAttributesAsList();

  /**
   * Adds the specified {@link TermAttribute}.
   * @param ta a {@link TermAttribute} to add
   */
  public void addTermAttribute(TermAttribute ta);

  /**
   * Removes all {@link TermAttribute}s.
   */
  public void clearTermAttributes();

  /**
   * Removes the specified {@link TermAttribute}.
   * @param ta the {@link TermAttribute} to remove
   */
  public void removeTermAttribute(TermAttribute ta);

  /**
   * Replaces the {@link TermAttribute}s with the specified {@link TermAttribute} objects.
   * @param tas an array of {@link TermAttribute} objects
   */
  public void setTermAttributes(TermAttribute [] tas);

  /**
   * Replaces {@link TermAttribute}s with the specified {@link List}
   * @param tas a {@link List} of {@link TermAttribute} objects
   */
  public void setTermAttributes(List tas);


  /**
   * Indicates whether or not this {@link Term} is the preferred {@link Term}
   * of the {@link Concept}.  In other words, does this {@link Term} have
   * the same LUI as the preferred {@link Term} of the {@link Concept}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredTermIdentifier();

  /**
   * Indicates whether or not this {@link Term} is the preferred {@link Term}
   * of the {@link Concept}.  This is shorthand for {@link
   * #hasPreferredTermIdentifier()}
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredLui();

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from key {@link Term} fields
   */
  public String getKeyFields();
}