/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta
 * Class:       TermTypeGroup.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Represents a grouping of {@link TermType}s. This class ensures that each unique
 * {@link TermType} group is represented by a singleton object.  In order for
 * this to work correctly, the class must be properly initialized and must be
 * initialized after Term Types is initialized.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public final class TermTypeGroup extends AbstractExpandedForm  {


  // Map by abbreviation to the actual classes
  private static final Map tty_groups = new HashMap();
  // ttys in this group
  private Set ttys = new HashSet();
  // Map by tty to the group(s) that tty belongs to
  private static Map tty_to_groups = new HashMap();

  // Abbreviation and description for this individual instance.
  private String abbrev;
  private String dsc;

  /**
   * Instantiates an empty {@link TermTypeGroup}.
   */
  public TermTypeGroup() {};

  /**
   * Instantiates an {@link TermTypeGroup} from the
   * specified {@link MetaDescription}.
   * @param meta_dsc a {@link MetaDescription} representing a term type
   */
  private TermTypeGroup(MetaDescription meta_dsc) {
    abbrev = meta_dsc.getExplain();
    dsc = meta_dsc.getExplain();
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link TermTypeGroup}s that exist at one time.  You may
   * pass all {@link MetaDescription}s as this method only
   * creates {@link TermTypeGroup}s for those with a type of "TTY".
   * @param meta_dsc an array of {@link MetaDescription}s
   * @throws IllegalStateException if an attempt is made to call this
   * method more than once
   */
  public static synchronized void initialize(MetaDescription[] meta_dsc) {
    for (int i=0; i<meta_dsc.length; i++) {
      TermTypeGroup tty_group = null;
      if (meta_dsc[i].getKey().equals("TTY") &&
          meta_dsc[i].getType().equals("tty_class")) {
	if (tty_groups.containsKey(meta_dsc[i].getExplain())) {
	   tty_group = (TermTypeGroup)tty_groups.get(meta_dsc[i].getExplain());
           //tty.dsc = meta_dsc[i].getExplain();
        } else {
          tty_group = new TermTypeGroup(meta_dsc[i]);
          tty_groups.put(meta_dsc[i].getExplain(), tty_group);
        }

        TermType tty = TermType.getTermType(meta_dsc[i].getValue());
        tty_group.ttys.add(tty);

        Set groups = (Set)tty_to_groups.get(tty);
        if (groups == null) {
          groups = new HashSet();
          tty_to_groups.put(tty, groups);
        }
        groups.add(tty_group);
      }
    }
  }

  /**
   * Initializes the {@link Map} used to track all of the
   * {@link TermTypeGroup}s that exist at one time.  This
   * is implemented as a call to {@link #initialize(MetaDescription[])}.
   * @param meta_dsc an array of {@link MetaDescription}s
   */
  public static synchronized void initialize(List meta_dsc) {
    initialize((MetaDescription[])meta_dsc.toArray(new MetaDescription[0]));
  }

  /**
   * Returns the {@link TermTypeGroup} for the specified abbreviation.
   * This method will return <code>null</code> if the <code>initialize</code>
   * method has not been called or if the specified abbreviation
   * was not one of those used during initialization.
   * @param tty_group the term type abbreviation to look up
   * @return the {@link TermTypeGroup} for the specified abbreviation
   */
  public static TermTypeGroup getTermTypeGroup(String tty_group) {
    return (TermTypeGroup)tty_groups.get(tty_group);
  }

  /**
   * Return all {@link TermTypeGroup}s for which the given {@link TermType}
   * is a member.
   * @param tty {@link TermType}
   * @return an array of {@link TermTypeGroup}s
   */
  public static TermTypeGroup[] getTermTypeGroups(TermType tty) {
    return (TermTypeGroup[])((Set)tty_to_groups.get(tty)).toArray(new TermTypeGroup[0]);
  }

  /**
   * Return all {@link TermTypeGroup}s for which the given {@link TermType} is
   * a member as a {@link List}.
   * @param tty {@link TermType}
   * @return all {@link TermTypeGroup}s as a {@link List}
   */
  public static List getTermTypeGroupsAsList(TermType tty) {
    return new ArrayList(((Set)tty_to_groups.get(tty)));
  }

  /**
   * Return all {@link TermTypeGroup}s.
   * @return an array of {@link TermTypeGroup}s
   */
  public static TermTypeGroup[] getTermTypeGroups() {
    return (TermTypeGroup[])tty_groups.values().toArray(new TermTypeGroup[0]);
  }

  /**
   * Return all {@link TermTypeGroup}s as a {@link List}.
   * @return all {@link TermTypeGroup}s as a {@link List}
   */
  public static List getTermTypeGroupsAsList() {
    return new ArrayList(tty_groups.values());
  }

  /**
   * Return all {@link TermType}s in this {@link TermTypeGroup}.
   * @return an array of {@link TermTypeGroup}s
   */
  public TermType[] getTermTypes() {
    return (TermType[])ttys.toArray(new TermType[0]);
  }

  /**
   * Return all {@link TermType}s in this {@link TermTypeGroup} as a {@link List}.
   * @return all {@link TermType}s as a {@link List}
   */
  public List getTermTypesAsList() {
    return new ArrayList(ttys);
  }

  /**
   * Returns the abbreviation.
   * @return the abbreviation
   */
  public String getAbbreviation() {
    return abbrev;
  }

  /**
   * Returns the description
   * @return the description
   */
  public String getDescription() {
    return dsc;
  }

}
