/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       AtomCoocurrenceImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomCoocurrence;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Coocurrence}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class AtomCoocurrenceImpl extends CoocurrenceImpl implements AtomCoocurrence  {

  //
  // Fields
  //
  private Atom atom;
  private Atom related_atom;

  /**
   * Instantiates an empty {@link AtomCoocurrence}.
   */
  public AtomCoocurrenceImpl() {};

  /**
   * Instantiates an {@link AtomCoocurrence} from the specified {@link AtomCoocurrence}.
   * @param co {@link AtomCoocurrence} to be copied
   */
  public AtomCoocurrenceImpl(AtomCoocurrence co) {
    super(co);
    setAtom(co.getAtom());
    setRelatedAtom(co.getRelatedAtom());
  }

  /**
   * Returns the related atom involved.
   * @return the related atom involved
   */
  public Atom getRelatedAtom() {
    return related_atom;
  }

  /**
   * Sets the second atom involved.
   * @param related_atom the related concept involved
   */
  public void setRelatedAtom(Atom related_atom) {
    this.related_atom = related_atom;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link Atom} this {@link AtomCoocurrence} is connected to.
   * @return the {@link Atom} this {@link AtomCoocurrence} is connected to
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the {@link Atom} this {@link AtomCoocurrence} is connected to.
   * @param atom  the {@link Atom} this {@link AtomCoocurrence} is connected to
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, cui2, aui2,
   * versioned source, and cot.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      final StringBuffer sb = new StringBuffer(100);
      sb.append(getConcept() != null ? getConcept().getCui() : "")
        .append(getAtom() != null ? getAtom().getAui() : "")
        .append(getRelatedConcept() != null ? getRelatedConcept().getCui() : "")
        .append(getRelatedAtom() != null ? getRelatedAtom().getAui() : "")
        .append(getSource() != null ? getSource().getVersionedSourceAbbreviation() : "")
        .append(getCot() != null ? getCot().getAbbreviation() : "");
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link AtomCoocurrence} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof AtomCoocurrence)) return false;
    return getKeyFields().equals(((AtomCoocurrence)o).getKeyFields());
  }

  /**
   * Compares this {@link AtomCoocurrence} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof AtomCoocurrence)
      return getKeyFields().compareTo(((AtomCoocurrence)o).getKeyFields());
    else
      return 0;
  }

}
