/**************************************************************************
 * Package:  gov.nih.nlm.umls.meta.impl
 * Class:    AtomMappingImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.AtomMapping;
import gov.nih.nlm.umls.meta.RelationshipLabel;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.AtomMapping}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class AtomMappingImpl implements AtomMapping {

  //
  // Fields
  //
  private String aui1;
  private String aui2;
  private String cui1;
  private String cui2;
  private String ver;
  private RelationshipLabel rel;
  private AdditionalRelationshipLabel rela;
  private String mapin;
  private String reason;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link AtomMapping}.
   */
  public AtomMappingImpl() {};

  /**
   * Instantiates a {@link AtomMapping} from the specified {@link AtomMapping}.
   * @param am {@link AtomMapping} to be copied
   */
  public AtomMappingImpl(AtomMapping am) {
    setAui1(am.getAui1());
    setAui2(am.getAui2());
    setCui2(am.getCui2());
    setVer(am.getVer());
    setRel(am.getRel());
    setRela(am.getRela());
    setMapin(am.getMapin());
    setMapReason(am.getMapReason());
  }

  /**
   * Returns the concept identifier for the first atom (CUI1).
   * @return the concept identifier for the first atom
   */
  public String getConceptIdentifier() {
    return cui1;
  }

  /**
   * Returns the concept identifier for the first atom (CUI1).
   * Shorthand for {@link #getConceptIdentifier()}.
   * @return the result of <code>getConceptIdentifier()</code>
   */
  public String getCui1() {
    return getConceptIdentifier();
  }

  /**
   * Sets the concept identifier for the first atom (CUI1).
   * @param cui1 the concept identifier for the first atom
   */
  public void setConceptIdentifier(String cui1) {
    this.cui1 = cui1;
    key_fields_flag = false;
  }

  /**
   * Sets the concept identifier for the first atom (CUI1).
   * Shorthand for {@link #setConceptIdentifier(String)}.
   * @param cui1 the concept identifier for the first atom
   */
  public void setCui1(String cui1) {
    setConceptIdentifier(cui1);
  }

  /**
   * Returns the atom identifier for the first atom (AUI1).
   * @return the atom identifier for the first atom
   */
  public String getAtomIdentifier() {
    return aui1;
  }

  /**
   * Shorthand for {@link #getAtomIdentifier()}.
   * @return the result of <code>getAtomIdentifier()</code>
   */
  public String getAui1() {
    return getAtomIdentifier();
  }

  /**
   * Sets the atom identifier for the first atom (AUI1).
   * @param aui1 the atom identifier for the first atom
   */
  public void setAtomIdentifier(String aui1) {
    this.aui1 = aui1;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAtomIdentifier(String)}.
   * @param aui1 the atom identifier for the first atom
   */
  public void setAui1(String aui1) {
    setAtomIdentifier(aui1);
  }

 /**
  * Returns the release version in which the indicated change took place (VER).
  * @return the release version in which the indicated change took place
  */
  public String getReleaseVersion() {
    return ver;
  }

  /**
   * Returns the release version in which the indicated change took place (VER).
   * This method is shortand for {@link #getReleaseVersion()}.
   * @return the release version in which the indicated change took place
   */
  public String getVer() {
    return getReleaseVersion();
  }

  /**
   * Sets the release version in which the indicated change took place (VER).
   * @param ver the release version in which the indicated change took place
   */
  public void setReleaseVersion(String ver) {
    this.ver = ver;
    key_fields_flag = false;
  }

  /**
   * Sets the release version in which the indicated change took place (VER).
   * This method is shorthand for (@link #setReleaseVersion(String)}.
   * @param ver the release version in which the indicated change took place
   */
  public void setVer(String ver) {
    setReleaseVersion(ver);
  }

  /**
   * Returns the relationship between the two concepts (REL).
   * @return the relationship between the two concepts
   */
  public RelationshipLabel getRelationshipLabel() {
    return rel;
  }

  /**
   * Shorthand for {@link #getRelationshipLabel()}.
   * @return the result of <code>getRelationshipLabel()</code>
   */
  public RelationshipLabel getRel() {
    return getRelationshipLabel();
  }

  /**
   * Sets the relationship between the two concepts (REL).
   * @param rel the relationship between the two concepts
   */
  public void setRelationshipLabel(RelationshipLabel rel) {
    this.rel = rel;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setRelationshipLabel(RelationshipLabel)}.
   * @param rel the relationship between the two concepts
   */
  public void setRel(RelationshipLabel rel) {
    setRelationshipLabel(rel);
  }

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel() {
    return rela;
  }

  /**
   * Shorthand for {@link #getAdditionalRelationshipLabel}).
   * @return the result of <code>getAdditionalRelationshipLabel()</code>
   */
  public AdditionalRelationshipLabel getRela() {
    return getAdditionalRelationshipLabel();
  }

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Returns the concept identifier for the second atom (CUI2).
   * @return the concept identifier for the second atom
   */
  public String getRelatedConceptIdentifier() {
    return cui2;
  }

  /**
   * Returns the concept identifier for the second atom (CUI2).
   * Shorthand for {@link #getRelatedConceptIdentifier()}.
   * @return the result of <code>getRelatedConceptIdentifier()</code>
   */
  public String getCui2() {
    return getRelatedConceptIdentifier();
  }

  /**
   * Sets the concept identifier for the second atom (CUI2).
   * @param cui2 the concept identifier for the second atom
   */
  public void setRelatedConceptIdentifier(String cui2) {
    this.cui2 = cui2;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setRelatedConceptIdentifier(String)}.
   * @param cui2 the concept identifier for the second atom
   */
  public void setCui2(String cui2) {
    setRelatedConceptIdentifier(cui2);
  }


  /**
   * Returns the atom identifier for the second atom (AUI2).
   * @return the atom identifier for the second atom
   */
  public String getRelatedAtomIdentifier() {
    return aui2;
  }

  /**
   * Shorthand for {@link #getRelatedAtomIdentifier()}.
   * @return the result of <code>getRelatedAtomIdentifier()</code>
   */
  public String getAui2() {
    return getRelatedAtomIdentifier();
  }

  /**
   * Sets the atom identifier for the second atom (AUI2).
   * @param aui2 the atom identifier for the second atom
   */
  public void setRelatedAtomIdentifier(String aui2) {
    this.aui2 = aui2;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setRelatedAtomIdentifier(String)}.
   * @param aui2 the atom identifier for the second atom
   */
  public void setAui2(String aui2) {
    setRelatedAtomIdentifier(aui2);
  }

  /**
   * Returns the reason for this {@link AtomMapping} (MAPREASON).
   * @return the reason for this {@link AtomMapping}
   */
  public String getMapReason() {
    return reason;
  }

  /**
   * Sets the reason for this {@link AtomMapping} (MAPREASON).
   * @param reason the reason for this {@link AtomMapping}
   */
  public void setMapReason(String reason) {
    this.reason = reason;
    key_fields_flag = false;
  }

   /**
   * Returns the "in current subset" value (MAPIN).
   * @return the "in current subset" value
   */
  public String getMapin() {
    return mapin;
  }

  /**
   * Indicates whether or not this {@link AtomMapping} is in the current subset (MAPIN).
   * @return <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public boolean inCurrentSubset() {
    return (getMapin() != null && getMapin().equals("Y"));
  }

  /**
   * Sets the "in current subset" value (MAPIN).
   * @param mapin the "in current subset" value
   */
  public void setMapin(String mapin) {
    this.mapin = mapin;
    key_fields_flag = false;
  }

  /**
   * Sets the flag indicating whether or not this {@link AtomMapping}
   * is in the current subset (MAPIN).
   * @param mapin <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public void setInCurrentSubset(boolean mapin) {
    setMapin(mapin ? "Y" : "N");
    key_fields_flag = false;
  }

  //
  // Overridden Object Methods
  //

  /**
   * Returns a hash code based on all fields.
   * @return a hash code based on all fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link AtomMapping} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal,
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof AtomMapping)) return false;
    return getKeyFields().equals(((AtomMapping)o).getKeyFields());
  }

  /**
   * Compares this {@link AtomMapping} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the Related ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof AtomMapping)
      return getKeyFields().compareTo(((AtomMapping)o).getKeyFields());
    else
      return 0;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui1, ver, rel,
   * rela, mapreason, cui2, and mapin.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(50);
      sb1.append(getAui1()).append(getCui1())
         .append(getAui2()).append(getCui2())
         .append(getRel() != null ? getRel().getAbbreviation() : "")
         .append(getRela() != null ? getRela().getAbbreviation() : "")
         .append(getVer())
         .append(getMapReason())
         .append(getMapin());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

}
