/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       AttributeContextImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AttributeContext;
import gov.nih.nlm.umls.meta.AttributeValue;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceIdentifierType;
import gov.nih.nlm.umls.meta.StringAttribute;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.AttributeContext}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class AttributeContextImpl implements AttributeContext {

  //
  // Fields
  //
  private Source source;
  private List attribute_values = new ArrayList();
  private SourceIdentifierType source_id_type;
  private StringAttribute string_attribute;
  private Atom string_source;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link AttributeContext}.
   */
  public AttributeContextImpl() {};

  /**
   * Instantiates a {@link AttributeContext} from the specified {@link AttributeContext}.
   * @param att {@link AttributeContext} to be copied
   */
  public AttributeContextImpl(AttributeContext att) {
    setAtom(att.getAtom());
    setSourceIdentifierType(att.getSourceIdentifierType());
    setStringAttribute(att.getStringAttribute());
    setSource(att.getSource());
    setAttributeValues(att.getAttributeValues());
  }



  /**
   * Returns the {@link StringAttribute} that holds this {@link AttributeContext}.
   * @return the {@link StringAttribute} that holds this {@link AttributeContext}
   */
  public StringAttribute getStringAttribute() {
    return string_attribute;
  }

  /**
   * Sets the {@link StringAttribute} that holds this {@link AttributeContext}.
   * @param sa the {@link StringAttribute} that holds this {@link AttributeContext}
   */
  public void setStringAttribute(StringAttribute sa) {
    string_attribute = sa;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link Atom} for this {@link AttributeContext}.
   * @return {@link Atom}
   */
  public Atom getAtom() {
    return string_source;
  }

  /**
   * Sets the {@link Atom} for this {@link AttributeContext}.
   * @param string_source {@link Atom}
   */
  public void setAtom(Atom string_source) {
    this.string_source = string_source;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link SourceIdentifierType} for this {@link AttributeContext}.
   * @return {@link SourceIdentifierType}
   */
  public SourceIdentifierType getSourceIdentifierType() {
    return source_id_type;
  }

  /**
   * Sets the {@link SourceIdentifierType} for this {@link AttributeContext}.
   * @param source_id_type {@link SourceIdentifierType}
   */
  public void setSourceIdentifierType(SourceIdentifierType source_id_type) {
    this.source_id_type = source_id_type;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link AttributeContext}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link AttributeContext}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns all {@link AttributeValue}s.
   * @return a array of {@link AttributeValue}s
   */
  public AttributeValue [] getAttributeValues() {
    return (AttributeValue[])attribute_values.toArray(new AttributeValue[0]);
  }

  /**
   * Returns all {@link AttributeValue}s as a {@link List}.
   * @return a {@link List} of {@link AttributeValue}s
   */
  public List getAttributeValuesAsList() {
    return new ArrayList(attribute_values);
  }

  /**
   * Adds the specified {@link AttributeValue}.
   * @param av the {@link AttributeValue} to be added
   */
  public void addAttributeValue(AttributeValue av) {
    attribute_values.add(av);
  }

  /**
   * Removes the specified {@link AttributeValue}.
   * @param av the {@link AttributeValue} to be removed
   */
  public void removeAttributeValue(AttributeValue av) {
    attribute_values.remove(av);
    //List attribute_list = getConcept().getAttributesAsList();
    // iterate through Attributes and remove those with same code &
  }

  /**
   * Removes all {@link AttributeValue}s.
   */
  public void clearAttributeValues() {
    attribute_values.clear();
  }

  /**
   * Replaces the {@link AttributeValue}s with the ones in the specified array.
   * @param avs a {@link AttributeValue}<code>[]</code>
   */
  public void setAttributeValues(AttributeValue [] avs) {
    attribute_values.clear();
    for (int i=0; i<avs.length; i++) {
      attribute_values.add(avs[i]);
    }
  }

  /**
   * Replaces the {@link AttributeValue}s with the ones in the specified {@link List}.
   * @param avs a {@link List} of {@link AttributeValue}s
   */
  public void setAttributeValues(List avs) {
    attribute_values = new ArrayList(avs);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from aui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      key_fields = (getAtom() != null) ?
          getAtom().getAui().toString() : "";
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link AttributeContext} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof AttributeContext)) return false;
    return getKeyFields().equals(((AttributeContext)o).getKeyFields());
  }

  /**
   * Compares this {@link AttributeContext} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof AttributeContext)
      return getKeyFields().compareTo(((AttributeContext)o).getKeyFields());
    else
      return 0;
  }

}
