/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       AttributeImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceIdentifierType;
import gov.nih.nlm.umls.meta.Suppressible;

/**
 * Default implementation of the {@link Attribute} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class AttributeImpl implements Attribute {

  //
  // Fields
  //
  private Concept concept;
  private String atui;
  private String satui;
  private Source source;
  private ContentView cvf;
  private String atv;
  private AttributeName atn;
  private Suppressible suppress;
  private SourceIdentifierType source_id_type;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiate an empty {@link Attribute}.
   */
  public AttributeImpl() {};

  /**
   * This is a copy constructor for {@link Attribute}.
   * @param att {@link Attribute} to be copied
   */
  public AttributeImpl(Attribute att) {
    setAtn(att.getAtn());
    setAtv(att.getAtv());
    setSuppress(att.getSuppress());
    setAtui(att.getAtui());
    setSatui(att.getSatui());
    setConcept(att.getConcept());
    setSource(att.getSource());
    setCvf(att.getCvf());
    setSourceIdentifierType(att.getSourceIdentifierType());
  }

  /**
   * Returns the {@link SourceIdentifierType} to which this element is connected.
   * @return the {@link SourceIdentifierType} to which this element is connected
   */
  public SourceIdentifierType getSourceIdentifierType() {
    return source_id_type;
  }

  /**
   * Sets the {@link SourceIdentifierType} to which this element is connected.
   * @param sit the {@link SourceIdentifierType} to which this element is connected.
   */
  public void setSourceIdentifierType(SourceIdentifierType sit) {
    this.source_id_type = sit;
  }
  /**
   * Returns the unique identifier (ATUI).
   * @return the ATUI
   */
  public String getAttributeIdentifier() {
    return atui;
  }

  /**
   * Returns the unique identifier (ATUI). This
   * is shorthand for {@link #getAttributeIdentifier()}.
   * @return the ATUI
   */
  public String getAtui() {
    return getAttributeIdentifier();
  }

  /**
   * Sets the unique identifier (ATUI).
   * @param atui the ATUI
   */
  public void setAttributeIdentifier(String atui) {
    this.atui = atui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeIdentifier(String)}.
   * @param atui the ATUI
   */
  public void setAtui(String atui) {
    setAttributeIdentifier(atui);
  }

  /**
   * Returns the source asserted attribute identifier (SATUI).
   * @return the SATUI
   */
  public String getSourceAttributeIdentifier() {
    return satui;
  }

  /**
   * Shorthand for {@link #getSourceAttributeIdentifier()}.
   * @return the SATUI
   */
  public String getSatui() {
    return getSourceAttributeIdentifier();
  }

  /**
   * Sets the source asserted attribute identifier (SATUI).
   * @param satui the SATUI
   */
  public void setSourceAttributeIdentifier(String satui) {
    this.satui = satui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceAttributeIdentifier(String)}.
   * @param satui the SATUI
   */
  public void setSatui(String satui) {
    setSourceAttributeIdentifier(satui);
  }

  /**
   * Returns the {@link Concept} that holds this {@link Attribute}.
   * @return the {@link Concept} that holds this {@link Attribute}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Attribute}.
   * @param concept the {@link Concept} that holds this {@link Attribute}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link Attribute}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link Attribute}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Attribute} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Attribute} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns the {@link AttributeName} (ATN).
   * @return the {@link AttributeName}
   */
  public AttributeName getAttributeName() {
    return atn;
  }

  /**
   * Shorthand for {@link #getAttributeName()}.
   * @return the {@link AttributeName}
   */
  public AttributeName getAtn() {
    return getAttributeName();
  }

  /**
   * Sets the {@link AttributeName} (ATN).
   * @param atn the {@link AttributeName}
   */
  public void setAttributeName(AttributeName atn) {
    this.atn = atn;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeName(AttributeName)}.
   * @param atn the {@link AttributeName}
   */
  public void setAtn(AttributeName atn) {
    setAttributeName(atn);
  }

  /**
   * Returns the attribute value (ATV).
   * @return the attribute value
   */
  public String getAttributeValue() {
    return atv;
  }

  /**
   * Shorthand for {@link #getAttributeValue()}.
   * @return the attribute value
   */
  public String getAtv() {
    return getAttributeValue();
  }

  /**
   * Sets the attribute value (ATV).
   * @param atv the attribute value
   */
  public void setAttributeValue(String atv) {
    this.atv = atv;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeValue(String)}.
   * @param atv the attribute value
   */
  public void setAtv(String atv) {
    setAttributeValue(atv);
  }


  /**
   * Returns the suppressibility value (SUPPRESS).
   * @return the {@link Suppressible} value
   */
  public Suppressible getSuppress() {
    return suppress;
  }

  /**
   * Indicates whether or not this {@link Attribute} is suppressible.
   * @return <code>true</code> if suppressible,
   *         <code>false</code> if not suppressible
   */
  public boolean isSuppressible() {
    return (suppress != null && !suppress.getAbbreviation().equals("N"));
  }

  /**
   * Sets the suppressibility value (SUPPRESS).
   * @param suppress the {@link Suppressible} value
   */
  public void setSuppress(Suppressible suppress) {
    this.suppress = suppress;
    key_fields_flag = false;
  }

  /**
   * Indicates whether or not this {@link AttributeImpl} is obsolete.
   * @return <code>true</code> if obsolete,
   *         <code>false</code> if not obsolete
   */
  public boolean isObsolete() {
    if (getSuppress().getAbbreviation().equals("O"))
      return true;
    return false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, atui,
   * versioned source, atn, atv, and satui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb = new StringBuffer(50);
      sb.append(getConcept() != null ? getConcept().getCui() : "")
        .append(getSource() != null ? getSource().getVsab() : "")
        .append(getAtui())
        .append(getAtn() != null ? getAtn().getAbbreviation() : "")
        .append(getSatui());
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link Attribute} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Attribute)) return false;
    return getKeyFields().equals(((Attribute)o).getKeyFields());
  }

  /**
   * Compares this {@link Attribute} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Attribute)
      return getKeyFields().compareTo(((Attribute)o).getKeyFields());
    else
      return 0;
  }

}
