/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       ConceptImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.AttributeContext;
import gov.nih.nlm.umls.meta.AttributeValue;
import gov.nih.nlm.umls.meta.ComponentHistory;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextGroup;
import gov.nih.nlm.umls.meta.ContextMember;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.CoocurrenceContext;
import gov.nih.nlm.umls.meta.CoocurrenceGroup;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.QualifierFrequency;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.RelationshipAttribute;
import gov.nih.nlm.umls.meta.Restrictor;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.umls.meta.SourceContext;
import gov.nih.nlm.umls.meta.StringAttribute;
import gov.nih.nlm.umls.meta.StringContext;
import gov.nih.nlm.umls.meta.StringInfo;
import gov.nih.nlm.umls.meta.Term;
import gov.nih.nlm.umls.meta.TermAttribute;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Concept} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ConceptImpl extends PlaceholderConcept {

  //
  // Fields
  //
  protected List attributes = null;
  protected List component_histories = null;
  protected List contexts = null;
  protected List coocurrences = null;
  protected List definitions = null;
  protected List locators = null;
  protected List mappings = null;
  protected List relationships = null;
  protected List semantic_types = null;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link Concept}.
   */
  public ConceptImpl() {};

  /**
   * Instantiates a {@link Concept} from the specified {@link Concept}.
   * @param cpt {@link Concept} to be copied
   */
  public ConceptImpl(Concept cpt) {
    setAtoms(cpt.getAtoms());
    setAttributes(cpt.getAttributes());
    setContexts(cpt.getContexts());
    setCoocurrences(cpt.getCoocurrences());
    setDefinitions(cpt.getDefinitions());
    setLocators(cpt.getLocators());
    setMappings(cpt.getMappings());
    setRelationships(cpt.getRelationships());
    setSemanticTypes(cpt.getSemanticTypes());
    setCui(cpt.getCui());
  }

  /**
   * This removes all of the {@link Concept}'s components.
   */
  public void clear() {
    // don't call clearAtoms because it is slow if we are clearing the whole concept
    atoms = null;
    clearAttributes();
    clearContexts();
    clearCoocurrences();
    clearDefinitions();
    clearLocators();
    clearMappings();
    relationships = null;
    clearSemanticTypes();
    clearComponentHistories();
  }

  /**
   * This indicates if the {@link Concept} is empty, meaning
   * that either the {@link SemanticType} was removed or
   * all of the {@link Atom}s were removed.
   * @return boolean if empty
   */
  public boolean isEmpty() {
    return (atoms == null || atoms.size() == 0) ||
           (semantic_types == null || semantic_types.size() == 0);
  }

  //
  // Atoms
  //

  /**
   * Returns the {@link Atom}s.
   * @return an {@link Atom} <code>[]</code>
   */
  public Atom[] getAtoms() {
    return atoms == null ? new Atom[0] :
      (Atom[])getAtomsAsList().toArray(new Atom[0]);
  }

  /**
   * Returns the {@link Atom}s as a {@link List}.
   * @return the {@link Atom}s as a {@link List}
   */
  public List getAtomsAsList() {
    if (atoms != null)
      return new ArrayList(atoms);
    else
      return new ArrayList(0);
  }

  /**
   * Adds the specified {@link Atom}.
   * @param atom the {@link Atom} to be added
   */
  public void addAtom(Atom atom) {
    if (atoms == null)
      atoms = new ArrayList();
    atoms.add(atom);
  }

  /**
   * Removes the specified {@link Atom} and the
   * {@link AtomAttribute}s attached to it.
   * @param atom the {@link Atom} to be removed
   */
  public void removeAtom(Atom atom) {
    if (atoms != null) {
      int i = atoms.lastIndexOf(atom);
      if (i != -1)
        atoms.remove(i);
      Attribute[] atts = atom.getAttributes();
      for (i = atts.length - 1; i >= 0; i--) {
        removeAttribute(atts[i]);
      }
      atom.clearAttributes();

      Coocurrence[] cocs = atom.getCoocurrences();
      for (i = cocs.length-1; i >= 0; i--) {
        removeCoocurrence(cocs[i]);
      }
      atom.clearCoocurrences();

      Context[] cxts = atom.getContexts();
      for (i = cxts.length-1; i >= 0; i--) {
        removeContext(cxts[i]);
      }
      atom.clearContexts();

      Definition[] defs = atom.getDefinitions();
      for (i = defs.length-1; i >= 0; i--) {
        removeDefinition(defs[i]);
      }
      atom.clearDefinitions();

      Locator[] locs = atom.getLocators();
      for (i = locs.length-1; i >= 0; i--) {
          removeLocator(locs[i]);
      }
      atom.clearLocators();

      Relationship[] rels = atom.getRelationships();
      for (i = rels.length-1; i >= 0; i--) {
        removeRelationship(rels[i]);
      }
      atom.clearRelationships();

    }
  }

  /**
   * Remove all {@link Atom}s.
   */
  public void clearAtoms() {
    if (atoms != null) {
      for (int i=0; i<atoms.size(); i++) {
        final Atom atom = (Atom)atoms.get(i);
        Attribute[] atts = atom.getAttributes();
        for (int j = atts.length-1; j >= 0; j--)
          removeAttribute(atts[j]);
        atom.clearAttributes();

        Coocurrence[] cocs = atom.getCoocurrences();
        for (int j = cocs.length-1; j >= 0; j--)
          removeCoocurrence(cocs[j]);
        atom.clearCoocurrences();

        Context[] cxts = atom.getContexts();
        for (int j = cxts.length-1; j >= 0; j--)
          removeContext(cxts[j]);
        atom.clearContexts();

        Definition[] defs = atom.getDefinitions();
        for ( int j = defs.length-1; j >= 0; j--)
          removeDefinition(defs[j]);
        atom.clearDefinitions();

        Locator[] locs = atom.getLocators();
        for (int j = locs.length-1; j >= 0; j--)
            removeLocator(locs[j]);
        atom.clearLocators();

        Relationship[] rels = atom.getRelationships();
        for (int j = rels.length-1; j >= 0; j--)
          removeRelationship(rels[j]);
        atom.clearRelationships();

      }
      atoms = null;
    }
  }

  /**
   * Replace any {@link Atom}s with the
   * ones in the specified array.
   * @param atoms a {@link Atom} <code>[]</code>
   */
  public void setAtoms(Atom atoms[]) {
    this.atoms = new ArrayList();
    for (int i=0; i<atoms.length; i++) {
      this.atoms.add(atoms[i]);
    }
  }

  /**
   * Replace {@link Atom}s with the
   * ones in the apecified {@link List}.
   * @param atoms a {@link List} of {@link Atom}s
   */
  public void setAtoms(List atoms) {
    this.atoms = new ArrayList(atoms);
  }

  /**
   * Returns the preferred atom of the concept.
   * It has TS=P, STT=PF, and ISPREF=Y.
   * @return {@link Atom}
   */
  public Atom getPreferredAtom() {
    Atom preferred = null;
    Atom[] atoms = getAtoms();
    for (int i=0; i<atoms.length; i++) {
      if(atoms[i].hasPreferredAtomIdentifier() &&
         atoms[i].hasPreferredStringIdentifier() &&
         atoms[i].hasPreferredTermIdentifier()) {
        if (atoms[i].getLanguage().getAbbreviation().equals("ENG")) {
          preferred = atoms[i];
          break;
        } else if (preferred == null) {
          preferred = atoms[i];
        }
      }
    }
    return preferred;
  }

  /**
   * Returns the preferred atom of the concept.
   * Uses the given {@link Comparator} to determine the preferred atom.
   * @param comparator {@link Comparator}
   * @return {@link Atom}
   */
  public Atom getPreferredAtom(Comparator comparator) {
    Atom[] atoms = getAtoms();
    Arrays.sort(atoms, comparator);
    return atoms[0];
  }

  //
  // Locators
  //

  /**
   * Returns the {@link Locator}s.
   * @return a {@link Locator} <code>[]</code>
   */
  public Locator[] getLocators() {
     return locators == null ? new Locator[0] :
       (Locator[])locators.toArray(new Locator[0]);
  }

  /**
   * Returns the {@link Locator}s as a {@link List}.
   * @return the {@link Locator}s as a {@link List}
   */
  public List getLocatorsAsList() {
    return locators == null ? new ArrayList(0) : new ArrayList(locators);
  }

  /**
   * Add the specified {@link Locator}.
   * @param loc the {@link Locator} to be added
   */
  public void addLocator(Locator loc) {
    if (locators == null) locators = new ArrayList();
    locators.add(loc);
  }

  /**
   * Removes the specified {@link Locator}.
   * @param loc the {@link Locator} to be removed
   */
  public void removeLocator(Locator loc) {
    if (locators != null) locators.remove(loc);
  }

  /**
   * Remove all {@link Locator}s.
   */
  public void clearLocators() {
    locators = null;
  }

  /**
   * Replace {@link Locator}s with the ones in the specified array.
   * @param locs a {@link Locator} <code>[]</code>
   */
  public void setLocators(Locator locs[]) {
    locators = new ArrayList();
    for (int i=0; i<locs.length; i++) {
      locators.add(locs[i]);
    }
  }

  /**
   * Replace {@link Locator}s with the ones in the specified {@link List}.
   * @param locs a {@link List} of {@link Locator}s
   */
  public void setLocators(List locs) {
    locators = new ArrayList(locs);
  }


  //
  // Attributes
  //

  /**
   * Returns the {@link Attribute}s.
   * @return a {@link Attribute} <code>[]</code>
   */
  public Attribute[] getAttributes() {
    return attributes == null ? new Attribute[0] : (Attribute[])attributes.toArray(new Attribute[0]);
  }

  /**
   * Returns the {@link Attribute}s as a {@link List}.
   * @return the {@link Attribute}s as a {@link List}
   */
  public List getAttributesAsList() {
    if (attributes != null)
      return new ArrayList(attributes);
    else
      return new ArrayList(0);
  }

  /**
   * Adds the specified {@link Attribute}.
   * @param att the {@link Attribute} to be added
   */
  public void addAttribute(Attribute att) {
    if (attributes == null)
      attributes = new ArrayList();
    attributes.add(att);
  }

  /**
   * Removes the specified {@link Attribute}.
   * @param att the {@link Attribute} to be removed
   */
  public void removeAttribute(Attribute att) {
    if (attributes != null) {
      int i = attributes.lastIndexOf(att);
      if (i != -1)
        attributes.remove(i);
    }
  }

  /**
   * Remove all {@link Attribute}s.
   */
  public void clearAttributes() {
    attributes = null;
  }

  /**
   * Replace {@link Attribute}s with the ones in the specified array.
   * @param atts an {@link Attribute} <code>[]</code>
   */
  public void setAttributes(Attribute atts[]) {
    attributes = new ArrayList();
    for (int i=0; i<atts.length; i++) {
      attributes.add(atts[i]);
    }
  }

  /**
   * Replace {@link Attribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link Attribute}s
   */
  public void setAttributes(List atts) {
    attributes = new ArrayList(atts);
  }

  //
  // Contexts
  //

  /**
   * Return the {@link Context}s.
   * @return a {@link Context} <code>[]</code>
   */
  public Context[] getContexts() {
    return contexts == null ? new Context[0] :
      (Context[])contexts.toArray(new Context[0]);
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Context}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Context[]
   */
  public Context[] getContexts(Restrictor restrictor) {
    List cxts = getContextsAsList();
    for (int i=cxts.size()-1; i>=0; i--) {
      if (!restrictor.keep(cxts.get(i)))
        cxts.remove(i);
    }
    Collections.sort(cxts, restrictor);
    return (Context[])cxts.toArray(new Context[0]);
  }

  /**
   * Returns the {@link Context}s as a {@link List}.
   * @return the {@link Context}s as a {@link List}
   */
  public List getContextsAsList() {
    return contexts == null ? new ArrayList(0) : new ArrayList(contexts);
  }

  /**
   * Adds the specified {@link Context}.
   * @param cxt the {@link Context} to be added
   */
  public void addContext(Context cxt) {
    if (contexts == null)
      contexts = new ArrayList();
    contexts.add(cxt);
  }

  /**
   * Removes the specified {@link Context}.
   * @param cxt the {@link Context} to be removed
   */
  public void removeContext(Context cxt) {
    if (contexts != null) {
      int i = contexts.lastIndexOf(cxt);
      if (i != -1)
        contexts.remove(i);
    }
  }

  /**
   * Remove all {@link Context}s.
   */
  public void clearContexts() {
    contexts = null;
  }

  /**
   * Replace {@link Context}s with the ones in the specified array.
   * @param cxts a {@link Context} <code>[]</code>
   */
  public void setContexts(Context cxts[]) {
    contexts = new ArrayList();
    for (int i=0; i<cxts.length; i++) {
      contexts.add(cxts[i]);
    }
  }

  /**
   * Replace {@link Context}s with the ones in the specified {@link List}.
   * @param cxts a {@link List} of {@link Context}s
   */
  public void setContexts(List cxts) {
    contexts = new ArrayList(cxts);
  }

  //
  // Coocurrences
  //

  /**
   * Returns the {@link Coocurrence}s.
   * @return a {@link Coocurrence} <code>[]</code>
   */
  public Coocurrence[] getCoocurrences() {
    return coocurrences == null ? new Coocurrence[0] :
      (Coocurrence[])coocurrences.toArray(new Coocurrence[0]);
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Coocurrence}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Coocurrence[]
   */
  public Coocurrence[] getCoocurrences(Restrictor restrictor) {
    List cocs = getCoocurrencesAsList();
    for (int i=cocs.size()-1; i>=0; i--) {
      if (!restrictor.keep(cocs.get(i)))
        cocs.remove(i);
    }
    Collections.sort(cocs, restrictor);
    return (Coocurrence[])cocs.toArray(new Coocurrence[0]);
  }


  /**
   * Returns the {@link Coocurrence}s as a {@link List}.
   * @return the {@link Coocurrence}s as a {@link List}
   */
  public List getCoocurrencesAsList() {
    return coocurrences == null ? new ArrayList(0) :
      new ArrayList(coocurrences);
  }

  /**
   * Adds the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be added
   */
  public void addCoocurrence(Coocurrence coc) {
    if (coocurrences == null)
      coocurrences = new ArrayList();
    coocurrences.add(coc);
  }

  /**
   * Removes the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be removed
   */
  public void removeCoocurrence(Coocurrence coc) {
    if (coocurrences != null) {
      int i = coocurrences.lastIndexOf(coc);
      if (i != -1)
        coocurrences.remove(i);
    }
  }

  /**
   * Remove all {@link Coocurrence}s.
   */
  public void clearCoocurrences() {
    coocurrences = null;
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified array.
   * @param cocs a {@link Coocurrence} <code>[]</code>
   */
  public void setCoocurrences(Coocurrence cocs[]){
    coocurrences = new ArrayList();
    for (int i=0; i<cocs.length; i++) {
      coocurrences.add(cocs[i]);
    }
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified {@link List}.
   * @param cocs a {@link List} of {@link Coocurrence}s
   */
  public void setCoocurrences(List cocs) {
    coocurrences = new ArrayList(cocs);
  }

  //
  // Definitions
  //

  /**
   * Returns the {@link Definition}s.
   * @return a {@link Definition} <code>[]</code>
   */
  public Definition[] getDefinitions() {
    return definitions == null ? new Definition[0] : (Definition[])definitions.toArray(new Definition[0]);
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Definition}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Definition[]
   */
  public Definition[] getDefinitions(Restrictor restrictor) {
    List defs = getDefinitionsAsList();
    for (int i=defs.size()-1; i>=0; i--) {
      if (!restrictor.keep(defs.get(i)))
        defs.remove(i);
    }
    Collections.sort(defs, restrictor);
    return (Definition[])defs.toArray(new Definition[0]);
  }

  /**
   * Returns the {@link Definition}s as a {@link List}.
   * @return the {@link Definition}s as a {@link List}
   */
  public List getDefinitionsAsList() {
    return definitions == null ? new ArrayList(0) : new ArrayList(definitions);
  }

  /**
   * Adds the specified {@link Definition}.
   * @param def the {@link Definition} to be added
   */
  public void addDefinition(Definition def) {
    if (definitions == null)
      definitions = new ArrayList();
    definitions.add(def);
  }

  /**
   * Remove the specified {@link Definition}.
   * @param def the {@link Definition} to be removed
   */
  public void removeDefinition(Definition def) {
    if (definitions != null) definitions.remove(def);
  }

  /**
   * Remove all {@link Definition}s.
   */
  public void clearDefinitions() {
    definitions = null;
  }

  /**
   * Replace {@link Definition}s with the ones in the specified array.
   * @param defs a {@link Definition} <code>[]</code>
   */
  public void setDefinitions(Definition defs[]) {
    definitions = new ArrayList();
    for (int i=0; i<defs.length; i++) {
      definitions.add(defs[i]);
    }
  }

  /**
   * Replace {@link Definition}s with the ones in the specified {@link List}.
   * @param defs a {@link List} of {@link Definition}s
   */
  public void setDefinitions(List defs) {
    definitions = new ArrayList(defs);
  }

  //
  // Relationships
  //

  /**
   * Returns the {@link Relationship}s.
   * @return a {@link Relationship} <code>[]</code>
   */
  public Relationship[] getRelationships() {
    return relationships == null ? new Relationship[0] :
        (Relationship[]) relationships.toArray(new Relationship[0]);
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Relationship}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Relationship[]
   */
  public Relationship[] getRelationships(Restrictor restrictor) {
    List rels = getRelationshipsAsList();
    for (int i = rels.size() - 1; i >= 0; i--) {
      if (!restrictor.keep(rels.get(i)))
        rels.remove(i);
    }
    Collections.sort(rels, restrictor);
    return (Relationship[]) rels.toArray(new Relationship[0]);
  }

  /**
   * Returns the {@link Relationship}s as a {@link List}.
   * @return the {@link Relationship}s as a {@link List}
   */
  public List getRelationshipsAsList() {
    return relationships == null ? new ArrayList(0) :
        new ArrayList(relationships);
  }

  /**
   * Adds the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be added
   */
  public void addRelationship(Relationship rel) {
    if (relationships == null)
      relationships = new ArrayList();
    relationships.add(rel);
  }

  /**
   * Removes the specified {@link Relationship} and the
   * {@link RelationshipAttribute}s attached to it.
   * @param rel the {@link Relationship} to be removed
   */
  public void removeRelationship(Relationship rel) {
    if (relationships != null) {
      int i = relationships.lastIndexOf(rel);
      if (i != -1)
        relationships.remove(i);
      Attribute[] atts = rel.getAttributes();
      for (i = atts.length-1; i >= 0; i--) {
        removeAttribute(atts[i]);
      }
      rel.clearAttributes();
    }
  }

  /**
   * Remove all {@link Relationship}s.
   */
  public void clearRelationships() {
    if (relationships != null) {
      for (int i=0; i<relationships.size(); i++) {
        Attribute[] atts = ((Relationship)relationships.get(i)).getAttributes();
        for (int j = atts.length-1; j >= 0; j--) {
          removeAttribute(atts[j]);
        }
      }
      relationships = null;
    }
  }

  /**
   * Replace {@link Relationship}s with the ones in the specified array.
   * @param rels a {@link Relationship} <code>[]</code>
   */
  public void setRelationships(Relationship rels[]) {
    relationships = new ArrayList();
    for (int i=0; i<rels.length; i++) {
      relationships.add(rels[i]);
    }
  }

  /**
   * Replace {@link Relationship}s of the ones in the specified {@link List}.
   * @param rels a {@link List} of {@link Relationship}s
   */
  public void setRelationships(List rels) {
    relationships = new ArrayList(rels);
  }

  //
  // SemanticTypes
  //

   /**
   * Returns the {@link SemanticType}s.
   * @return a {@link SemanticType} <code>[]</code>
   */
  public SemanticType[] getSemanticTypes() {
    return semantic_types == null ? new SemanticType[0] :
      (SemanticType[])semantic_types.toArray(new SemanticType[0]);
  }

  /**
   * Returns the {@link SemanticType}s as a {@link List}.
   * @return the {@link SemanticType}s as a {@link List}
   */
  public List getSemanticTypesAsList() {
    return semantic_types == null ? new ArrayList(0) :
      new ArrayList(semantic_types);
  }

  /**
   * Adds a {@link SemanticType}.
   * @param sty the {@link SemanticType} to be added
   */
  public void addSemanticType(SemanticType sty) {
    if (semantic_types == null) semantic_types = new ArrayList();
    semantic_types.add(sty);
  }

  /**
   * Removes the specified {@link SemanticType}.
   * @param sty the {@link SemanticType} to be removed
   */
  public void removeSemanticType(SemanticType sty) {
    if (semantic_types != null) semantic_types.remove(sty);
  }

  /**
   * Removes all {@link SemanticType}s.
   */
  public void clearSemanticTypes() {
    semantic_types = null;
  }

  /**
   * Replace {@link SemanticType}s with the ones in the specified array.
   * @param stys a {@link SemanticType} <code>[]</code>
   */
  public void setSemanticTypes(SemanticType stys[]) {
    semantic_types = new ArrayList();
    for (int i=0; i<stys.length; i++) {
      semantic_types.add(stys[i]);
    }
  }

  /**
   * Replace {@link SemanticType}s with the ones in the specified {@link List}.
   * @param stys a {@link List} of {@link SemanticType}s
   */
  public void setSemanticTypes(List stys) {
    semantic_types = new ArrayList(stys);
  }

  //
  // Component Histories
  //

   /**
   * Returns the {@link ComponentHistory}s.
   * @return a {@link ComponentHistory} <code>[]</code>
   */
  public ComponentHistory[] getComponentHistories() {
    return component_histories == null ?
      new ComponentHistory[0] :
      (ComponentHistory[])component_histories.toArray(new ComponentHistory[0]);
  }

  /**
   * Returns the {@link ComponentHistory}s as a {@link List}.
   * @return the {@link ComponentHistory}s as a {@link List}
   */
  public List getComponentHistoriesAsList() {
    return component_histories == null ?
      new ArrayList(0) : new ArrayList(component_histories);
  }

  /**
   * Adds a {@link ComponentHistory}.
   * @param map the {@link ComponentHistory} to be added
   */
  public void addComponentHistory(ComponentHistory map) {
    if (component_histories == null)
      component_histories = new ArrayList();
    component_histories.add(map);
  }

  /**
   * Removes the specified {@link ComponentHistory}.
   * @param chs the {@link ComponentHistory} to be removed
   */
  public void removeComponentHistory(ComponentHistory chs) {
    if (component_histories != null) {
      int i = component_histories.lastIndexOf(chs);
      if (i != -1)
        component_histories.remove(i);
    }
  }

  /**
   * Removes all {@link ComponentHistory}s.
   */
  public void clearComponentHistories() {
    component_histories = null;
  }

  /**
   * Replace {@link ComponentHistory}s with the ones in the specified array.
   * @param maps a {@link ComponentHistory} <code>[]</code>
   */
  public void setComponentHistories(ComponentHistory maps[]) {
    component_histories = new ArrayList();
    for (int i=0; i<maps.length; i++) {
      component_histories.add(maps[i]);
    }
  }

  /**
   * Replace {@link ComponentHistory}s with the ones in the specified {@link List}.
   * @param maps a {@link List} of {@link ComponentHistory}s
   */
  public void setComponentHistories(List maps) {
    component_histories = new ArrayList(maps);
  }

  //
  // Mappings
  //

   /**
   * Returns the {@link Mapping}s.
   * @return a {@link Mapping} <code>[]</code>
   */
  public Mapping[] getMappings() {
    return mappings == null ? new Mapping[0] :
      (Mapping[])mappings.toArray(new Mapping[0]);
  }

  /**
   * Returns the {@link Mapping}s as a {@link List}.
   * @return the {@link Mapping}s as a {@link List}
   */
  public List getMappingsAsList() {
    return mappings == null ? new ArrayList(0) :
      new ArrayList(mappings);
  }

  /**
   * Adds a {@link Mapping}.
   * @param map the {@link Mapping} to be added
   */
  public void addMapping(Mapping map) {
    if (mappings == null) mappings = new ArrayList();
    mappings.add(map);
  }

  /**
   * Removes the specified {@link Mapping}.
   * @param map the {@link Mapping} to be removed
   */
  public void removeMapping(Mapping map) {
    if (mappings != null) {
      int i = mappings.lastIndexOf(map);
      if (i != -1)
        mappings.remove(i);
    }
  }

  /**
   * Removes all {@link Mapping}s.
   */
  public void clearMappings() {
    mappings = null;
  }

  /**
   * Replace {@link Mapping}s with the ones in the specified array.
   * @param maps a {@link Mapping} <code>[]</code>
   */
  public void setMappings(Mapping maps[]) {
    mappings = new ArrayList();
    for (int i=0; i<maps.length; i++) {
      mappings.add(maps[i]);
    }
  }

  /**
   * Replace {@link Mapping}s with the ones in the specified {@link List}.
   * @param maps a {@link List} of {@link Mapping}s
   */
  public void setMappings(List maps) {
    mappings = new ArrayList(maps);
  }

 /**
   * Returns the attributes with aui that matches that of the given {@link Atom}.
   * @param atom {@link Atom}
   * @return {@link List} of {@link Attribute}s
   */
  public Attribute[] getAttributesForAtom(Atom atom) {
    return (Attribute[])getAttributesForAtomAsList(atom).toArray(new Attribute[0]);
  }

  /**
   * Returns the attributes with aui that matches that of the given {@link Atom}
   * in a {@link List}.
   * @param atom {@link Atom}
   * @return {@link List} of {@link AtomAttribute}s
   */
  public List getAttributesForAtomAsList(Atom atom) {
    if (attributes == null) return new ArrayList(0);
    ArrayList atts = new ArrayList(attributes.size());
    for (int i=0; i<attributes.size(); i++) {
      AtomAttribute l_att = (AtomAttribute)attributes.get(i);
      if (l_att.getAtom().getAui().equals(atom.getAui())) {
        atts.add(l_att);
      }
    }
    return atts;
  }

  /**
   * Returns the attributes with aui that matches that of the given {@link Relationship}.
   * @param rel {@link AtomRelationship}
   * @return {@link List} of {@link Attribute}s
   */
  public AtomAttribute[] getAttributesForRelationship(AtomRelationship rel) {
    return (AtomAttribute[])getAttributesForRelationshipAsList(rel).toArray(new AtomAttribute[0]);
  }

  /**
   * Returns the attributes with aui that matches that of the given {@link Relationship}
   * in a {@link List}.
   * @param rel {@link AtomRelationship}
   * @return {@link List} of {@link Attribute}s
   */
  public List getAttributesForRelationshipAsList(AtomRelationship rel) {
    if (attributes == null) return new ArrayList(0);
    ArrayList atts = new ArrayList(attributes.size());
    for (int i=0; i<attributes.size(); i++) {
      AtomAttribute l_att = (AtomAttribute)attributes.get(i);
      if (l_att.getAtom().getAui().equals(rel.getAtom().getAui())) {
        atts.add(l_att);
      }
    }
    return atts;
  }

  /**
   * Shorthand for {@link #getConceptIdentifier()}.
   * @return the CUI
   */
  public String getCui() {
    return getConceptIdentifier();
  }

  /**
   * Shorthand for {@link #setConceptIdentifier(String)}.
   * @param cui the CUI
   */
  public void setCui(String cui) {
    setConceptIdentifier(cui);
  }

  /**
   * Returns the concept unique identifier (CUI) for the {@link Concept}
   * that holds this {@link Concept}.
   * @return the CUI
   */
  public String getConceptIdentifier() {
    return cui;
  }

  /**
   * Sets the concept unique identifier (CUI) for the {@link Concept}
   * that holds this {@link Concept}.
   * @param cui the CUI
   */
  public void setConceptIdentifier(String cui) {
    this.cui = cui;
    key_fields_flag = false;
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Atom}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Atom[]
   */
  public Atom[] getAtoms(Restrictor restrictor) {
    List atoms = getAtomsAsList();
    for (int i=atoms.size()-1; i>=0; i--) {
      if (!restrictor.keep(atoms.get(i)))
        atoms.remove(i);
    }
    Collections.sort(atoms, restrictor);
    return (Atom[])atoms.toArray(new Atom[0]);
  }


  //
  // UMLSKS View Methods
  //

  /**
   * Return the {@link ContextGroup}s.
   * @return a {@link ContextGroup} <code>[]</code>
   */
  public ContextGroup[] getContextGroups() {
    return (ContextGroup[])getContextGroupsAsList().toArray(new ContextGroup[0]);
  }

  /**
   * Returns the {@link ContextGroup}s as a {@link List}.
   * @return the {@link ContextGroup}s as a {@link List}
   */
  public List getContextGroupsAsList() {
    Map cxt_grp_hm = new HashMap();
    Map string_cxt_hm = new HashMap();
    Map string_info_hm = new HashMap();
    Map atom_hm = new HashMap();
    List terms = getTermsAsList();

    // before doing anything else, populate string_info_hm and atom_hm
    ArrayList terms_list = (ArrayList) getTermsAsList();
    for (int i = 0; i < terms_list.size(); i++) {
      StringInfo[] info = ( (Term) terms.get(i)).getTermVariants();
      for (int j = 0; j < info.length; j++) {
        StringInfo this_info = info[j];
        string_info_hm.put(this_info.getSui(), info[j]);
        Atom[] atoms = this_info.getAtoms();
        for (int k = 0; k < atoms.length; k++) {
          atom_hm.put(atoms[k].getAui(), atoms[k]);
        }
      }
    }

    Context l_cxt;
    ContextGroup l_cxt_grp;
    for (int i = 0; i < contexts.size(); i++) {
      l_cxt = (Context) contexts.get(i);

      StringBuffer sb = new StringBuffer(200);
      sb.append(l_cxt.getAtom().getSui()).append(l_cxt.getConcept().getCui()).
          append(
          l_cxt.getCvf());
      String cxt_grp_key = sb.toString();
      sb.append(l_cxt.getAtom().getCode()).append(l_cxt.getSource());
      String string_cxt_key = sb.toString();

      if (cxt_grp_hm.containsKey(cxt_grp_key)) {
        l_cxt_grp = (ContextGroup) cxt_grp_hm.get(cxt_grp_key);
        StringContext l_sc;
        if (string_cxt_hm.containsKey(string_cxt_key)) {
          l_sc = (StringContext) string_cxt_hm.get(string_cxt_key);
        }
        else { // no matching string_context found
          l_sc = new StringContextImpl();
          if (atom_hm.containsKey(l_cxt.getAtom().getAui()))
            l_sc.setAtom( (Atom) atom_hm.get(l_cxt.getAtom().getAui()));
            //attention need Atom
          l_sc.setSource(l_cxt.getSource());
          string_cxt_hm.put(string_cxt_key, l_sc);
          l_cxt_grp.addStringContext(l_sc);
        }
        SourceContext source_context = new SourceContextImpl();
        source_context.setAtom(l_cxt.getAtom());
        source_context.setContextMembers(l_cxt.getContextMembers());
        source_context.setCxn(l_cxt.getCxn());
        source_context.setPaui(l_cxt.getPaui());
        source_context.setPtr(l_cxt.getPathToRootAsString());
        l_sc.addSourceContext(source_context);
      }
      else { // no matching context group
        l_cxt_grp = new ContextGroupImpl();
        if (string_info_hm.containsKey(l_cxt.getAtom().getSui()))
          l_cxt_grp.setStringInfo( (StringInfo) string_info_hm.get(l_cxt.
              getAtom().getSui()));
          // attention need StringInfo
        l_cxt_grp.getConcept().setCui(l_cxt.getConcept().getCui());
        l_cxt_grp.setConcept(l_cxt.getConcept());

        StringContext l_sc = new StringContextImpl();
        if (atom_hm.containsKey(l_cxt.getAtom().getAui()))
          l_sc.setAtom( (Atom) atom_hm.get(l_cxt.getAtom().getAui()));
        l_sc.setSource(l_cxt.getSource());
        string_cxt_hm.put(string_cxt_key, l_sc);

        SourceContext source_context = new SourceContextImpl();
        source_context.setAtom(l_cxt.getAtom());
        source_context.setContextMembers( ( (ContextMember[]) l_cxt.
                                           getContextMembers()));
        source_context.setCxn(l_cxt.getCxn());
        source_context.setPaui(l_cxt.getPaui());
        source_context.setPtr(l_cxt.getPathToRootAsString());
        l_sc.addSourceContext(source_context);
        l_cxt_grp.addStringContext(l_sc);
        cxt_grp_hm.put(cxt_grp_key, l_cxt_grp);
      }
    } // for each context
    return new ArrayList(cxt_grp_hm.values());

  }

  /**
   * Returns the {@link CoocurrenceGroup}s.
   * @return a {@link CoocurrenceGroup} <code>[]</code>
   */
  public CoocurrenceGroup[] getCoocurrenceGroups() {
    return (CoocurrenceGroup[])getCoocurrenceGroupsAsList().toArray(new CoocurrenceGroup[0]);
  }

  /**
   * Returns the {@link CoocurrenceGroup}s as a {@link List}.
   * @return the {@link CoocurrenceGroup}s as a {@link List}
   */
  public List getCoocurrenceGroupsAsList() {
    HashMap coc_grp_hm = new HashMap();
    HashMap coc_cxt_hm = new HashMap();
    Coocurrence l_coc;
    CoocurrenceGroup l_coc_grp;
    for (int i = 0; i < coocurrences.size(); i++) {
      l_coc = (Coocurrence) coocurrences.get(i);

      StringBuffer sb = new StringBuffer(200);
      sb.append(l_coc.getConcept().getCui()).append(
          l_coc.getRelatedConcept().getCui());
      String coc_grp_key = sb.toString();
      // not including cof or coa in key b/c not needed for uniqueness
      sb.append(l_coc.getSource()).append(l_coc.getCot());
      String coc_cxt_key = sb.toString();

      if (coc_grp_hm.containsKey(coc_grp_key)) {
        l_coc_grp = (CoocurrenceGroup) coc_grp_hm.get(coc_grp_key);
        CoocurrenceContext l_coc_cxt;
        if (coc_cxt_hm.containsKey(coc_cxt_key)) {
          l_coc_cxt = (CoocurrenceContext) coc_cxt_hm.get(coc_cxt_key);
        }
        else { // no matching coocurrence context found
          l_coc_cxt = new CoocurrenceContextImpl();
          l_coc_cxt.setCot(l_coc.getCot());
          l_coc_cxt.setSource(l_coc.getSource());
          l_coc_cxt.setCof(l_coc.getCof());
          l_coc_cxt.setCvf(l_coc.getCvf());
          coc_cxt_hm.put(coc_cxt_key, l_coc_cxt);
          l_coc_grp.addCoocurrenceContext(l_coc_cxt);
        }
        List qual_freq_list = getQualifierFrequenciesAsList(l_coc.getCoa());
        l_coc_cxt.setQualifiers(qual_freq_list);
      }
      else { // no matching coocurrence group found
        l_coc_grp = new CoocurrenceGroupImpl();
        l_coc_grp.setRelatedConcept(l_coc.getRelatedConcept());
        //l_coc_grp.setAtom(l_coc.getAtom());
        //l_coc_grp.setRelatedAtom(l_coc.getRelatedAtom());
        l_coc_grp.setConcept(l_coc.getConcept());
        // don't need to copy cui1 b/c it is in concept

        CoocurrenceContext l_coc_cxt = new CoocurrenceContextImpl();
        l_coc_cxt.setCot(l_coc.getCot());
        l_coc_cxt.setSource(l_coc.getSource());
        l_coc_cxt.setCof(l_coc.getCof());
        l_coc_cxt.setCvf(l_coc.getCvf());
        coc_cxt_hm.put(coc_cxt_key, l_coc_cxt);

        List qual_freq_list = getQualifierFrequenciesAsList(l_coc.getCoa());
        l_coc_cxt.setQualifiers(qual_freq_list);
        l_coc_grp.addCoocurrenceContext(l_coc_cxt);
        coc_grp_hm.put(coc_grp_key, l_coc_grp);
      }
    } // for each coocurrence
    return new ArrayList(coc_grp_hm.values());
  }

  /**
   * Returns the {@link TermAttribute}s.
   * @return a {@link TermAttribute} <code>[]</code>
   */
  public TermAttribute[] getTermAttributes() {
    return (TermAttribute[])getTermAttributesAsList().toArray(new TermAttribute[0]);
  }

  /**
   * Returns the {@link TermAttribute}s as a {@link List}.
   * @return the {@link TermAttribute}s as a {@link List}
   */
  public List getTermAttributesAsList() {
    HashMap term_att_hm = new HashMap();
    HashMap string_att_hm = new HashMap();
    HashMap att_context_hm = new HashMap();
    HashMap string_info_hm = new HashMap();
    HashMap atom_hm = new HashMap();
    List terms = getTermsAsList();

    // before doing anything else, populate string_info_hm and atom_hm
    ArrayList terms_list = (ArrayList) getTermsAsList();
    for (int i = 0; i < terms_list.size(); i++) {
      StringInfo[] info = ( (Term) terms.get(i)).getTermVariants();
      for (int j = 0; j < info.length; j++) {
        StringInfo this_info = info[j];
        string_info_hm.put(this_info.getSui(), info[j]);
        Atom[] atoms = this_info.getAtoms();
        for (int k = 0; k < atoms.length; k++) {
          atom_hm.put(atoms[k].getAui(), atoms[k]);
        }
      }
    }

    AtomAttribute l_att;
    TermAttribute l_term_att;

    for (int i = 0; i < attributes.size(); i++) {
      // check if AtomAttribute
      if (attributes.get(i)instanceof AtomAttribute) {
        l_att = ( (AtomAttribute) attributes.get(i));

        StringBuffer sb = new StringBuffer(200);
        sb.append(l_att.getAtom().getLui()).append(l_att.getConcept().getCui()).
            append(l_att.getCvf());
        String term_att_key = sb.toString();
        sb.append(l_att.getAtom().getSui());
        String string_att_key = sb.toString();
        sb.append(l_att.getAtom().getCode()).append(l_att.getAtom().getAui()).
            append(l_att.getSource());
        String attribute_str_key = sb.toString();

        if (l_att.getAtom().getLui() == null)
          continue;
        if (term_att_hm.containsKey(term_att_key)) {
          l_term_att = (TermAttribute) term_att_hm.get(term_att_key);
          if (string_att_hm.containsKey(string_att_key)) {
            StringAttribute l_sa = (StringAttribute) string_att_hm.get(
                string_att_key);
            AttributeContext att_context;
            if (att_context_hm.containsKey(attribute_str_key)) {
              att_context = (AttributeContext) att_context_hm.get(
                  attribute_str_key);
            }
            else {
              att_context = new AttributeContextImpl();
              if (atom_hm.containsKey(l_att.getAtom().getAui()))
                att_context.setAtom( (Atom) atom_hm.get(l_att.getAtom().getAui()));
                // get atom_hm and use in place of getAtom()
                // attention don't need this? att_context.getAtom().setCode(l_att.getAtom().getCode());
              att_context.setSource(l_att.getSource());
              att_context.setSourceIdentifierType(l_att.getSourceIdentifierType());
              att_context.setStringAttribute(l_sa);
              l_sa.addAttributeContext(att_context);
              att_context_hm.put(attribute_str_key, att_context);
            }
            AttributeValue att_value = new AttributeValueImpl();
            att_value.setAtn(l_att.getAtn());
            att_value.setAtv(l_att.getAtv());
            att_value.setAtui(l_att.getAtui());
            att_value.setSatui(l_att.getSatui());
            att_context.addAttributeValue(att_value);

          }
          else { // no matching string_attribute found
            StringAttribute l_sa = new StringAttributeImpl();
            l_sa.setTermAttribute(l_term_att);
            if (string_info_hm.containsKey(l_att.getAtom().getSui()))
              l_sa.setStringInfo( (StringInfo) string_info_hm.get(l_att.getAtom().
                  getSui()));
            string_att_hm.put(string_att_key, l_sa);

            AttributeContext att_context = new AttributeContextImpl();
            if (atom_hm.containsKey(l_att.getAtom().getAui()))
              att_context.setAtom( (Atom) atom_hm.get(l_att.getAtom().getAui()));
              //att_context.setAtom(l_att.getAtom());
              // do as above with setAtom
            att_context.setSource(l_att.getSource());
            att_context.setSourceIdentifierType(l_att.getSourceIdentifierType());
            att_context.setStringAttribute(l_sa);
            l_sa.addAttributeContext(att_context);
            l_term_att.addStringAttribute(l_sa);
            att_context_hm.put(attribute_str_key, att_context);

            AttributeValue att_value = new AttributeValueImpl();
            att_value.setAtn(l_att.getAtn());
            att_value.setAtv(l_att.getAtv());
            att_value.setAtui(l_att.getAtui());
            att_value.setSatui(l_att.getSatui());
            att_context.addAttributeValue(att_value);

            term_att_hm.put(term_att_key, l_term_att);
          }
        }
        else { // no term_attribute with matching lui
          l_term_att = new TermAttributeImpl();
          l_term_att.setLui(l_att.getAtom().getLui());
          l_term_att.getConcept().setCui(l_att.getConcept().getCui());
          l_term_att.setConcept(l_att.getConcept());

          StringAttribute l_sa = new StringAttributeImpl();
          l_sa.setTermAttribute(l_term_att);
          if (string_info_hm.containsKey(l_att.getAtom().getSui()))
            l_sa.setStringInfo( (StringInfo) string_info_hm.get(l_att.getAtom().
                getSui()));
          string_att_hm.put(l_att.getAtom().getSui(), l_sa);

          AttributeContext att_context = new AttributeContextImpl();
          if (atom_hm.containsKey(l_att.getAtom().getAui()))
            att_context.setAtom( (Atom) atom_hm.get(l_att.getAtom().getAui()));
            //att_context.setAtom(l_att.getAtom());
            // setAtom as above
          att_context.setSource(l_att.getSource());
          att_context.setSourceIdentifierType(l_att.getSourceIdentifierType());
          att_context.setStringAttribute(l_sa);
          l_sa.addAttributeContext(att_context);
          l_term_att.addStringAttribute(l_sa);
          att_context_hm.put(attribute_str_key, att_context);

          AttributeValue att_value = new AttributeValueImpl();
          att_value.setAtn(l_att.getAtn());
          att_value.setAtv(l_att.getAtv());
          att_value.setAtui(l_att.getAtui());
          att_value.setSatui(l_att.getSatui());
          att_context.addAttributeValue(att_value);

          term_att_hm.put(l_att.getAtom().getLui(), l_term_att);
        }
      }
    } // for each attribute
    return new ArrayList(term_att_hm.values());
  }

  /**
   * Returns the {@link Term}s.
   * @return an {@link Term} <code>[]</code>
   */
  public Term[] getTerms() {
    return (Term[])getTermsAsList().toArray(new Term[0]);
  }

  /**
   * Returns the {@link Term}s as a {@link List}.
   * @return the {@link Term}s as a {@link List}
   */
  public List getTermsAsList() {
    HashMap term_hm = new HashMap();
    HashMap string_info_hm = new HashMap();
    Atom l_atom;
    Term l_term;

    //
    // Iterate through atoms and build terms
    //
    for (int i = 0; i < atoms.size(); i++) {

      //
      // Get atom
      //
      l_atom = (Atom) atoms.get(i);

      //
      // Construct the term key
      //
      StringBuffer sb = new StringBuffer(200);
      sb.append(l_atom.getLui()).append(l_atom.getTs()).append(
          l_atom.getLat()).append(l_atom.getCvf());
      String term_key = sb.toString();

      //
      // construct the string_info key
      //
      sb.append(l_atom.getSui()).append(l_atom.getStr()).append(l_atom.getStt());
      String string_info_key = sb.toString();

      //
      // We can only construct terms for non-null LUI atoms
      //
      if (l_atom.getLui() == null)
        continue;

      //
      // If the term exists already, add this atom's info to it
      //
      if (term_hm.containsKey(term_key)) {

        l_term = (Term) term_hm.get(term_key);
        StringInfo l_si;

        //
        // If the string info exists already, get reference
        //
        if (string_info_hm.containsKey(string_info_key)) {
          l_si = (StringInfo) string_info_hm.get(string_info_key);
        }

        //
        // No matching string info, create new one
        else {
          l_si = new StringInfoImpl();
          l_si.setSui(l_atom.getSui());
          l_si.setStr(l_atom.getStr());
          l_si.setStt(l_atom.getStt());
          string_info_hm.put(string_info_key, l_si);
          l_term.addTermVariant(l_si);
          l_si.setTerm(l_term);
        }

        //
        // Add atom to term
        //
        Atom atom = l_atom;
        l_si.addAtom(atom);
      }

      //
      // No term for this LUI, make one, populate it
      //
      else {
        l_term = new TermImpl();
        l_term.setLui(l_atom.getLui());
        l_term.getConcept().setCui(l_atom.getConcept().getCui());
        l_term.setStr(l_atom.getStr());
        l_term.setLat(l_atom.getLat());
        l_term.setTs(l_atom.getTs());
        l_term.setConcept(l_atom.getConcept());

        StringInfo l_si = new StringInfoImpl();
        l_si.setSui(l_atom.getSui());
        l_si.setStr(l_atom.getStr());
        l_si.setStt(l_atom.getStt());
        //
        // Put string info in map
        //
        string_info_hm.put(string_info_key, l_si);

        Atom atom = l_atom;
        l_si.addAtom(atom);
        l_term.addTermVariant(l_si);
        l_si.setTerm(l_term);

        //
        // Put term in map
        //
        term_hm.put(term_key, l_term);
      }
    } // for each atom
    return new ArrayList(term_hm.values());
  }


  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      key_fields = getCui();
      key_fields_flag = true;
    }
    return key_fields;
  }


  //
  // Private helper methods
  //

  /**
   * Takes a string representing a COA and returns a {@link List}
   * of {@link QualifierFrequency} elements.
   * @param coa a string representing a COA value
   * @return a {@link List} of {@link QualifierFrequency}
   */
  private List getQualifierFrequenciesAsList(String coa) {
    List qualifier_list = new ArrayList();
    StringTokenizer st = new StringTokenizer(coa, ",");
    while (st.hasMoreTokens()) {
      QualifierFrequency qf = new QualifierFrequencyImpl();
      String this_token = st.nextToken();
      qf.setQualifier(this_token.substring(0, this_token.indexOf("=")));
      qf.setFrequency(new Integer(this_token.substring(this_token.indexOf("=") + 1)).intValue());
      qualifier_list.add(qf);
    }
    return qualifier_list;
  }

}
