/**************************************************************************
 * Package:  gov.nih.nlm.umls.meta.impl
 * Class:    ConceptMappingImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.ConceptMapping;
import gov.nih.nlm.umls.meta.RelationshipLabel;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.ConceptMapping}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ConceptMappingImpl implements ConceptMapping {

  //
  // Fields
  //
  private String cui1 = null;
  private String cui2 = null;
  private String ver = null;
  private RelationshipLabel rel = null;
  private AdditionalRelationshipLabel rela = null;
  private String mapin = null;
  private String reason = null;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag = false;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link ConceptMapping}.
   */
  public ConceptMappingImpl() {};

  /**
   * Instantiates a {@link ConceptMapping} from the specified {@link ConceptMapping}.
   * @param cm {@link ConceptMapping} to be copied
   */
  public ConceptMappingImpl(ConceptMapping cm) {
    setCui2(cm.getCui2());
    setVer(cm.getVer());
    setRel(cm.getRel());
    setRela(cm.getRela());
    setMapin(cm.getMapin());
    setMapReason(cm.getMapReason());
  }

  /**
   * Returns the concept identifier for the first concept (CUI1).
   * @return the concept identifier for the first concept
   */
  public String getConceptIdentifier() {
    return cui1;
  }

  /**
   * Shorthand for {@link #getConceptIdentifier()}.
   * @return the result of <code>getConceptIdentifier()</code>
   */
  public String getCui1() {
    return getConceptIdentifier();
  }

  /**
   * Sets the concept identifier for the first concept (CUI1).
   * @param cui1 the concept identifier for the first concept
   */
  public void setConceptIdentifier(String cui1) {
    this.cui1 = cui1;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setConceptIdentifier(String)}.
   * @param cui1 the concept identifier for the first concept
   */
  public void setCui1(String cui1) {
    setConceptIdentifier(cui1);
  }

  /**
   * Returns the last release version in which the CUI1 was valid (VER).
   * @return the last release version in which the CUI1 was valid
   */
  public String getLastReleaseVersion() {
    return ver;
  }

  /**
   * Shorthand for {@link #getLastReleaseVersion()}.
   * @return the result of <code>getLastReleaseVersion()</code>
   */
  public String getVer() {
    return getLastReleaseVersion();
  }

  /**
   * Sets the last release version in which the CUI1 was valid (VER).
   * @param ver the last release version in which the CUI1 was valid
   */
  public void setLastReleaseVersion(String ver) {
    this.ver = ver;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setLastReleaseVersion(String)}
   * @param ver the last release version in which the CUI1 was valid
   */
  public void setVer(String ver) {
    setLastReleaseVersion(ver);
  }

  /**
   * Returns the relationship between the two concepts (REL).
   * @return the relationship between the two concepts
   */
  public RelationshipLabel getRelationship() {
    return rel;
  }

  /**
   * Shorthand for {@link #getRelationship()}.
   * @return the result of <code>getRelationship()</code>
   */
  public RelationshipLabel getRel() {
    return getRelationship();
  }

  /**
   * Sets the relationship between the two concepts (REL).
   * @param rel the relationship between the two concepts
   */
  public void setRelationship(RelationshipLabel rel) {
    this.rel = rel;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setRelationship(RelationshipLabel)}.
   * @param rel the relationship between the two concepts
   */
  public void setRel(RelationshipLabel rel) {
    setRelationship(rel);
  }

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel() {
    return rela;
  }

  /**
   * Shorthand for {@link #getAdditionalRelationshipLabel}).
   * @return the result of <code>getAdditionalRelationshipLabel()</code>
   */
  public AdditionalRelationshipLabel getRela() {
    return getAdditionalRelationshipLabel();
  }

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Returns the concept identifier for the second concept (CUI2).
   * @return the concept identifier for the second concept
   */
  public String getRelativeConceptIdentifier() {
    return cui2;
  }

  /**
   * Shorthand for {@link #getRelativeConceptIdentifier()}.
   * @return the result of <code>getRelativeConceptIdentifier()</code>
   */
  public String getCui2() {
    return getRelativeConceptIdentifier();
  }

  /**
   * Sets the concept identifier for the second concept (CUI2).
   * @param cui2 the concept identifier for the second concept
   */
  public void setRelativeConceptIdentifier(String cui2) {
    this.cui2 = cui2;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setRelativeConceptIdentifier(String)}.
   * @param cui2 the concept identifier for the second concept
   */
  public void setCui2(String cui2) {
    setRelativeConceptIdentifier(cui2);
  }

  /**
   * Returns the reason for this {@link ConceptMapping} (MAPREASON).
   * @return the reason for this {@link ConceptMapping}
   */
  public String getMapReason() {
    return reason;
  }

  /**
   * Sets the reason for this {@link ConceptMapping} (MAPREASON).
   * @param reason the reason for this {@link ConceptMapping}
   */
  public void setMapReason(String reason) {
    this.reason = reason;
    key_fields_flag = false;
  }

   /**
   * Returns the "in current subset" value (MAPIN).
   * @return the "in current subset" value
   */
  public String getMapin() {
    return mapin;
  }

  /**
   * Indicates whether or not this {@link ConceptMapping} is in the current subset (MAPIN).
   * @return <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public boolean inCurrentSubset() {
    return (getMapin() != null && getMapin().equals("Y"));
  }

  /**
   * Sets the "in current subset" value (MAPIN).
   * @param mapin the "in current subset" value
   */
  public void setMapin(String mapin) {
    this.mapin = mapin;
    key_fields_flag = false;
  }

  /**
   * Sets the flag indicating whether or not this {@link ConceptMapping}
   * is in the current subset (MAPIN).
   * @param mapin <code>boolean</code> true if in current subset,
   *         <code>boolean</code> false if not in current subset
   */
  public void setInCurrentSubset(boolean mapin) {
    setMapin(mapin ? "Y" : "N");
    key_fields_flag = false;
  }

  /**
   * Return the reason for this {@link ConceptMapping}.
   * @return reason
   */
  public String getReason() {
    return reason;
  }

  /**
   * Set the reason for this {@link ConceptMapping}.
   * @param reason the reason for the mapping
   */
  public void setReason(String reason) {
    this.reason = reason;
    key_fields_flag = false;
  }

  //
  // Overridden Object Methods
  //

  /**
   * Returns a hash code based on all fields.
   * @return a hash code based on all fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link ConceptMapping} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal,
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof ConceptMapping)) return false;
    return getKeyFields().equals(((ConceptMapping)o).getKeyFields());
  }

  /**
   * Compares this {@link ConceptMapping} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof ConceptMapping)
      return getKeyFields().compareTo(((ConceptMapping)o).getKeyFields());
    else
      return 0;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui1, ver, rel,
   * rela, mapreason, cui2, and mapin.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(25);
      sb1.append(getCui1())
         .append(getCui2() != null ? getCui2() : "")
         .append(getRel() != null ? getRel().getAbbreviation() : "")
         .append(getVer());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

}
