/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       CoocurrenceContextImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.CoocurrenceContext;
import gov.nih.nlm.umls.meta.CoocurrenceGroup;
import gov.nih.nlm.umls.meta.CoocurrenceType;
import gov.nih.nlm.umls.meta.QualifierFrequency;
import gov.nih.nlm.umls.meta.Source;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.CoocurrenceContext}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class CoocurrenceContextImpl implements CoocurrenceContext {

  //
  // Fields
  //
  private CoocurrenceGroup coocurrence_group;
  private CoocurrenceType cot;
  private int cof = 0;
  private List qualifiers = new ArrayList();
  private Source source;
  private ContentView cvf;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link CoocurrenceContext}.
   */
  public CoocurrenceContextImpl() {};

  /**
   * Instantiates a {@link CoocurrenceContext} from the specified {@link CoocurrenceContext}.
   * @param cc {@link CoocurrenceContext} to be copied
   */
  public CoocurrenceContextImpl(CoocurrenceContext cc) {
    setCoocurrenceGroup(cc.getCoocurrenceGroup());
    setCoocurrenceType(cc.getCoocurrenceType());
    setCof(cc.getCof());
    setQualifiers(cc.getQualifiers());
    setSource(cc.getSource());
    setCvf(cc.getCvf());
  }

  /**
   * Returns the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}.
   * @return the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}
   */
  public CoocurrenceGroup getCoocurrenceGroup() {
    return coocurrence_group;
  }

  /**
   * Sets the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}.
   * @param group the {@link CoocurrenceGroup} that holds this {@link CoocurrenceContext}
   */
  public void setCoocurrenceGroup(CoocurrenceGroup group) {
    coocurrence_group = group;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link CoocurrenceType} (COT).
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCoocurrenceType() {
    return cot;
  }

  /**
   * Shorthand for {@link #getCoocurrenceType()}.
   * @return the {@link CoocurrenceType}
   */
  public CoocurrenceType getCot() {
    return getCoocurrenceType();
  }

  /**
   * Sets the {@link CoocurrenceType} (COT).
   * @param cot the {@link CoocurrenceType}
   */
  public void setCoocurrenceType(CoocurrenceType cot) {
    this.cot = cot;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setCoocurrenceType(CoocurrenceType)}.
   * @param cot the {@link CoocurrenceType}
   */
  public void setCot(CoocurrenceType cot) {
    setCoocurrenceType(cot);
  }

  /**
   * Returns the CoocurrenceContext frequency (COF).
   * @return the CoocurrenceContext frequency
   */
  public int getCoocurrenceContextFrequency() {
    return cof;
  }

  /**
   * Shorthand for {@link #getCoocurrenceContextFrequency()}.
   * @return the CoocurrenceContext frequency
   */
  public int getCof() {
    return getCoocurrenceContextFrequency();
  }

  /**
   * Sets the CoocurrenceContext frequency (COF).
   * @param cof the CoocurrenceContext frequency
   */
  public void setCoocurrenceContextFrequency(int cof) {
    this.cof = cof;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setCoocurrenceContextFrequency(int)}.
   * @param cof the CoocurrenceContext frequency
   */
  public void setCof(int cof) {
    setCoocurrenceContextFrequency(cof);
  }

  /**
   * Returns the {@link QualifierFrequency} list.
   * @return an array of {@link QualifierFrequency}s
   */
  public QualifierFrequency [] getQualifiers() {
    return (QualifierFrequency[])qualifiers.toArray(new QualifierFrequency[0]);
  }

  /**
   * Returns the {@link QualifierFrequency} values as a {@link List}.
   * @return the {@link QualifierFrequency} values as a {@link List}
   */
  public List getQualifierAsList() {
    return new ArrayList(qualifiers);
  }

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified array.
   * @param quals an array of {@link QualifierFrequency} values
   */
  public void setQualifiers(QualifierFrequency [] quals) {
    qualifiers.clear();
    for (int i=0; i<quals.length; i++) {
      qualifiers.add(quals[i]);
    }
  }

  /**
   * Replaces the {@link QualifierFrequency} values with those in the
   * specified {@link List}
   * @param quals a {@link List} of {@link QualifierFrequency} values
   */
  public void setQualifiers(List quals) {
    qualifiers = new ArrayList(quals);
  }

  /**
   * Adds the specified {@link QualifierFrequency}.
   * @param qual the {@link QualifierFrequency} to be added
   */
  public void addQualifier(QualifierFrequency qual) {
    qualifiers.add(qual);
  }

  /**
   * Removes the specified {@link QualifierFrequency}.
   * @param qual the {@link QualifierFrequency} to be removed
   */
  public void removeQualifier(QualifierFrequency qual) {
    qualifiers.remove(qual);
  }

  /**
   * Removes all {@link QualifierFrequency}s.
   */
  public void clearQualifiers() {
    qualifiers.clear();
  }

  /**
   * Returns the source for this {@link CoocurrenceContext}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link CoocurrenceContext}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link CoocurrenceContext} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link CoocurrenceContext} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cot and
   * versioned source.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(100);
      sb1.append(getSource() != null ? getSource().getVsab() : "")
        .append(getCot() != null ? getCot().getAbbreviation() : "");
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link CoocurrenceContext} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof CoocurrenceContext)) return false;
    return getKeyFields().equals(((CoocurrenceContext)o).getKeyFields());
  }

  /**
   * Compares this {@link Source} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof CoocurrenceContext)
      return getKeyFields().compareTo(((Atom)o).getKeyFields());
    else
      return 0;
  }

}