/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       CoocurrenceGroupImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.CoocurrenceContext;
import gov.nih.nlm.umls.meta.CoocurrenceGroup;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.CoocurrenceGroup}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class CoocurrenceGroupImpl implements CoocurrenceGroup {

  //
  // Fields
  //
  private Atom string_source;
  private Atom related_string_source;
  private Concept related_concept;
  private Concept concept;
  private List coocurrence_contexts = new ArrayList();

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link CoocurrenceGroup}.
   */
  public CoocurrenceGroupImpl() {};

  /**
   * Instantiates a {@link CoocurrenceGroup} from the specified {@link CoocurrenceGroup}.
   * @param cg {@link CoocurrenceGroup} to be copied
   */
  public CoocurrenceGroupImpl(CoocurrenceGroup cg) {
    setAtom(cg.getAtom());
    setRelatedAtom(cg.getRelatedAtom());
    setRelatedConcept(cg.getRelatedConcept());
    setConcept(cg.getConcept());
    setCoocurrenceContexts(cg.getCoocurrenceContexts());
  }

  /**
   * Returns the {@link Atom} this {@link CoocurrenceGroup} is connected to.
   * @return the {@link Atom} this {@link CoocurrenceGroup} is connected to
   */
  public Atom getAtom() {
    return string_source;
  }

  /**
   * Sets the {@link Atom} this {@link CoocurrenceGroup} is connected to.
   * @param string_source  the {@link Atom} this {@link CoocurrenceGroup} is connected to
   */
  public void setAtom(Atom string_source) {
    this.string_source = string_source;
    key_fields_flag = false;
  }

  /**
   * Returns the related concept involved.
   * @return the related concept involved
   */
  public Concept getRelatedConcept() {
    return related_concept;
  }

  /**
   * Sets the second concept involved.
   * @param related_concept the related concept involved
   */
  public void setRelatedConcept(Concept related_concept) {
    this.related_concept = related_concept;
    key_fields_flag = false;
  }

  /**
   * Returns the related {@link Atom} involved.
   * @return the related {@link Atom} involved
   */
  public Atom getRelatedAtom() {
    return related_string_source;
  }

  /**
   * Sets the related {@link Atom} involved.
   * @param related_string_source the related {@link Atom} involved
   */
  public void setRelatedAtom(Atom related_string_source) {
    this.related_string_source = related_string_source;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link CoocurrenceContext}s .
   * @return an array of {@link CoocurrenceContext}s
   */
  public CoocurrenceContext [] getCoocurrenceContexts() {
    return (CoocurrenceContext[])coocurrence_contexts.toArray(
      new CoocurrenceContext[0]);
  }

  /**
   * Returns the {@link CoocurrenceContext}s as a {@link List}.
   * @return the {@link CoocurrenceContext}s as a {@link List}
   */
  public List getCoocurrenceContextsAsList() {
    return new ArrayList(coocurrence_contexts);
  }

  /**
   * Sets the coocurrence attribute value. This is a comma separated
   * list of name=value pairs.
   * @param cxt_contexts the coocuurence attribute value
   */
  public void setCoocurrenceContexts(String cxt_contexts) {
    StringTokenizer st = new StringTokenizer(cxt_contexts, ",");
     while (st.hasMoreTokens()) {
         coocurrence_contexts.add(st.nextToken());
     }
    key_fields_flag = false;
  }

  /**
   * Replaces the {@link CoocurrenceContext}s with those in the
   * specified array.
   * @param coc_contexts an array of {@link CoocurrenceContext}s
   */
  public void setCoocurrenceContexts(CoocurrenceContext [] coc_contexts) {
    coocurrence_contexts.clear();
    for (int i=0; i<coc_contexts.length; i++) {
      coocurrence_contexts.add(coc_contexts[i]);
    }
  }

  /**
   * Replaces the {@link CoocurrenceContext}s with those in the
   * specified {@link List}
   * @param coc_contexts a {@link List} of {@link CoocurrenceContext}s
   */
  public void setCoocurrenceContexts(List coc_contexts) {
    coocurrence_contexts = new ArrayList(coc_contexts);
  }

  /**
   * Adds the specified {@link CoocurrenceContext}.
   * @param coc_context the {@link CoocurrenceContext} to be added
   */
  public void addCoocurrenceContext(CoocurrenceContext coc_context) {
    coocurrence_contexts.add(coc_context);
  }

  /**
   * Removes the specified {@link CoocurrenceContext}.
   * @param coc_context the {@link CoocurrenceContext} to be removed
   */
  public void removeCoocurrenceContext(CoocurrenceContext coc_context) {
    coocurrence_contexts.remove(coc_context);
  }

  /**
   * Removes all {@link CoocurrenceContext}s.
   */
  public void clearCoocurrenceContexts() {
    coocurrence_contexts.clear();
  }

  /**
   * Returns the {@link Concept} that holds this {@link CoocurrenceGroup}.
   * @return the {@link Concept} that holds this {@link CoocurrenceGroup}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link CoocurrenceGroup}.
   * @param concept the {@link Concept} that holds this {@link CoocurrenceGroup}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui1, aui1, cui2,
   * and aui2.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(50);
      sb1.append(getConcept().getCui())
         .append(getAtom() != null ? getAtom().getAui() : "")
         .append(getRelatedConcept() != null ? getRelatedConcept().getCui() : "")
         .append(getRelatedAtom() != null ? getRelatedAtom().getAui() : "");
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link CoocurrenceGroup} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof CoocurrenceGroup)) return false;
    return getKeyFields().equals(((CoocurrenceGroup)o).getKeyFields());
  }

  /**
   * Compares this {@link CoocurrenceGroup} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof CoocurrenceGroup)
      return getKeyFields().compareTo(((CoocurrenceGroup)o).getKeyFields());
    else
      return 0;
  }

}
