/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       DefinitionImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.Suppressible;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.Definition}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class DefinitionImpl implements Definition {

  //
  // Fields
  //
  private Concept concept;
  private Atom atom;
  private String atui;
  private String satui;
  private Source source;
  private String def;
  private Suppressible suppress;
  private ContentView cvf;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link Definition}.
   */
  public DefinitionImpl() {};

  /**
   * Instantiates a {@link Definition} from the specified {@link Definition}.
   * @param def {@link Definition} to be copied
   */
  public DefinitionImpl(Definition def) {
    setAtom(def.getAtom());
    setAtui(def.getAtui());
    setSatui(def.getSatui());
    setConcept(def.getConcept());
    setSource(def.getSource());
    setCvf(def.getCvf());
    setSuppress(def.getSuppress());
    setDef(def.getDef());
  }

  /**
   * Returns the {@link Atom} related to this {@link Definition}.
   * @return the {@link Atom} related to this {@link Definition}
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the {@link Atom} related to this {@link Definition}.
   * @param atom the {@link Atom} related to this {@link Definition}
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }

  /**
   * Returns the unique identifier (ATUI).
   * @return the ATUI
   */
  public String getAttributeIdentifier() {
    return atui;
  }

  /**
   * Shorthand for {@link #getAttributeIdentifier()}.
   * @return the ATUI
   */
  public String getAtui() {
    return getAttributeIdentifier();
  }

  /**
   * Sets the unique identifier (ATUI).
   * @param atui the ATUI
   */
  public void setAttributeIdentifier(String atui) {
    this.atui = atui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeIdentifier(String)}.
   * @param atui the ATUI
   */
  public void setAtui(String atui) {
    setAttributeIdentifier(atui);
  }

  /**
   * Returns the source asserted attribute identifier (SATUI).
   * @return the SATUI
   */
  public String getSourceAttributeIdentifier() {
    return satui;
  }

  /**
   * Shorthand for {@link #getSourceAttributeIdentifier()}.
   * @return the SATUI
   */
  public String getSatui() {
    return getSourceAttributeIdentifier();
  }

  /**
   * Sets the source asserted attribute identifier (SATUI).
   * @param satui the SATUI
   */
  public void setSourceAttributeIdentifier(String satui) {
    this.satui = satui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceAttributeIdentifier(String)}.
   * @param satui the SATUI
   */
  public void setSatui(String satui) {
    setSourceAttributeIdentifier(satui);
  }

  /**
   * Returns the {@link Concept} that holds this {@link Definition}.
   * @return the {@link Concept} that holds this {@link Definition}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Definition}.
   * @param concept the {@link Concept} that holds this {@link Definition}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link Definition}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link Definition}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns the definition value (DEF).
   * @return the definition value (DEF).
   */
  public String getDefinition() {
    return def;
  }

  /**
   * Shorthand for {@link #getDefinition()}.
   * @return the definition value (DEF).
   */
  public String getDef() {
    return getDefinition();
  }

  /**
   * Sets the definition value (DEF).
   * @param def the definition value
   */
  public void setDefinition(String def) {
    this.def = def;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setDefinition(String)}.
   * @param def the definition value
   */
  public void setDef(String def) {
    setDefinition(def);
  }


  /**
   * Returns the suppressibility value (SUPPRESS).
   * @return the {@link Suppressible} value
   */
  public Suppressible getSuppress() {
    return suppress;
  }

  /**
   * Indicates whether or not this {@link Atom} is suppressible.
   * @return <code>true</code> if suppressible,
   *         <code>false</code> if not suppressible
   */
  public boolean isSuppressible() {
    return (suppress != null && !suppress.getAbbreviation().equals("N"));
  }

  /**
   * Sets the suppressibility value (SUPPRESS).
   * @param suppress the {@link Suppressible} value
   */
  public void setSuppress(Suppressible suppress) {
    this.suppress = suppress;
    key_fields_flag = false;
  }

  /**
   * Indicates whether or not this {@link DefinitionImpl} is obsolete.
   * @return <code>true</code> if obsolete,
   *         <code>false</code> if not obsolete
   */
  public boolean isObsolete() {
    if (getSuppress().getAbbreviation().equals("O"))
      return true;
    return false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Definition} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Definition} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, atui,
   * versioned source, satui, and def.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(200);
      sb1.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getAtom() != null ? getAtom().getAui() : "")
         .append(getSource() != null ? getSource().getVsab() : "")
         .append(getAtui())
         .append(getSatui())
         .append(getDef());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link Definition} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Definition)) return false;
    return getKeyFields().equals(((Definition)o).getKeyFields());
  }

  /**
   * Compares this {@link Definition} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Definition)
      return getKeyFields().compareTo(((Definition)o).getKeyFields());
    else
      return 0;
  }

}
