/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       MappingImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.MapObject;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceIdentifierType;

/**
 * Default implementation of {@link gov.nih.nlm.umls.meta.Mapping}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class MapObjectImpl implements MapObject {

  //
  // Fields
  //
  private String id;
  private String sid;
  private String expr;
  private SourceIdentifierType type;
  private String rule;
  private String res;
  private Source source;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link MappingImpl}.
   */
  public MapObjectImpl() {};

  /**
   * Instantiates a {@link MapObject} from the specified {@link MapObject}.
   * @param mapping the {@link MapObject} to copy
   */
  public MapObjectImpl(MapObject mapping) {
    id = mapping.getId();
    sid = mapping.getSid();
    expr = mapping.getExpr();
    type = mapping.getType();
    rule = mapping.getRule();
    res = mapping.getRestriction();
  }

  /**
   * Returns the source asserted unique identifier (FROMSID or TOSID).
   * @return the source asserted unique identifier
   */
  public String getSourceIdentifier() {
    return sid;
  }

  /**
   * Sets the source asserted unique identifier (FROMSID or TOSID).
   * @param sid source asserted unique identifier
   */
  public void setSourceIdentifier(String sid) {
    this.sid = sid;
    key_fields_flag = false;
  }

  /**
   * Returns the source asserted unique identifier (FROMSID or TOSID).
   * This method is shorthand for {@link #getSourceIdentifier()}.
   * @return the source asserted unique identifier
   */
  public String getSid() {
    return getSourceIdentifier();
  }

  /**
   * Sets the source asserted unique identifier (FROMSID or TOSID).
   * This method is shorthand for {@link #setSourceIdentifier(String)}.
   * @param sid source asserted unique identifier
   */
  public void setSid(String sid) {
    setSourceIdentifier(sid);
  }

  /**
   * Returns the unique identifier (FROMID or TOID).
   * @return the unique identifier
   */
  public String getIdentifier() {
    return id;
  }

  /**
   * Sets the unique identifier (FROMID or TOID).
   * @param id unique identifier
   */
  public void setIdentifier(String id) {
    this.id = id;
    key_fields_flag = false;
  }

  /**
   * Returns the unique identifier (FROMID or TOID).
   * This method is shorthand for {@link #getIdentifier()}.
   * @return the unique identifier
   */
  public String getId() {
    return getIdentifier();
  }

  /**
   * Sets the unique identifier (FROMID or TOID).
   * This method is shorthand for {@link #setIdentifier(String)}.
   * @param id unique identifier
   */
  public void setId(String id) {
    setIdentifier(id);
  }

  /**
   * Returns the expression (FROMEXPR or TOEXPR).
   * @return the expression
   */
  public String getExpression() {
    return expr;
  }

  /**
   * Sets the expression (FROMEXPR or TOEXPR).
   * @param expr expression
   */
  public void setExpression(String expr) {
    this.expr = expr;
    key_fields_flag = false;
  }

  /**
   * Returns the expression (FROMEXPR or TOEXPR).
   * This method is shorthand for {@link #getExpression()}.
   * @return the expression
   */
  public String getExpr() {
    return getExpression();
  }

  /**
   * Sets the expression (FROMEXPR or TOEXPR).
   * This method is shorthand for {@link #setExpression(String)}.
   * @param expr expression
   */
  public void setExpr(String expr) {
    setExpression(expr);
  }

  /**
   * Returns the type {@link SourceIdentifierType} (FROMTYPE or TOTYPE).
   * @return the type
   */
  public SourceIdentifierType getType() {
    return type;
  }

  /**
   * Sets the type (FROMTYPE or TOTYPE).
   * @param type {@link SourceIdentifierType}
   */
  public void setType(SourceIdentifierType type) {
    this.type = type;
    key_fields_flag = false;
  }

  /**
   * Returns the machine processable rule for when mapping
   * is valid (FROMRULE or TORULE).
   * @return the rule
   */
  public String getRule() {
    return rule;
  }

  /**
   * Sets the machine processable rule for when mapping is
   * valid (FROMRULE or TORULE).
   * @param rule the rule
   */
  public void setRule(String rule) {
    this.rule = rule;
    key_fields_flag = false;
  }

  /**
   * Returns the restriction (FROMRES or TORES).
   * @return the restriction
   */
  public String getRestriction() {
    return res;
  }

  /**
   * Sets the restriction (FROMRES or TORES).
   * @param res restriction
   */
  public void setRestriction(String res) {
    this.res = res;
    key_fields_flag = false;
  }

  /**
   * Returns the restriction (FROMRES or TORES).
   * This method is shorthand for {@link #getRestriction()}.
   * @return the restriction
   */
  public String getRes() {
    return getRestriction();
  }

  /**
   * Sets the restriction (FROMRES or TORES).
   * This method is shorthand for {@link #setRestriction(String)}.
   * @param res restriction
   */
  public void setRes(String res) {
    setRestriction(res);
  }

  /**
   * Returns the source for this {@link MapObject}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link MapObject}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the ui, expr, type,
   * rule, restriction and source fields
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb = new StringBuffer(100);
      sb.append(getId())
         .append(getSid())
         .append(getExpr())
         .append(getType() != null ? getType().getAbbreviation() : "")
         .append(getRule())
         .append(getRestriction())
         .append(getSource() != null ? getSource().getVsab() : "");
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Returns <code>true</code> if the specified object is equal to this object.
   * The specified object must be a {@link MapObject} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal;
   *                 <code>false</code> otherwise.
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof MapObject)) return false;
    return getKeyFields().equals(((MapObject) o).getKeyFields());
  }

  /**
   * Returns a hash code based on key fields.
   * @return a hash code based on key fields.
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Compares this {@link MapObject} with the specified object for order.
   * Returns a negative integer, zero, or a positive integer as this object is
   * less than, equal to, or greater than the specified object.  Results
   * are indeterminate if the specified parameter is not of the same class.
   * @param o the Object to be compared.
   * @return a negative integer, zero, or a positive integer as this object
   *         is less than, equal to, or greater than the specified object.
   *         Returns zero if the specified object's type prevents it from being
   *         compared to this Object.
   */
  public int compareTo(Object o) {
    if (o instanceof MapObject)
      return getKeyFields().compareTo(((MapObject) o).getKeyFields());
    else
      return 0;
  }
}