/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       MappingImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AdditionalRelationshipLabel;
import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.MapObject;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.RelationshipLabel;
import gov.nih.nlm.umls.meta.Source;

/**
 * Default implementation of {@link gov.nih.nlm.umls.meta.Mapping}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class MappingImpl implements Mapping {

  //
  // Fields
  //
  private RelationshipLabel rel;
  private AdditionalRelationshipLabel rela;
  private Concept concept;
  private Source source;
  private String subsetid;
  private String rank;
  private String id;
  private String sid;
  private MapObject from;
  private MapObject to;
  private String rule;
  private String res;
  private String type;
  private AttributeName atn;
  private String atv;
  private ContentView content_view;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link MappingImpl}.
   */
  public MappingImpl() {};

  /**
   * Instantiates a {@link Mapping} from the specified {@link Mapping}.
   * @param mapping the {@link Mapping} to copy
   */
  public MappingImpl(Mapping mapping) {
    rel = mapping.getRel();
    rela = mapping.getRela();
    concept = mapping.getConcept();
    source = mapping.getSource();
    subsetid = mapping.getSubsetId();
    rank = mapping.getRank();
    from = new MapObjectImpl(mapping.getFrom());
    to = new MapObjectImpl(mapping.getTo());
    rule = mapping.getRule();
    type = mapping.getType();
    atn = mapping.getAttributeName();
    atv = mapping.getAttributeValue();
    content_view = mapping.getContentView();
  }

  /**
   * Returns the {@link Concept} that holds this {@link Mapping} (CUI).
   * @return the {@link Concept} that holds this {@link Mapping}
   */
  public Concept getConcept() { return concept; }

  /**
   * Sets the {@link Concept} that holds this {@link Mapping} (CUI).
   * @param concept the {@link Concept} that holds this {@link Mapping}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getRelationshipName()}.
   * @return the result of <code>getRelationshipName()</code>
   */
  public RelationshipLabel getRel() { return getRelationshipName(); }

  /**
   * Shorthand for {@link #setRelationshipName(RelationshipLabel)}.
   * @param rel the {@link RelationshipLabel}
   */
  public void setRel(RelationshipLabel rel) { setRelationshipName(rel); }

  /**
   * Returns the {@link RelationshipLabel} (REL).
   * @return the {@link RelationshipLabel}
   */
  public RelationshipLabel getRelationshipName() { return rel; }

  /**
   * Sets the {@link RelationshipLabel} (REL).
   * @param rel the {@link RelationshipLabel}
   */
  public void setRelationshipName(RelationshipLabel rel) { this.rel = rel; }

  /**
   * Returns the {@link AdditionalRelationshipLabel} (RELA).
   * @return the {@link AdditionalRelationshipLabel}
   */
  public AdditionalRelationshipLabel getAdditionalRelationshipLabel() { return rela; }

  /**
   * Shorthand for {@link #getAdditionalRelationshipLabel()}.
   * @return the result of <code>getAdditionalRelationshipLabel()</code>
   */
  public AdditionalRelationshipLabel getRela() { return rela; }

  /**
   * Sets the {@link AdditionalRelationshipLabel} (RELA).
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setAdditionalRelationshipLabel(AdditionalRelationshipLabel rela) {
    this.rela = rela;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAdditionalRelationshipLabel(AdditionalRelationshipLabel)}.
   * @param rela the {@link AdditionalRelationshipLabel}
   */
  public void setRela(AdditionalRelationshipLabel rela) {
    setAdditionalRelationshipLabel(rela);
    key_fields_flag = false;
  }

  /**
   * Returns the mapped from {@link MapObject}.
   * @return the mapped from {@link MapObject}
   */
  public MapObject getFrom() { return from; }

  /**
   * Sets the mapped from {@link MapObject}.
   * @param from the mapped from {@link MapObject}
   */
  public void setFrom(MapObject from) {
    this.from = from;
    key_fields_flag = false;
  }

  /**
   * Returns the mapped to {@link MapObject}.
   * @return the mapped to {@link MapObject}
   */
  public MapObject getTo() { return to; }

  /**
   * Sets the mapped to {@link MapObject}.
   * @param to the mapped to {@link MapObject}
   */
  public void setTo(MapObject to) {
    this.to = to;
    key_fields_flag = false;
  }

  /**
   * Returns the subset identifier (MAPSUBSETID).
   * @return the subset identifier
   */
  public String getSubsetIdentifier() { return subsetid; }

  /**
   * Shorthand for {@link #getSubsetIdentifier()}.
   * @return the result of <code>getSubsetIdentifier()</code>
   */
  public String getSubsetId() { return getSubsetIdentifier(); }

  /**
   * Sets the subset identifier (MAPSUBSETID).
   * @param subsetid subset identifier
   */
  public void setSubsetIdentifier(String subsetid) {
    this.subsetid = subsetid;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSubsetIdentifier(String)}.
   * @param subsetid subset identifier
   */
  public void setSubsetId(String subsetid) { setSubsetIdentifier(subsetid); }

  /**
   * Returns the rank (RANK).
   * @return the rank
   */
  public String getRank() { return rank; }

  /**
   * Sets the rank (RANK).
   * @param rank priority
   */
  public void setRank(String rank) {
    this.rank = rank;
    key_fields_flag = false;
  }

  /**
   * Returns the source asserted unique identifier (MAPSID).
   * @return the source asserted unique identifier
   */
  public String getSourceIdentifier() {
    return sid;
  }

  /**
   * Sets the source asserted unique identifier (MAPSID).
   * @param sid source asserted unique identifier
   */
  public void setSourceIdentifier(String sid) {
    this.sid = sid;
    key_fields_flag = false;
  }

  /**
   * Returns the source asserted unique identifier (MAPSID).
   * This method is shorthand for {@link #getSourceIdentifier()}.
   * @return the source asserted unique identifier
   */
  public String getSid() { return getSourceIdentifier();}

  /**
   * Sets the source asserted unique identifier (MAPSID).
   * This method is shorthand for {@link #setSourceIdentifier(String)}.
   * @param sid source asserted unique identifier
   */
  public void setSid(String sid) {setSourceIdentifier(sid);}

  /**
   * Returns the unique identifier (MAPID).
   * @return the unique identifier
   */
  public String getIdentifier() {
    return id;
  }

  /**
   * Sets the unique identifier (MAPID).
   * @param id unique identifier
   */
  public void setIdentifier(String id) {
    this.id = id;
    key_fields_flag = false;
  }

  /**
   * Returns the unique identifier (MAPID).
   * This method is shorthand for {@link #getIdentifier()}.
   * @return the unique identifier
   */
  public String getId() { return getIdentifier(); }

  /**
   * Sets the unique identifier (MAPID).
   * This method is shorthand for {@link #setIdentifier(String)}.
   * @param id unique identifier
   */
  public void setId(String id) {setIdentifier(id);}

  /**
   * Returns the machine processable rule for when to apply mapping (MAPRULE).
   * @return the rule
   */
  public String getRule() { return rule; }

  /**
   * Sets the machine processable rule for when to apply mapping (MAPRULE).
   * @param rule rule for when to apply mapping
   */
  public void setRule(String rule) {
    this.rule = rule;
    key_fields_flag = false;
  }

  /**
   * Returns the human readable rule for use of mapping (MAPRES).
   * @return the map restriction
   */
  public String getRestriction() {
    return res;
  }

  /**
   * Sets the human readable rule for use of mapping (MAPRES).
   * @param res {@link String}
   */
  public void setRestriction(String res) {
    this.res = res;
    key_fields_flag = false;
  }

  /**
   * Returns the human readable rule for use of mapping (MAPRES).
   * This is shorthand for {@link #getRestriction()}.
   * @return the human readable rule for use of mapping
   */
  public String getRes() {return getRestriction();}

  /**
   * Sets the human readable rule for use of mapping (MAPRES).
   * This is shorthand for {@link #setRestriction(String)}.
   * @param res the human readable rule for use of mapping
   */
  public void setRes(String res) {setRestriction(res);}

  /**
   * Returns the {@link AttributeName} (MAPATN).
   * @return the {@link AttributeName}
   */
  public AttributeName getAttributeName() {
    return atn;
  }

  /**
   * Sets the {@link AttributeName} (MAPATN).
   * @param atn the {@link AttributeName}
   */
  public void setAttributeName(AttributeName atn) {
    this.atn = atn;
    key_fields_flag = false;
  }

  /**
   * Returns the attribute value (MAPATV).
   * @return the attribute value
   */
  public String getAttributeValue() {
    return atv;
  }

  /**
   * Sets the attribute value (MAPATV).
   * @param atv the attribute value
   */
  public void setAttributeValue(String atv) {
    this.atv = atv;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link Source}.
   * @return the source
   */
  public Source getSource() { return source; }

  /**
   * Sets the {@link Source}.
   * @param source the source
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns the type of mapping (MAPTYPE).
   * @return the type of mapping
   */
  public String getType() {
    return type;
  }

  /**
   * Sets the type of mapping (MAPTYPE).
   * @param type the type of mapping
   */
  public void setType(String type) {
    this.type = type;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Mapping} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() { return content_view; }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Mapping} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    content_view = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return the result of <code>getContentView()</code>
   */
  public ContentView getCvf() { return getContentView(); }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) { setContentView(cvf); }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the mapsetcui, mapsetsab,
   * mapsubsetid, maprank, from fields, rel, rela, to fields,
   * maprule, maptype, mapatn, mapatv, cvf  fields
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb = new StringBuffer(100);
      sb.append(getConcept() != null ? getConcept().getCui() : "")
         /*.append(getSource() != null ? getSource().getVsab() : "")
         .append(getSubsetId())
         .append(getRank())
         .append(getFrom() != null ? getFrom().getKeyFields() : "")
         .append(getRel() != null ? getRel().getAbbreviation() : "")
         .append(getRela() != null ? getRela().getAbbreviation() : "")
         .append(getTo() != null ? getTo().getKeyFields() : "")
         .append(getRule())
         .append(getType())
         .append(getAttributeName() != null ?
            getAttributeName().getAbbreviation() : "")
         .append(getAttributeValue());*/
         .append(getIdentifier());
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link Mapping} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Mapping)) return false;
    return getKeyFields().equals(((Mapping)o).getKeyFields());
  }

  /**
   * Compares this {@link Mapping} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Mapping)
      return getKeyFields().compareTo(((Mapping)o).getKeyFields());
    else
      return 0;
  }


}