/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       PlaceholderAtom.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.Restrictor;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.StringType;
import gov.nih.nlm.umls.meta.Suppressible;
import gov.nih.nlm.umls.meta.TermStatus;
import gov.nih.nlm.umls.meta.TermType;
import gov.nih.nlm.umls.meta.TermTypeGroup;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * Implementation of the {@link gov.nih.nlm.umls.meta.Atom}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class PlaceholderAtom implements Atom {

  //
  // Fields
  //
  protected Concept concept;
  protected TermType tty;
  protected Language lat;
  protected TermStatus ts;
  protected String lui;
  protected StringType stt;
  protected String sui;
  protected String str;
  protected String saui;
  protected String scui;
  protected String sdui;
  protected String code;
  protected String aui;
  protected Suppressible suppress;
  protected String ispref;

  protected Source source;
  protected ContentView cvf;

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link PlaceholderAtom}.
   */
  public PlaceholderAtom() {};

  /**
   * Instantiates a {@link PlaceholderAtom} from the specified {@link Atom}.
   * @param atom {@link Atom} to be copied
   */
  public PlaceholderAtom(Atom atom) {
    setTty(atom.getTty());
    setSaui(atom.getSaui());
    setAui(atom.getAui());
    setScui(atom.getScui());
    setSdui(atom.getSdui());
    setSuppress(atom.getSuppress());
    setSui(atom.getSui());
    setConcept(atom.getConcept());
    setSource(atom.getSource());
    setCvf(atom.getCvf());
    setLui(atom.getLui());
    setCode(atom.getCode());
    setTs(atom.getTs());
    setLat(atom.getLat());
    setStt(atom.getStt());
    setStr(atom.getStr());
    setIspref(atom.getIspref());
  }

  /**
   * Returns the {@link Concept} that holds this {@link Atom}.
   * @return the {@link Concept} that holds this {@link Atom}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Atom}.
   * @param concept the {@link Concept} that holds this {@link Atom}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns the {@link TermType} (TTY).
   * @return the {@link TermType}
   */
  public TermType getTermType() {
    return tty;
  }

  /**
   * Shorthand for {@link #getTermType()}.
   * @return the {@link TermType}
   */
  public TermType getTty() {
    return getTermType();
  }

  /**
   * Sets the {@link TermType} (TTY).
   * @param tty the {@link TermType}
   */
  public void setTermType(TermType tty) {
    this.tty = tty;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setTermType(TermType)}.
   * @param tty the {@link TermType}
   */
  public void setTty(TermType tty) {
    setTermType(tty);
  }

  /**
   * Returns the {@link Language} (LAT).
   * @return the {@link Language}
   */
  public Language getLanguage() {
    return lat;
  }

  /**
   * Shorthand for {@link #getLanguage()}.
   * @return {@link Language}
   */
  public Language getLat() {
    return getLanguage();
  }

  /**
   * Sets the {@link Language} (LAT).
   * @param lat the {@link Language}
   */
  public void setLanguage(Language lat) {
    this.lat = lat;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setLanguage(Language)}.
   * @param lat the {@link Language}
   */
  public void setLat(Language lat) {
    setLanguage(lat);
  }

  /**
   * Indicates whether or not this atom is part of the preferred lexical class
   * of the {@link Concept}.  In other words, does this {@link Atom} have
   * the same LUI as the preferred {@link Atom} of the {@link Concept}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredTermIdentifier() {
    return getTs().getAbbreviation().toLowerCase().equals("p");
  }

  /**
   * Shorthand for {@link #hasPreferredTermIdentifier()}
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredLui() {
    return hasPreferredTermIdentifier();
  }

  /**
   * Returns the {@link TermStatus} (TS).
   * @return the {@link TermStatus}
   */
  public TermStatus getTermStatus() {
    return ts;
  }

  /**
   * Shorthand for {@link #getTermStatus()}.
   * @return the {@link TermStatus}
   */
  public TermStatus getTs() {
    return getTermStatus();
  }

  /**
   * Sets the {@link TermStatus} (TS).
   * @param ts the {@link TermStatus}
   */
  public void setTermStatus(TermStatus ts) {
    this.ts = ts;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setTermStatus(TermStatus)}.
   * @param ts the {@link TermStatus}
   */
  public void setTs(TermStatus ts) {
    setTermStatus(ts);
  }

  /**
   * Shorthand for {@link #getTermIdentifier()}.
   * @return the LUI
   */
  public String getLui() {
    return getTermIdentifier();
  }

  /**
   * Shorthand for {@link #setTermIdentifier(String)}.
   * @param lui the LUI
   */
  public void setLui(String lui) {
    setTermIdentifier(lui);
  }

  /**
   * Returns the term identifier (LUI).
   * @return the LUI
   */
  public String getTermIdentifier() {
    return lui;
  }

  /**
   * Sets the term identifier (LUI).
   * @param lui the LUI
   */
  public void setTermIdentifier(String lui) {
    this.lui = lui;
    key_fields_flag = false;
  }

  /**
   * Indicates whether or not this atom has the preferred string identifier
   * of the lexical class it is part of.  In other words, does this {@link Atom} have
   * the same SUI as the preferred {@link Atom} of the lexical class.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredStringIdentifier() {
    return getStt().getAbbreviation().equals("PF");
  }

  /**
   * Shorthand for {@link #hasPreferredStringIdentifier()}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredSui() {
    return hasPreferredStringIdentifier();
  }

  /**
   * Returns the {@link StringType} (STT).
   * @return the {@link StringType}
   */
  public StringType getStringType() {
    return stt;
  }

  /**
   * Shorthand for {@link #getStringType()}.
   * @return the {@link StringType}
   */
  public StringType getStt() {
    return getStringType();
  }

  /**
   * Sets the {@link StringType} (STT).
   * @param stt the STT
   */
  public void setStringType(StringType stt) {
    this.stt = stt;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setStringType(StringType)}.
   * @param stt the STT
   */
  public void setStt(StringType stt) {
    setStringType(stt);
  }

  /**
   * Shorthand for {@link #getStringIdentifier()}.
   * @return the SUI
   */
  public String getSui() {
    return getStringIdentifier();
  }

  /**
   * Shorthand for {@link #setStringIdentifier(String)}.
   * @param sui the SUI
   */
  public void setSui(String sui) {
    setStringIdentifier(sui);
  }

  /**
   * Return the string identifier (SUI).
   * @return the SUI
   */
  public String getStringIdentifier() {
    return sui;
  }

  /**
   * Sets the string identifier (SUI).
   * @param sui the SUI
   */
  public void setStringIdentifier(String sui) {
    this.sui = sui;
    key_fields_flag = false;
  }

  /**
   * Sets the name (STR).
   * @return the name
   */
  public String getName() {
    return str;
  }

  /**
   * Shorthand for {@link #getName()}.
   * @return the name
   */
  public String getStr() {
    return getName();
  }

  /**
   * Sets the name (STR).
   * @param str the name
   */
  public void setName(String str) {
    this.str = str;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setName(String)}.
   * @param str the name
   */
  public void setStr(String str) {
    setName(str);
  }

  //
  // Definitions
  //

  /**
   * Returns the {@link Definition}s.
   * @return a {@link Definition} <code>[]</code>
   */
  public Definition[] getDefinitions() {
    return new Definition[0];
  }

  /**
   * Returns the {@link Definition}s as a {@link List}.
   * @return the {@link Definition}s as a {@link List}
   */
  public List getDefinitionsAsList() {
    return new ArrayList();
  }

  /**
   * Adds the specified {@link Definition}.
   * @param def the {@link Definition} to be added
   */
  public void addDefinition(Definition def) {
  }

  /**
   * Remove the specified {@link Definition}.
   * @param def the {@link Definition} to be removed
   */
  public void removeDefinition(Definition def) {
  }

  /**
   * Remove all {@link Definition}s.
   */
  public void clearDefinitions() {
  }

  /**
   * Replace {@link Definition}s with the ones in the specified array.
   * @param defs a {@link Definition} <code>[]</code>
   */
  public void setDefinitions(Definition defs[]) {
  }

  /**
   * Replace {@link Definition}s with the ones in the specified {@link List}.
   * @param defs a {@link List} of {@link Definition}s
   */
  public void setDefinitions(List defs) {
  }

  //
  // Locators
  //

  /**
   * Returns the {@link Locator}s.
   * @return a {@link Locator} <code>[]</code>
   */
  public Locator[] getLocators() {
     return new Locator[0];
  }

  /**
   * Returns the {@link Locator}s as a {@link List}.
   * @return the {@link Locator}s as a {@link List}
   */
  public List getLocatorsAsList() {
    return new ArrayList();
  }

  /**
   * Add the specified {@link Locator}.
   * @param loc the {@link Locator} to be added
   */
  public void addLocator(Locator loc) {
  }

  /**
   * Removes the specified {@link Locator}.
   * @param loc the {@link Locator} to be removed
   */
  public void removeLocator(Locator loc) {
  }

  /**
   * Remove all {@link Locator}s.
   */
  public void clearLocators() {
  }

  /**
   * Replace {@link Locator}s with the ones in the specified array.
   * @param locs a {@link Locator} <code>[]</code>
   */
  public void setLocators(Locator locs[]) {
  }

  /**
   * Replace {@link Locator}s with the ones in the specified {@link List}.
   * @param locs a {@link List} of {@link Locator}s
   */
  public void setLocators(List locs) {
  }

  //
  // Relationships
  //

  /**
   * Returns the {@link Relationship}s.
   * @return a {@link Relationship} <code>[]</code>
   */
  public Relationship[] getRelationships() {
    return new Relationship[0];
  }

  /**
   * Returns the {@link Relationship}s as a {@link List}.
   * @return the {@link Relationship}s as a {@link List}
   */
  public List getRelationshipsAsList() {
    return new ArrayList();
  }

  /**
   * Adds the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be added
   */
  public void addRelationship(Relationship rel) {
  }

  /**
   * Removes the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be removed
   */
  public void removeRelationship(Relationship rel) {
  }

  /**
   * Remove all {@link Relationship}s.
   */
  public void clearRelationships() {
  }

  /**
   * Replace {@link Relationship}s with the ones in the specified array.
   * @param rels a {@link Relationship} <code>[]</code>
   */
  public void setRelationships(Relationship rels[]) {
  }

  /**
   * Replace {@link Relationship}s of the ones in the specified {@link List}.
   * @param rels a {@link List} of {@link Relationship}s
   */
  public void setRelationships(List rels) {
  }

  //
  // Coocurrences
  //

  /**
   * Returns the {@link Coocurrence}s.
   * @return a {@link Coocurrence} <code>[]</code>
   */
  public Coocurrence[] getCoocurrences() {
    return new Coocurrence[0];
  }


  /**
   * Returns the {@link Coocurrence}s as a {@link List}.
   * @return the {@link Coocurrence}s as a {@link List}
   */
  public List getCoocurrencesAsList() {
    return new ArrayList();
  }

  /**
   * Adds the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be added
   */
  public void addCoocurrence(Coocurrence coc) {
  }

  /**
   * Removes the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be removed
   */
  public void removeCoocurrence(Coocurrence coc) {
  }

  /**
   * Remove all {@link Coocurrence}s.
   */
  public void clearCoocurrences() {
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified array.
   * @param cocs a {@link Coocurrence} <code>[]</code>
   */
  public void setCoocurrences(Coocurrence cocs[]){
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified {@link List}.
   * @param cocs a {@link List} of {@link Coocurrence}s
   */
  public void setCoocurrences(List cocs) {
  }

  //
  // Contexts
  //

  /**
   * Return the {@link Context}s.
   * @return a {@link Context} <code>[]</code>
   */
  public Context[] getContexts() {
    return new Context[0];
  }

  /**
   * Returns the {@link Context}s as a {@link List}.
   * @return the {@link Context}s as a {@link List}
   */
  public List getContextsAsList() {
    return new ArrayList();
  }

  /**
   * Adds the specified {@link Context}.
   * @param cxt the {@link Context} to be added
   */
  public void addContext(Context cxt) {
  }

  /**
   * Removes the specified {@link Context}.
   * @param cxt the {@link Context} to be removed
   */
  public void removeContext(Context cxt) {
  }

  /**
   * Remove all {@link Context}s.
   */
  public void clearContexts() {
  }

  /**
   * Replace {@link Context}s with the ones in the specified array.
   * @param cxts a {@link Context} <code>[]</code>
   */
  public void setContexts(Context cxts[]) {
  }

  /**
   * Replace {@link Context}s with the ones in the specified {@link List}.
   * @param cxts a {@link List} of {@link Context}s
   */
  public void setContexts(List cxts) {
  }

  //
  // AtomAttributes
  //

  /**
   * Returns the {@link AtomAttribute}s.
   * @return a {@link AtomAttribute} <code>[]</code>
   */
  public AtomAttribute[] getAttributes() {
    return new AtomAttribute[0];
  }

  /**
   * Returns the {@link AtomAttribute}s as a {@link List}.
   * @return the {@link AtomAttribute}s as a {@link List}
   */
  public List getAttributesAsList() {
      return new ArrayList();
  }

  /**
   * Adds the specified {@link AtomAttribute}.
   * @param att the {@link AtomAttribute} to be added
   */
  public void addAttribute(AtomAttribute att) {
  }

  /**
   * Removes the specified {@link AtomAttribute}.
   * @param att the {@link AtomAttribute} to be removed
   */
  public void removeAttribute(AtomAttribute att) {
  }

  /**
   * Remove all {@link AtomAttribute}s.
   */
  public void clearAttributes() {
  }

  /**
   * Replace {@link AtomAttribute}s with the ones in the specified array.
   * @param atts an {@link AtomAttribute} <code>[]</code>
   */
  public void setAttributes(AtomAttribute atts[]) {
  }

  /**
   * Replace {@link AtomAttribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link AtomAttribute}s
   */
  public void setAttributes(List atts) {
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, sui,
   * versioned source, tty, and code.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(80);
      sb1.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getSui())
         .append(getAui())
         .append(getSource() != null ? getSource().getVsab() : "")
         .append(getTty() != null ? getTty().getAbbreviation() : "")
         .append(getCode());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link Atom} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Atom)) return false;
    return getKeyFields().equals(((Atom)o).getKeyFields());
  }

  /**
   * Compares this {@link Atom} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Atom)
      return getKeyFields().compareTo(((Atom)o).getKeyFields());
    else
      return 0;
  }

  /**
   * Shorthand for {@link #getSourceAtomIdentifier()}.
   * @return the SAUI
   */
  public String getSaui() {
    return getSourceAtomIdentifier();
  }

  /**
   * Returns the source asserted identifier (SAUI) for this {@link Atom}.
   * @return the SAUI
   */
  public String getSourceAtomIdentifier() {
    return saui;
  }

  /**
   * Sets the source asserted identifier (SAUI) for this {@link Atom}.
   * @param saui the SAUI
   */
  public void setSourceAtomIdentifier(String saui) {
    this.saui = saui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceAtomIdentifier(String)}.
   * @param saui the SAUI
   */
  public void setSaui(String saui) {
    setSourceAtomIdentifier(saui);
  }

  /**
   * Shorthand for {@link #getSourceDescriptor()}.
   * @return the SDUI
   */
  public String getSdui() {
    return getSourceDescriptor();
  }

  /**
   * Returns the source asserted descriptor (SDUI).
   * @return the SDUI
   */
  public String getSourceDescriptor() {
    return sdui;
  }

  /**
   * Sets the source asserted descriptor (SDUI).
   * @param sdui the SDUI
   */
  public void setSourceDescriptor(String sdui) {
    this.sdui = sdui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setSourceDescriptor(String)}.
   * @param sdui the SDUI
   */
  public void setSdui(String sdui) {
    setSourceDescriptor(sdui);
  }

  /**
   * Shorthand for {@link #getSourceConceptIdentifier()}.
   * @return the SCUI
   */
  public String getScui() {
    return getSourceConceptIdentifier();
  }

  /**
   * Returns the source attributed concept identifier (SCUI).
   * @return the SCUI
   */
  public String getSourceConceptIdentifier() {
    return scui;
  }

  /**
   * Shorthand for {@link #setSourceConceptIdentifier(String)}.
   * @param scui the SCUI
   */
  public void setScui(String scui) {
    setSourceConceptIdentifier(scui);
  }

  /**
   * Sets the source attributed concept identifier (SCUI).
   * @param scui the SCUI
   */
  public void setSourceConceptIdentifier(String scui) {
    this.scui = scui;
    key_fields_flag = false;
  }

  /**
   * Return the code (CODE).
   * @return the code
   */
  public String getCode() {
    return code;
  }

  /**
   * Sets the code (CODE).
   * @param code the code
   */
  public void setCode(String code) {
    this.code = code;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getAtomIdentifier()}.
   * @return the AUI
   */
  public String getAui() {
    return getAtomIdentifier();
  }

  /**
   * Shorthand for {@link #setAtomIdentifier(String)}.
   * @param aui the AUI
   */
  public void setAui(String aui) {
    setAtomIdentifier(aui);
  }

  /**
   * Returns the atom unique identifier (AUI) related to this
   * {@link Atom}.
   * @return the AUI
   */
  public String getAtomIdentifier() {
    return aui;
  }

  /**
   * Sets the atom unique identifier (AUI) related to this
   * {@link Atom}.
   * @param aui the AUI
   */
  public void setAtomIdentifier(String aui) {
    this.aui = aui;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link Atom}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link Atom}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns the suppressibility value (SUPPRESS).
   * @return the {@link Suppressible} value
   */
  public Suppressible getSuppress() {
    return suppress;
  }

  /**
   * Indicates whether or not this {@link Atom} is suppressible.
   * @return <code>true</code> if suppressible,
   *         <code>false</code> if not suppressible
   */
  public boolean isSuppressible() {
    return (suppress != null && !suppress.getAbbreviation().equals("N"));
  }

  /**
   * Sets the suppressibility value (SUPPRESS).
   * @param suppress the {@link Suppressible} value
   */
  public void setSuppress(Suppressible suppress) {
    this.suppress = suppress;
    key_fields_flag = false;
  }

  /**
   * Returns the flag indicating whether or not this {@link Atom} is
   * preferred (ISPREF).
   * @return the flag indicating whether or not this {@link Atom} is
   * preferred
   */
  public String getIspref() {
    return ispref;
  }


  /**
   * Sets the flag indicating whether or not this {@link Atom} is
   * preferred (ISPREF).
   * @param ispref the flag indicating whether or not this {@link Atom} is
   * preferred
   */
  public void setIspref(String ispref) {
    this.ispref = ispref;
    key_fields_flag = false;
  }

  /**
   * Indicates whether or not this {@link Atom} is the preferred atom identifier
   * for its {@link Concept} and string identifier.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredAtomIdentifier() {
    return "Y".equals(ispref);
  }

  /**
   * Shorthand for {@link #hasPreferredAtomIdentifier()}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredAui() {
    return hasPreferredAtomIdentifier();
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link Atom} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link Atom} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns a subset of this {@link Atom}'s {@link Relationship}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Relationship[]
   */
  public Relationship[] getRelationships(Restrictor restrictor) {
    List rels = getRelationshipsAsList();
    for (int i=rels.size()-1; i>=0; i--) {
      if (!restrictor.keep(rels.get(i)))
        rels.remove(i);
    }
    Collections.sort(rels, restrictor);
    return (Relationship[])rels.toArray(new Atom[0]);
  }

  /**
   * Indicates whether or not this {@link AtomImpl} is obsolete.
   * @return <code>true</code> if obsolete,
   *         <code>false</code> if not obsolete
   */
  public boolean isObsolete() {
    if (getSuppress().getAbbreviation().equals("O"))
      return true;
    TermTypeGroup[] tty_groups = TermTypeGroup.getTermTypeGroups(getTermType());
    for (int i=0; i<tty_groups.length; i++) {
      if (tty_groups[i].getDescription().equals("obsolete"))
        return true;
    }
    return false;
  }

}
