/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       PlaceholderConcept.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.AtomAttribute;
import gov.nih.nlm.umls.meta.AtomRelationship;
import gov.nih.nlm.umls.meta.Attribute;
import gov.nih.nlm.umls.meta.ComponentHistory;
import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Context;
import gov.nih.nlm.umls.meta.ContextGroup;
import gov.nih.nlm.umls.meta.Coocurrence;
import gov.nih.nlm.umls.meta.CoocurrenceGroup;
import gov.nih.nlm.umls.meta.Definition;
import gov.nih.nlm.umls.meta.Locator;
import gov.nih.nlm.umls.meta.Mapping;
import gov.nih.nlm.umls.meta.Relationship;
import gov.nih.nlm.umls.meta.Restrictor;
import gov.nih.nlm.umls.meta.SemanticType;
import gov.nih.nlm.umls.meta.Term;
import gov.nih.nlm.umls.meta.TermAttribute;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

/**
 * Implementation of the {@link gov.nih.nlm.umls.meta.Concept} interface.
 * Used when {@link gov.nih.nlm.umls.meta.Concept} components do not need to be saved.
 * Uses less space than {@link gov.nih.nlm.umls.meta.Concept}.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class PlaceholderConcept implements Concept {

  //
  // Fields
  //
  protected List atoms = null;
  protected String cui;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link PlaceholderConcept}.
   */
  public PlaceholderConcept() {};

  /**
   * This removes all of the {@link PlaceholderConcept}'s components.
   */
  public void clear() {
    clearAtoms();
  }

  public boolean isEmpty() {
    return (getAtoms().length == 0);
  }

  //
  // Atoms
  //

  /**
   * Returns the {@link Atom}s.
   * @return an {@link Atom} <code>[]</code>
   */
  public Atom[] getAtoms() {
    return (Atom[])getAtomsAsList().toArray(new Atom[0]);
  }

  /**
   * Returns the {@link Atom}s as a {@link List}.
   * @return the {@link Atom}s as a {@link List}
   */
  public List getAtomsAsList() {
    if (atoms != null)
      return new ArrayList(atoms);
    else
      return new ArrayList();
  }

  /**
   * Adds the specified {@link Atom}.
   * @param atom the {@link Atom} to be added
   */
  public void addAtom(Atom atom) {
    if (atoms == null)
      atoms = new ArrayList();
    atoms.add(atom);
  }

  /**
   * Removes the specified {@link Atom}.
   * @param atom the {@link Atom} to be removed
   */
  public void removeAtom(Atom atom) {
    if (atoms != null) atoms.remove(atom);
  }

  /**
   * Remove all {@link Atom}s.
   */
  public void clearAtoms() {
    if (atoms != null) atoms.clear();
  }

  /**
   * Replace any {@link Atom}s with the
   * ones in the specified array.
   * @param atoms a {@link Atom} <code>[]</code>
   */
  public void setAtoms(Atom atoms[]) {
    this.atoms = new ArrayList();
    for (int i=0; i<atoms.length; i++) {
      this.atoms.add(atoms[i]);
    }
  }

  /**
   * Replace {@link Atom}s with the
   * ones in the specified {@link List}.
   * @param atoms a {@link List} of {@link Atom}s
   */
  public void setAtoms(List atoms) {
    this.atoms = new ArrayList(atoms);
  }

  /**
   * Returns the preferred atom of the concept.
   * It has TS=P, STT=PF, and ISPREF=Y.
   * @return {@link Atom}
   */
  public Atom getPreferredAtom() {
    Atom[] atoms = getAtoms();
    for (int i=0; i<atoms.length; i++) {
      if(atoms[i].hasPreferredAtomIdentifier() &&
         atoms[i].hasPreferredStringIdentifier() &&
         atoms[i].hasPreferredTermIdentifier()) {
        return atoms[i];
      }
    }
    return null;
  }

  /**
   * Returns the preferred atom of the concept.
   * Uses the given {@link Comparator} to determine the preferred atom.
   * @param comparator {@link Comparator}
   * @return {@link Atom}
   */
  public Atom getPreferredAtom(Comparator comparator) {
    Atom[] atoms = getAtoms();
    Arrays.sort(atoms, comparator);
    return atoms[0];
  }

  //
  // Locators
  //

  /**
   * Returns the {@link Locator}s.
   * @return a {@link Locator} <code>[]</code>
   */
  public Locator[] getLocators() {
     return new Locator[0];
  }

  /**
   * Returns the {@link Locator}s as a {@link List}.
   * @return the {@link Locator}s as a {@link List}
   */
  public List getLocatorsAsList() {
    return new ArrayList();
  }

  /**
   * Add the specified {@link Locator}.
   * @param loc the {@link Locator} to be added
   */
  public void addLocator(Locator loc) {
  }

  /**
   * Removes the specified {@link Locator}.
   * @param loc the {@link Locator} to be removed
   */
  public void removeLocator(Locator loc) {
  }

  /**
   * Remove all {@link Locator}s.
   */
  public void clearLocators() {
  }

  /**
   * Replace {@link Locator}s with the ones in the specified array.
   * @param locs a {@link Locator} <code>[]</code>
   */
  public void setLocators(Locator locs[]) {
  }

  /**
   * Replace {@link Locator}s with the ones in the specified {@link List}.
   * @param locs a {@link List} of {@link Locator}s
   */
  public void setLocators(List locs) {
  }


  //
  // Attributes
  //

  /**
   * Returns the {@link Attribute}s.
   * @return a {@link Attribute} <code>[]</code>
   */
  public Attribute[] getAttributes() {
    return new Attribute[0];
  }

  /**
   * Returns the {@link Attribute}s as a {@link List}.
   * @return the {@link Attribute}s as a {@link List}
   */
  public List getAttributesAsList() {
      return new ArrayList();
  }

  /**
   * Adds the specified {@link Attribute}.
   * @param att the {@link Attribute} to be added
   */
  public void addAttribute(Attribute att) {
  }

  /**
   * Removes the specified {@link Attribute}.
   * @param att the {@link Attribute} to be removed
   */
  public void removeAttribute(Attribute att) {
  }

  /**
   * Remove all {@link Attribute}s.
   */
  public void clearAttributes() {
  }

  /**
   * Replace {@link Attribute}s with the ones in the specified array.
   * @param atts an {@link Attribute} <code>[]</code>
   */
  public void setAttributes(Attribute atts[]) {
  }

  /**
   * Replace {@link Attribute}s with the ones in the specified {@link List}.
   * @param atts a {@link List} of {@link Attribute}s
   */
  public void setAttributes(List atts) {
  }

  //
  // Contexts
  //

  /**
   * Return the {@link Context}s.
   * @return a {@link Context} <code>[]</code>
   */
  public Context[] getContexts() {
    return new Context[0];
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Context}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Context[]
   */
  public Context[] getContexts(Restrictor restrictor) {
    List cxts = getDefinitionsAsList();
    for (int i=cxts.size()-1; i>=0; i--) {
      if (!restrictor.keep(cxts.get(i)))
        cxts.remove(i);
    }
    Collections.sort(cxts, restrictor);
    return (Context[])cxts.toArray(new Context[0]);
  }

  /**
   * Returns the {@link Context}s as a {@link List}.
   * @return the {@link Context}s as a {@link List}
   */
  public List getContextsAsList() {
    return new ArrayList();
  }

  /**
   * Adds the specified {@link Context}.
   * @param cxt the {@link Context} to be added
   */
  public void addContext(Context cxt) {
  }

  /**
   * Removes the specified {@link Context}.
   * @param cxt the {@link Context} to be removed
   */
  public void removeContext(Context cxt) {
  }

  /**
   * Remove all {@link Context}s.
   */
  public void clearContexts() {
  }

  /**
   * Replace {@link Context}s with the ones in the specified array.
   * @param cxts a {@link Context} <code>[]</code>
   */
  public void setContexts(Context cxts[]) {
  }

  /**
   * Replace {@link Context}s with the ones in the specified {@link List}.
   * @param cxts a {@link List} of {@link Context}s
   */
  public void setContexts(List cxts) {
  }

  //
  // Coocurrences
  //

  /**
   * Returns the {@link Coocurrence}s.
   * @return a {@link Coocurrence} <code>[]</code>
   */
  public Coocurrence[] getCoocurrences() {
    return new Coocurrence[0];
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Coocurrence}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Coocurrence[]
   */
  public Coocurrence[] getCoocurrences(Restrictor restrictor) {
    List cocs = getDefinitionsAsList();
    for (int i=cocs.size()-1; i>=0; i--) {
      if (!restrictor.keep(cocs.get(i)))
        cocs.remove(i);
    }
    Collections.sort(cocs, restrictor);
    return (Coocurrence[])cocs.toArray(new Coocurrence[0]);
  }


  /**
   * Returns the {@link Coocurrence}s as a {@link List}.
   * @return the {@link Coocurrence}s as a {@link List}
   */
  public List getCoocurrencesAsList() {
      return new ArrayList();
  }

  /**
   * Adds the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be added
   */
  public void addCoocurrence(Coocurrence coc) {
  }

  /**
   * Removes the specified {@link Coocurrence}.
   * @param coc the {@link Coocurrence} to be removed
   */
  public void removeCoocurrence(Coocurrence coc) {
  }

  /**
   * Remove all {@link Coocurrence}s.
   */
  public void clearCoocurrences() {
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified array.
   * @param cocs a {@link Coocurrence} <code>[]</code>
   */
  public void setCoocurrences(Coocurrence cocs[]){
  }

  /**
   * Replace {@link Coocurrence}s with the ones in the specified {@link List}.
   * @param cocs a {@link List} of {@link Coocurrence}s
   */
  public void setCoocurrences(List cocs) {
  }

  //
  // Definitions
  //

  /**
   * Returns the {@link Definition}s.
   * @return a {@link Definition} <code>[]</code>
   */
  public Definition[] getDefinitions() {
    return new Definition[0];
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Definition}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Definition[]
   */
  public Definition[] getDefinitions(Restrictor restrictor) {
    List defs = getDefinitionsAsList();
    for (int i=defs.size()-1; i>=0; i--) {
      if (!restrictor.keep(defs.get(i)))
        defs.remove(i);
    }
    Collections.sort(defs, restrictor);
    return (Definition[])defs.toArray(new Definition[0]);
  }

  /**
   * Returns the {@link Definition}s as a {@link List}.
   * @return the {@link Definition}s as a {@link List}
   */
  public List getDefinitionsAsList() {
    return new ArrayList();
  }

  /**
   * Adds the specified {@link Definition}.
   * @param def the {@link Definition} to be added
   */
  public void addDefinition(Definition def) {
  }

  /**
   * Remove the specified {@link Definition}.
   * @param def the {@link Definition} to be removed
   */
  public void removeDefinition(Definition def) {
  }

  /**
   * Remove all {@link Definition}s.
   */
  public void clearDefinitions() {
  }

  /**
   * Replace {@link Definition}s with the ones in the specified array.
   * @param defs a {@link Definition} <code>[]</code>
   */
  public void setDefinitions(Definition defs[]) {
  }

  /**
   * Replace {@link Definition}s with the ones in the specified {@link List}.
   * @param defs a {@link List} of {@link Definition}s
   */
  public void setDefinitions(List defs) {
  }

  //
  // Relationships
  //

  /**
   * Returns the {@link Relationship}s.
   * @return a {@link Relationship} <code>[]</code>
   */
  public Relationship[] getRelationships() {
    return new Relationship[0];
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Relationship}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Relationship[]
   */
  public Relationship[] getRelationships(Restrictor restrictor) {
    List rels = getDefinitionsAsList();
    for (int i=rels.size()-1; i>=0; i--) {
      if (!restrictor.keep(rels.get(i)))
        rels.remove(i);
    }
    Collections.sort(rels, restrictor);
    return (Relationship[])rels.toArray(new Relationship[0]);
  }

  /**
   * Returns the {@link Relationship}s as a {@link List}.
   * @return the {@link Relationship}s as a {@link List}
   */
  public List getRelationshipsAsList() {
    return new ArrayList();
  }

  /**
   * Adds the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be added
   */
  public void addRelationship(Relationship rel) {
  }

  /**
   * Removes the specified {@link Relationship}.
   * @param rel the {@link Relationship} to be removed
   */
  public void removeRelationship(Relationship rel) {
  }

  /**
   * Remove all {@link Relationship}s.
   */
  public void clearRelationships() {
  }

  /**
   * Replace {@link Relationship}s with the ones in the specified array.
   * @param rels a {@link Relationship} <code>[]</code>
   */
  public void setRelationships(Relationship rels[]) {
  }

  /**
   * Replace {@link Relationship}s of the ones in the specified {@link List}.
   * @param rels a {@link List} of {@link Relationship}s
   */
  public void setRelationships(List rels) {
  }

  //
  // SemanticTypes
  //

   /**
   * Returns the {@link SemanticType}s.
   * @return a {@link SemanticType} <code>[]</code>
   */
  public SemanticType[] getSemanticTypes() {
    return new SemanticType[0];
  }

  /**
   * Returns the {@link SemanticType}s as a {@link List}.
   * @return the {@link SemanticType}s as a {@link List}
   */
  public List getSemanticTypesAsList() {
    return new ArrayList();
  }

  /**
   * Adds a {@link SemanticType}.
   * @param sty the {@link SemanticType} to be added
   */
  public void addSemanticType(SemanticType sty) {
  }

  /**
   * Removes the specified {@link SemanticType}.
   * @param sty the {@link SemanticType} to be removed
   */
  public void removeSemanticType(SemanticType sty) {
  }

  /**
   * Removes all {@link SemanticType}s.
   */
  public void clearSemanticTypes() {
  }

  /**
   * Replace {@link SemanticType}s with the ones in the specified array.
   * @param stys a {@link SemanticType} <code>[]</code>
   */
  public void setSemanticTypes(SemanticType stys[]) {
  }

  /**
   * Replace {@link SemanticType}s with the ones in the specified {@link List}.
   * @param stys a {@link List} of {@link SemanticType}s
   */
  public void setSemanticTypes(List stys) {
  }

  //
  // Component Histories
  //

   /**
   * Returns the {@link ComponentHistory}s.
   * @return a {@link ComponentHistory} <code>[]</code>
   */
  public ComponentHistory[] getComponentHistories() {
    return new ComponentHistory[0];
  }

  /**
   * Returns the {@link ComponentHistory}s as a {@link List}.
   * @return the {@link ComponentHistory}s as a {@link List}
   */
  public List getComponentHistoriesAsList() {
    return new ArrayList();
  }

  /**
   * Adds a {@link ComponentHistory}.
   * @param map the {@link ComponentHistory} to be added
   */
  public void addComponentHistory(ComponentHistory map) {
  }

  /**
   * Removes the specified {@link ComponentHistory}.
   * @param map the {@link ComponentHistory} to be removed
   */
  public void removeComponentHistory(ComponentHistory map) {
  }

  /**
   * Removes all {@link ComponentHistory}s.
   */
  public void clearComponentHistories() {
  }

  /**
   * Replace {@link ComponentHistory}s with the ones in the specified array.
   * @param maps a {@link ComponentHistory} <code>[]</code>
   */
  public void setComponentHistories(ComponentHistory maps[]) {
  }

  /**
   * Replace {@link ComponentHistory}s with the ones in the specified {@link List}.
   * @param maps a {@link List} of {@link ComponentHistory}s
   */
  public void setComponentHistories(List maps) {
  }

  //
  // Mappings
  //

   /**
   * Returns the {@link Mapping}s.
   * @return a {@link Mapping} <code>[]</code>
   */
  public Mapping[] getMappings() {
    return new Mapping[0];
  }

  /**
   * Returns the {@link Mapping}s as a {@link List}.
   * @return the {@link Mapping}s as a {@link List}
   */
  public List getMappingsAsList() {
    return new ArrayList();
  }

  /**
   * Adds a {@link Mapping}.
   * @param map the {@link Mapping} to be added
   */
  public void addMapping(Mapping map) {
  }

  /**
   * Removes the specified {@link Mapping}.
   * @param map the {@link Mapping} to be removed
   */
  public void removeMapping(Mapping map) {
  }

  /**
   * Removes all {@link Mapping}s.
   */
  public void clearMappings() {
  }

  /**
   * Replace {@link Mapping}s with the ones in the specified array.
   * @param maps a {@link Mapping} <code>[]</code>
   */
  public void setMappings(Mapping maps[]) {
  }

  /**
   * Replace {@link Mapping}s with the ones in the specified {@link List}.
   * @param maps a {@link List} of {@link Mapping}s
   */
  public void setMappings(List maps) {
  }

 /**
   * Returns the attributes with aui that matches that of the given {@link Atom}.
   * @param atom {@link Atom}
   * @return {@link List} of {@link Attribute}s
   */
  public Attribute[] getAttributesForAtom(Atom atom) {
    return (Attribute[])getAttributesForAtomAsList(atom).toArray(new Attribute[0]);
  }

  /**
   * Returns the attributes with aui that matches that of the given {@link Atom}
   * in a {@link List}.
   * @param atom {@link Atom}
   * @return {@link List} of {@link AtomAttribute}s
   */
  public List getAttributesForAtomAsList(Atom atom) {
    return new ArrayList();
  }

  /**
   * Returns the attributes with aui that matches that of the given {@link Relationship}.
   * @param rel {@link AtomRelationship}
   * @return {@link List} of {@link Attribute}s
   */
  public AtomAttribute[] getAttributesForRelationship(AtomRelationship rel) {
    return new AtomAttribute[0];
  }

  /**
   * Returns the attributes with aui that matches that of the given {@link Relationship}
   * in a {@link List}.
   * @param rel {@link AtomRelationship}
   * @return {@link List} of {@link Attribute}s
   */
  public List getAttributesForRelationshipAsList(AtomRelationship rel) {
    return new ArrayList();
  }

  /**
   * Shorthand for {@link #getConceptIdentifier()}.
   * @return the CUI
   */
  public String getCui() {
    return getConceptIdentifier();
  }

  /**
   * Shorthand for {@link #setConceptIdentifier(String)}.
   * @param cui the CUI
   */
  public void setCui(String cui) {
    setConceptIdentifier(cui);
  }

  /**
   * Returns the concept unique identifier (CUI) for the {@link Concept}
   * that holds this {@link Concept}.
   * @return the CUI
   */
  public String getConceptIdentifier() {
    return cui;
  }

  /**
   * Sets the concept unique identifier (CUI) for the {@link Concept}
   * that holds this {@link Concept}.
   * @param cui the CUI
   */
  public void setConceptIdentifier(String cui) {
    this.cui = cui;
    key_fields_flag = false;
  }


  //
  // UMLSKS View Methods
  //

  /**
   * Return the {@link ContextGroup}s.
   * @return a {@link ContextGroup} <code>[]</code>
   */
  public ContextGroup[] getContextGroups() { return new ContextGroup[0]; }

  /**
   * Returns the {@link ContextGroup}s as a {@link List}.
   * @return the {@link ContextGroup}s as a {@link List}
   */
  public List getContextGroupsAsList() { return new ArrayList(); }

  /**
   * Returns the {@link CoocurrenceGroup}s.
   * @return a {@link CoocurrenceGroup} <code>[]</code>
   */
  public CoocurrenceGroup[] getCoocurrenceGroups() { return new CoocurrenceGroup[0]; }

  /**
   * Returns the {@link CoocurrenceGroup}s as a {@link List}.
   * @return the {@link CoocurrenceGroup}s as a {@link List}
   */
  public List getCoocurrenceGroupsAsList() { return new ArrayList(); }

  /**
   * Returns the {@link TermAttribute}s.
   * @return a {@link TermAttribute} <code>[]</code>
   */
  public TermAttribute[] getTermAttributes() { return new TermAttribute[0]; }

  /**
   * Returns the {@link TermAttribute}s as a {@link List}.
   * @return the {@link TermAttribute}s as a {@link List}
   */
  public List getTermAttributesAsList() { return new ArrayList(); }

  /**
   * Returns the {@link Term}s.
   * @return an {@link Term} <code>[]</code>
   */
  public Term[] getTerms() { return new Term[0]; }

  /**
   * Returns the {@link Term}s as a {@link List}.
   * @return the {@link Term}s as a {@link List}
   */
  public List getTermsAsList() { return new ArrayList(); }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      key_fields = getCui();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Returns a subset of this {@link Concept}'s {@link Atom}s.
   * Restricts the return based on the given {@link Restrictor}.
   * @param restrictor {@link Restrictor}
   * @return Atom[]
   */
  public Atom[] getAtoms(Restrictor restrictor) {
    List atoms = getAtomsAsList();
    for (int i=atoms.size()-1; i>=0; i--) {
      if (!restrictor.keep(atoms.get(i)))
        atoms.remove(i);
    }
    Collections.sort(atoms, restrictor);
    return (Atom[])atoms.toArray(new Atom[0]);
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link Concept} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Concept)) return false;
    return getKeyFields().equals(((Concept)o).getKeyFields());
  }

  /**
   * Compares this {@link Concept} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Concept)
      return getKeyFields().compareTo(((Concept)o).getKeyFields());
    else
      return 0;
  }

}
