/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       QualifierFrequencyImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.QualifierFrequency;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.QualifierFrequency}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class QualifierFrequencyImpl implements QualifierFrequency  {

  //
  // Fields
  //
  private String qualifier;
  private int frequency = 0;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;


  /**
   * Instantiates an empty {@link QualifierFrequency}.
   */
  public QualifierFrequencyImpl() {};

  /**
   * Instantiates a {@link QualifierFrequency} from the specified {@link QualifierFrequency}.
   * @param qf {@link QualifierFrequency} to be copied
   */
  public QualifierFrequencyImpl(QualifierFrequency qf) {
    setFrequency(qf.getFrequency());
    setQualifier(qf.getQualifier());
  }

  /**
   * Returns the qualifier value.
   * @return the qualifier value
   */
  public String getQualifier() {
    return qualifier;
  }

  /**
   * Sets the qualifier value.
   * @param qual the qualifier value
   */
  public void setQualifier(String qual) {
    qualifier = qual;
    key_fields_flag = false;
  }

  /**
   * Returns the frequency.
   * @return the frequency
   */
  public int getFrequency() {
    return frequency;
  }

  /**
   * Sets the frequency;
   * @param freq the frequency
   */
  public void setFrequency(int freq) {
    frequency = freq;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the frequency and qualifier.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(200);
      sb1.append(getFrequency())
        .append(getQualifier());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link QualifierFrequency} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof QualifierFrequency)) return false;
    return getKeyFields().equals(((QualifierFrequency)o).getKeyFields());
  }

  /**
   * Compares this {@link QualifierFrequency} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof QualifierFrequency) {
      QualifierFrequency qf = (QualifierFrequency)o;
      if (getFrequency() != qf.getFrequency())
        return getFrequency() - qf.getFrequency();
      else
      return getKeyFields().compareTo(((QualifierFrequency)o).getKeyFields());
    } else {
      return 0;
    }
  }

}
