/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       ReleaseMetadataImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AtomMapping;
import gov.nih.nlm.umls.meta.ConceptMapping;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.ReleaseMetadata;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.SourceTermType;

import java.io.Reader;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.ReleaseMetadata}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class ReleaseMetadataImpl implements ReleaseMetadata {

  //
  // Fields
  //
  private String release_version;
  private Date release_date;
  private String release_description;
  private List current_sources = new ArrayList(100);
  private List concept_mappings = new ArrayList(100000);
  private List atom_mappings = new ArrayList(2000);
  private List source_term_types = new ArrayList(1000);
  private List meta_descriptions = new ArrayList(1000);
  private Map change_files = new HashMap();
  private Map ancillary_files = new HashMap();
  private Set subx_cuis = new HashSet();

  // Used to compute key fields more efficiently
  protected boolean key_fields_flag;
  protected String key_fields = null;

  /**
   * Instantiates an empty {@link ReleaseMetadataImpl}.
   */
  public ReleaseMetadataImpl() {};

  /**
   * Instantiates a {@link ReleaseMetadataImpl} from the specified
   * {@link ReleaseMetadataImpl}.
   * @param rm {@link ReleaseMetadataImpl} to be copied
   */
  public ReleaseMetadataImpl(ReleaseMetadataImpl rm) {
    setAtomMappings(rm.getAtomMappings());
    setConceptMappings(rm.getConceptMappings());
    setMetadata(rm.getMetadata());
    setReleaseDate(rm.getReleaseDate());
    setReleaseDescription(rm.getReleaseDescription());
    setReleaseVersion(rm.getReleaseVersion());
    setSourceMetadata(rm.getSourceMetadata());
    setSourceTermTypePrecedence(rm.getSourceTermTypePrecedence());
  }

  /**
   * Returns the release version.
   * @return the release version
   */
  public String getReleaseVersion() {
    return release_version;
  }

  /**
   * Sets the release version.
   * @param rv the release version
   */
  public void setReleaseVersion(String rv) {
    release_version = rv;
    key_fields_flag = false;
  }

  /**
   * Returns the release {@link Date}.
   * @return the release {@link Date}.
   */
  public Date getReleaseDate() {
    return release_date;
  }

  /**
   * Sets the release {@link Date}.
   * @param rd the release {@link Date}
   */
  public void setReleaseDate(Date rd) {
    release_date = rd;
    key_fields_flag = false;
  }

  /**
   * Returns the release description.
   * @return the release description
   */
  public String getReleaseDescription() {
    return release_description;
  }

  /**
   * Sets the release description.
   * @param rd the release description
   */
  public void setReleaseDescription(String rd) {
    release_description = rd;
    key_fields_flag = false;
  }

  /**
   * Returns {@link SourceMetaDescription}s in the current subset.
   * @return an array of all {@link SourceMetaDescription} in the current subset
   */
  public SourceMetaDescription[] getSourceMetadata(){
    return (SourceMetaDescription[])current_sources.toArray(new SourceMetaDescription[0]);
  }

  /**
   * Returns {@link SourceMetaDescription} in the current subset as a {@link List}.
   * @return {@link SourceMetaDescription} in the current subset as a {@link List}
   */
  public List getSourceMetadataAsList() {
    return new ArrayList(current_sources);
  }

  /**
   * Sets {@link SourceMetaDescription} in the current subset.
   * @param sabs the {@link SourceMetaDescription} in the current subset
   */
  public void setSourceMetadata(SourceMetaDescription[] sabs) {
    current_sources.clear();
    for (int i=0; i< sabs.length; i++ ) {
      current_sources.add(sabs[i]);
    }
  }

  /**
   * Sets {@link SourceMetaDescription}s in the current subset.
   * @param sabs a {@link List} of {@link SourceMetaDescription}s in the current subset
   */
  public void setSourceMetadata(List sabs) {
    current_sources = new ArrayList(sabs);
  }

  /**
   * Adds the specified {@link SourceMetaDescription}.
   * @param smd the {@link SourceMetaDescription} to be added
   */
  public void addSourceMetaDescription(SourceMetaDescription smd) {
    current_sources.add(smd);
  }

  /**
   * Removes the specified {@link SourceMetaDescription}.
   * @param smd the {@link SourceMetaDescription} to be removed
   */
  public void removeSourceMetaDescription(SourceMetaDescription smd) {
    current_sources.remove(smd);
  }

  /**
   * Removes all {@link SourceMetaDescription}s.
   */
  public void clearSourceMetaDescription() {
    current_sources.clear();
  }

  /**
   * Returns all {@link ConceptMapping}s.
   * @return an array of all {@link ConceptMapping}s
   */
  public ConceptMapping[] getConceptMappings() {
    return (ConceptMapping[])concept_mappings.toArray(new ConceptMapping[0]);
  }

  /**
   * Returns all {@link ConceptMapping}s as a {@link List}.
   * @return all {@link ConceptMapping}s
   */
  public List getConceptMappingsAsList() {
    return new ArrayList(concept_mappings);
  }

  /**
   * Sets the {@link ConceptMapping}s.
   * @param cm the {@link ConceptMapping}s
   */
  public void setConceptMappings(ConceptMapping[] cm) {
    concept_mappings.clear();
    for (int i=0; i<cm.length; i++) {
      concept_mappings.add(cm[i]);
    }
  }

  /**
   * Adds the specified {@link ConceptMapping}.
   * @param cm the {@link ConceptMapping} to be added
   */
  public void addConceptMapping(ConceptMapping cm) {
    concept_mappings.add(cm);
  }

  /**
   * Removes the specified {@link ConceptMapping}.
   * @param cm the {@link ConceptMapping} to be removed
   */
  public void removeConceptMapping(ConceptMapping cm) {
    concept_mappings.remove(cm);
  }

  /**
   * Removes all {@link ConceptMapping}s.
   */
  public void clearConceptMappings() {
    concept_mappings.clear();
  }

  /**
   * Sets the {@link ConceptMapping}s.
   * @param cm a {@link List} of the {@link ConceptMapping}s
   */
  public void setConceptMappings(List cm) {
    concept_mappings = new ArrayList(cm);
  }

  /**
   * List of SUBX cuis.
   * @return array of SUBX cuis
   */
  public String[] getCUIsRemovedFromSubset() {
    return (String[])subx_cuis.toArray(new String[0]);
  }

  /**
   * Add SUBX cui to list of cuis removed from subset.
   * @param cui {@link String}
   */
  public void addCUIRemovedFromSubset(String cui) {
    subx_cuis.add(cui);
  }

  /**
   * Returns all {@link AtomMapping}s.
   * @return an array of all {@link AtomMapping}s
   */
  public AtomMapping[] getAtomMappings() {
    return (AtomMapping[])atom_mappings.toArray(new AtomMapping[0]);
  }

  /**
   * Returns all {@link AtomMapping}s as a {@link List}.
   * @return all {@link AtomMapping}s
   */
  public List getAtomMappingsAsList() {
    return new ArrayList(atom_mappings);
  }

  /**
   * Sets the {@link AtomMapping}s.
   * @param cm the {@link AtomMapping}s
   */
  public void setAtomMappings(AtomMapping[] cm) {
    atom_mappings.clear();
    for (int i=0; i<cm.length; i++) {
      atom_mappings.add(cm[i]);
    }
  }

  /**
   * Adds the specified {@link AtomMapping}.
   * @param cm the {@link AtomMapping} to be added
   */
  public void addAtomMapping(AtomMapping cm) {
    atom_mappings.add(cm);
  }

  /**
   * Removes the specified {@link AtomMapping}.
   * @param cm the {@link AtomMapping} to be removed
   */
  public void removeAtomMapping(AtomMapping cm) {
    atom_mappings.remove(cm);
  }

  /**
   * Removes all {@link AtomMapping}s.
   */
  public void clearAtomMappings() {
    atom_mappings.clear();
  }

  /**
   * Sets the {@link AtomMapping}s.
   * @param cm a {@link List} of the {@link AtomMapping}s
   */
  public void setAtomMappings(List cm) {
    atom_mappings = new ArrayList(cm);
  }

  /**
   * Returns all {@link SourceTermType}s in rank order.
   * @return an array of all {@link SourceTermType}s in rank order
   */
  public SourceTermType[] getSourceTermTypePrecedence() {
    return (SourceTermType[])source_term_types.toArray(new SourceTermType[0]);
  }

  /**
   * Returns all {@link SourceTermType}s in rank order as a {@link List}.
   * @return all {@link SourceTermType}s in rank order as a {@link List}
   */
  public List getSourceTermTypePrecedenceAsList() {
    return new ArrayList(source_term_types);
  }

  /**
   * Sets all {@link SourceTermType}s in rank order.
   * @param sab_ttys an array of all {@link SourceTermType}s in rank order
   */
  public void setSourceTermTypePrecedence(SourceTermType[] sab_ttys) {
    source_term_types.clear();
    for (int i=0; i<sab_ttys.length; i++) {
      source_term_types.add(sab_ttys[i]);
    }
  }

  /**
   * Sets all {@link SourceTermType}s in rank order.
   * @param sab_ttys a {@link List} of all {@link SourceTermType}s in rank order
   */
  public void setSourceTermTypePrecedence(List sab_ttys) {
    source_term_types = new ArrayList(sab_ttys);
  }

  /**
   * Returns {@link MetaDescription}s in the current subset.
   * @return an array of all {@link MetaDescription}s in the current subset
   */
  public MetaDescription[] getMetadata(){
    return (MetaDescription[])meta_descriptions.toArray(new MetaDescription[0]);
  }

  /**
   * Returns {@link MetaDescription} in the current subset as a {@link List}.
   * @return {@link MetaDescription} in the current subset as a {@link List}
   */
  public List getMetadataAsList() {
    return new ArrayList(meta_descriptions);
  }

  /**
   * Sets {@link MetaDescription} in the current subset.
   * @param docs the {@link MetaDescription} in the current subset
   */
  public void setMetadata(MetaDescription[] docs) {
    meta_descriptions.clear();
    for (int i=0; i< docs.length; i++ ) {
      meta_descriptions.add(docs[i]);
    }
  }

  /**
   * Sets {@link MetaDescription}s in the current subset.
   * @param docs a {@link List} of {@link MetaDescription}s in the current subset
   */
  public void setMetadata(List docs) {
    meta_descriptions = new ArrayList(docs);
  }

  /**
   * Adds the specified {@link MetaDescription}.
   * @param md the {@link MetaDescription} to be added
   */
  public void addMetadata(MetaDescription md) {
    meta_descriptions.add(md);
  }

  /**
   * Removes the specified {@link MetaDescription}.
   * @param md the {@link MetaDescription} to be removed
   */
  public void removeMetadata(MetaDescription md) {
    meta_descriptions.remove(md);
  }

  /**
   * Removes all {@link MetaDescription}s.
   */
  public void clearMetadata() {
    meta_descriptions.clear();
  }

  /**
   * Returns a {@link String}[] of all the file names in the CHANGE directory.
   * @return array of all the file names in the CHANGE directory
   */
  public String[] getChangeFileNames() {
    return (String[])(change_files.keySet().toArray(new String[0]));
  }

  /**
   * Returns the {@link Reader} for the given CHANGE file.
   * @param file_name for CHANGE file
   * @return {@link Reader} for the given CHANGE file
   */
  public Reader getChangeFile(String file_name) {
    return (Reader)change_files.get(file_name);
  }

  /**
   * Returns a {@link String}[] of all the ancillary file names.
   * @return array of all the ancillary file names
   */
  public String[] getAncillaryFileNames() {
    return (String[])(ancillary_files.keySet().toArray(new String[0]));
  }

  /**
   * Returns a {@link Reader} for the given ancillary file.
   * @param file_name for given ancillary file
   * @return a {@link Reader} for the given ancillary file
   */
  public Reader getAncillaryFile(String file_name) {
    return (Reader)ancillary_files.get(file_name);
  }

  /**
   * Adds change file to {@link HashMap} of change files.
   * @param file_name name of change file
   * @param reader a {@link Reader} over the file
   */
  public void addChangeFile(String file_name, Reader reader) {
    change_files.put(file_name, reader);
  }

  /**
   * Adds ancillary file to {@link HashMap} of ancillary files.
   * @param file_name name of ancillary file
   * @param reader a {@link Reader} over the file
   */
  public void addAncillaryFile(String file_name, Reader reader) {
    ancillary_files.put(file_name, reader);
  }

  /**
   * Returns a {@link String} key which can be used by equals, compareTo, and
   * hashCode methods.
   *
   * @return a {@link String} composed from the release version, date, and
   * description fields.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb = new StringBuffer(100);
      sb.append(getReleaseVersion())
          .append(getReleaseDate())
          .append(getReleaseDescription());
      key_fields = sb.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Returns <code>true</code> if the specified object is equal to this object.
   * The specified object must be a {@link ReleaseMetadata} with matching key
   * fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal;
   *                 <code>false</code> otherwise.
   */
  public boolean equals(Object o) {
    return getKeyFields().equals(((ReleaseMetadata) o).getKeyFields());
  }

  /**
   * Returns a hash code based on key fields.
   * @return a hash code based on key fields.
   */
  public int hashCode() {
    return (getKeyFields()).hashCode();
  }

  /**
   * Compares this {@link ReleaseMetadata} with the specified object for order.
   * Returns a negative integer, zero, or a positive integer as this object is
   * less than, equal to, or greater than the specified object.  Results
   * are indeterminate if the specified parameter is not of the same class.
   * @param o the Object to be compared.
   * @return a negative integer, zero, or a positive integer as this object
   *         is less than, equal to, or greater than the specified object.
   *         Returns zero if the specified object's type prevents it from being
   *         compared to this Object.
   */
  public int compareTo(Object o) {
    if (o instanceof ReleaseMetadata)
      return getKeyFields().compareTo(((ReleaseMetadata) o).getKeyFields());
    else
      return 0;
  }
}
