/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       SemanticTypeImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.ContentView;
import gov.nih.nlm.umls.meta.HierarchicalSemanticType;
import gov.nih.nlm.umls.meta.SemanticType;

/**
 * Default implementation of the {@link gov.nih.nlm.umls.meta.SemanticType}
 * interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class SemanticTypeImpl implements SemanticType {

  //
  // Fields
  //
  private HierarchicalSemanticType sty;
  private Concept concept;
  private String atui;
  private ContentView cvf;


  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;

  /**
   * Instantiates an empty {@link SemanticType}.
   */
  public SemanticTypeImpl() {};

  /**
   * Instantiates a {@link SemanticType} from the specified {@link SemanticType}.
   * @param sty {@link SemanticType} to be copied
   */
  public SemanticTypeImpl(SemanticType sty) {
    setSty(sty.getSty());
    setAtui(sty.getAtui());
    setConcept(sty.getConcept());
    setCvf(sty.getCvf());
  }

  /**
   * Returns the {@link HierarchicalSemanticType} (STY).
   * @return the {@link HierarchicalSemanticType}
   */
  public HierarchicalSemanticType getSemanticType() {
    return sty;
  }

  /**
   * Sets the {@link HierarchicalSemanticType} (STY).
   * @param sty the {@link HierarchicalSemanticType}
   */
  public void setSemanticType(HierarchicalSemanticType sty) {
    this.sty = sty;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getSemanticType()}
   * @return the {@link HierarchicalSemanticType}
   */
  public HierarchicalSemanticType getSty() {
    return getSemanticType();
  }

  /**
   * Shorthand for {@link #setSemanticType(HierarchicalSemanticType)}.
   * @param sty the {@link HierarchicalSemanticType}
   */
  public void setSty(HierarchicalSemanticType sty) {
    setSemanticType(sty);
  }

  /**
   * Returns the unique identifier (ATUI).
   * @return the ATUI
   */
  public String getAttributeIdentifier() {
    return atui;
  }

  /**
   * Shorthand for {@link #getAttributeIdentifier()}.
   * @return the ATUI
   */
  public String getAtui() {
    return getAttributeIdentifier();
  }

  /**
   * Sets the unique identifier (ATUI).
   * @param atui the ATUI
   */
  public void setAttributeIdentifier(String atui) {
    this.atui = atui;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setAttributeIdentifier(String)}.
   * @param atui the ATUI
   */
  public void setAtui(String atui) {
    setAttributeIdentifier(atui);
  }

  /**
   * Returns the {@link Concept} that holds this {@link SemanticType}.
   * @return the {@link Concept} that holds this {@link SemanticType}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link SemanticType}.
   * @param concept the {@link Concept} that holds this {@link SemanticType}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns tbe {@link ContentView} definition for this {@link SemanticType} (CVF).
   * @return tbe {@link ContentView} definition
   */
  public ContentView getContentView() {
    return cvf;
  }

  /**
   * Sets tbe {@link ContentView} definition for this {@link SemanticType} (CVF).
   * @param cvf the {@link ContentView} definition
   */
  public void setContentView(ContentView cvf) {
    this.cvf = cvf;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #getContentView()}.
   * @return tbe {@link ContentView} definition
   */
  public ContentView getCvf() {
    return getContentView();
  }

  /**
   * Shorthand for {@link #setContentView(ContentView)}.
   * @param cvf the {@link ContentView} definition
   */
  public void setCvf(ContentView cvf) {
    setContentView(cvf);
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, and sty ui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(100);
      sb1.append(getConcept() != null ? getConcept().getCui() : "")
         .append(getSty() != null ? getSty().getUi() : "");
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }
  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link SemanticType} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof SemanticType)) return false;
    return getKeyFields().equals(((SemanticType)o).getKeyFields());
  }

  /**
   * Compares this {@link SemanticType} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof SemanticType)
      return getKeyFields().compareTo(((SemanticType)o).getKeyFields());
    else
      return 0;
  }

}
