/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       SourceMetaDescriptionImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.AttributeName;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.MetaDescription;
import gov.nih.nlm.umls.meta.SourceMetaDescription;
import gov.nih.nlm.umls.meta.TermType;
import gov.nih.nlm.util.FieldedStringTokenizer;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Default implementation of the {@link SourceMetaDescription} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class SourceMetaDescriptionImpl implements SourceMetaDescription {

  //
  // Fields
  //
  private String vcui;
  private String rcui;
  private String vsab;
  private String rsab;
  private String son;
  private String sf;
  private String sver;
  private Date mstart;
  private Date mend;
  private String imeta;
  private String rmeta;
  private String slc;
  private String scc;
  private String srl;
  private int tfr;
  private int cfr;
  private String cxty;
  private List ttys = new ArrayList();
  private List atns = new ArrayList();
  private Language lat;
  private String cenc;
  private boolean curver;
  private boolean sabin;
  private String ssn;
  private String scit;

  /**
   * Instantiates an empty {@link SourceMetaDescriptionImpl}.
   */
  public SourceMetaDescriptionImpl() {};

  /**
   * Instantiates the {@link SourceMetaDescription} with the fields
   * from a MRSAB row.
   * @param mrsab_row row from MRSAB
   */
  public SourceMetaDescriptionImpl(String mrsab_row) {
    SimpleDateFormat df = new SimpleDateFormat("yyyy_MM_dd");

    String [] fields  = FieldedStringTokenizer.split(mrsab_row, "|", 26);

    vcui = fields[0];
    rcui = fields[1];
    vsab = fields[2];
    rsab = fields[3];
    son = fields[4];
    sf = fields[5];
    sver = fields[6];
    String date_token = "";
    try {
      date_token = fields[7];
      if(!date_token.equals("")) {
        mstart = df.parse(date_token);
      }

      date_token = fields[8];
      if(!date_token.equals("")) {
        mend = df.parse(date_token);
      }
    } catch(Exception e) {
    }
    imeta = fields[9];
    rmeta = fields[10];
    slc = fields[11];
    scc = fields[12];
    srl = fields[13];
    String token = fields[14];
    if(!token.equals(""))
      tfr = new Integer(token).intValue();
    token = fields[15];
    if(!token.equals(""))
      cfr = new Integer(token).intValue();
    cxty = fields[16];
    String ttyl_token = fields[17];
    StringTokenizer ttyl_st = new StringTokenizer(ttyl_token, ",");
    TermType termType;
    while(ttyl_st.hasMoreTokens()) {
      termType = TermType.getTermType(ttyl_st.nextToken());
      if (termType != null) {
        ttys.add(termType);
      }
    }
    String atnl_token = fields[18];
    StringTokenizer atnl_st = new StringTokenizer(atnl_token, ",");
    while(atnl_st.hasMoreTokens()) {
      String att_token = atnl_st.nextToken();
      AttributeName att = AttributeName.getAttributeName(att_token);
      if (att == null) {
        MetaDescription[] mdsc_array = new MetaDescription[1];
        mdsc_array[0] = new MetaDescriptionImpl("ATN|" +
                                              att_token + "|expanded_form||");
        AttributeName.initialize(mdsc_array);
      }
      atns.add(AttributeName.getAttributeName(att_token));
    }
    lat = Language.getLanguage(fields[19]);
    cenc = fields[20];
    if (fields[21].equals("Y"))
      curver = true;
    else
      curver = false;
    if (fields[22].equals("Y"))
      sabin = true;
    else
      sabin = false;
    if (fields.length >= 25) {
      ssn = fields[23];
      scit = fields[24];
    }
  }


  /**
   * Shorthand for {@link #getVersionedConceptIdentifier()}.
   * @return the concept identifier for the versioned SRC concept
   */
  public String getVcui() {
    return getVersionedConceptIdentifier();
  }

  /**
   * Returns the concept identifier for the versioned SRC concept (VCUI).
   * @return the concept identifier for the versioned SRC concept
   */
  public String getVersionedConceptIdentifier() {
    return vcui;
  }

  /**
   * Shorthand for {@link #getRootConceptIdentifier()}.
   * @return the concept identifier for the root SRC concept
   */
  public String getRcui() {
    return getRootConceptIdentifier();
  }

  /**
   * Returns the concept identifier for the root SRC concept (VCUI).
   * @return the concept identifier for the root SRC concept
   */
  public String getRootConceptIdentifier() {
    return rcui;
  }

  /**
   * Shorthand for (@ link #getVersionedSourceAbbreviation()}.
   * @return the versioned source abbreviation
   */
  public String getVsab() {
    return getVersionedSourceAbbreviation();
  }

  /**
   * Returns the versioned source abbreviation (VSAB).
   * @return the versioned source abbreviation
   */
  public String getVersionedSourceAbbreviation() {
    return vsab;
  }

  /**
   * Shorthand for {@link #getRootSourceAbbreviation()}.
   * @return the root source abbreviation
   */
  public String getRsab() {
    return getRootSourceAbbreviation();
  }

  /**
   * Returns the root source abbreviation (RSAB).
   * @return the root source abbreviation
   */
  public String getRootSourceAbbreviation() {
    return rsab;
  }

  /**
   * Shorthand for {@link #getOfficialName()}.
   * @return the source official name
   */
  public String getSon() {
    return getOfficialName();
  }

  /**
   * Returns the source official name (SON).
   * @return the source official name
   */
  public String getOfficialName() {
    return son;
  }

  /**
   * Shorthand for {@link #getShortName()}.
   * @return the source short name
   */
  public String getSsn() {
    return getShortName();
  }

  /**
   * Returns the source short name (SSN).
   * @return the source short name
   */
  public String getShortName() {
    return ssn;
  }

  /**
   * Shorthand for {@link #getCitation()}.
   * @return the source citation
   */
  public String getScit() {
    return getCitation();
  }

  /**
   * Returns the source citation (SCIT).
   * @return the source citation
   */
  public String getCitation() {
    return scit;
  }

  /**
   * Shorthand for {@link #getSourceFamily()}.
   * @return the source family
   */
  public String getSf() {
    return getSourceFamily();
  }

  /**
   * Returns the source family (SF).
   * @return the source family
   */
  public String getSourceFamily() {
    return sf;
  }

  /**
   * Shorthand for {@link #getVersion()}.
   * @return the source version
   */
  public String getSver() {
    return getVersion();
  }

  /**
   * Returns source version (SVER).
   * @return the source version
   */
  public String getVersion() {
    return sver;
  }

  /**
   * Shorthand for {@link #getValidStartDate()}.
   * @return the date when this source becomes valid in the Metathesaurus
   */
  public Date getMstart() {
    return getValidStartDate();
  }

  /**
   * Returns the date when this source becomes valid in the Metathesaurus (MSTART).
   * @return the date when this source becomes valid in the Metathesaurus
   */
  public Date getValidStartDate() {
    return mstart;
  }

  /**
   * Shorthand for {@link #getValidEndDate()}.
   * @return the date when this source is no longer valid in the Metathesaurus
   */
  public Date getMend() {
    return getValidEndDate();
  }

  /**
   * Returns the date when this source is no longer valid in the Metathesaurus (MEND).
   * @return the date when this source is no longer valid in the Metathesaurus
   */
  public Date getValidEndDate() {
    return mend;
  }

  /**
   * Returns the version of the Metathesaurus in which this source
   * first appeared (IMETA).
   * @return the version of the Metathesaurus in which this source
   * first appeared
   */
  public String getInsertMetaVersion() {
    return imeta;
  }

  /**
   * Shorthand for {@link #getInsertMetaVersion()}
   * @return the version of the Metathesaurus in which this source
   * first appeared
   */
  public String getImeta() {
    return getInsertMetaVersion();
  }

  /**
   * Shorthand for {@link #getRemoveMetaVersion()}
   * @return the first version of the Metathesaurus in which this source
   * no longer
   */
  public String getRmeta() {
    return getRemoveMetaVersion();
  }

  /**
   * Returns the first version of the Metathesaurus in which this source
   * no longer (RMETA).
   * @return the first version of the Metathesaurus in which this source
   * no longer
   */
  public String getRemoveMetaVersion() {
    return rmeta;
  }

  /**
   * Shorthand for {@link #getLicenseContact()}.
   * @return the license contact information
   */
  public String getSlc() {
    return getLicenseContact();
  }

  /**
   * Returns the license contact information (SLC).
   * @return the license contact information
   */
  public String getLicenseContact() {
    return slc;
  }

  /**
   * Returns the content contact information (SCC).
   * @return the content contact information
   */
  public String getContentContact() {
    return scc;
  }

  /**
   * Shorthand for {@link #getContentContact()}.
   * @return the content contact information
   */
  public String getScc() {
    return getContentContact();
  }

  /**
   * Shorthand for {@link #getRestrictionLevel()}.
   * @return the restriction level
   */
  public String getSrl() {
    return getRestrictionLevel();
  }

  /**
   * Returns the restriction level (SRL).
   * @return the restriction level
   */
  public String getRestrictionLevel() {
    return srl;
  }

  /**
   * Shorthand for {@link #getTermFrequency()}
   * @return the frequency of terms
   */
  public int getTfr() {
    return getTermFrequency();
  }

  /**
   * Returns the frequency of terms (TFR).
   * @return the frequency of terms
   */
  public int getTermFrequency() {
    return tfr;
  }

  /**
   * Shorthand for {@link #getConceptFrequency()}
   * @return the frequency of concepts
   */
  public int getCfr() {
    return getConceptFrequency();
  }

  /**
   * Returns the frequency of concepts (CFR).
   * @return the frequency of concepts
   */
  public int getConceptFrequency() {
    return cfr;
  }

  /**
   * Shorthand for {@link #getContextType()}.
   * @return the context type
   */
  public String getCxty() {
    return getContextType();
  }

  /**
   * Returns the context type (CXTY).
   * @return the context type
   */
  public String getContextType() {
    return cxty;
  }

  /**
   * Returns the {@link TermType}s.
   * @return an array of {@link TermType}s
   */
  public TermType[] getTermTypes() {
    return (TermType[])ttys.toArray(new TermType[0]);
  }

  /**
   * Returns the {@link TermType}s as a {@link List}.
   * @return the {@link TermType}s as a {@link List}
   */
  public List getTermTypesAsList() {
    return new ArrayList(ttys);
  }

  /**
   * Shorthand for {@link #getTermTypesAsString()}.
   * @return the term type values
   */
  public String getTtyl() {
    return getTermTypesAsString();
  }

  /**
   * Returns the term types as a comma (,) separated
   * list of term type abbreviations (TTYL).
   * @return the term type values
   */
  public String getTermTypesAsString() {
    StringBuffer ttyl = new StringBuffer(200);
    for (int i=0; i<ttys.size(); i++) {
      if (i <= (ttys.size() - 2))
        ttyl.append(ttys.get(i)).append(",");
      else
        ttyl.append(ttys.get(i));
    }
    return ttyl.toString();
  }


  /**
   * Get the {@link AttributeName}s for this {@link SourceMetaDescription}.
   * Returns the {@link AttributeName}s.
   * @return an array of {@link AttributeName}s
   */
  public AttributeName[] getAttributeNames() {
    return (AttributeName[])atns.toArray(new AttributeName[0]);
  }

  /**
   * Returns the {@link AttributeName}s as a {@link List}.
   * @return the {@link AttributeName}s as a {@link List}
   */
  public List getAttributeNamesAsList() {
    return new ArrayList(atns);
  }

  /**
   * Shorthand for {@link #getAttributeNamesAsString()}.
   * @return the attribute name values
   */
  public String getAtnl() {
    return getAttributeNamesAsString();
  }

  /**
   * Returns the attribute names as a comma (,) separated
   * list of attribute name abbreviations (ATNL).
   * @return the attribute name values
   */
  public String getAttributeNamesAsString() {
    StringBuffer atnl = new StringBuffer(200);
    for (int i=0; i<atns.size(); i++) {
      if (i <= (atns.size() - 2))
        atnl.append(atns.get(i)).append(",");
      else
        atnl.append(atns.get(i));
    }
    return atnl.toString();
  }

  /**
   * Shorthand for {@link #getLanguage()}.
   * @return the {@link Language}
   */
  public Language getLat() {
    return getLanguage();
  }

  /**
   * Returns the {@link Language} (LAT).
   * @return the {@link Language}
   */
  public Language getLanguage() {
    return lat;
  }

  /**
   * Shorthand for {@link #getCharacterEncoding()}.
   * @return the character encoding
   */
  public String getCenc() {
    return getCharacterEncoding();
  }

  /**
   * Returns the ISO character encoding (CENC).
   * @return the character encoding
   */
  public String getCharacterEncoding() {
    return cenc;
  }

  /**
   * Returns the "is current version" value (CURVER).
   * @return the "is current version" value
   */
  public String getCurver() {
    return (curver ? "Y" : "N");
  }

  /**
   * Indicates whether or not this represents the current version
   * of the source.
   * @return <code>true</code> if it does, <code>false</code> otherwise.
   */
  public boolean isCurrentVersion() {
    return curver;
  }

  /**
   * Returns the "in current subset" value (SABIN).
   * @return the "in current subset" value
   */
  public String getSabin() {
    return (sabin ? "Y" : "N");
  }

  /**
   * Indicates whether or not this source is represented
   * in the current subset.
   * @return <code>true</code> if it is, <code>false</code> otherwise.
   */
  public boolean inCurrentSubset() {
    return sabin;
  }

  /**
   * Sets whether or not this represents the current version
   * of the source.
   * @param sabin <code>boolean</code>
   */
  public void setInCurrentSubset(boolean sabin) {
    this.sabin = sabin;
  }

  /**
   * Sets the "in current subset" value (SABIN).
   * @param sabin "in current subset" value
   */
  public void setSabin(String sabin) {
    this.sabin = (sabin != null && sabin.equals("Y"));
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getVersionedSourceAbbreviation().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be a
   * {@link SourceMetaDescription} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof SourceMetaDescription)) return false;
    SourceMetaDescription source = (SourceMetaDescription)o;
    return
      source.getVersionedSourceAbbreviation() == getVersionedSourceAbbreviation();
  }

  /**
   * Compares this {@link SourceMetaDescription} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof SourceMetaDescription)
      return getVersionedSourceAbbreviation().compareTo(
        ((SourceMetaDescription)o).getVersionedSourceAbbreviation());
    else
      return 0;
  }

}
