/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       StringContextImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Atom;
import gov.nih.nlm.umls.meta.ContextGroup;
import gov.nih.nlm.umls.meta.Source;
import gov.nih.nlm.umls.meta.SourceContext;
import gov.nih.nlm.umls.meta.StringContext;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link StringContext} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class StringContextImpl implements StringContext {

  //
  // Fields
  //
  private Atom atom;
  private List source_contexts = new ArrayList();
  private ContextGroup context_group;
  private Source source;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;


  /**
   * Instantiates an empty {@link StringContext}.
   */
  public StringContextImpl() {};

  /**
   * Instantiates a {@link StringContext} from the specified {@link StringContext}.
   * @param sc {@link StringContext} to be copied
   */
  public StringContextImpl(StringContext sc) {
    setAtom(sc.getAtom());
    setSourceContexts(sc.getSourceContexts());
    setContextGroup(sc.getContextGroup());
    setSource(sc.getSource());
  }

  /**
   * Returns the {@link ContextGroup} that holds this {@link StringContext}.
   * @return the {@link ContextGroup} that holds this {@link StringContext}
   */
  public ContextGroup getContextGroup() {
    return context_group;
  }

  /**
   * Sets the {@link ContextGroup} that holds this {@link StringContext}.
   * @param cg the {@link ContextGroup} that holds this {@link StringContext}
   */
  public void setContextGroup(ContextGroup cg) {
    context_group = cg;
    key_fields_flag = false;
  }

  /**
   * Returns all {@link SourceContext}s.
   * @return an array of {@link SourceContext}s
   */
  public SourceContext[] getSourceContexts() {
    return (SourceContext[])source_contexts.toArray(new SourceContext[0]);
  }

  /**
   * Returns all {@link SourceContext}s as a {@link List}.
   * @return a List of {@link SourceContext}s
   */
  public List getSourceContextsAsList() {
    return new ArrayList(source_contexts);
  }

  /**
   * Adds the specified {@link SourceContext}.
   * @param sc the {@link SourceContext} to be added
   */
  public void addSourceContext(SourceContext sc) {
    source_contexts.add(sc);
  }

  /**
   * Removes the specified {@link SourceContext}.
   * @param sc the {@link SourceContext} to be removed
   */
  public void removeSourceContext(SourceContext sc) {
    source_contexts.remove(sc);
  }

  /**
   * Removes all {@link SourceContext}s.
   */
  public void clearSourceContexts() {
    source_contexts.clear();
  }

  /**
   * Replace the {@link SourceContext}s with the ones in the specified array.
   * @param scs a {@link SourceContext}<code>[]</code>
   */
  public void setSourceContexts(SourceContext [] scs) {
    source_contexts.clear();
    for (int i= 0; i<scs.length; i++) {
      source_contexts.add(scs[i]);
    }
  }

  /**
   * Replace the {@link SourceContext}s with the ones in the specified {@link List}.
   * @param scs a {@link List} of {@link SourceContext}s
   */
  public void setSourceContexts(List scs) {
    source_contexts = new ArrayList(scs);
  }

  /**
   * Returns the {@link Atom} for this {@link StringContext}.
   * @return {@link Atom}
   */
  public Atom getAtom() {
    return atom;
  }

  /**
   * Sets the source for this {@link StringContext}.
   * @param atom {@link Source}
   */
  public void setAtom(Atom atom) {
    this.atom = atom;
    key_fields_flag = false;
  }

  /**
   * Returns the source for this {@link StringContext}.
   * @return {@link Source}
   */
  public Source getSource() {
    return source;
  }

  /**
   * Sets the source for this {@link StringContext}.
   * @param source {@link Source}
   */
  public void setSource(Source source) {
    this.source = source;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui, aui, sui,
   * versioned source, tty, and code.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      key_fields = getAtom().getCode();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link StringContext} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof StringContext)) return false;
    return getKeyFields().equals(((StringContext)o).getKeyFields());
  }

  /**
   * Compares this {@link StringContext} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof StringContext)
      return getKeyFields().compareTo(((StringContext)o).getKeyFields());
    else
      return 0;
  }

}
