/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       TermAttributeImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.StringAttribute;
import gov.nih.nlm.umls.meta.Term;
import gov.nih.nlm.umls.meta.TermAttribute;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link TermAttribute} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class TermAttributeImpl implements TermAttribute {

  //
  // Fields
  //
  private String lui;
  private List string_attributes = new ArrayList();
  private Concept concept;
  private Term term;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;


  /**
   * Instantiates an empty {@link TermAttribute}.
   */
  public TermAttributeImpl() {};

  /**
   * Instantiates a {@link TermAttribute} from the specified {@link TermAttribute}.
   * @param ta {@link TermAttribute} to be copied
   */
  public TermAttributeImpl(TermAttribute ta) {
    setLui(ta.getLui());
    setStringAttributes(ta.getStringAttributes());
    setConcept(ta.getConcept());
  }

  /**
   * Returns the {@link Term}.
   * @return {@link Term}
   */
  public Term getTerm() {
    return term;
  }

  /**
   * Sets the {@link Term} to which this {@link TermAttribute} refers.
   * @param term {@link Term} to which this {@link TermAttribute} refers
   */
  public void setTerm(Term term) {
    this.term = term;
  }

  /**
   * Shorthand for {@link #getTermIdentifier()}.
   * @return the LUI
   */
  public String getLui() {
    return getTermIdentifier();
  }

  /**
   * Shorthand for {@link #setTermIdentifier(String)}.
   * @param lui the LUI
   */
  public void setLui(String lui) {
    setTermIdentifier(lui);
  }

  /**
   * Returns the term identifier (LUI).
   * @return the LUI
   */
  public String getTermIdentifier() {
    return lui;
  }

  /**
   * Sets the term identifier (LUI).
   * @param lui the LUI
   */
  public void setTermIdentifier(String lui) {
    this.lui = lui;
    key_fields_flag = false;
  }

  /**
   * Returns all {@link StringAttribute}s.
   * @return a array of {@link StringAttribute}s
   */
  public StringAttribute [] getStringAttributes() {
    return (StringAttribute[])string_attributes.toArray(new StringAttribute[0]);
  }

  /**
   * Returns all {@link StringAttribute}s as a {@link List}.
   * @return a {@link List} of {@link StringAttribute}s
   */
  public List getStringAttributesAsList() {
    return new ArrayList(string_attributes);
  }

  /**
   * Adds the specified {@link StringAttribute}.
   * @param sa the {@link StringAttribute} to be added
   */
  public void addStringAttribute(StringAttribute sa) {
    string_attributes.add(sa);
    // need to add Attribute to concept?
  }

  /**
   * Removes the specified {@link StringAttribute}.
   * @param sa the {@link StringAttribute} to be removed
   */
  public void removeStringAttribute(StringAttribute sa) {
    string_attributes.remove(sa);
  }

  /**
   * Removes all {@link StringAttribute}s.
   */
  public void clearStringAttributes() {
    string_attributes.clear();
  }

  /**
   * Replaces the {@link StringAttribute}s with the ones in the specified array.
   * @param sa a {@link StringAttribute}<code>[]</code>
   */
  public void setStringAttributes(StringAttribute [] sa) {
    string_attributes.clear();
    for (int i=0; i<sa.length; i++) {
      string_attributes.add(sa[i]);
    }
  }

  /**
   * Replaces the {@link StringAttribute}s with the ones in the specified {@link List}.
   * @param sa a {@link List} of {@link StringAttribute}s
   */
  public void setStringAttributes(List sa) {
    string_attributes = new ArrayList(sa);
  }

  /**
   * Returns the {@link Concept} that holds this {@link TermAttribute}.
   * @return the {@link Concept} that holds this {@link TermAttribute}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link TermAttribute}.
   * @param concept the {@link Concept} that holds this {@link TermAttribute}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui and lui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(20);
      sb1.append(getConcept().getCui())
         .append(getLui());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link TermAttribute} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof TermAttribute)) return false;
    return getKeyFields().equals(((TermAttribute)o).getKeyFields());
  }

  /**
   * Compares this {@link TermAttribute} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof TermAttribute)
      return getKeyFields().compareTo(((TermAttribute)o).getKeyFields());
    else
      return 0;
  }


}
