/**************************************************************************
 * Package:     gov.nih.nlm.umls.meta.impl
 * Class:       TermImpl.java
 *
 **************************************************************************/
package gov.nih.nlm.umls.meta.impl;

import gov.nih.nlm.umls.meta.Concept;
import gov.nih.nlm.umls.meta.Language;
import gov.nih.nlm.umls.meta.StringInfo;
import gov.nih.nlm.umls.meta.Term;
import gov.nih.nlm.umls.meta.TermAttribute;
import gov.nih.nlm.umls.meta.TermStatus;

import java.util.ArrayList;
import java.util.List;

/**
 * Default implementation of the {@link Term} interface.
 *
 * @author Brian Carlsen, Deborah Shapiro
 */
public class TermImpl implements Term {

  //
  // Fields
  //
  private Language lat;
  private TermStatus ts;
  private String lui;
  private String str;
  private List term_variants = new ArrayList();
  private List term_attributes = new ArrayList();
  private Concept concept;

  // Used to compute key fields more efficiently
  private boolean key_fields_flag;
  private String key_fields = null;


  /**
   * Instantiates an empty {@link Term}.
   */
  public TermImpl() {};

  /**
   * Instantiates a {@link Term} from the specified {@link Term}.
   * @param term {@link Term} to be copied
   */
  public TermImpl(Term term) {
    setLui(term.getLui());
    setTs(term.getTs());
    setLat(term.getLat());
    setStr(term.getStr());
    setTermVariants(term.getTermVariants());
    setConcept(term.getConcept());
  }

  /**
   * Returns the {@link Language} (LAT).
   * @return the {@link Language}
   */
  public Language getLanguage() {
    return lat;
  }

  /**
   * Shorthand for {@link #getLanguage()}.
   * @return {@link Language}
   */
  public Language getLat() {
    return getLanguage();
  }

  /**
   * Sets the {@link Language} (LAT).
   * @param lat the {@link Language}
   */
  public void setLanguage(Language lat) {
    this.lat = lat;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setLanguage(Language)}.
   * @param lat the {@link Language}
   */
  public void setLat(Language lat) {
    setLanguage(lat);
  }

  /**
   * Returns the {@link TermStatus} (TS).
   * @return the {@link TermStatus}
   */
  public TermStatus getTermStatus() {
    return ts;
  }

  /**
   * Shorthand for {@link #getTermStatus()}.
   * @return the {@link TermStatus}
   */
  public TermStatus getTs() {
    return getTermStatus();
  }

  /**
   * Sets the {@link TermStatus} (TS).
   * @param ts the {@link TermStatus}
   */
  public void setTermStatus(TermStatus ts) {
    this.ts = ts;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setTermStatus(TermStatus)}.
   * @param ts the {@link TermStatus}
   */
  public void setTs(TermStatus ts) {
    setTermStatus(ts);
  }

  /**
   * Shorthand for {@link #getTermIdentifier()}.
   * @return the LUI
   */
  public String getLui() {
    return getTermIdentifier();
  }

  /**
   * Shorthand for {@link #setTermIdentifier(String)}.
   * @param lui the LUI
   */
  public void setLui(String lui) {
    setTermIdentifier(lui);
  }

  /**
   * Returns the term identifier (LUI).
   * @return the LUI
   */
  public String getTermIdentifier() {
    return lui;
  }

  /**
   * Sets the term identifier (LUI).
   * @param lui the LUI
   */
  public void setTermIdentifier(String lui) {
    this.lui = lui;
    key_fields_flag = false;
  }

  /**
   * Sets the name (STR).
   * @return the name
   */
  public String getName() {
    return str;
  }

  /**
   * Shorthand for {@link #getName()}.
   * @return the name
   */
  public String getStr() {
    return getName();
  }

  /**
   * Sets the name (STR).
   * @param str the name
   */
  public void setName(String str) {
    this.str = str;
    key_fields_flag = false;
  }

  /**
   * Shorthand for {@link #setName(String)}.
   * @param str the name
   */
  public void setStr(String str) {
    setName(str);
  }

  /**
   * Returns all {@link StringInfo} term variants.
   * @return an array of {@link StringInfo} objects
   */
  public StringInfo [] getTermVariants() {
    return (StringInfo[])term_variants.toArray(new StringInfo[0]);
  }

  /**
   * Returns all {@link StringInfo} term variants as a {@link List}.
   * @return all {@link StringInfo} term variants as a {@link List}
   */
  public List getTermVariantsAsList() {
    return new ArrayList(term_variants);
  }

  /**
   * Adds the specified term variant.
   * @param tv a {@link StringInfo} term variant to add
   */
  public void addTermVariant(StringInfo tv) {
    term_variants.add(tv);
  }

  /**
   * Removes all term variants.
   */
  public void clearTermVariants() {
    term_variants.clear();
  }

  /**
   * Removes the specified term variant.
   * @param tv a {@link StringInfo} term variant to remove
   */
  public void removeTermVariant(StringInfo tv) {
    term_variants.remove(tv);
  }

  /**
   * Replaces term variants with the specified {@link StringInfo} objects.
   * @param tvs an array of {@link StringInfo} objects
   */
  public void setTermVariants(StringInfo [] tvs) {
    term_variants.clear();
    for (int i=0; i<tvs.length; i++) {
      term_variants.add(tvs[i]);
    }
  }

  /**
   * Replaces term variants with the specified {@link List}
   * @param tvs a {@link List} of {@link StringInfo} objects
   */
  public void setTermVariants(List tvs) {
    term_variants = tvs;
  }

  /**
   * Returns all {@link TermAttribute}s.
   * @return an array of {@link TermAttribute} objects
   */
  public TermAttribute [] getTermAttributes() {
    return (TermAttribute[])term_attributes.toArray(new TermAttribute[0]);
  }

  /**
   * Returns all {@link TermAttribute}s as a {@link List}.
   * @return all {@link TermAttribute}s as a {@link List}
   */
  public List getTermAttributesAsList() {
    return new ArrayList(term_attributes);
  }

  /**
   * Adds the specified {@link TermAttribute}.
   * @param ta a {@link TermAttribute} to add
   */
  public void addTermAttribute(TermAttribute ta) {
    term_attributes.add(ta);
  }

  /**
   * Removes all {@link TermAttribute}s.
   */
  public void clearTermAttributes() {
    term_attributes.clear();
  }

  /**
   * Removes the specified {@link TermAttribute}.
   * @param ta the {@link TermAttribute} to remove
   */
  public void removeTermAttribute(TermAttribute ta) {
    term_attributes.remove(ta);
  }

  /**
   * Replaces the {@link TermAttribute}s with the specified {@link TermAttribute} objects.
   * @param tas an array of {@link TermAttribute} objects
   */
  public void setTermAttributes(TermAttribute [] tas) {
    term_attributes.clear();
    for (int i=0; i<tas.length; i++) {
      term_attributes.add(tas[i]);
    }
  }

  /**
   * Replaces {@link TermAttribute}s with the specified {@link List}
   * @param tas a {@link List} of {@link TermAttribute} objects
   */
  public void setTermAttributes(List tas) {
    term_attributes = tas;
  }

  /**
   * Returns the {@link Concept} that holds this {@link Term}.
   * @return the {@link Concept} that holds this {@link Term}
   */
  public Concept getConcept() {
    return concept;
  }

  /**
   * Sets the {@link Concept} that holds this {@link Term}.
   * @param concept the {@link Concept} that holds this {@link Term}
   */
  public void setConcept(Concept concept) {
    this.concept = concept;
    key_fields_flag = false;
  }

  /**
   * Returns a {@link String} key which can be used
   * by equals, compareTo, and hashCode methods.
   * @return a {@link String} composed from the cui and lui.
   */
  public String getKeyFields() {
    if (key_fields == null || !key_fields_flag) {
      StringBuffer sb1 = new StringBuffer(20);
      sb1.append(getConcept().getCui())
         .append(getLui());
      key_fields = sb1.toString();
      key_fields_flag = true;
    }
    return key_fields;
  }

  /**
   * Return a hash code based on key fields.
   * @return a hash code based on key fields
   */
  public int hashCode() {
    return getKeyFields().hashCode();
  }

  /**
   * Returns <code>true</code> if the specified object is
   * equal to this object.  The specified object must be an
   * {@link Term} with matching key fields.
   * @param o an {@link Object} to compare against
   * @return boolean <code>true</code> if Object o is equal
   *                 <code>false</code> otherwise
   */
  public boolean equals(Object o) {
    if (o == null || !(o instanceof Term)) return false;
    return getKeyFields().equals(((Term)o).getKeyFields());
  }

  /**
   * Compares this {@link Term} to another one.  Results
   * are indeterminate if the specified parameter is not of the
   * same class.
   * @param o the {@link Object} to compare to
   * @return <code>int</code> representing the relative ordering
   *         of this and the compare-to object
   */
  public int compareTo(Object o) {
    if (o instanceof Term)
      return getKeyFields().compareTo(((Term)o).getKeyFields());
    else
      return 0;
  }

  /**
   * Indicates whether or not this {@link Term} is the preferred {@link Term}
   * of the {@link Concept}.  In other words, does this {@link Term} have
   * the same LUI as the preferred {@link Term} of the {@link Concept}.
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredTermIdentifier() {
    return getTs().getAbbreviation().toLowerCase().equals("p");
  }

  /**
   * Shorthand for {@link #hasPreferredTermIdentifier()}
   * @return <code>true</code> if so, <code>false</code> otherwise
   */
  public boolean hasPreferredLui() {
    return hasPreferredTermIdentifier();
  }
}
